/*  $Id: 3dscroller.C,v 1.1 90/06/17 18:44:52 gjb Exp $
 *
 *  3dscroller.C:							*
 *									*
 *  "3D" versions of the vertical and horizontal scroll bars.		*
 *									*
 *  By Greg Brail. Copyright 1990 Brown University.			*/

#include "3dscroller.h"
#include <InterViews/painter.h>
#include <InterViews/canvas.h>
#include <InterViews/color.h>
#include <InterViews/perspective.h>
#include <InterViews/pattern.h>

const int inset = 1;		// Space between scroller and bar

inline void Scroller::Background (Coord x1, Coord y1, Coord x2, Coord y2) {
    output->FillRect(canvas, x1, y1, x2, y2);
}

ThreeDVScroller::ThreeDVScroller(Interactor* i, int size)
: (i, size)

{
  SetClassName("ThreeDVScroller");
  border = 4;
}

ThreeDHScroller::ThreeDHScroller(Interactor* i, int size)
: (i, size)

{
  SetClassName("ThreeDHScroller");
  border = 4;
}

void ThreeDHScroller::GetBarInfo (register Perspective* s, Coord& left, int& width) {
    Coord maxwidth = xmax + 1;

    if (s->width == 0) {
        scale = 0.0;
        left = -1;
        width = maxwidth + 2;
    } else {
	scale = double(maxwidth) / double(s->width);
        left = round(double(s->curx - s->x0) * scale);
        width = max(round(double(s->curwidth) * scale), 5);
    }
    if (width < (border * 3))
      border = width / 3;
}

void ThreeDVScroller::GetBarInfo (register Perspective* s, Coord& bot, int& height) {
    Coord maxheight = ymax + 1;

    if (s->height == 0) {
        scale = 0.0;
        bot = -1;
        height = maxheight + 2;
    } else {
	scale = double(maxheight) / double(s->height);
        bot = round(double(s->cury - s->y0) * scale);
        height = max(round(double(s->curheight) * scale), 5);
    }
    if (height < (border * 3))
      border = height / 3;
}

inline void ThreeDHScroller::Outline (Coord x, int width) {
    output->Rect(canvas, x, inset, x+width-1, ymax-inset);
}

inline void ThreeDVScroller::Outline (Coord y, int height) {
    output->Rect(canvas, inset, y, xmax-inset, y+height-1);
}

inline void ThreeDHScroller::Border (Coord x) {
    output->Line(canvas, x, inset, x, ymax-inset);
}

inline void ThreeDVScroller::Border (Coord y) {
    output->Line(canvas, inset, y, xmax-inset, y);
}

inline void ThreeDHScroller::Sides (Coord x1, Coord x2) {
    output->Line(canvas, x1, inset, x2, inset);
    output->Line(canvas, x1, ymax-inset, x2, ymax-inset);
}

inline void ThreeDVScroller::Sides (Coord y1, Coord y2) {
    output->Line(canvas, inset, y1, inset, y2);
    output->Line(canvas, xmax-inset, y1, xmax-inset, y2);
}

void ThreeDVScroller::Redraw(Coord x1, Coord y1, Coord x2, Coord y2)

{
  Coord bot;
  int height;

  Background(x1, y1, x2, y2);
  GetBarInfo(shown, bot, height);
  Bar(bot, height);
  Outline(bot, height);
}

void ThreeDHScroller::Redraw(Coord x1, Coord y1, Coord x2, Coord y2)

{
  Coord left;
  int width;
  Background(x1, y1, x2, y2);
  GetBarInfo(shown, left, width);
  Bar(left, width);
  Outline(left, width);
}

void ThreeDVScroller::Bar(Coord bot, int height)

{
  Pattern* oldPat = output->GetPattern();
  output->SetPattern(solid);
  Color* origFg = output->GetFgColor();
  Color* origBg = output->GetBgColor();
  output->ClearRect(canvas, inset + 1, bot, xmax - inset - 1, bot + height - 1);
  output->SetColors(white, origBg);
  output->FillRect(canvas, inset + 1, bot, 
		   inset + border + 1, bot + height - 1);
  output->SetColors(black, origBg);
  output->FillRect(canvas, xmax - inset - 1, bot,
		   xmax - inset - border - 1, bot + height - 1);
  output->SetColors(origFg, origBg);
  output->SetPattern(oldPat);
}

void ThreeDHScroller::Bar(Coord x, int width)

{
  Pattern* oldPat = output->GetPattern();
  output->SetPattern(solid);
  Color* origFg = output->GetFgColor();
  Color* origBg = output->GetBgColor();
  output->ClearRect(canvas, x, inset + 1, x + width - 1, ymax - inset - 1);
  output->SetColors(white, origBg);
  output->FillRect(canvas, x, ymax - inset - 1, 
		   x + width - 1, ymax - inset - border - 1);
  output->SetColors(black, origBg);
  output->FillRect(canvas, x, inset + 1, 
		   x + width - 1, inset + border + 1);
  output->SetColors(origFg, origBg);
  output->SetPattern(oldPat);
}

void ThreeDHScroller::Update () {
    Coord oldleft, oldright, newleft, newright;
    int oldwidth, newwidth;
    Perspective* p;

    if (canvas == nil) {
	return;
    }
    p = view;
    GetBarInfo(shown, oldleft, oldwidth);
    GetBarInfo(p, newleft, newwidth);
    if (oldleft != newleft || oldwidth != newwidth) {
	oldright = oldleft+oldwidth-1;
	newright = newleft+newwidth-1;
	if (oldright >= newleft && newright >= oldleft) {
	    if (oldright > newright) {
		Background(newright+1, inset, oldright, ymax-inset);
		Border(newright);
	    } else if (oldright < newright) {
		Bar(oldright, newright-oldright);
		Sides(oldright, newright);
		Border(newright);
	    }
	    if (oldleft > newleft) {
		Bar(newleft+1, oldleft-newleft);
		Sides(newleft, oldleft);
		Border(newleft);
	    } else if (oldleft < newleft) {
		Background(oldleft, inset, newleft-1, ymax-inset);
		Border(newleft);
	    }
	} else {
	    Background(oldleft, inset, oldright, ymax-inset);
	    Bar(newleft, newwidth);
	    Outline(newleft, newwidth);
	}
    }
    *shown = *p;
}

void ThreeDVScroller::Update () {
    Coord oldbottom, oldtop, newbottom, newtop;
    int oldheight, newheight;
    Perspective* p;

    if (canvas == nil) {
	return;
    }
    p = view;
    GetBarInfo(shown, oldbottom, oldheight);
    GetBarInfo(p, newbottom, newheight);
    if (oldbottom != newbottom || oldheight != newheight) {
	oldtop = oldbottom+oldheight-1;
	newtop = newbottom+newheight-1;
	if (oldtop >= newbottom && newtop >= oldbottom) {
	    if (oldtop > newtop) {
		Background(inset, newtop+1, xmax-inset, oldtop);
		Border(newtop);
	    } else if (oldtop < newtop) {
		Bar(oldtop, newtop-oldtop);
		Sides(oldtop, newtop);
		Border(newtop);
	    }
	    if (oldbottom > newbottom) {
		Bar(newbottom+1, oldbottom-newbottom);
		Sides(newbottom, oldbottom);
		Border(newbottom);
	    } else if (oldbottom < newbottom) {
		Background(inset, oldbottom, xmax-inset, newbottom-1);
		Border(newbottom);
	    }
	} else {
	    Background(inset, oldbottom, xmax-inset, oldtop);
	    Bar(newbottom, newheight);
	    Outline(newbottom, newheight);
	}
    }
    *shown = *p;
}
