/*
 * Copyright (c) 1987, 1988, 1989, 1990, 1991 Stanford University
 * Copyright (c) 1991 Silicon Graphics, Inc.
 * Copyright (c) 1991 Yufan Hu (yufan@cs.keele.ac.uk)
 *		Dept. of Computer Science, University of Keele, UK
 * 
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * Stanford and Silicon Graphics may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of Stanford and Silicon Graphics.
 *
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
 *
 * IN NO EVENT SHALL STANFORD OR SILICON GRAPHICS OR KEELE UNIVERSITY OR
 * Yufan Hu BE LIABLE FOR ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL 
 * DAMAGES OF ANY KIND, OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF 
 * USE, DATA OR PROFITS, WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, 
 * AND ON ANY THEORY OF LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE 
 * USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include <InterViews/textbuffer.h>
#include <InterViews/regexp.h>

#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/param.h>

#include "Iediteditor.h"
#include "filemgr.h"

FileManager::FileManager(){
	_buffer = new char [MINTEXTSIZE];
	_text = new TextBuffer(_buffer, 0, MINTEXTSIZE);
	filename = nil; _size = 0;
}

FileManager::~FileManager(){
	delete _text;
	delete _buffer;
}

void FileManager::CheckBuffer(SelEditor* editor, int inc){
	int len = _text->Length();
	if(len + inc + MINFREESIZE <= _size) return;
	int newSize = round(_size+inc+MINTEXTSIZE);
	char* new_buffer = new char[newSize];
	if(new_buffer == nil) return;
	strncpy(new_buffer, _text->Text(), len);
	int dot = editor->Dot();
	delete _buffer;
	delete _text;
	_buffer = new_buffer;
	_size = newSize;
	_text = new TextBuffer(_buffer, len, _size);
	boolean dirty = editor->Dirty();
	editor->Edit(_text);
	editor->Dirty(dirty);
	editor->Select(dot);
}

boolean FileManager::Open(const char* fname){
	delete _buffer;
	delete _text;
	FILE* f = fopen(fname, "r");
	boolean readonly = false;

	delete filename;
	filename = strdup(fname);
	if(f != nil){
		struct stat filestats;
		stat(fname, &filestats);
		_size = round (filestats.st_size + MINTEXTSIZE);
		_buffer = new char[_size];
		char* b = _buffer;
		int remaining = _size;
		while(remaining > 1 && fgets(b,  remaining, f) != nil){
			int l = strlen(b);
			remaining -= l;
			b += l;
		}
		fclose(f);
		_text = new TextBuffer(_buffer, b-_buffer, _size);
		readonly = access(fname, W_OK) != 0;
	} else {
		_size = MINTEXTSIZE;
		_buffer = new char[_size];
		_text = new TextBuffer(_buffer, 0, _size);
	}
	return readonly;
}

boolean FileManager::Include(SelEditor* editor, FILE* f, int inc){
	int newSize = round(_size+inc+MINTEXTSIZE);
	char* new_buffer = new char[newSize];
	if(new_buffer == nil) return false;
	_size = newSize;
	int dot = editor->Dot();
	strcpy(new_buffer, _text->Text(0,dot));
	char* b = &new_buffer[dot];
	int remaining = _size - dot;
	while(remaining > 1 && fgets(b,  remaining, f) != nil){
		int l = strlen(b);
		remaining -= l;
		b += l;
	}
	strcpy(b, _text->Text(dot, _text->EndOfText()));
	int len = b-new_buffer + strlen(b);
	dot = b-new_buffer;
	delete _buffer;
	delete _text;
	_buffer = new_buffer;
	_text = new TextBuffer(_buffer, len, _size);
	editor->Edit(_text);
	editor->Dirty(true);
	editor->Select(dot);
	return true;
}

boolean FileManager::Save(const char* name){
	delete filename;
	filename = strdup(name);
	FILE* f = fopen(name,"w");
	if(f != nil){
		for (int i=0; i < _text->Length(); ++i){
			putc(_text->Char(i), f);
		}
	} else return false;
	fclose(f);
	return true;
}

boolean FileManager::FwdSearch(const char* string, int& beg, int&end){
	Regexp re(string);

	boolean found =
		_text->ForwardSearch(&re, beg) >= 0 ||
		_text->ForwardSearch(&re, _text->BeginningOfText()) >= 0;

	if( found ) {
		beg = re.BeginningOfMatch();
		end = re.EndOfMatch();
		if( end > _text->EndOfText() ) /* Bug in REG match */
			found = false;
	}
	return found;
}

boolean FileManager::BwdSearch(const char* string, int& beg, int& end){
	Regexp re(string);

	boolean found =
		_text->BackwardSearch(&re, beg) >= 0 ||
		_text->BackwardSearch(&re, _text->EndOfText()) >= 0;

	if( found ) {

		beg = re.BeginningOfMatch();
		end = re.EndOfMatch();
		if( end > _text->EndOfText() ) /* Bug in REG match */
			found = false;
	}
	return found;
}
