static char _rcs_id[] = "DATAP $Source: /usr/datap/src/Iv/RCS/RadioButtonSet.cc,v $ $Revision: 1.4 $";

//
// Copyright (c) DATAP Systems, 1991. All rights reserved.
// License material - Property of DATAP Systems.
//
// Project:
//   Iris 77000
//
// Configuration:
//
// Revision History:
//   $Log: RadioButtonSet.cc,v $
// Revision 1.4  1991/10/31  23:23:57  pbrown
// Resolved action items for inspection #202
//
// Revision 1.3  1991/10/17  19:43:28  pbrown
// Pre-inspection cleanup.
//
// Revision 1.2  1991/10/11  18:51:02  dmorriso
// Updated to use new IV 3.0 release.
// Now use the function 'choose' instead of 'select'.
//
// Revision 1.1  1991/07/15  19:50:28  morrison
// Initial revision
//
// Description:
//   This class provides the mechanisms for grouping 'radioButtons' 
//   together and remembering which one is currently chosen.
//
// Hints:
//   There are two steps involved in using this class.  The constructor
//   should be passed a guess at the number of RadioButtons that will be
//   attached.  A call to addButton() will then be made as each new
//   RadioButton is constructed for the button set.

#include <Iv/RadioButtonSet.h>
#include <Rtap/ThisProcess.h>

#include <stream.h>

// Function:
//   This constructor reserves space for the default number of buttons.
//   It also initializes several variables.

RadioButtonSet::RadioButtonSet(int defaultNbrOfButtons)
{
    ThisProcess::constructorActivated("RadioButtonSet(int)",
				      (int) this);

    // The minimum number of radio buttons is one.
    // To simplify coding a button zero has been placed in position zero
    // in the _buttons array that has nil for its button pointer.

    _totalNbrOfButtons = ((defaultNbrOfButtons > 0) ?
        defaultNbrOfButtons + 1 : 1);

    // Reserve space for the default number of buttons.

    _buttons = (RadioButton**) new char[sizeof(RadioButton*)
        * _totalNbrOfButtons];

    // Initialize to indicate no buttons chosen.

    _currentNbrOfButtons = 1;
    _chosenButton = 0;
    _buttons[_chosenButton] = nil;
}

// Function:
//   This destructor unreferences each RadioButton in the set and then
//   deletes the array pointing to them.

RadioButtonSet::~RadioButtonSet()
{
    ThisProcess::destructorActivated("~RadioButtonSet()",
				     (int) this);
    for (int i = 1; i < _currentNbrOfButtons; i++) {
        Resource::unref(_buttons[i]);
    }
    delete _buttons;
}

// Function:
//   This function returns the state of the button by indicating
//   whether it is enabled or not.  The button is described by its
//   button set index.

boolean RadioButtonSet::buttonEnabled(int button) const
{
    if ((button > 0) && (button < _currentNbrOfButtons)) {
	return _buttons[button]->enabled();
    } else {
	return false;
    }
}

// Function:
//   This function is to display information about the button
//   set for debugging purposes.

void RadioButtonSet::display() const
{
    cout << "Radio Button Set\n"
         << "\tTotal buttons:   " << _totalNbrOfButtons << "\n"
         << "\tButtons used:    " << _currentNbrOfButtons << "\n"
         << "\tChosen button: " << chosenButtonIndex() << "\n\n";
    cout.flush();
}

// Function:
//   This function provides the ability to preselect a radio button
//   by button set index before buttons are displayed.

void RadioButtonSet::chooseButton(int button)
{
    if ((button > 0) && (button < _currentNbrOfButtons)) {
	chooseButton(_buttons[button]);
    } else {
	chooseButton((RadioButton*) nil);
    }
}

// Function:
//   This function provides the ability to enable (state == true) or
//   disable (state == false) a radio button by button set index.  If
//   the "chosen" button is disabled it will stay chosen.

void RadioButtonSet::enableButton(int button, boolean state)
{
    if ((button > 0) && (button < _currentNbrOfButtons)) {
	_buttons[button]->enable(state);
    }
}

// Function:
//   This function provides the ability to enable (state == true) or
//   disable (state == false) all radio buttons in the set.  The 
//   chosen button will remain chosen.

void RadioButtonSet::enableAllButtons(boolean state)
{
    for (int i = 1; i < _currentNbrOfButtons; i++) {
        enableButton(i, state);
    }
}

// Private Function:
//   This routine adds a new radio button to the set.  It is assumed
//   that nil is NOT passed.

void RadioButtonSet::addButton(RadioButton* button)
{
    // If there is no room left in the _buttons array then expand
    // it to accomodate the new RadioButton

    if (_totalNbrOfButtons == _currentNbrOfButtons) {
        RadioButton** newButtons = (RadioButton**)
            new char[sizeof(RadioButton*) * (++_totalNbrOfButtons)];
        for (int i = 0; i < _currentNbrOfButtons; i++) {
            newButtons[i] = _buttons[i];
	}
        delete _buttons;
        _buttons = newButtons;
    }
    _buttons[_currentNbrOfButtons++] = button;
    Resource::ref(button);
}

// Private Function:
//   This function provides the 'car radio button' functionality.
//   It is passed a pointer to the newly chosen button.  The routine 
//   loops through the list of existing buttons looking for a match.
//   When the chosen button is found the index is saved.  All other 
//   buttons are 'unchosen'.

void RadioButtonSet::chooseButton(const RadioButton* button)
{
    int i;
    RadioButton* aButton;
    for (i = 0; i < _currentNbrOfButtons; i++) {
        if ((aButton = _buttons[i]) == button) {
            _chosenButton = i;
	    if ((aButton != nil) && (!aButton->chosen())) {
		aButton->choose(true);
	    }
	} else {
            if (aButton != nil) {
                aButton->choose(false);
            }
	}
    }
}
