/*
 * Copyright (c) 1992 U.S. Geological Survey (USGS)
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of USGS not be used in advertising or
 * publicity pertaining to distribution of the software without specific,
 * written prior permission.  USGS makes no representations about the
 * suitability of this software for any purpose.  It is provided "as is"
 * without express or implied warranty.
 *
 * USGS DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL USGS
 * BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION
 * OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
 * CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 */

#include "Command.h"
#include "HyperView.h"
#include "TextView.h"
#include "PageBorder.h"
#include "doc-listener.h"

#include <IV-look/button.h>
#include <IV-look/kit.h>
#include <InterViews/background.h>
#include <InterViews/bitmap.h>
#include <InterViews/border.h>
#include <InterViews/hit.h>
#include <InterViews/label.h>
#include <InterViews/layout.h>
#include <InterViews/patch.h>
#include <InterViews/style.h>
#include <InterViews/style.h>
#include <InterViews/session.h>
#include <OS/list.h>
#include <OS/string.h>
#include <string.h>
#include <stdlib.h>

/*-------------------------------------------------------------------------
   Class HyperView

   The HyperView class provides a window for viewing TextView instances
   and a set of buttons for paging or printing the current TextView.
   TextViews are loaded on an AS NEEDED basis.

   Members:
      command	- processes a button press
      event	- required to receive events
      hypertext - skip to page containing hypertext section
      jump_to	- jump to page containing a given character item
      nextview  - view the next TextView item
      prevview  - view the previous TextView item
      page_to	- changes the current page
      pop	- pop last hypertext source
      readTOC	- reads the Table of Contents
      _toc	- list of Table of Contents items
      updateTOC - update the TOC offset variable
      TOC       - finds TOC index for a given tag

 *-------------------------------------------------------------------------*/

const Coord TEXTWIDTH  = 450.0;
const Coord TEXTHEIGHT = 240.0;
 
//
// *** create TOC structure (see list.h)
//
declareList(TextItem_List, TextItem)
implementList(TextItem_List, TextItem)

//
// *** create TOC structure (see list.h)
//
declareList(TOC_List, TOCItem)
implementList(TOC_List, TOCItem)

//
// *** create Hypertext source stack
//
declareList(Hypertext_Stack, HyperSource)
implementList(Hypertext_Stack, HyperSource)

const int MAXSTACKDEPTH = 5;

//
//  *** read of line of text from an MML file
//
int getline(FILE* fd, char* line)
   {
   int   i = 0;
   int   c;

   //
   //   *** skip leading blanks
   //
   while (((c = fgetc(fd)) != EOF) && ((c == ' ') || (c == '\n')));

   line[i++] = c;

   if ((c == EOF) || (c == '\n'))
      return(c);

   while (((c = fgetc(fd)) != EOF) && (c != '\n'))
         line[i++] = (char)c;

   line[i] = '\0';
   return((char)c);
   }

HyperViewer::HyperViewer (const char* htag=NULL)
          : ApplicationWindow(_top = new Patch(nil)), Handler() 
    {
    FILE   *fd;
    int    i;
    char   fullname[256];
    void   init_fonts();
    String v;
    Style  *style = Session::instance()->style();
    float  convert_metric(const char*);

    _page_width  = TEXTWIDTH;
    _page_height = TEXTHEIGHT;

    if (style->find_attribute("pagewidth", v))
       _page_width = (Coord)convert_metric(v.string());
    if (style->find_attribute("pageheight", v))
       _page_height = (Coord)convert_metric(v.string());
    if (style->find_attribute("book", v))
       {
       _home = new char[strlen(v.string()) + strlen(getenv("HYPERDOC")) + 2];
       strcpy(_home, getenv("HYPERDOC"));
       strcat(_home, "/");
       strcat(_home, v.string());
       }
    else
       _home = strcpy(new char[strlen(getenv("HYPERDOC")) + 1], 
                      getenv("HYPERDOC"));

    //
    //  *** get default colors and font
    //
    WidgetKit& kit = *WidgetKit::instance();
    const LayoutKit& layout = *LayoutKit::instance();
    const Color* fg = kit.foreground();
    const Color* bg = kit.background();
    const Font* font = kit.font();

    //
    //  *** initialize fonts needed by textview 
    //
    init_fonts();

    _textlist   = new TextItem_List(20);
    _hyperstack = new Hypertext_Stack(MAXSTACKDEPTH);

    //
    //   *** initialize Table of contents
    //
    _toc  = new TOC_List(10);

    if (_home != NULL)
       {
       strcpy(fullname, _home);
       strcat(fullname, "/TOC");
       }
    else
       strcpy(fullname, "TOC");

    if ((fd = fopen(fullname, "r")) == (FILE*)NULL)
       {
       printf("%s not found\n", fullname);
       exit(0);
       return;
       }

    readTOC(fd);
    fclose(fd);

    //
    //  *** create textview of given hyper tag
    //
    if (htag == NULL)
       htag = _toc->item(0)._tag;

    i = 0;
    while ((i < _toc->count()) &&
           (strcasecmp(_toc->item(i)._tag, htag) != 0))
       i++;

    if (i < _toc->count())
       {
       TextItem  t;

       t._view = new TextView(this, _home, _toc->item(i)._filename, i,
                              _page_width,_page_height);
       Resource::ref((Glyph*)t._view);
       _textlist->append(t);
       _item = 0;
       }
    else
       {
       printf("Unknown HyperText:  %s\n", htag);
       exit(0);
       }

    _prev_page = kit.push_button("Prev",  new Command(this, "page backward"));
    _next_page = kit.push_button("Next",  new Command(this, "page forward"));
    _pop = kit.push_button("Pop",  new Command(this, "pop"));
    _print = kit.push_button("Print",  new Command(this, "print"));
    _quit = kit.push_button("Quit", new Command(this, "quit"));

    Resource::ref(_prev_page);
    Resource::ref(_next_page);
    Resource::ref(_pop);
    Resource::ref(_print);
    Resource::ref(_quit);

    //
    //  *** align buttons left to right
    //
    _footer_patch = new Patch(
        layout.hbox_first_aligned(
                 layout.hglue(), 
                 _prev_page, 
                 _next_page,
                 _pop,
                 _print,
                 _quit,
                 layout.hglue()
        )
    );

    Resource::ref(_footer_patch);


    _page_patch = new Patch(nil);
    Resource::ref(_page_patch);
   
    _body_patch = new Patch(
        layout.vbox_first_aligned(
            layout.margin(
                new PageBorder(
                    layout.margin(_page_patch, 1, 5, 5, 1), fg, bg
                ),
                5,   0, 0,
                5, fil, 0,
                5, fil, 0,
                5,   0, 0
            ),
            layout.margin(_footer_patch, 0, 0, 5, 5)
        )
    );
    Resource::ref(_body_patch);
 
    Listener* listener = new Listener(
        new Background(_body_patch, bg), this
    );

    listener->button(true);

    _top->body(listener);

    _page_patch->body(
        layout.margin(_textlist->item(_item)._view, 10.0, 10.0, 10.0, 10.0)
        );

   //
   //  *** skip to requested hypertext
   //
   jump_to(_toc->item(i)._offset);
   }

HyperViewer::~HyperViewer()
   {
   _top->body(nil);
   _page_patch->body(nil);

   Resource::unref(_prev_page);
   Resource::unref(_next_page);
   Resource::unref(_pop);
   Resource::unref(_print);

   Resource::unref(_footer_patch);
   Resource::unref(_page_patch);
   Resource::unref(_body_patch);

   while (_textlist->count())
      {
      Resource::unref((Glyph*)_textlist->item(0)._view);
      _textlist->remove(0);
      }
   delete _textlist;

   while (_toc->count())
      {
      delete _toc->item(0)._section;
      delete _toc->item(0)._filename;
      delete _toc->item(0)._tag;
      _toc->remove(0);
      }
   delete _toc;

   _hyperstack->remove_all();
   delete _hyperstack;

   if (_home != nil)
      delete _home;
   }

//
//  *** command is invoked by the HyperView buttons
//
void HyperViewer::command(const char* command)
   {
   if (strcmp(command, "page forward") == 0)
      {
      if (_current_page < _textlist->item(_item)._view->Count())
         _textlist->item(_item)._view->view_page(++_current_page);
      else
         nextview();
      }
   else if (strcmp(command, "page backward") == 0)
      {
      if (_current_page > 1)
         _textlist->item(_item)._view->view_page(--_current_page);
      else
         prevview();
      }
   else if (strcmp(command, "pop") == 0)
      pop();
   else if (strcmp(command, "print") == 0)
      {
      _textlist->item(_item)._view->printout();
      }
   else if (strcmp(command, "quit") == 0)
      {
      void  destroy_fonts();
      
      destroy_fonts();
      delete this;
      Session::instance()->quit();
      }
   }


boolean HyperViewer::event (Event& e) 
   {
   return true;
   }

//
//  *** page within the current textview
//
void HyperViewer::page_to(int p)
   {
   _current_page = p;
   _textlist->item(_item)._view->view_page(_current_page);
   }

//
//  *** jump to given character offset
//
void HyperViewer::jump_to(int offset)
   {
   _textlist->item(_item)._view->jump_to(offset);
   }

void HyperViewer::readTOC(FILE* fd)
   {
   TOCItem tocitem;
   char    filename[80];
   char    section[20];
   char    tag[80];
   char    line[100];
   char    c;

   while (getline(fd, line) != EOF)
      {
      switch (line[0])
         {
         case ' ':
         case '\n':
            break;

         //
         // *** assume comment
         //
         case '<':
            if (strchr(line, '>') == NULL)
               {
               while ((c = getc(fd)) != '>');
               while ((c = getc(fd)) != '\n');
               }
            break;

         default:
            sscanf(line, "%s %s %s", section, filename, tag);
            tocitem._section  = strcpy(new char[strlen(section)+1], section);
            tocitem._filename = strcpy(new char[strlen(filename)+1], filename);
            tocitem._tag      = strcpy(new char[strlen(tag)+1], tag);
            tocitem._offset   = -1;
            _toc->append(tocitem);
            break;
         }
      }
   }

void HyperViewer::updateTOC(FILE* fd, int tcount)
   {
   long  offset;
   char  sect[20];
   char  line[100];
   int   i;

   offset = ftell(fd);

   if (getline(fd, line) != EOF)
      {
      sscanf(line, "%s", sect);

     /*
      *  *** find TOC entry for this section and update offset
      */
      for (i = 0; i < _toc->count(); i++)
         {
         if (strcmp(_toc->item(i)._section, sect) == 0)
            {
            TOCItem ti = _toc->item(i);
            ti._offset = tcount;
             _toc->remove(i);  
            _toc->insert(i, ti);
            break;
            }
         }
      }

   //
   //  *** reset file to original position
   //
   fseek(fd, offset, 0);
   }

int HyperViewer::TOC(char* tag)
   {
   int   i;

   for (i = 0; i < _toc->count(); i++)
      {
      if (strcasecmp(tag, _toc->item(i)._tag) == 0)
         return((int)i);
      }

   return(255);
   }

void HyperViewer::hypertext(int index)
   {
   TextItem    titem;
   HyperSource hypsrc;
   int         newitem;
   int         i;

   const LayoutKit& layout = *LayoutKit::instance();

   //
   //  *** add current location to hypertext stack
   //
   hypsrc._item  = _item;
   hypsrc._page = _current_page;

   //
   //  *** append to list, if at maximum stack depth, remove first item
   //
   if (_hyperstack->count() < MAXSTACKDEPTH)
      _hyperstack->append(hypsrc);
   else
      {
      _hyperstack->remove(0);
      _hyperstack->append(hypsrc);
      }
   
   //
   // *** if hypertext in current textview
   //
   if (strcmp(_toc->item(index)._filename, 
              _textlist->item(_item)._view->MMLFile()) == 0)
      {
      jump_to(_toc->item(index)._offset);
      return;
      }

   //
   //  *** check loaded file for hypertext
   //
   for (i = 0; i < _textlist->count(); i++)
      {
      if (strcmp(_textlist->item(i)._view->MMLFile(), 
                 _toc->item(index)._filename) == 0)
         break;
      }

   //
   //  *** if view with hypertext is not loaded, load new MML file
   //
   if (i < _textlist->count())
      newitem = i;
   else
      {
      titem._view = new TextView(this,  _home, _toc->item(index)._filename, 
                                 index, _page_width, _page_height);
      Resource::ref((Glyph*)titem._view);
      newitem = (int)_textlist->count();
      _textlist->append(titem);
      }

   //
   //  *** jump to hypertext page
   //
   _item = newitem;
   _page_patch->body( 
           layout.margin(_textlist->item(newitem)._view, 10.0, 10.0, 10.0, 10.0)
           );
   _page_patch->reallocate();
   jump_to(_toc->item(index)._offset);
   }

void HyperViewer::nextview()
   {
   TextItem titem;
   TOCItem  ti;
   int      i, j;
   int      newitem;
   char     *current;

   current = (char*)_textlist->item(_item)._view->MMLFile();

   //
   //  *** find  next MML file name
   //
   i = 0;
   while (strcmp(_toc->item(i)._filename, current) != 0)
      {
      ti = _toc->item(i);
      i++;
      }

   while ((i < _toc->count()) &&
          (strcmp(_toc->item(i)._filename, current) == 0)) 
      i++;

   //
   //  *** if at end, return
   //
   if (i ==_toc->count())
      return;

   j = 0;
   for (j = 0; j < _textlist->count(); j++)
       {
       if (strcmp(_textlist->item(j)._view->MMLFile(), 
                  _toc->item(i)._filename) == 0)
          {
          newitem = j;
          break;
          }
       }

   //
   //  *** if next section is not loaded, create new TextView
   //
   if (j == _textlist->count())
      {
      titem._view = new TextView(this,  _home, _toc->item(i)._filename, i,
                                 _page_width, _page_height);
      Resource::ref((Glyph*)titem._view);
      newitem = (int)_textlist->count();
      _textlist->append(titem);
      }

   //
   //  *** change text view and view first page
   //
   if (newitem != _item)
      {
      const LayoutKit& layout = *LayoutKit::instance();

      _item = newitem;
      _page_patch->body( 
           layout.margin(_textlist->item(newitem)._view, 10.0, 10.0, 10.0, 10.0)
      );
      _page_patch->reallocate();
      page_to(1);
      }
   }

void HyperViewer::prevview()
   {
   TextItem titem;
   int      i, j;
   int      newitem;
   char     *current;

   current = (char*)_textlist->item(_item)._view->MMLFile();

   //
   //  *** find previous MML file name
   //
   i = 0;
   while (strcmp(_toc->item(i)._filename, current) != 0)
      i++;

   while ((i > 0) && (strcmp(_toc->item(i)._filename, current) == 0)) 
      i--;

   if (i < 0)
      return;

   j = 0;
   for (j = 0; j < _textlist->count(); j++)
       {
       if (strcmp(_textlist->item(j)._view->MMLFile(), 
                  _toc->item(i)._filename) == 0)
          {
          newitem = j;
          break;
          }
       }

   //
   //  *** if not loaded, create new TextView
   //
   if (j == _textlist->count())
      {
      titem._view = new TextView(this, _home, (char*)_toc->item(i)._filename, i,
                                 _page_width, _page_height);
      Resource::ref((Glyph*)titem._view);
      newitem = (int)_textlist->count();
      _textlist->append(titem);
      }

   //
   //  *** view last page of previous file
   //
   if (newitem != _item)
      {
      const LayoutKit& layout = *LayoutKit::instance();

      _item = newitem;
      _page_patch->body( 
           layout.margin(_textlist->item(newitem)._view, 10.0, 10.0, 10.0, 10.0)
      );
      _page_patch->reallocate();
      _current_page = _textlist->item(newitem)._view->last_page();
      }
   }

void HyperViewer::pop()
   {
   long         top;
   HyperSource  hypersrc;
   
   //
   //  *** if nothing in stack, return
   //
   if ((top = _hyperstack->count()) == 0)
      return;

   //
   //  *** pop stack
   //
   hypersrc = _hyperstack->item(top - 1);
   _hyperstack->remove(top-1);

   //
   //  *** jump back to the hypertext source
   //
   if (_item != hypersrc._item)
      {
      const LayoutKit& layout = *LayoutKit::instance();

      _item = hypersrc._item;
      _page_patch->body(
           layout.margin(_textlist->item(_item)._view, 10.0, 10.0, 10.0, 10.0)
           );
      _page_patch->reallocate();
      if (hypersrc._page != 0)
         page_to(hypersrc._page);
      }
   else
      page_to(hypersrc._page);
   }

