// This file is part of xplot.
// 
// Copyright (C) 1989 Rene' Seindal (seindal@diku.dk)
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.


#include "xplot.hh"
#include "util.hh"
#include "plot.hh"
#include "canvas.hh"

#include <InterViews/sensor.h>
#include <InterViews/painter.h>
#include <InterViews/font.h>
#include <InterViews/shape.h>
#include <InterViews/world.h>
#include <InterViews/rubrect.h>
#include <InterViews/event.h>
#include <InterViews/cursor.h>
#include <InterViews/canvas.h>
#include <InterViews/perspective.h>
#include <InterViews/transformer.h>

static const int plotview_width = 512;	// width of canvas, if none specified.
static const int plotview_height = 512;	// height of canvas, if none specified.

void PlotCanvas::Init(PlotGraphSet *gset)
{
    SetClassName("PlotCanvas");
    graphset = gset;
}

PlotCanvas::PlotCanvas(PlotCanvas *v)
{
    Init(v->graphset);
}

PlotCanvas::PlotCanvas(PlotGraphSet *gset)
{
    Init(gset);
}

void PlotCanvas::Reconfig()
{
    const char *attr;
    
    attr = GetAttribute("plotGeometry");
    Coord dummy;
    unsigned int mask;
    
    mask = world->ParseGeometry((char *)attr, dummy, dummy, shape->width, shape->height);
    if ((mask & GeomWidthValue) == 0)
	shape->width = plotview_width;
    if ((mask & GeomHeightValue) == 0)
	shape->height = plotview_height;
    
    if ((attr = GetAttribute("labelFont")) != nil) {
	Font *labelfont = new Font(attr);
	
	if (labelfont->Valid()) {
	    Painter *new_output = new Painter(output);
	    delete output;
	    output = new_output;
	    output->SetFont(labelfont);
	} else {
	    fprintf(stderr, "xplot: %s is not valid font.\n", (char *)attr);
	    delete labelfont;
	}
    }
    if (IsOn(GetAttribute("backingStore"))) {
	SetCanvasType(CanvasSaveContents);
    }
    SetCursor(arrow);
}

void PlotCanvas::SelectRect(Coord &x1, Coord &y1, Coord &x2, Coord &y2)
{
    RubberRect *rect;
    Event e;
    Sensor *s = new Sensor;
    Cursor *corners[] = { upperright, upperleft, lowerright, lowerleft };
    
    s->Catch(UpEvent);
    s->Catch(MotionEvent);
    s->Catch(DownEvent);
    
    Listen(s);
    SetCursor(upperleft);
    
    do {
	Read(e);
	if (e.target != this)
	    e.target->Handle(e);
    } while (e.eventType != DownEvent);
    
    x1 = x2 = e.x;
    y1 = y2 = e.y;

    if (e.target == this)
	rect = new RubberRect(output, canvas, x1, y1, x2, y2);
    else {
	rect = nil;
	UnRead(e);
    }
    
    while (rect != nil) {
	Read(e);
	
	if (e.target != this) {
	    rect->Erase();
	    delete rect;
	    rect = nil;
	    UnRead(e);
	    break;
	}

	switch(e.eventType) {
	case MotionEvent:
	    SetCursor(corners[(e.x < x1) | (e.y < y1)<<1]);
	    if (rect != nil)
		rect->Track(e.x, e.y);
	    break;
	    
	case UpEvent:
	    if (rect != nil) {
		rect->GetCurrent(x1, y1, x2, y2);
		rect->Erase();
		delete rect;
		rect = nil;
	    }
	    break;
	default:
	    break;
	}
    }

    SetCursor(arrow);
    Listen(input);
    delete s;
}

void PlotCanvas::Redraw(Coord x1, Coord y1, Coord x2, Coord y2) // Redraw a rectangel
{
    PlotGraph *graph = graphset->Current();
    PlotCmd *cmd;
    Painter *tmp;
    int curx, cury;
    int maxx, maxy, minx, miny;

    if (!graph->Valid())
	return;

    graph->GetArea(minx, miny, maxx, maxy);
    graph->SetTextScale((float)(maxy-miny)/shape->height);

    tmp = new Painter(output);
    tmp->Translate((float)-minx, (float)-miny);
    tmp->Scale(float(canvas->Width())/(float)(maxx-minx),
	       float(canvas->Height())/(float)(maxy-miny));
    tmp->GetTransformer()->InvTransform(x1, y1);
    tmp->GetTransformer()->InvTransform(x2, y2);
    tmp->ClearRect(canvas, x1, y1, x2, y2);

    PlotGraphIterator trav = graph;
    while ((cmd = trav.Next()) != nil)
	cmd->Draw(canvas, tmp, curx, cury, x1, y1, x2, y2);
    delete tmp;
}

