/*
 * Copyright (c) 1990 Stanford University
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided
 * that the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of Stanford not be used in advertising or
 * publicity pertaining to distribution of the software without specific,
 * written prior permission.  Stanford makes no representations about
 * the suitability of this software for any purpose.  It is provided "as is"
 * without express or implied warranty.
 *
 * STANFORD DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS.
 * IN NO EVENT SHALL STANFORD BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 * DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
 * OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION
 * WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

/*
 * Drawing editor main class implementation.
 */

#include "dwgclasses.h"
#include "dwgcmds.h"
#include "dwged.h"
#include "dwgkybd.h"
#include "iv.peel"

#include <Unidraw/catalog.h>
#include <Unidraw/ctrlinfo.h>
#include <Unidraw/editor.h>
#include <Unidraw/editorinfo.h>
#include <Unidraw/grid.h>
#include <Unidraw/keymap.h>
#include <Unidraw/kybd.h>
#include <Unidraw/upage.h>
#include <Unidraw/selection.h>
#include <Unidraw/statevars.h>
#include <Unidraw/stateviews.h>
#include <Unidraw/uctrls.h>
#include <Unidraw/unidraw.h>
#include <Unidraw/viewer.h>

#include <Unidraw/Commands/align.h>
#include <Unidraw/Commands/brushcmd.h>
#include <Unidraw/Commands/catcmds.h>
#include <Unidraw/Commands/colorcmd.h>
#include <Unidraw/Commands/edit.h>
#include <Unidraw/Commands/font.h>
#include <Unidraw/Commands/patcmd.h>
#include <Unidraw/Commands/struct.h>
#include <Unidraw/Commands/transforms.h>
#include <Unidraw/Commands/viewcmds.h>

#include <Unidraw/Components/ellipse.h>
#include <Unidraw/Components/line.h>
#include <Unidraw/Components/link.h>
#include <Unidraw/Components/pad.h>
#include <Unidraw/Components/pin.h>
#include <Unidraw/Components/polygon.h>
#include <Unidraw/Components/rastercomp.h>
#include <Unidraw/Components/rect.h>
#include <Unidraw/Components/slot.h>
#include <Unidraw/Components/spline.h>
#include <Unidraw/Components/stencilcomp.h>
#include <Unidraw/Components/text.h>

#include <Unidraw/Graphic/ellipses.h>
#include <Unidraw/Graphic/picture.h>
#include <Unidraw/Graphic/polygons.h>
#include <Unidraw/Graphic/rasterrect.h>
#include <Unidraw/Graphic/splines.h>
#include <Unidraw/Graphic/ustencil.h>

#include <Unidraw/Tools/connect.h>
#include <Unidraw/Tools/grcomptool.h>
#include <Unidraw/Tools/magnify.h>
#include <Unidraw/Tools/move.h>
#include <Unidraw/Tools/reshape.h>
#include <Unidraw/Tools/rotate.h>
#include <Unidraw/Tools/scale.h>
#include <Unidraw/Tools/select.h>
#include <Unidraw/Tools/stretch.h>

#include <InterViews/bitmap.h>
#include <InterViews/border.h>
#include <InterViews/box.h>
#include <InterViews/button.h>
#include <InterViews/frame.h>
#include <InterViews/glue.h>
#include <InterViews/menu.h>
#include <InterViews/panner.h>
#include <InterViews/raster.h>
#include <InterViews/tray.h>
#include <InterViews/window.h>
#include <InterViews/world.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/*****************************************************************************/

static const char* brAttrib = "brush";
static const char* fontAttrib = "font";
static const char* patAttrib = "pattern";
static const char* fgAttrib = "fgcolor";
static const char* bgAttrib = "bgcolor";

static const char* initBrAttrib = "initialbrush";
static const char* initFontAttrib = "initialfont";
static const char* initPatAttrib = "initialpattern";
static const char* initFgAttrib = "initialfgcolor";
static const char* initBgAttrib = "initialbgcolor";

/*****************************************************************************/

DrawingEditor::DrawingEditor (GraphicComp* comp) {
    SetClassName("DrawingEditor");

    ManagedWindow* window = new ApplicationWindow(this);
    SetWindow(window);

    _comp = (comp == nil) ? (new GraphicComps) : comp;
    _keymap = new KeyMap;
    _curCtrl = new ControlState;
    _selection = new Selection;
    _tray = new Tray;

    InitStateVars();
    InitEditorInfo();
    InitViewer();

    Insert(new Frame(Interior()));
    GetKeyMap()->Execute(CODE_SELECT);

    if (_initEdInfo) {
        Catalog* catalog = unidraw->GetCatalog();
        const char* attrib = catalog->GetAttribute("setup");
        catalog->Save(_edInfo, attrib);
    }
}

void DrawingEditor::InitEditorInfo () {
    Catalog* catalog = unidraw->GetCatalog();
    const char* edInfoFile = catalog->GetAttribute("setup");

    if (edInfoFile == nil) {
        _edInfo = nil;
        _initEdInfo = false;

    } else if (catalog->Retrieve(edInfoFile, _edInfo)) {
        _initEdInfo = false;

    } else {
        _edInfo = new EditorInfo;
        _initEdInfo = true;
    }
}

void DrawingEditor::InitStateVars () {
    _name = new CompNameVar(_comp);
    _modifStatus = new ModifStatusVar(_comp);
    _gravity = new GravityVar;
    _magnif = new MagnifVar;
    _font = new FontVar;
    _brush = new BrushVar;
    _pattern = new PatternVar;
    _color = new ColorVar;

    Catalog* catalog = unidraw->GetCatalog();

    const char* brIndex = catalog->GetAttribute(initBrAttrib);
    const char* fontIndex = catalog->GetAttribute(initFontAttrib);
    const char* patIndex = catalog->GetAttribute(initPatAttrib);
    const char* fgIndex = catalog->GetAttribute(initFgAttrib);
    const char* bgIndex = catalog->GetAttribute(initBgAttrib);

    _brush->SetBrush(catalog->ReadBrush(brAttrib, atoi(brIndex)));
    _font->SetFont(catalog->ReadFont(fontAttrib, atoi(fontIndex)));
    _pattern->SetPattern(catalog->ReadPattern(patAttrib, atoi(patIndex)));
    _color->SetColors(
        catalog->ReadColor(fgAttrib, atoi(fgIndex)),
        catalog->ReadColor(bgAttrib, atoi(bgIndex))
    );
}

DrawingEditor::~DrawingEditor () {
    delete _keymap;
    delete _selection;

    if (_edInfo != nil) {
        delete _edInfo;
    }
}

Component* DrawingEditor::GetComponent () { return _comp; }
Viewer* DrawingEditor::GetViewer (int id) { return (id == 0) ? _viewer : nil; }
KeyMap* DrawingEditor::GetKeyMap () { return _keymap; }
Selection* DrawingEditor::GetSelection () { return _selection; }

StateVar* DrawingEditor::GetState (const char* name) {
    if (strcmp(name, "CompNameVar") == 0) {
        return _name;
    } else if (strcmp(name, "ModifStatusVar") == 0) {
        return _modifStatus;
    } else if (strcmp(name, "MagnifVar") == 0) {
        return _magnif;
    } else if (strcmp(name, "GravityVar") == 0) {
        return _gravity;
    } else if (strcmp(name, "FontVar") == 0) {
        return _font;
    } else if (strcmp(name, "BrushVar") == 0) {
        return _brush;
    } else if (strcmp(name, "PatternVar") == 0) {
        return _pattern;
    } else if (strcmp(name, "ColorVar") == 0) {
        return _color;
    } else {
        return Editor::GetState(name);
    }
}

Tool* DrawingEditor::GetCurTool () {
    UControl* c = (UControl*) _curCtrl->Selection();
    return (Tool*) c->GetControlInfo()->GetOwner();
}

void DrawingEditor::SetComponent (Component* comp) {
    if (comp == nil || comp->IsA(GRAPHIC_COMP)) {
        _comp = (GraphicComp*) comp;
    }
}

void DrawingEditor::SetViewer (Viewer* v, int i) { 
    if (i == 0) {
        _viewer = v;
    }
}

void DrawingEditor::SetKeyMap (KeyMap* k) { _keymap = k; }
void DrawingEditor::SetSelection (Selection* s) { _selection = s; }

Interactor* DrawingEditor::Interior () {
    Interactor* tools = Tools();
    Interactor* commands = Commands();
    HBorder* hborder = new HBorder;
    VBorder* vborder = new VBorder;
    int gap = round(.1*cm);

    HBox* indicators = new HBox(
        new BrushVarView(_brush, _color),
        new VBorder,
        new PatternVarView(_pattern, _color)
    );
    VBox* status = new VBox(
        new HBox(
            new HGlue(gap, 0, 0),
            new ModifStatusVarView(_modifStatus),
            new CompNameVarView(_name, Left),
            new MagnifVarView(_magnif),
            new GravityVarView(_gravity, Right),
            new FontVarView(_font, Right)
        ),
        new HBorder 
    );
    VBox* panel = new VBox(
        tools,
        new VGlue,
        new HBorder,
        new Panner(_viewer)
    );
    panel->Propagate(false);
    _tray->Propagate(false);
    
    _tray->HBox(_tray, status, _tray);
    _tray->HBox(_tray, indicators, vborder, commands, new HGlue, _tray);
    _tray->HBox(_tray, hborder, _tray);
    _tray->HBox(_tray, panel, vborder, _viewer, _tray);
    _tray->VBox(_tray, status, indicators, hborder, panel, _tray);
    _tray->VBox(_tray, status, vborder, _tray);
    _tray->VBox(_tray, status, commands, hborder, _viewer, _tray);

    return _tray;
}

Interactor* DrawingEditor::Commands () {
    return (_edInfo == nil || _initEdInfo) ? InitCommands() : ReadCommands();
}

Interactor* DrawingEditor::InitCommands () {
    MenuBar* commands = new MenuBar;

    commands->Include(FileMenu());
    commands->Include(EditMenu());
    commands->Include(StructureMenu());
    commands->Include(FontMenu());
    commands->Include(BrushMenu());
    commands->Include(PatternMenu());
    commands->Include(ColorMenu("FgColor", fgAttrib));
    commands->Include(ColorMenu("BgColor", bgAttrib));
    commands->Include(AlignMenu());
    commands->Include(ViewMenu());

    return commands;
}

static EditorInfo* FindCommands (EditorInfo* edInfo) {
    EditorInfo* cmds = new EditorInfo;
    
    for (int i = 0; i < edInfo->Count(); ++i) {
        const char* info = edInfo->GetInfo(i);

        if (*info != '\0') {
            cmds->Register(edInfo->GetName(i), info);
        }
    }
    return cmds;
}

static PulldownMenu* MakePulldown (const char* name) {
    return new PulldownMenu(
        new PanelInteractor(new ControlInfo(name), Horizontal)
    );
}

Interactor* DrawingEditor::ReadCommands () {
    MenuBar* menuBar = new MenuBar;
    EditorInfo* cmds = FindCommands(_edInfo);

    while (cmds->Count() != 0) {
        char* menuName = strnew(cmds->GetInfo(0));
        const char* cmdName;

        PulldownMenu* pdm = MakePulldown(menuName);
        menuBar->Include(pdm);

        while ((cmdName = cmds->GetName(menuName)) != nil) {
            if (strcmp(cmdName, "-") == 0) {
                Separator(pdm);

            } else {
                Command* cmd;
                unidraw->GetCatalog()->Retrieve(cmdName, cmd);
                Include(cmd->Copy(), pdm, cmdName, menuName);
            }
            cmds->UnregisterName(cmdName);
        }
        delete menuName;
    }
    delete cmds;
    return menuBar;
}

static const unit = 15;

static Coord xClosed[] = { unit/5, unit, unit, unit*3/5, 0 };
static Coord yClosed[] = { 0, unit/5, unit*3/5, unit, unit*2/5 };
static const int nClosed = 5;

static Coord xOpen[] = { 0, unit/2, unit/2, unit };
static Coord yOpen[] = { 0, unit/4, unit*3/4, unit };
static const int nOpen = 4;

Interactor* DrawingEditor::Tools () {
    return (_edInfo == nil || _initEdInfo) ? InitTools() : ReadTools();
}

static Raster* CreateRaster () {
    const int w = 16;
    const int h = 15;
    const ColorIntensity limit = 1.0;
    const ColorIntensity iinc = limit/w;
    const ColorIntensity jinc = limit/h;
    const ColorIntensity kinc = limit/(w*h);

    Raster* raster = new Raster(w, h);
    int k = 0;

    for (int i = 0; i < w; ++i) {
        for (int j = 0; j < h; ++j) {
            raster->poke(i, j, i*iinc, j*jinc, limit - k*kinc, 1.0);
//            float g = 0.3*i*iinc + 0.59*j*jinc + 0.11*k*kinc;
//            raster->poke(i, j, g, g, g, 1.0);
            ++k;
        }
    }
    raster->flush();

    return raster;
}

Interactor* DrawingEditor::InitTools () {
    VBox* tools = new VBox;

    Include(
        new SelectTool(new ControlInfo("Select", KLBL_SELECT, CODE_SELECT)),
        tools, "Select.tool"
    );
    Include(
        new MoveTool(new ControlInfo("Move", KLBL_MOVE, CODE_MOVE)),
        tools, "Move.tool"
    );
    Include(
        new ScaleTool(new ControlInfo("Scale", KLBL_SCALE, CODE_SCALE)),
        tools, "Scale.tool"
    );
    Include(
        new StretchTool(new ControlInfo("Stretch", KLBL_STRETCH,CODE_STRETCH)),
        tools, "Stretch.tool"
    );
    Include(
        new RotateTool(new ControlInfo("Rotate", KLBL_ROTATE, CODE_ROTATE)),
        tools, "Rotate.tool"
    );
    Include(
        new ReshapeTool(new ControlInfo("Reshape", KLBL_RESHAPE,CODE_RESHAPE)),
        tools, "Reshape.tool"
    );
    Include(
        new MagnifyTool(new ControlInfo("Magnify", KLBL_MAGNIFY,CODE_MAGNIFY)),
        tools, "Magnify.tool"
    );

    TextGraphic* text = new TextGraphic("Text", stdgraphic);
    TextComp* textComp = new TextComp(text);
    Include(
        new GraphicCompTool(
            new ControlInfo("Text", KLBL_TEXT, CODE_TEXT), textComp
        ), tools, "Text.tool"
    );

    Line* line = new Line(0, 0, unit, unit, stdgraphic);
    LineComp* lineComp = new LineComp(line);
    Include(
        new GraphicCompTool(
            new ControlInfo(lineComp, KLBL_LINE, CODE_LINE), lineComp
        ), tools, "Line.tool"
    );

    SF_MultiLine* ml = new SF_MultiLine(xOpen, yOpen, nOpen, stdgraphic);
    ml->SetPattern(psnonepat);
    MultiLineComp* mlComp = new MultiLineComp(ml);
    Include(
        new GraphicCompTool(
            new ControlInfo(mlComp, KLBL_MULTILINE, CODE_MULTILINE), mlComp
        ), tools, "MultiLine.tool"
    );

    SFH_OpenBSpline* spline = new SFH_OpenBSpline(
        xOpen, yOpen, nOpen, stdgraphic
    );
    spline->SetPattern(psnonepat);
    SplineComp* splineComp = new SplineComp(spline);
    Include(
        new GraphicCompTool(
            new ControlInfo(splineComp, KLBL_SPLINE, CODE_SPLINE), splineComp
        ), tools, "Spline.tool"
    );

    SF_Rect* rect = new SF_Rect(0, 0, unit, unit*4/5, stdgraphic);
    rect->SetPattern(psnonepat);
    RectComp* rectComp = new RectComp(rect);
    Include(
        new GraphicCompTool(
            new ControlInfo(rectComp, KLBL_RECT, CODE_RECT), rectComp
        ), tools, "Rect.tool"
    );

    SF_Ellipse* ellipse = new SF_Ellipse(0, 0, unit*2/3, unit*2/5, stdgraphic);
    ellipse->SetPattern(psnonepat);
    EllipseComp* ellipseComp = new EllipseComp(ellipse);
    Include(
        new GraphicCompTool(
            new ControlInfo(ellipseComp, KLBL_ELLIPSE, CODE_ELLIPSE),
            ellipseComp
        ), tools, "Ellipse.tool"
    );

    SF_Polygon* polygon = new SF_Polygon(xClosed, yClosed, nClosed,stdgraphic);
    polygon->SetPattern(psnonepat);
    PolygonComp* polygonComp = new PolygonComp(polygon);
    Include(
        new GraphicCompTool(
            new ControlInfo(polygonComp, KLBL_POLY, CODE_POLY), polygonComp
        ), tools, "Polygon.tool"
    );

    SFH_ClosedBSpline* cspline = new SFH_ClosedBSpline(
        xClosed, yClosed, nClosed, stdgraphic
    );
    cspline->SetPattern(psnonepat);
    ClosedSplineComp* csplineComp = new ClosedSplineComp(cspline);
    Include(
        new GraphicCompTool(
            new ControlInfo(csplineComp, KLBL_CSPLINE,CODE_CSPLINE),csplineComp
        ), tools, "ClosedSpline.tool"
    );

    Bitmap* bm = new Bitmap(iv_bits, iv_width, iv_height);
    UStencil* fstencil = new UStencil(bm, nil, stdgraphic);
    StencilComp* fstencilComp = new StencilComp(fstencil);
    Include(
        new GraphicCompTool(
            new ControlInfo(fstencilComp), fstencilComp
        ), tools, "FStencil.tool"
    );
    UStencil* sstencil = new UStencil(bm, bm, stdgraphic);
    StencilComp* sstencilComp = new StencilComp(sstencil);
    Include(
        new GraphicCompTool(
            new ControlInfo(sstencilComp), sstencilComp
        ), tools, "SStencil.tool"
    );

    RasterRect* raster = new RasterRect(CreateRaster(), stdgraphic);
    RasterComp* rasterComp = new RasterComp(raster);
    Include(
        new GraphicCompTool(
            new ControlInfo(rasterComp), rasterComp
        ), tools, "RasterRect.tool"
    );

    Include(
        new GraphicCompTool(
            new ControlInfo("Pin", "*", "*"),
            new PinComp(new PinGraphic(0, 0, stdgraphic))
        ),
        tools, "Pin.tool"
    );
    Include(
        new GraphicCompTool(
            new ControlInfo("HSlot", ")", ")"),
            new HSlotComp(new SlotGraphic(0, 0, 10, stdgraphic))
        ),
        tools, "HSlot.tool"
    );
    Include(
        new GraphicCompTool(
            new ControlInfo("VSlot", "(", "("),
            new VSlotComp(new SlotGraphic(0, 0, 10, stdgraphic))
        ),
        tools, "VSlot.tool"
    );
    Include(
        new GraphicCompTool(
            new ControlInfo("Pad", "&", "&"),
            new PadComp(new PadGraphic(0, 0, 10, 10, stdgraphic))
        ),
        tools, "Pad.tool"
    );
    Include(
        new GraphicCompTool(
            new ControlInfo("Link", "^", "^"),
            new LinkComp(new Line(0, 0, unit, unit, stdgraphic))
        ),
        tools, "Link.tool"
    );
    Include(
        new ConnectTool(new ControlInfo("Connect", KLBL_CONNECT,CODE_CONNECT)),
        tools, "Connect.tool"
    );
    return tools;
}

Interactor* DrawingEditor::ReadTools () {
    VBox* tools = new VBox;

    for (int i = 0; i < _edInfo->Count(); ++i) {
        const char* toolName = _edInfo->GetName(i);
        const char* info = _edInfo->GetInfo(i);

        if (*info == '\0') {
            Tool* tool;
            unidraw->GetCatalog()->Retrieve(toolName, tool);
            Include(tool->Copy(), tools, toolName);
        }
    }
    return tools;
}

static const float PAGE_WIDTH = 8.5;
static const float PAGE_HEIGHT = 11;
static const float DEFAULT_XINCR = 8;
static const float DEFAULT_YINCR = 8;

void DrawingEditor::InitViewer () {
    const float w = PAGE_WIDTH * inches;
    const float h = PAGE_HEIGHT * inches;

    GraphicView* view = (GraphicView*) _comp->Create(COMPONENT_VIEW);
    _comp->Attach(view);
    view->Update();

    UPage* page = new UPage(w, h);
    Grid* grid = new Grid(w, h, DEFAULT_XINCR, DEFAULT_YINCR);
    grid->Visibility(false);

    _viewer = new Viewer(this, view, page, grid);
}

void DrawingEditor::Include (
    Command* cmd, PulldownMenu* pdm, const char* cmdName, const char* menuName
) {
    ControlInfo* ctrlInfo = cmd->GetControlInfo();
    UControl* ctrl = new CommandControl(ctrlInfo);
    _keymap->Register(ctrl);
    pdm->Include(ctrl);
    cmd->SetEditor(this);

    if (_initEdInfo) {
        _edInfo->Register(cmdName, menuName);
        unidraw->GetCatalog()->Save(cmd, cmdName);
    }
}

void DrawingEditor::Separator (PulldownMenu* pdm, const char* menuName) {
    pdm->GetScene()->Insert(
        new VBox(
            new VGlue(2, 0, 0),
            new HBorder,
            new VGlue(2, 0, 0)
        )
    );
    if (menuName != nil && _initEdInfo) {
        _edInfo->Register("-", menuName);
    }
}

void DrawingEditor::Include (Tool* tool, Box* box, const char* toolName) {
    ControlInfo* ctrlInfo = tool->GetControlInfo();
    UControl* ctrl = new VPanelControl(ctrlInfo, _curCtrl);
    _keymap->Register(ctrl);
    box->Insert(ctrl);

    if (_initEdInfo) {
        _edInfo->Register(toolName);
        unidraw->GetCatalog()->Save(tool, toolName);
    }
}

PulldownMenu* DrawingEditor::FileMenu () {
    const char* menuName = "File";
    PulldownMenu* pdm = MakePulldown(menuName);

    Include(
        new NewCompCmd(
            new ControlInfo("New", KLBL_NEWCOMP, CODE_NEWCOMP),
            new GraphicComps
        ),
        pdm, "NewComp.cmd", menuName
    );
    Include(
        new RevertCmd(new ControlInfo("Revert", KLBL_REVERT, CODE_REVERT)),
        pdm, "Revert.cmd", menuName
    );
    Separator(pdm, menuName);
    Include(
        new ViewCompCmd(
            new ControlInfo("Open...", KLBL_VIEWCOMP, CODE_VIEWCOMP)
        ), pdm, "ViewComp.cmd", menuName
    );
    Include(
        new SaveCompCmd(new ControlInfo("Save", KLBL_SAVECOMP, CODE_SAVECOMP)),
        pdm, "SaveComp.cmd", menuName
    );
    Include(
        new SaveCompAsCmd(
            new ControlInfo("Save As...", KLBL_SAVECOMPAS, CODE_SAVECOMPAS)
        ), pdm, "SaveCompAs.cmd", menuName
    );
    Include(
        new PrintCmd(
            new ControlInfo("Print...", KLBL_PRINT, CODE_PRINT)
        ), pdm, "Print.cmd", menuName
    );
    Separator(pdm, menuName);
    Include(
        new QuitCmd(
            new ControlInfo("Quit", KLBL_QUIT, CODE_QUIT)
        ), pdm, "Quit.cmd", menuName
    );

    return pdm;
}

PulldownMenu* DrawingEditor::EditMenu () {
    const char* menuName = "Edit";
    PulldownMenu* pdm = MakePulldown(menuName);

    Include(
        new UndoCmd(
            new ControlInfo("Undo", KLBL_UNDO, CODE_UNDO)
        ), pdm, "Undo.cmd", menuName
    );
    Include(
        new RedoCmd(
            new ControlInfo("Redo", KLBL_REDO, CODE_REDO)
        ), pdm, "Redo.cmd", menuName
    );
    Separator(pdm, menuName);
    Include(
        new CutCmd(
            new ControlInfo("Cut", KLBL_CUT, CODE_CUT)
        ), pdm, "Cut.cmd", menuName
    );
    Include(
        new CopyCmd(
            new ControlInfo("Copy", KLBL_COPY, CODE_COPY)
        ), pdm, "Copy.cmd", menuName
    );
    Include(
        new PasteCmd(
            new ControlInfo("Paste", KLBL_PASTE,CODE_PASTE)
        ), pdm, "Paste.cmd", menuName
    );
    Include(
        new DupCmd(
            new ControlInfo("Duplicate", KLBL_DUP, CODE_DUP)
        ), pdm, "Duplicate.cmd", menuName
    );
    Include(
        new DeleteCmd(
            new ControlInfo("Delete", KLBL_DEL, CODE_DEL)
        ), pdm, "Delete.cmd", menuName
    );
    Include(
        new SlctAllCmd(
            new ControlInfo("Select All", KLBL_SLCTALL, CODE_SLCTALL)
        ), pdm, "SelectAll.cmd", menuName
    );
    Separator(pdm, menuName);
    Include(
        new ScaleCmd(
            new ControlInfo("Flip Horizontal", KLBL_HFLIP, CODE_HFLIP),
            -1.0, 1.0
        ), pdm, "HFlip.cmd", menuName
    );
    Include(
        new ScaleCmd(
            new ControlInfo("Flip Vertical", KLBL_VFLIP, CODE_VFLIP),
            1.0, -1.0
        ), pdm, "VFlip.cmd", menuName
    );
    Include(
        new RotateCmd(
            new ControlInfo("90 Clockwise", KLBL_CW90, CODE_CW90), -90.0
        ), pdm, "90CW.cmd", menuName
    );
    Include(
        new RotateCmd(
            new ControlInfo("90 CounterCW", KLBL_CCW90, CODE_CCW90), 90.0
        ), pdm, "90CCW.cmd", menuName
    );
    Separator(pdm, menuName);
    Include(
        new MobilityCmd(
            new ControlInfo("Fix"), Fixed
        ), pdm, "Fix.cmd", menuName
    );
    Include(
        new MobilityCmd(
            new ControlInfo("Float"), Floating
        ), pdm, "Float.cmd", menuName
    );

    return pdm;
}

PulldownMenu* DrawingEditor::StructureMenu () {
    const char* menuName = "Structure";
    PulldownMenu* pdm = MakePulldown(menuName);

    Include(
        new GroupCmd(
            new ControlInfo("Group", KLBL_GROUP, CODE_GROUP)
        ), pdm, "Group.cmd", menuName
    );
    Include(
        new UngroupCmd(
            new ControlInfo("Ungroup", KLBL_UNGROUP, CODE_UNGROUP)
        ), pdm, "Ungroup.cmd", menuName
    );
    Include(
        new FrontCmd(
            new ControlInfo("Bring to Front", KLBL_FRONT, CODE_FRONT)
        ), pdm, "Front.cmd", menuName
    );
    Include(
        new BackCmd(
            new ControlInfo("Send to Back", KLBL_BACK, CODE_BACK)
        ), pdm, "Back.cmd", menuName
    );

    return pdm;
}

PulldownMenu* DrawingEditor::FontMenu () {
    Catalog* catalog = unidraw->GetCatalog();
    char buf[CHARBUFSIZE];
    const char* menuName = "Font";
    PulldownMenu* pdm = MakePulldown(menuName);

    int i = 1;
    PSFont* font = catalog->ReadFont(fontAttrib, i);

    while (font != nil) {
        TextGraphic* text = new TextGraphic(
            font->GetPrintFontAndSize(), stdgraphic
        );
        text->SetFont(font);
        sprintf(buf, "%s%d%s", "Font", i, ".cmd");

        Include(
            new FontCmd(
                new ControlInfo(new TextComp(text)), font
            ), pdm, buf, menuName
        );
        font = catalog->ReadFont(fontAttrib, ++i);
    }

    return pdm;
}

static float MENU_WIDTH = 1;   /* in cm */
static float MENU_HEIGHT = 0.5;

PulldownMenu* DrawingEditor::BrushMenu () {
    Catalog* catalog = unidraw->GetCatalog();
    char buf[CHARBUFSIZE];
    const char* menuName = "Brush";
    PulldownMenu* pdm = MakePulldown(menuName);

    int i = 1;
    PSBrush* br = catalog->ReadBrush(brAttrib, i);

    while (br != nil) {
        ControlInfo* ctrlInfo;

        if (br->None()) {
            ctrlInfo = new ControlInfo("None");

        } else {
            Line* line = new Line(0, 0, round(MENU_WIDTH*cm), 0, stdgraphic);
            line->SetBrush(br);
            ctrlInfo = new ControlInfo(new LineComp(line));
        }
        sprintf(buf, "%s%d%s", "Brush", i, ".cmd");
        
        Include(new BrushCmd(ctrlInfo, br), pdm, buf, menuName);
        br = catalog->ReadBrush(brAttrib, ++i);
    }

    return pdm;
}

PulldownMenu* DrawingEditor::PatternMenu () {
    Catalog* catalog = unidraw->GetCatalog();
    char buf[CHARBUFSIZE];
    const char* menuName = "Pattern";
    PulldownMenu* pdm = MakePulldown(menuName);

    int i = 1;
    PSPattern* pat = catalog->ReadPattern(patAttrib, i);
    
    while (pat != nil) {
        ControlInfo* ctrlInfo;

        if (pat->None()) {
            ctrlInfo = new ControlInfo("None");

        } else {
            SF_Rect* sfr = new SF_Rect(
                0, 0, round(MENU_WIDTH*cm), round(MENU_HEIGHT*cm), stdgraphic
            );
            sfr->SetPattern(pat);
            ctrlInfo = new ControlInfo(new RectComp(sfr));
        }
        sprintf(buf, "%s%d%s", "Pattern", i, ".cmd");
        
        Include(new PatternCmd(ctrlInfo, pat), pdm, buf, menuName);
        pat = catalog->ReadPattern(patAttrib, ++i);
    }

    return pdm;
}

PulldownMenu* DrawingEditor::ColorMenu (const char* name, const char* attrib) {
    Catalog* catalog = unidraw->GetCatalog();
    char buf[CHARBUFSIZE];
    PulldownMenu* pdm = MakePulldown(name);

    int i = 1;
    PSColor* color = catalog->ReadColor(attrib, i);
    
    while (color != nil) {
        ControlInfo* ctrlInfo;
        Coord w = round(MENU_WIDTH*cm);
        Coord h = round(MENU_HEIGHT*cm);

        SF_Rect* sfr = new SF_Rect(0, 0, w, h, stdgraphic);
	sfr->SetColors(color, color);
	ctrlInfo = new ControlInfo(new RectComp(sfr), color->GetName());
	sprintf(buf, "%s%d%s", name, i, ".cmd");

        if (strcmp(attrib, fgAttrib) == 0) {
            Include(new ColorCmd(ctrlInfo, color, nil), pdm, buf, name);
        } else {
            Include(new ColorCmd(ctrlInfo, nil, color), pdm, buf, name);
        }
	color = catalog->ReadColor(attrib, ++i);
    }
    return pdm;
}

PulldownMenu* DrawingEditor::AlignMenu () {
    const char* menuName = "Align";
    PulldownMenu* pdm = MakePulldown(menuName);

    Include(
        new AlignCmd(
            new ControlInfo("Left Sides", KLBL_ALGNLEFT, CODE_ALGNLEFT), 
            Left, Left
        ), pdm, "AlignLeft.cmd", menuName
    );
    Include(
        new AlignCmd(
            new ControlInfo("Right Sides", KLBL_ALGNRIGHT, CODE_ALGNRIGHT),
            Right, Right
        ), pdm, "AlignRight.cmd", menuName
    );
    Include(
        new AlignCmd(
            new ControlInfo("Tops", KLBL_ALGNTOP, CODE_ALGNTOP), Top, Top
        ), pdm, "AlignTops.cmd", menuName
    );
    Include(
        new AlignCmd(
            new ControlInfo("Bottoms", KLBL_ALGNBOT, CODE_ALGNBOT),	
            Bottom, Bottom
        ), pdm, "AlignBots.cmd", menuName
    );
    Include(
        new AlignCmd(
            new ControlInfo("Horiz Centers", KLBL_ALGNHCTR, CODE_ALGNHCTR),
            HorizCenter, HorizCenter
        ), pdm, "AlignHCtrs.cmd", menuName
    );
    Include(
        new AlignCmd(
            new ControlInfo("Vert Centers", KLBL_ALGNVCTR, CODE_ALGNVCTR),
            VertCenter, VertCenter
        ), pdm, "AlignVCtrs.cmd", menuName
    );
    Include(
        new AlignCmd(
            new ControlInfo("Centers", KLBL_ALGNCTR, CODE_ALGNCTR),	
            Center, Center
        ), pdm, "AlignCtrs.cmd", menuName
    );
    Include(
        new AlignCmd(
            new ControlInfo("Abut Left", KLBL_ABUTLEFT, CODE_ABUTLEFT), 
            Left, Right
        ), pdm, "AbutLeft.cmd", menuName
    );
    Include(
        new AlignCmd(
            new ControlInfo("Abut Right", KLBL_ABUTRIGHT, CODE_ABUTRIGHT),
            Right, Left
        ), pdm, "AbutRight.cmd", menuName
    );
    Include(
        new AlignCmd(
            new ControlInfo("Abut Up", KLBL_ABUTUP, CODE_ABUTUP),
            Top, Bottom
        ), pdm, "AbutUp.cmd", menuName
    );
    Include(
        new AlignCmd(
            new ControlInfo("Abut Down", KLBL_ABUTDOWN, CODE_ABUTDOWN),
            Bottom, Top
        ), pdm, "AbutDown.cmd", menuName
    );
    Separator(pdm, menuName);
    Include(
        new AlignToGridCmd(
            new ControlInfo(
                "Align to Grid", KLBL_ALGNTOGRID, CODE_ALGNTOGRID
            )
        ), pdm, "AlignToGrid.cmd", menuName
    );

    return pdm;
}

PulldownMenu* DrawingEditor::ViewMenu () {
    const char* menuName = "View";
    PulldownMenu* pdm = MakePulldown(menuName);

    Include(
        new NewViewCmd(
            new ControlInfo("New View", KLBL_NEWVIEW, CODE_NEWVIEW)
        ), pdm, "NewView.cmd", menuName
    );
    Include(
        new CloseEditorCmd(
            new ControlInfo("Close View", KLBL_CLOSEEDITOR, CODE_CLOSEEDITOR)
        ), pdm, "CloseView.cmd", menuName
    );
    Separator(pdm, menuName);
    Include(
        new NormSizeCmd(
            new ControlInfo("Normal Size", KLBL_NORMSIZE, CODE_NORMSIZE)
        ), pdm, "NormalSize.cmd", menuName
    );
    Include(
        new RedToFitCmd(
            new ControlInfo("Reduce to Fit", KLBL_REDTOFIT, CODE_REDTOFIT)
        ), pdm, "RedToFit.cmd", menuName
    );
    Include(
        new CenterCmd(
            new ControlInfo("Center Page", KLBL_CENTER, CODE_CENTER)
        ), pdm, "CtrPage.cmd", menuName
    );
    Include(
        new GridCmd(
            new ControlInfo("Grid on/off", KLBL_GRID, CODE_GRID)
        ), pdm, "Grid.cmd", menuName
    );
    Include(
        new GridSpacingCmd(
            new ControlInfo("Grid Spacing...", KLBL_GRIDSPC, CODE_GRIDSPC)
        ), pdm, "GridSpc.cmd", menuName
    );
    Include(
        new GravityCmd(
            new ControlInfo("Gravity on/off", KLBL_GRAVITY, CODE_GRAVITY)
        ), pdm, "Gravity.cmd", menuName
    );
    Include(
        new OrientationCmd(
            new ControlInfo("Orientation", KLBL_ORIENTATION, CODE_ORIENTATION)
        ), pdm, "Orientation.cmd", menuName
    );

    return pdm;
}
