// This file is part of xplot.
// 
// Copyright (C) 1989 Rene' Seindal (seindal@diku.dk)
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.


#include "xplot.hh"
#include "util.hh"
#include "frame.hh"
#include "view.hh"
#include "canvas.hh"
#include "plot.hh"
#include "panel.hh"

#include <InterViews/world.h>
#include <InterViews/sensor.h>
#include <InterViews/panner.h>
#include <InterViews/interactor.h>
#include <InterViews/banner.h>
#include <InterViews/border.h>
#include <InterViews/box.h>
#include <InterViews/glue.h>

void PlotFrame::Init(PlotGraphSet *gset, const char *name, const char *file)
{
    SetClassName("PlotFrame");

    graphset = gset;
    graphset->Attach(this);
    
    PlotFrame::name = StringDup(name);
    PlotFrame::file = StringDup(file);
    
    char buf[1024];
    sprintf(buf, "%s: %s", name, file);
    SetName(StringDup(buf));

    Propagate(false);
    
    headline = new Banner((char *)version.PrettyVersion(),
			  (char *)file,
			  (char *)version.Date()
			  );
    
    pcanvas = new PlotCanvas(graphset);
    view_port = new PlotView(pcanvas);
    
    state = new ButtonState;
    
    PanelBox *panelbox = new PanelBox();
    
    panelbox->Insert(new ButtonReplot("Redraw", state, this));
    
    prev_button = new ButtonPrevious("Prev. page", state, this);
    panelbox->Insert(prev_button);
    next_button = new ButtonNext("Next page", state, this);
    panelbox->Insert(next_button);
    ManageButtons();
    
    panelbox->Insert(new ButtonNormalize("Normalize", state, this));
    panelbox->Insert(new ButtonMagnify("Magnify", state, this));
    panelbox->Insert(new ButtonOneToOne("One to one", state, this));
    
    panelbox->Insert(new ButtonNewView("New view", state, this));
    panelbox->Insert(new ButtonDeleteView("Delete", state, this));
    panelbox->Insert(new ButtonDeleteAll("Delete views", state, this));
    
    panelbox->Insert(new ButtonQuit("Quit", state, this));
    
    HBox *panner = new HBox(new Panner(view_port));
    
    HBox *panel;
    if (IsOn(GetAttribute("leftPanel")))
	panel = new HBox(new VBox(panelbox, new VGlue, new HBorder, panner),
			 new VBorder,
			 view_port);
    else
	panel = new HBox(view_port,
			 new VBorder,
			 new VBox(panelbox, new VGlue, new HBorder, panner));
    Insert(new VBox(headline, new HBorder, panel));

    if (GetGeometry() == nil) {
	const char *geometry = GetAttribute("geometry");
	if (geometry != nil)
	    SetGeometry(geometry);
    }
    
    delete input;
    input = noEvents;
    input->Reference();
}

PlotFrame::PlotFrame(PlotGraphSet *gset, const char *name, const char *file)
{
    Init(gset, name, file);
}

PlotFrame::PlotFrame(PlotFrame *p, const char *geom)
{
    SetGeometry(geom == nil ? p->GetGeometry() : geom);
    SetName(p->GetName());
    Init(p->graphset, p->name, p->file);
}


PlotFrame::~PlotFrame()
{
    graphset->Detach(this);

    delete state;
    delete file;
    delete name;
}

void PlotFrame::ManageButtons()
{
    if (graphset->Next() == nil)
	next_button->Disable();
    else
	next_button->Enable();

    if (graphset->Previous() == nil)
	prev_button->Disable();
    else
	prev_button->Enable();
}

void PlotFrame::Update()
{
    ManageButtons();
    Replot();
}

void PlotFrame::Replot()
{
    view_port->Draw();
}

void PlotFrame::Previous()
{
    graphset->Backward();
}

void PlotFrame::Next()
{
    graphset->Forward();
}

void PlotFrame::NewView()
{
    PlotFrame *p;

    if (GetAttribute("copyGeometry") != nil) {
	char buf[256];
	sprintf(buf, "%dx%d", xmax+1, ymax+1);
	p = new PlotFrame(this, buf);
    } else
	p = new PlotFrame(this);

    world->InsertToplevel(p, nil);
}

void PlotFrame::DeleteView()
{
    if (graphset->NViews() == 1)
	DeleteAll();
    else
	delete this;
}

void PlotFrame::DeleteAll()
{
    int nsets = graphset->NSets();

    delete graphset;
    if (nsets == 1)
	Quit();
}

void PlotFrame::Quit()
{
    exit(0);
}

void PlotFrame::Normalize()
{
    view_port->Normalize();
}

void PlotFrame::Magnify()
{
    view_port->Magnify(pcanvas);
}

void PlotFrame::OneToOne()
{
    view_port->OneToOne();
}
















































