/*  GlyphPort: A Glyph-based Viewport.
 *
 * $Id$
Copyright 1991, Brown University, Providence, RI

Permission to use and modify this software and its documentation for any
purpose other than its incorporation into a commercial product is hereby
granted without fee.  Permission to copy and distribute this software and its
documentation for non-commercial use only is also granted without fee,
provided, however, that the above copyright notice appear in all copies, that
both that copyright notice and this permission notice appear in supporting
documentation, that the name of Brown University not be used in advertising or
publicity pertaining to distribution of the software without specific, written
prior permission, and that Brown University be notified by the distributor when
such distributions are done outside of his or her organization. Brown
University makes no representations about the suitability of this software for
any purpose.  It is provided "as is" without express or implied warranty.
Brown University requests notification of any modifications to this software
or its documentation.

Please send the name, organization, and address (postal and electronic) of
recipients of software outside of the receiving organization to:

	Software Librarian
	Computer Science Department, Box 1910
	Brown University
	Providence, RI 02912
		or
	brusd@cs.brown.edu

All electronic notifications will be acknowledged.
*/

#include "glyphport.h"

#include <InterViews/canvas.h>
#include <InterViews/printer.h>
#include <InterViews/background.h>
#include <InterViews/color.h>
#include <InterViews/transformer.h>
#include <InterViews/world.h>
#include <InterViews/sensor.h>
#include <InterViews/hit.h>

GlyphPort::GlyphPort(Glyph* g) : MonoGlyph(g), Adjustable()

{
  horizontal = new Requirement(4 * inch, fil, 3 * inch, 0.0);
  vertical = new Requirement(4 * inch, fil, 3 * inch, 0.0);
  Init();
}

GlyphPort::GlyphPort(Glyph* g, 
		     Coord x, Coord y) : MonoGlyph(g), Adjustable()

{
  horizontal = new Requirement(x, fil, 0, 0.0);
  vertical = new Requirement(y, fil, 0, 0.0);
  Init();
}

void GlyphPort::Init()

{
  scale = 1.0;
  transX = transY = 0;
  xScrollStep = yScrollStep = 5;
  difference = new Transformer();
  difference->ref();
}

GlyphPort::~GlyphPort()

{
  difference->unref();
}

void GlyphPort::request(Requisition& r)

{
  r.require(Dimension_X, *horizontal);
  r.require(Dimension_Y, *vertical);
  if (body() != nil)
    body()->request(interiorReq);
}

void GlyphPort::allocate(Canvas* c, Allocation& a, Extension& e)

{
  if (body() != nil)
    AllocateChild(c, a);
  e.xy_extents(a);
  alloc = a;
  canvas = c;
  xScrollStep = (a.allotment(Dimension_X).span() / 8);
  yScrollStep = (a.allotment(Dimension_Y).span() / 8);
}

void GlyphPort::ReTransform()

{
  if (transX < 0)
    transX = 0;
  else if (transX > interiorReq.requirement(Dimension_X).natural())
    transX = interiorReq.requirement(Dimension_X).natural();
  if (transY < 0)
    transY = 0;
  else if (transY > interiorReq.requirement(Dimension_Y).natural())
    transY = transY > interiorReq.requirement(Dimension_Y).natural();
  delete difference;
  difference = new Transformer();
  difference->translate((0 - transX), (0 - transY));
  difference->scale(scale, scale);
}

void GlyphPort::AllocateChild(Canvas* c, Allocation& a)

{
  body()->request(interiorReq);
  Allotment childHoriz(a.x(),interiorReq.requirement(Dimension_X).natural(),
		       0.0);
  Allotment childVert(a.y(), interiorReq.requirement(Dimension_Y).natural(),
		      0.0);
  interiorAlloc.allot(Dimension_X, childHoriz);
  interiorAlloc.allot(Dimension_Y, childVert);
  if (body() != nil)
    body()->allocate(c, interiorAlloc, interiorExt);
}

void GlyphPort::redraw()

{
  if (canvas != nil)
    {
      canvas->damage(alloc.left(), alloc.bottom(), alloc.right(), alloc.top());
      draw(canvas, alloc);
    }
}

void GlyphPort::draw(Canvas* c, Allocation& a)

{
  if (body() != nil)
    {
      c->push_clipping();
      c->clip_rect(a.left(), a.bottom(), a.right(), a.top());
      c->push_transform();
      c->transform(*difference);
      body()->draw(c, interiorAlloc);
      c->pop_clipping();
      c->pop_transform();
    }
}

void GlyphPort::print(Printer* p, Allocation&)

{
  if (body() != nil)
    {
      p->push_transform();
      p->transform(*difference);
      body()->print(p, interiorAlloc);
      p->pop_transform();
    }
}

boolean GlyphPort::pick(Canvas* c, Allocation&, Coord x, Coord y, 
			Hit& h, int depth)

{
  Coord x0 = x;
  Coord y0 = y;
  difference->transform(x0, y0);
  c->push_transform();
  c->transform(*difference);
  boolean retVal = body()->pick(c, interiorAlloc, x0, y0, h, depth);
  c->pop_transform();
  return retVal;
}

Allotment& GlyphPort::allotment(int, DimensionName dim)

{
  return alloc.allotment(dim);
}

Coord GlyphPort::lower(DimensionName dim)

{
  register Coord x = 0;
  register Coord y = 0;
  if (dim == Dimension_X)
    x = interiorAlloc.allotment(Dimension_X).begin();
  else
    y = interiorAlloc.allotment(Dimension_Y).begin();
  difference->transform(x, y);
  if (dim == Dimension_X)
    return x;
  else
    return y;
}

Coord GlyphPort::upper(DimensionName dim)

{
  register Coord x = 0;
  register Coord y = 0;
  if (dim == Dimension_X)
    x = interiorAlloc.allotment(Dimension_X).end();
  else
    y = interiorAlloc.allotment(Dimension_Y).end();
  difference->transform(x, y);
  if (dim == Dimension_X)
    return x;
  else
    return y;
}

Coord GlyphPort::length(DimensionName dim)

{
  register Coord x = 0;
  register Coord y = 0;
  if (dim == Dimension_X)
    x = interiorAlloc.allotment(Dimension_X).span();
  else
    y = interiorAlloc.allotment(Dimension_Y).span();
  difference->transform(x, y);
  if (dim == Dimension_X)
    return x;
  else
    return y;
}

Coord GlyphPort::cur_lower(DimensionName dim)

{
  return alloc.allotment(dim).begin();
}

Coord GlyphPort::cur_upper(DimensionName dim)

{
  return alloc.allotment(dim).end();
}

Coord GlyphPort::cur_length(DimensionName dim)

{
  return alloc.allotment(dim).span();
}

void GlyphPort::scroll_forward(DimensionName dim)

{
  if (dim == Dimension_X)
    transX += xScrollStep;
  else
    transY += yScrollStep;
  ReTransform();
  redraw();
  notify(dim);
}

void GlyphPort::scroll_backward(DimensionName dim)

{
  if (dim == Dimension_X)
    transX -= xScrollStep;
  else
    transY -= yScrollStep;
  ReTransform();
  redraw();
  notify(dim);
}

void GlyphPort::page_forward(DimensionName dim)

{
  if (dim == Dimension_X)
    transX += (alloc.allotment(dim).span() - xScrollStep);
  else
    transY += (alloc.allotment(dim).span() - yScrollStep);
  ReTransform();
  redraw();
  notify(dim);
}

void GlyphPort::page_backward(DimensionName dim)

{
  if (dim == Dimension_X)
    transX -= (alloc.allotment(dim).span() - xScrollStep);
  else
    transY -= (alloc.allotment(dim).span() - yScrollStep);
  ReTransform();
  redraw();
  notify(dim);
}

void GlyphPort::scroll_to(DimensionName dim, Coord loc)

{
  if (dim == Dimension_X)
    transX = loc - lower(dim) - (cur_length(dim) / 2);
  else
    transY = loc - lower(dim) - (cur_length(dim) / 2);
  ReTransform();
  redraw();
  notify(dim);
}

void GlyphPort::scale_to(DimensionName, float fraction_visible)

{
  zoom_to(fraction_visible);
}

void GlyphPort::zoom_to(float magnification)

{
  scale = magnification;
  ReTransform();
  redraw();
  notify(Dimension_X);
  notify(Dimension_Y);
}



