/*  $Id: hiergraph.C,v 1.3 91/04/13 16:45:12 gjb Exp $
 *									*
 *    HierGraph:							*
 *  This is the class that draws the graph.				*
Copyright 1991, Brown University, Providence, RI

Permission to use and modify this software and its documentation for any
purpose other than its incorporation into a commercial product is hereby
granted without fee.  Permission to copy and distribute this software and its
documentation for non-commercial use only is also granted without fee,
provided, however, that the above copyright notice appear in all copies, that
both that copyright notice and this permission notice appear in supporting
documentation, that the name of Brown University not be used in advertising or
publicity pertaining to distribution of the software without specific, written
prior permission, and that Brown University be notified by the distributor when
such distributions are done outside of his or her organization. Brown
University makes no representations about the suitability of this software for
any purpose.  It is provided "as is" without express or implied warranty.
Brown University requests notification of any modifications to this software
or its documentation.

Please send the name, organization, and address (postal and electronic) of
recipients of software outside of the receiving organization to:

	Software Librarian
	Computer Science Department, Box 1910
	Brown University
	Providence, RI 02912
		or
	brusd@cs.brown.edu

All electronic notifications will be acknowledged.

*/

#include "hiergraph.h"
#include <InterViews/geometry.h>
#include <InterViews/minmax.h>
#include <InterViews/world.h>
#include <InterViews/event.h>
#include <InterViews/sensor.h>
#include <InterViews/brush.h>
#include <InterViews/font.h>
#include <InterViews/cursor.h>
#include <InterViews/bitmap.h>
#include <InterViews/color.h>
#include <InterViews/action.h>
#include <InterViews/box.h>
#include <InterViews/glue.h>
#include <InterViews/canvas.h>
#include <InterViews/printer.h>
#include <InterViews/listener.h>
#include <InterViews/label.h>
#include <InterViews/margin.h>
#include <IV-look/choice.h>
#include <IV-look/bevel.h>
#include <IV-look/telltale.h>
#include <Dispatch/dispatcher.h>
#include <Dispatch/iocallback.h>
declare(IOCallback,HierNode);
#include "glyphport.h"
#include "simplebutton.h"
#include <string.h>
#include <stdlib.h>

static const int CLICK_WAIT = 50000;

/*  Constructor */

HierGraph::HierGraph(int* s,			// state of display's button
		     int num,			// how many nodes in graph
		     HierNode** theNodes,	// the nodes themselves
		     int numR,		        // number of root nodes
		     HierNode** roots,		// the root nodes
		     boolean horiz)		// To make horizontal graph
     : Page(nil)

{
  state = s;
  displayerIndex = nil;
  edgeList = nil;
  autoUpdating = true;
  needsUpdating = false;
  showInteresting = false;
  horizontal = horiz;
  numNodes = num;
  allNodes = theNodes;
  numRoots = numR;
  rootNodes = roots;
  margins = 0;
  for (int inc = 0; inc < MAXLEVELS; inc++)
    this->rows[inc].allocSize = 0;
  singletons.allocSize = 0;
  nodeWidth = World::current()->font()->width("Bill Atkinson's mom", 19) + H_PAD;
  MakeDisplay();
  FindShape();
}

HierGraph::~HierGraph()

{
  EdgeNode* curEdge = edgeList;
  while (curEdge != nil)
    {
      delete curEdge;
      curEdge = curEdge->next;
    }
  IndexNode* curIndex = displayerIndex;
  while (curIndex != nil)
    {
      delete curIndex;
      curIndex = curIndex->next;
    }
}

/*  CleanUp: Remove data, but don't deconstruct the entire thing.	*/

void HierGraph::CleanUp()

{
  EdgeNode* curEdge = edgeList;
  EdgeNode* deadEdge;

  while (curEdge != nil)
    {
      deadEdge = curEdge;
      curEdge = curEdge->next;
      delete deadEdge;    
    }
  edgeList = nil;
  lastEdge = nil;
  for (int curLevel = 0; curLevel < MAXLEVELS; curLevel++)
    rows[curLevel].levelSize = 0;
  singletons.levelSize = 0;
  numLevels = 0;
  IndexNode* curIndex = displayerIndex;
  IndexNode* deadIndex;
  while (curIndex != nil)
    {
      deadIndex = curIndex;
      curIndex = curIndex->next;
      delete deadIndex;
    }
  displayerIndex = nil;
  maxRightPos = 0;
  numLevels = 0;
}

/*  ReGraph:							*
 *    Kill the old data and replace it with new stuff.		*/

void HierGraph::ReGraph(int num, HierNode** theNodes, int numR, HierNode** roots)

{
  CleanUp();
  numNodes = num;
  allNodes = theNodes;
  numRoots = numR;
  rootNodes = roots;
  margins = 0;
  for (int inc = 0; inc < count(); inc++)
    remove(inc);
  MakeDisplay();
}

/*  MakeDisplay:						*
 *    Make the entire display from scratch.			*/

void HierGraph::MakeDisplay()

{
  SetUp();
  for (int inc = 0; inc < numRoots; inc++)		// Level assignment
    {
      if (rootNodes[inc]->GetSuccessors()->Size() == 0)
	AddToLevel(&singletons, rootNodes[inc]);
      else
	MakeLists(0, rootNodes[inc], true);
    }
  MakeFakeNodes();					// Fake node construction
  CleanDisplay();					// Make the display
  DrawDisplay();					// Place nodes
}

/*  SetUp:							*
 *    Create index of all type objects, and initialize the	*
 *  edge list.							*/

void HierGraph::SetUp()

{
  int increment;
  int j;

  displayerIndex = new IndexNode;
  displayerIndex->type = nil;
  displayerIndex->typeDisp = nil;
  displayerIndex->next = nil;
  for (increment = 0; increment < MAXLEVELS; increment++)
    {
      this->rows[increment].levelSize = 0;
      if (this->rows[increment].allocSize == 0)
	{
	  this->rows[increment].allocSize = LEVEL_SIZE;
	  this->rows[increment].typeDisps = 
	    (HierNode**)malloc(sizeof(HierNode*) * 
			       this->rows[increment].allocSize);
	}
      for (j = 0; j < this->rows[increment].allocSize; j++)
	this->rows[increment].typeDisps[j] = nil;
    }
  singletons.levelSize = 0;	
  if (singletons.allocSize == 0)
    {
      singletons.allocSize = LEVEL_SIZE;
      singletons.typeDisps = (HierNode**)malloc(sizeof(HierNode*) * 
						singletons.allocSize);
    }
  for (j = 0; j < singletons.allocSize; j++)
    singletons.typeDisps[j] = nil;
  IndexNode* curIndex;
  IndexNode* newIndex;
  curIndex = displayerIndex;
  increment = 0;
  while (increment < numNodes)
    {
      newIndex = new IndexNode;
      newIndex->type = allNodes[increment]->GetType();
      newIndex->typeDisp = allNodes[increment];
      newIndex->next = nil;
      curIndex->next = newIndex;
      curIndex = newIndex;
      increment++;
    }
  EdgeNode* firstNode;
  firstNode = new EdgeNode;
  firstNode->parent = firstNode->child = nil;
  firstNode->next = nil;
  edgeList = firstNode;
  lastEdge = firstNode;
}

void HierGraph::AddToLevel(Level* whatLevel, HierNode* whatNode)

{
  if (whatLevel->levelSize == (whatLevel->allocSize - 1))
    {
      whatLevel->allocSize += LEVEL_SIZE;
      whatLevel->typeDisps = 
	(HierNode**)realloc((char*)whatLevel->typeDisps, 
			    sizeof(HierNode*) *  whatLevel->allocSize);
      for (int j = whatLevel->levelSize; j < whatLevel->allocSize; j++)
	whatLevel->typeDisps[j] = nil;
    }
  whatLevel->typeDisps[whatLevel->levelSize] = whatNode;
  whatLevel->levelSize++;
}
  

/*  MakeLists:							*
 *    Calls itself recursively to make the edge list and the	*
 *  list of types on each level. Assigns types to levels by	*
 *  waiting until all each type's parents are assigned a level	*
 *  and then putting the type one level deeper than the deepest	*
 *  parent.							*/

void HierGraph::MakeLists(int level,			// what level
			  HierNode* thisDisplay,	// what to put there
			  boolean goOn)			// whether to continue

{
  EdgeNode* newEdge;			// New node to add to edge list
  HierNode* childDisp;			// Child of the type we're dealing with
  AdjList* parents;			// Parents of the type
  int newLevel = -1;			// New level to put type on
  int parentLevel;			// Level of lowest parent
  char* typeName;			// name of this type
  boolean dontAdd = false;		// Don't add the type to the list just yet

  parents = thisDisplay->GetPredecessors();	
  parents->First();
  if (parents->Size() <= 1)
    newLevel = level;
  else
    for (int i = 1; i <= parents->Size(); i++)
      {
	parentLevel = parents->GetCur()->GetDisp()->GetLevel();
	if (parentLevel < 0)	// If parents don't have a level, don't assign one
	  newLevel = WAIT;
	if ((newLevel != WAIT) && (parentLevel >= newLevel))
	  newLevel = parentLevel + 1;
	parents->Next();
      }
  typeName = thisDisplay->getName();
  if ((showInteresting) && (!thisDisplay->IsInteresting()))
    dontAdd = true; 
  if ((newLevel != WAIT) && (!dontAdd))
    {
      if (newLevel >= 0)
	{
	  thisDisplay->SetLevel(newLevel);
	  AddToLevel(&(this->rows[newLevel]), thisDisplay);
	}
      if (thisDisplay->GetSuccessors()->Size() > 0)
	{
	  int inc;
	  AdjList* subtypes = thisDisplay->GetSuccessors();
	  subtypes->First();
	  for (inc = 0; inc < subtypes->Size(); inc++)
	    {
	      childDisp = subtypes->GetCur()->GetDisp();
	      if (childDisp != nil)
		{
		  newEdge = new EdgeNode;
		  newEdge->parent = thisDisplay;
		  newEdge->child = childDisp;
		  newEdge->next = nil;
		  lastEdge->next = newEdge;
		  lastEdge = newEdge;
		  if (goOn)
		    MakeLists(level + 1, childDisp, true);	// Recursive
		}
	      subtypes->Next();
	    }
	}
    }
}

/*  MakeFakeNodes:						*
 *    If any edges extend through more than one level, add	*
 *  fake nodes in between to make the display look nicer.	*/

void HierGraph::MakeFakeNodes()

{
  EdgeNode* curEdge = edgeList;			// Current edge
  EdgeNode* newEdge;				// New edge to add for fake node
  int curLevel;					// Current level
  AdjList* parents;				// Parents of current type
  HierNode* curParent;				// Current parent

  while (curEdge != nil)
    {
      if ((curEdge->child != nil) && (curEdge->parent != nil) && 
	  (curEdge->child->GetLevel() < 0) && 
	  (curEdge->parent->GetLevel() >= 0))
	{
	  curLevel = curEdge->parent->GetLevel();
	  parents = curEdge->child->GetPredecessors();
	  parents->First();
	  for (int i = 0; i < parents->Size(); i++)
	    {
	      curParent = parents->GetCur()->GetDisp();
	      if (curParent->GetLevel() < 0)
		{
		  MakeLists(curLevel - 1, curParent, false);	// Not recursive
		  newEdge = new EdgeNode;			// Add to edge list
		  newEdge->parent = FindDisp(curParent->GetType());
		  newEdge->child = curEdge->child;
		  newEdge->next = lastEdge->next;
		  lastEdge->next = newEdge;
		  lastEdge = newEdge;
		}
	      parents->Next();
	    }
	  MakeLists(curLevel, curEdge->child, false);
	}
      curEdge = curEdge->next;
    }
  int increment;
  HierNode* fakeDisplayer;
  HierNode* parentNode;
  HierNode* childNode;
  HierNode* lastNode;
  int difference;
  EdgeNode* firstNode = edgeList;
  EdgeNode* deadEdge;
  EdgeNode* oldLastEdge = lastEdge;
  curEdge = firstNode->next;
  while ((curEdge != nil) && (curEdge != oldLastEdge->next))
    {
      parentNode = curEdge->parent;
      childNode = curEdge->child;
      if ((parentNode == nil) || (childNode == nil))
	difference = 1;
      else
	difference = childNode->GetLevel() - parentNode->GetLevel();
      if (difference > 1)
	{
	  parentNode->RemoveSuccessor(childNode);
	  childNode->RemovePredecessor(parentNode);
	  deadEdge = curEdge;
	  for (curEdge = firstNode; curEdge->next != deadEdge; 
	       curEdge = curEdge->next);
	  curEdge->next = deadEdge->next;
	  curEdge = deadEdge->next;
	  delete deadEdge;
	  lastNode = parentNode;
	  for (increment = parentNode->GetLevel() + 1; 
	       increment <= childNode->GetLevel() - 1; increment++)
	    {	
	      if (horizontal)
		fakeDisplayer = new FakeDisplayer(increment, nil, 
						  horizontal, FAKE_HEIGHT);
	      else
		fakeDisplayer = new FakeDisplayer(increment, nil, 
						  horizontal, nodeWidth);
	      AddToLevel(&this->rows[increment], fakeDisplayer);
	      lastNode->AddSuccessor(fakeDisplayer);
	      fakeDisplayer->AddPredecessor(lastNode);
	      newEdge = new EdgeNode;
	      newEdge->parent = lastNode;
	      newEdge->child = fakeDisplayer;
	      newEdge->next = lastEdge->next;
	      lastEdge->next = newEdge;
	      lastEdge = newEdge;
	      lastNode = fakeDisplayer;
	    }
	  fakeDisplayer->AddSuccessor(childNode);
	  lastNode->AddPredecessor(fakeDisplayer);
	  newEdge = new EdgeNode;
	  newEdge->parent = lastNode;
	  newEdge->child = childNode;
	  newEdge->next = lastEdge->next;
	  lastEdge->next = newEdge;
	  lastEdge = newEdge;
	}
      else
	curEdge = curEdge->next;
    }
}

/*  CleanDisplay:							*
 *    This is the graph layout algorithm itself. This moves the type	*
 *  displayer objects back and forth to try and minimize edge		*
 *  crossings and generally make things look nice.			*/

void HierGraph::CleanDisplay()

{
  HierNode** curNodes;
  int curLevel = 0;
  int curNode;
  int pass;
  int inc;
  boolean goingDown = false;
  int levelCount = 0;

  while ((this->rows[curLevel].levelSize > 0) && (curLevel < MAXLEVELS))
    curLevel++;
  numLevels = curLevel;
  this->rows[numLevels++] = singletons;
  maxRightPos = 0;
  for (curLevel = 0; curLevel < numLevels; curLevel++)
    {
      curNodes = this->rows[curLevel].typeDisps;
      for (curNode = 0; curNode < this->rows[curLevel].levelSize; curNode++)
	curNodes[curNode]->SetPos(curNode);
      if (curNode > maxRightPos)
	maxRightPos = curNode;
    }
  maxRightPos += 5;
  margins = (maxRightPos / 4) + 1;
  maxRightPos += (margins * 2);
  for (pass = 0; pass < PASSES; pass++)			// Start Barycentre passes
    {
      if (goingDown)
	{
	  curLevel = numLevels - 1;
	  goingDown = false;
	}
      else
	{
	  curLevel = 0;
	  goingDown = true;
	}
      while ((curLevel >= 0) && (curLevel < numLevels))
	{
	  curNodes = this->rows[curLevel].typeDisps;
	  if (goingDown)
	    {
	      UpBarycentre(curNodes, this->rows[curLevel].levelSize);
	      UpSort(curNodes, this->rows[curLevel].levelSize);
	      for (curNode = 0; curNode < this->rows[curLevel].levelSize; 
		   curNode++)
		curNodes[curNode]->SetPos(curNode);
	      curLevel++;
	    }
	  else	
	    {
	      DownBarycentre(curNodes, this->rows[curLevel].levelSize);
	      DownSort(curNodes, this->rows[curLevel].levelSize);
	      for (curNode = 0; curNode < this->rows[curLevel].levelSize; 
		   curNode++)
		curNodes[curNode]->SetPos(curNode);
	      curLevel--;
	    }
	}
    }
  int prevPriority;
  int curPriority;
  int maxPriority;
  int difference;
  goingDown = false;
  for (curLevel = 0; curLevel < numLevels; curLevel++)
    {
      for (inc = 0; inc < this->rows[curLevel].levelSize; inc++)
	rows[curLevel].typeDisps[inc]->SetPos(inc + margins);
    }
  for (pass = 0; pass < PASSES; pass++)			// Start Barycentre passes
    {
      if (goingDown)
	{
	  curLevel = numLevels - 1;
	  goingDown = false;
	}
      else
	{
	  curLevel = 0;
	  goingDown = true;
	}
      while ((curLevel >= 0) && (curLevel < numLevels))
	{
	  curNodes = this->rows[curLevel].typeDisps;
	  UpBarycentre(curNodes, this->rows[curLevel].levelSize);
	  DownBarycentre(curNodes, this->rows[curLevel].levelSize);
	  prevPriority = FAKE_PRIORITY * 2;
	  inc = 0;
	  while (inc < this->rows[curLevel].levelSize)
	    {
	      maxPriority = 0;
	      for (curNode = 0; curNode < 
		   this->rows[curLevel].levelSize; curNode++)
		{
		  curPriority = curNodes[curNode]->GetConnectivity();
		  if ((curPriority > maxPriority) && (curPriority < prevPriority))
		    maxPriority = curNodes[curNode]->GetConnectivity();
		}
	      for (curNode = 0; curNode < 
		   this->rows[curLevel].levelSize; curNode++)
		{
		  if (curNodes[curNode]->GetConnectivity() == maxPriority)
		    {
		      if (curNodes[curNode]->GetPredecessors()->Size() == 0)
			difference = round(curNodes[curNode]->GetDownBary());
		      else if (curNodes[curNode]->GetSuccessors()->Size() == 0)
			difference = round(curNodes[curNode]->GetUpBary());
		      else
			difference = round((curNodes[curNode]->GetUpBary() + 
					    curNodes[curNode]->GetDownBary()) / 2);
		      difference -= curNodes[curNode]->GetPos();
		      if (difference != 0)
			MoveNode(curNodes, curNode, difference);
		      inc++;	      
		    }
		}
	      prevPriority = maxPriority;
	    }
	  if (goingDown)
	    curLevel++;
	  else
	    curLevel--;
	}
    }
}

void HierGraph::DownBarycentre(HierNode** level,	// Level to adjust
			       int numNodes)

{
  for (int inc = 0; inc < numNodes; inc++)
    level[inc]->ComputeDownBary();
}

void HierGraph::UpBarycentre(HierNode** level,
			     int numNodes)

{
  for (int inc = 0; inc < numNodes; inc++)
    level[inc]->ComputeUpBary();
}

int HierGraph::MoveNode(HierNode** level, int node, int delta)

{
  int possibleDelta;
  int difference;
  int myPos = level[node]->GetPos();
  if (((node == 0) && (delta < 0)) || 
      ((level[node + 1] == nil) && (delta > 0)))
    possibleDelta = delta;  /* At the beginning or the end */
  else
    {
      if (delta < 0)
	{
	  difference = level[node - 1]->GetPos() - myPos;
	  if (difference > (delta - 1))
	    {
	      possibleDelta = difference + 1;
	      if (level[node]->GetConnectivity() <= 
		  level[node - 1]->GetConnectivity())
		possibleDelta += MoveNode(level, node - 1, delta - difference - 1);
	    }
	  else
	    possibleDelta = delta;
	}
      else
	{
	  difference = level[node + 1]->GetPos() - myPos;
	  if (difference < (delta + 1))
	    {
	      possibleDelta = difference - 1;
	      if (level[node]->GetConnectivity() <= 
		  level[node + 1]->GetConnectivity())
		possibleDelta += MoveNode(level, node + 1, delta - difference + 1);
	    }
	  else
	    possibleDelta = delta;
	}
    }
  if ((myPos + possibleDelta) < 0)	// We can't go over the edge of the graph
    possibleDelta = 0 - myPos;
  else if ((myPos + possibleDelta) > maxRightPos)
    possibleDelta = maxRightPos - myPos;
  level[node]->SetPos(myPos + possibleDelta);
  return possibleDelta;
}	
	    

/*  DrawDisplay:						*
 *    Once the nodes have been laid out, place their actual	*
 *  canvasses in the window.					*/

void HierGraph::DrawDisplay()

{
  int curLevel;
  int curNode;
  HierNode** curNodes;
  Coord curX;
  Coord curY;
  Coord maxX;
  Coord topY;
  int curGlyph = 0;
  
  const Coord fudge = (nodeWidth / 2) + (2 * FUDGE);
  maxX = 0;
  if (horizontal)
    topY = top() - ROW_HEIGHT;
  else
    topY = FUDGE;		// Start at the bottom, you silly
  for (curLevel = 0; curLevel < numLevels; curLevel++)
    {
      curNodes = this->rows[curLevel].typeDisps;
      if (horizontal)
	{
	  curY = topY - (curLevel * ROW_HEIGHT) - 15;
	  for (curNode = 0; curNode < this->rows[curLevel].levelSize; curNode++)
	    {
	      curX = (curNodes[curNode]->GetPos() * nodeWidth) + fudge;
	      insert(curGlyph, curNodes[curNode]);
	      move(curGlyph, curX, curY);
	      show(curGlyph++, true);
	      if ((curX + nodeWidth) > maxX)
		maxX = curX + nodeWidth;
	    }
	}
      else
	{
	  curX = (curLevel * nodeWidth) + H_PAD;
	  for (curNode = 0; curNode < this->rows[curLevel].levelSize; curNode++)
	    {
	      curY = (curNodes[curNode]->GetPos() * (FAKE_HEIGHT + V_PAD)) 
		+ FAKE_HEIGHT;
	      insert(curGlyph, curNodes[curNode]);
	      move(curGlyph, curX, curY);
	      show(curGlyph++, true);
	      if ((curX + nodeWidth) > maxX)
		maxX = curX + nodeWidth;
	    }
	}
    }
}

HierNode* HierGraph::FindDisp(int what)

{
  IndexNode* cur = displayerIndex->next;
  while ((cur != nil) &&
	 (cur->typeDisp->GetType() != what))
    cur = cur->next;
  if (cur != nil)
    return cur->typeDisp;
  else
    return nil;
}

void HierGraph::FindShape()

{
  if (horizontal)
    {
      maxX = maxRightPos * (nodeWidth + H_PAD);
      maxY = ROW_HEIGHT * numLevels;
    }
  else
    {
      maxX = numLevels * (nodeWidth + H_PAD);
      maxY = maxRightPos * (FAKE_HEIGHT + V_PAD);
    }
}

void HierGraph::draw(Canvas* c, Allocation& a)

{
  Coord parentX, parentY, childX, childY;
  EdgeNode* curEdge = edgeList->next;
  HierNode* prevParent = nil;
  Color* fg = World::current()->foreground();
  Brush* brush = new Brush(1);

  Page::draw(c, a);
  while (curEdge != nil)
    {
      if ((curEdge->parent != nil) && (curEdge->child != nil))
	{
	  prevParent = curEdge->parent;
	  if (horizontal)
	    {
	      parentX = curEdge->parent->GetXCoord();
	      parentY = curEdge->parent->GetBotYCoord();
	      childX = curEdge->child->GetXCoord();
	      childY = curEdge->child->GetTopYCoord();
	    }	
	  else
	    {
	      parentX = curEdge->parent->GetRightXCoord();
	      parentY = curEdge->parent->GetYCoord();
	      childX = curEdge->child->GetLeftXCoord();
	      childY = curEdge->child->GetYCoord();
	    }
	  if ((parentX != -1) && (parentY != -1) && 
	      (childX != -1) && (childY != -1))
	    c->line(parentX, parentY, childX, childY, fg, brush);
	}
      curEdge = curEdge->next;
    }
}

void HierGraph::print(Printer* p, Allocation& a)

{
  Coord parentX, parentY, childX, childY;
  EdgeNode* curEdge = edgeList->next;
  HierNode* prevParent = nil;
  Color* fg = World::current()->foreground();
  Brush* brush = new Brush(1);

  Page::print(p, a);
  while (curEdge != nil)
    {
      if ((curEdge->parent != nil) && (curEdge->child != nil))
	{
	  prevParent = curEdge->parent;
	  if (horizontal)
	    {
	      parentX = curEdge->parent->GetXCoord();
	      parentY = curEdge->parent->GetBotYCoord();
	      childX = curEdge->child->GetXCoord();
	      childY = curEdge->child->GetTopYCoord();
	    }	
	  else
	    {
	      parentX = curEdge->parent->GetRightXCoord();
	      parentY = curEdge->parent->GetYCoord();
	      childX = curEdge->child->GetLeftXCoord();
	      childY = curEdge->child->GetYCoord();
	    }
	  if ((parentX != -1) && (parentY != -1) && 
	      (childX != -1) && (childY != -1))
	    p->line(parentX, parentY, childX, childY, fg, brush);
	}
      curEdge = curEdge->next;
    }
}

void HierGraph::request(Requisition& req)

{
  Requirement horiz(maxX, fil, 0, 0.0);
  Requirement vert(maxY, fil, 0, 0.0);
  req.require(Dimension_X, horiz);
  req.require(Dimension_Y, vert);
}

void HierGraph::allocate(Canvas* c, Allocation& a, Extension& e)

{
  Allotment xAllot;
  Allotment yAllot;
  for (int inc = 0; inc < count(); inc++)
    {
      xAllot = allotment(inc, Dimension_X);
      yAllot = allotment(inc, Dimension_Y);
      if (c->damaged(xAllot.begin(), yAllot.begin(), 
		     xAllot.end(), yAllot.end()))
	change(inc);
    }
  Page::allocate(c, a, e);
}

/***********

void HierGraph::Handle(Event& anEvent)

{
  if (anEvent.target != this)
    anEvent.target->Handle(anEvent);
  else if ((anEvent.eventType == DownEvent) && (anEvent.button == MIDDLEMOUSE))
    {
      Viewport* thePort = (Viewport*)Parent();
      Cursor* oldCursor = GetCursor();
      SetCursor(handCursor);
      float hierWidth = ((float)shape->width);
      float hierHeight = ((float)shape->height);
      Coord oldX = anEvent.x;
      Coord oldY = anEvent.y;
      float Xpos = thePort->XPos();
      float Ypos = thePort->YPos();
      do 
	{
	  Xpos += (oldX - anEvent.x) / hierWidth;
	  Ypos += (oldY - anEvent.y) / hierHeight;
	  thePort->ScrollTo(Xpos, Ypos);
	  oldX = anEvent.x;
	  oldY = anEvent.y;
	  Poll(anEvent);
	}
      while (anEvent.middlemouse);
      SetCursor(oldCursor);
    }
}

*********/

/*  NewRoot:						*
 *    Make a particular type the root of the hierarchy.	*/

void HierGraph::NewRoot(HierNode* t)

{
  for (int inc = 0; inc < count(); inc++)
    remove(inc);
  CleanUp();
  SetUp();
  MakeLists(0, t, true);
  MakeFakeNodes();
  CleanDisplay();
  DrawDisplay();
  FindShape();
}

/*  FullView:						*
 *    Lay out the entire hierarhcy.			*/

void HierGraph::FullView()

{
  for (int inc = 0; inc < count(); inc++)
    remove(inc);
  CleanUp();
  SetUp();
  for (inc = 0; inc < numRoots; inc++)
    {
      if (rootNodes[inc]->GetSuccessors()->Size() == 0)
	AddToLevel(&singletons, rootNodes[inc]);
      else
      	MakeLists(0, rootNodes[inc], true);
    }
  MakeFakeNodes();
  CleanDisplay();
  DrawDisplay();
  FindShape();
}

void HierGraph::SetAutoUpdate(boolean val)

{
  if (val == true)
    {
      autoUpdating = true;
      if (needsUpdating)
	{
	  FullView();
	  needsUpdating = false;
	}
    }
  else
    autoUpdating = false;
}

void HierGraph::SetShowInteresting(boolean val)

{
  if ((val == true) && (showInteresting == false))
    {
      showInteresting = true;
      FullView();
    }
  else if ((val == false) && (showInteresting == true))
    {
      showInteresting = false;
      FullView();
    }
}

void HierGraph::NewType(HierNode* newNode)

{
  allNodes[numNodes] = newNode;
  numNodes++;
  if (autoUpdating)
    {
      FullView();		// This should be smarter
      needsUpdating = false;
    }
  else
    needsUpdating = true;
}

void HierGraph::UpSort(HierNode** theList, int size)

{
  register int i;
  register int j;
  int min;
  HierNode* temp;

  i = 0;
  while (i < size - 1)
    {
       min = i;
       j = i + 1;
       while (theList[j] != nil)
	 {
	   if (theList[j]->GetUpBary() < theList[min]->GetUpBary())
	     min = j;
	   j++;
	 }
       temp = theList[min];
       theList[min] = theList[i];
       theList[i] = temp;
       i++;
     }
}

void HierGraph::DownSort(HierNode** theList, int size)

{
  register int i;
  register int j;
  int min;
  HierNode* temp;

  i = 0;
  while (i < size - 1)
    {
       min = i;
       j = i + 1;
       while (theList[j] != nil)
	 {
	   if (theList[j]->GetDownBary() < theList[min]->GetDownBary())
	     min = j;
	   j++;
	 }
       temp = theList[min];
       theList[min] = theList[i];
       theList[i] = temp;
       i++;
     }
}

HierNode::HierNode(char* name,
		   int* s,
		   Telltale** ref,
		   int t,
		   Action* toDo)
     : SimpleButton(nil, nil)

{
  World* theWorld = World::current();
  whatLevel = -1;
  whatPos = -1;
  myData = t;
  state = s;
  upBary = downBary = 0.0;
  predecessors = new AdjList();
  successors = new AdjList();
  myName = name;
  whatToDo = toDo;
  waitForClick = false;
  doubleCallback = new IOCallback(HierNode)(this, HierNode::SecondClick);
  pickedButton = ref;
  Allotment fake(-1, 0, 0.0);
  alloc.allot(Dimension_X, fake);
  alloc.allot(Dimension_Y, fake);
  Color* fg = theWorld->foreground();
  Color* bg = theWorld->background();
  Font* fn = nil;
/******* fuck it
  char fontName[255];
  bzero(fontName, 255);
  if (theWorld->property_value("hierarchyFont") != nil)
    strcpy(fontName, theWorld->property_value("*hierarchyFont"));
  else
    fn = theWorld->font();
  if ((fn == nil) && (fontName[0] == '\0'))
    fn = theWorld->font();
  else if (fn == nil)
    fn = new Font(fontName);
*******/
  fn = new Font("*-helvetica-bold-r-normal--10-*");
  Label* label = new Label(name, fn, fg);
  Margin* frame = new Margin(label, 2, 0, 2, 2, 0, 2, 2, 0, 2, 2, 0, 2);
  telltale_ = new ChoiceItem(
			     new Bevel(frame, new Outset(bg)),
			     new Bevel(frame, new Inset(bg))
			     );
  telltale_->ref();
  listener_->body(telltale_);
}

HierNode::~HierNode()

{
  delete predecessors;
  delete successors;
  if ((pickedButton != nil) && (*pickedButton == telltale_))
    *pickedButton = nil;
}


void HierNode::event(Event& e)

{
  listener_->sensor()->motion(true);
  do 
    {
      if (listener_->picks(e.pointer_x(), e.pointer_y())) 
	{
	  if (telltale_ != nil) 
	    telltale_->highlight(true);
	} 
      else 
	{
	  if (telltale_ != nil)
	    telltale_->highlight(false);
	}
      e.world()->read(e);
    } 
  while (e.type() != Event::up);
  if (telltale_ != nil) 
    telltale_->highlight(false);
  listener_->sensor()->motion(false);
  if (listener_->picks(e.pointer_x(), e.pointer_y())) 
    {
      if (whatToDo != nil) 
	{
	  if (waitForClick)
	    {
	      telltale_->highlight(true);
	      whatToDo->execute();
	      waitForClick = false;
	      Dispatcher::instance().stopTimer(doubleCallback);
	    }
	  else
	    {
	      if (state != nil)
		*state = myData;
	      if (*pickedButton != nil)
		(*pickedButton)->highlight(false);
	      *pickedButton = telltale_;
	      telltale_->highlight(true);
	      waitForClick = true;
	      Dispatcher::instance().startTimer(0, CLICK_WAIT, doubleCallback);
	    }
	}
    }
}

void HierNode::SetPick(Telltale* node)

{
  if (*pickedButton != nil)
    (*pickedButton)->highlight(false);
  *pickedButton = node;
  if (*pickedButton != nil)
    (*pickedButton)->highlight(true);
}

void HierNode::allocate(Canvas* c, Allocation& a, Extension& e)

{
  alloc = a;
  SimpleButton::allocate(c, a, e);
}

void HierNode::SecondClick(long, long)

{
  waitForClick = false;
  Dispatcher::instance().stopTimer(doubleCallback);
}

void HierNode::ClearLists()
     
{
  delete predecessors;
  delete successors;
  predecessors = new AdjList();
  successors = new AdjList();
}

float HierNode::ComputeDownBary()

{
  if (successors->Size() == 0)
    downBary = 0.0;
  else
    {
      HierNode* curNode = successors->First()->GetDisp();
      int totalPos = 0;
      for (int inc = 0; inc < successors->Size(); inc++)
	{
	  totalPos += successors->GetCur()->GetDisp()->GetPos();
	  successors->Next();
	}
      downBary = totalPos / successors->Size();
    }
  return downBary;
}

float HierNode::ComputeUpBary()

{
  if (predecessors->Size() == 0)
    upBary = 0.0;
  else
    {
      predecessors->First();
      int totalPos = 0;
      for (int inc = 0; inc < predecessors->Size(); inc++)
	{
	  totalPos += predecessors->GetCur()->GetDisp()->GetPos();
	  predecessors->Next();
	}
      upBary = totalPos / predecessors->Size();
    }
  return upBary;
}

int HierNode::GetConnectivity()

{
  return (predecessors->Size() + successors->Size());
}

Coord HierNode::GetXCoord()

{
  return ((alloc.right() - alloc.left()) / 2) + alloc.left();
}

Coord HierNode::GetTopYCoord()

{
  return alloc.top();
}

Coord HierNode::GetBotYCoord()

{
  return alloc.bottom();
}

Coord HierNode::GetYCoord()

{
  return ((alloc.top() - alloc.bottom()) / 2) + alloc.bottom();
}

Coord HierNode::GetLeftXCoord()

{
  return alloc.left();
}

Coord HierNode::GetRightXCoord()

{
  return alloc.right();
}

FakeDisplayer::FakeDisplayer(int level, int t, boolean horiz, Coord ext)
     : HierNode("X", nil, nil, t, nil)

{
  whatLevel = level;
  horizontal = horiz;
  extent = ext;
}

void FakeDisplayer::request(Requisition& r)

{
  if (horizontal)
    {
      Requirement width(1, 0, 0, 0.0);
      Requirement height(extent, 0, 0, 0.0);
      r.require(Dimension_X, width);
      r.require(Dimension_Y, height);
    }
  else
    {
      Requirement width(extent, 0, 0, 0.0);	
      Requirement height(1, 0, 0, 0.0);
      r.require(Dimension_X, width);
      r.require(Dimension_Y, height);
    }
}

void FakeDisplayer::allocate(Canvas*, Allocation& a, Extension &e)

{
  e.xy_extents(a);
  alloc = a;
}

void FakeDisplayer::draw(Canvas* c, Allocation& a)

{
  if (horizontal)
    c->line(a.left(), a.bottom(), a.left(), a.top(), 
	    World::current()->background(), new Brush(1));
  else
    c->line(a.left(), a.bottom(), a.right(), a.bottom(), 
	    World::current()->background(), new Brush(1));
}

void FakeDisplayer::print(Printer* p, Allocation& a)

{
  draw(p, a);
}

void FakeDisplayer::event(Event& anEvent)

{
}

AdjNode::AdjNode(HierNode* disp)
     
{
  displayer = disp;
  valuep = (void**)displayer;
}

void AdjList::Append(HierNode* newDisp)

{
  AdjNode* newNode = new AdjNode(newDisp);
  BaseList::Append(newNode);
}

void AdjList::Remove(HierNode* rmThis)

{
  First();
  while ((GetCur()->GetDisp() != rmThis) && (!AtEnd()))
    Next();
  if (GetCur()->GetDisp() == rmThis)
    DeleteCur();
}

implement(IOCallback,HierNode);
