/*  $Id: hiergraph.C,v 1.2 90/06/17 18:45:33 gjb Exp $
 *									*
 *    HierGraph:							*
 *  This is the class that draws the graph.				*
 *									
Copyright 1991, Brown University, Providence, RI

Permission to use and modify this software and its documentation for any
purpose other than its incorporation into a commercial product is hereby
granted without fee.  Permission to copy and distribute this software and its
documentation for non-commercial use only is also granted without fee,
provided, however, that the above copyright notice appear in all copies, that
both that copyright notice and this permission notice appear in supporting
documentation, that the name of Brown University not be used in advertising or
publicity pertaining to distribution of the software without specific, written
prior permission, and that Brown University be notified by the distributor when
such distributions are done outside of his or her organization. Brown
University makes no representations about the suitability of this software for
any purpose.  It is provided "as is" without express or implied warranty.
Brown University requests notification of any modifications to this software
or its documentation.

Please send the name, organization, and address (postal and electronic) of
recipients of software outside of the receiving organization to:

	Software Librarian
	Computer Science Department, Box 1910
	Brown University
	Providence, RI 02912
		or
	brusd@cs.brown.edu

All electronic notifications will be acknowledged.

*/ 

#include "hiergraph.h"
#include <InterViews/font.h>
#include <InterViews/cursor.h>
#include <InterViews/bitmap.h>
#include <InterViews/Bitmaps/hand.bm>
#include <InterViews/Bitmaps/handMask.bm>
#include <InterViews/color.h>
#include <InterViews/painter.h>
#include <InterViews/perspective.h>
#include "squarebutton.h"

/* Constants */

static const int H_PAD = 30;		// Horizontal padding (10)
static const int V_PAD = 25;		// Space between vertically-arranged nodes (5)
static const int PASSES = 6;		// Number of Barycentre passes to make
static const int ROW_HEIGHT = 75;	// The height of each row (55)
static const int WAIT = -99;		// Flag for the level of a type not to add yet (see MakeLists)
static const int FUDGE = 50;		// Fudge factor for drawing
static const int FAKE_HEIGHT = 18;
static const int LEVEL_SIZE = 16;	// Number to grow each level by with realloc

/*  Constructor */

HierGraph::HierGraph(ButtonState* s,		// state of display's button
		     int num,			// how many nodes in graph
		     HierNode** theNodes,	// the nodes themselves
		     int numR,		        // number of root nodes
		     HierNode** roots,		// the root nodes
		     boolean horiz)		// To make horizontal graph
     : Scene()

{
  SetClassName("HierGraph");
  state = s;
  displayerIndex = nil;
  edgeList = nil;
  autoUpdating = true;
  needsUpdating = false;
  showInteresting = false;
  horizontal = horiz;
  numNodes = num;
  allNodes = theNodes;
  numRoots = numR;
  rootNodes = roots;
  lastWidth = lastHeight = 0;
  margins = 0;
  scaleX = scaleY = 1.0;
  for (int inc = 0; inc < MAXLEVELS; inc++)
    this->rows[inc].allocSize = 0;
  singletons.allocSize = 0;
  MakeDisplay();
  SetClassName("HierGraph");
  nodeWidth = round(World::current()->font()->width("Bill Atkinson's mom", 19) + H_PAD);
}

HierGraph::~HierGraph()

{
  Level* curNode;
  EdgeNode* curEdge = edgeList;
  int inc;
  while (curEdge != nil)
    {
      delete curEdge;
      curEdge = curEdge->next;
    }
/***
  for (int curLevel = 0; curLevel < MAXLEVELS; curLevel++)
    {
      curNode = &(this->rows[curLevel]);
      for (inc = 0; inc < curNode->levelSize; inc++)
	delete curNode->typeDisps[inc];
    }
  for (inc = 0; inc < singletons.levelSize; inc++)
    delete singletons.typeDisps[inc];
**********/
  IndexNode* curIndex = displayerIndex;
  while (curIndex != nil)
    {
      delete curIndex;
      curIndex = curIndex->next;
    }
}

/*  CleanUp: Remove data, but don't deconstruct the entire thing.	*/

void HierGraph::CleanUp()

{
  EdgeNode* curEdge = edgeList;
  EdgeNode* deadEdge;
  Level* curNode;
  int inc;

  while (curEdge != nil)
    {
      deadEdge = curEdge;
      curEdge = curEdge->next;
      delete deadEdge;    
    }
  edgeList = nil;
  lastEdge = nil;
  for (int curLevel = 0; curLevel < MAXLEVELS; curLevel++)
    {
      curNode = &(this->rows[curLevel]);
      for (inc = 0; inc < curNode->levelSize; inc++)
	delete curNode->typeDisps[inc];
      curNode->levelSize = 0;
    }
  for (inc = 0; inc < singletons.levelSize; inc++)
    delete singletons.typeDisps[inc];
  singletons.levelSize = 0;
  numLevels = 0;
  IndexNode* curIndex = displayerIndex;
  IndexNode* deadIndex;
  while (curIndex != nil)
    {
      deadIndex = curIndex;
      curIndex = curIndex->next;
      delete deadIndex;
    }
  displayerIndex = nil;
  maxRightPos = 0;
  numLevels = 0;
}

/*  ReGraph:							*
 *    Kill the old data and replace it with new stuff.		*/

void HierGraph::ReGraph(int num, HierNode** theNodes, int numR, HierNode** roots)

{
  CleanUp();
  numNodes = num;
  allNodes = theNodes;
  numRoots = numR;
  rootNodes = roots;
  margins = 0;
  output->ClearRect(canvas, 0, 0, xmax, ymax);
  MakeDisplay();
}

/*  MakeDisplay:						*
 *    Make the entire display from scratch.			*/

void HierGraph::MakeDisplay()

{
  SetUp();
  for (int inc = 0; inc < numRoots; inc++)		// Level assignment
    if (rootNodes[inc]->GetSuccessors()->Size() == 0)
      AddToLevel(&singletons, rootNodes[inc]);
    else
      MakeLists(0, rootNodes[inc], true);
  MakeFakeNodes();					// Fake node construction
  CleanDisplay();					// Make the display
}

/*  SetUp:							*
 *    Create index of all type objects, and initialize the	*
 *  edge list.							*/

void HierGraph::SetUp()

{
  int increment;
  int j;

  displayerIndex = new IndexNode;
  displayerIndex->type = nil;
  displayerIndex->typeDisp = nil;
  displayerIndex->next = nil;
  for (increment = 0; increment < MAXLEVELS; increment++)
    {
      this->rows[increment].levelSize = 0;
      if (this->rows[increment].allocSize == 0)
	{
	  this->rows[increment].allocSize = LEVEL_SIZE;
	  this->rows[increment].typeDisps = 
	    (HierNode**)malloc(sizeof(HierNode*) * this->rows[increment].allocSize);
	}
      for (j = 0; j < this->rows[increment].allocSize; j++)
	this->rows[increment].typeDisps[j] = nil;
    }
  singletons.levelSize = 0;	
  if (singletons.allocSize == 0)
    {
      singletons.allocSize = LEVEL_SIZE;
      singletons.typeDisps = (HierNode**)malloc(sizeof(HierNode*) * singletons.allocSize);
    }
  for (j = 0; j < singletons.allocSize; j++)
    singletons.typeDisps[j] = nil;
  IndexNode* curIndex;
  IndexNode* newIndex;
  curIndex = displayerIndex;
  increment = 0;
  while (increment < numNodes)
    {
      newIndex = new IndexNode;
      newIndex->type = allNodes[increment]->GetType();
      newIndex->typeDisp = allNodes[increment];
      newIndex->next = nil;
      curIndex->next = newIndex;
      curIndex = newIndex;
      increment++;
    }
  EdgeNode* firstNode;
  firstNode = new EdgeNode;
  firstNode->parent = firstNode->child = nil;
  firstNode->next = nil;
  edgeList = firstNode;
  lastEdge = firstNode;
}

void HierGraph::AddToLevel(Level* whatLevel, HierNode* whatNode)

{
  if (whatLevel->levelSize == (whatLevel->allocSize - 1))
    {
      whatLevel->allocSize += LEVEL_SIZE;
      whatLevel->typeDisps = 
	(HierNode**)realloc((char*)whatLevel->typeDisps, 
			    sizeof(HierNode*) *  whatLevel->allocSize);
      for (int j = whatLevel->levelSize; j < whatLevel->allocSize; j++)
	whatLevel->typeDisps[j] = nil;
    }
  whatLevel->typeDisps[whatLevel->levelSize] = whatNode;
  whatLevel->levelSize++;
}
  

/*  MakeLists:							*
 *    Calls itself recursively to make the edge list and the	*
 *  list of types on each level. Assigns types to levels by	*
 *  waiting until all each type's parents are assigned a level	*
 *  and then putting the type one level deeper than the deepest	*
 *  parent.							*/

void HierGraph::MakeLists(int level,			// what level
			  HierNode* thisDisplay,	// what to put there
			  boolean goOn)			// whether to continue descending

{
  EdgeNode* newEdge;			// New node to add to edge list
  HierNode* childDisp;			// Child of the type we're dealing with
  AdjList* parents;			// Parents of the type
  int newLevel = -1;			// New level to put type on
  int parentLevel;			// Level of lowest parent
  char* typeName;			// name of this type
  boolean dontAdd = false;		// Don't add the type to the list just yet

  parents = thisDisplay->GetPredecessors();	
  parents->First();
  if (parents->Size() <= 1)
    newLevel = level;
  else
    for (int i = 1; i <= parents->Size(); i++)
      {
	parentLevel = parents->GetCur()->GetDisp()->GetLevel();
	if (parentLevel < 0)		// If all parents don't have a level, don't assign one
	  newLevel = WAIT;
	if ((newLevel != WAIT) && (parentLevel >= newLevel))
	  newLevel = parentLevel + 1;
	parents->Next();
      }
  typeName = thisDisplay->getName();
  if ((showInteresting) && (!thisDisplay->IsInteresting()))
    dontAdd = true; 
  if ((newLevel != WAIT) && (!dontAdd))
    {
      if (newLevel >= 0)
	{
	  thisDisplay->SetLevel(newLevel);
	  AddToLevel(&(this->rows[newLevel]), thisDisplay);
	}
      if (thisDisplay->GetSuccessors()->Size() > 0)
	{
	  int inc;
	  AdjList* subtypes = thisDisplay->GetSuccessors();
	  subtypes->First();
	  for (inc = 0; inc < subtypes->Size(); inc++)
	    {
	      childDisp = subtypes->GetCur()->GetDisp();
	      if (childDisp != nil)
		{
		  newEdge = new EdgeNode;
		  newEdge->parent = thisDisplay;
		  newEdge->child = childDisp;
		  newEdge->next = nil;
		  lastEdge->next = newEdge;
		  lastEdge = newEdge;
		  if (goOn)
		    MakeLists(level + 1, childDisp, true);	// Go to all my subtypes recursively
		}
	      subtypes->Next();
	    }
	}
    }
}

/*  MakeFakeNodes:						*
 *    If any edges extend through more than one level, add	*
 *  fake nodes in between to make the display look nicer.	*/

void HierGraph::MakeFakeNodes()

{
  EdgeNode* curEdge = edgeList;			// Current edge
  EdgeNode* newEdge;				// New edge to add for fake node
  int curLevel;					// Current level
  AdjList* parents;				// Parents of current type
  HierNode* curParent;				// Current parent

  while (curEdge != nil)
    {
      if ((curEdge->child != nil) && (curEdge->parent != nil) 
	  && (curEdge->child->GetLevel() < 0) && (curEdge->parent->GetLevel() >= 0))
	{
	  curLevel = curEdge->parent->GetLevel();
	  parents = curEdge->child->GetPredecessors();
	  parents->First();
	  for (int i = 0; i < parents->Size(); i++)
	    {
	      curParent = parents->GetCur()->GetDisp();
	      if (curParent->GetLevel() < 0)
		{
		  MakeLists(curLevel - 1, curParent, false);	// Add to level list non-recursively
		  newEdge = new EdgeNode;			// Add to edge list
		  newEdge->parent = FindDisp(curParent->GetType());
		  newEdge->child = curEdge->child;
		  newEdge->next = lastEdge->next;
		  lastEdge->next = newEdge;
		  lastEdge = newEdge;
		}
	      parents->Next();
	    }
	  MakeLists(curLevel, curEdge->child, false);
	}
      curEdge = curEdge->next;
    }
  int increment;
  HierNode* fakeDisplayer;
  HierNode* parentNode;
  HierNode* childNode;
  HierNode* lastNode;
  int difference;
  EdgeNode* firstNode = edgeList;
  EdgeNode* deadEdge;
  EdgeNode* oldLastEdge = lastEdge;
  curEdge = firstNode->next;
  while ((curEdge != nil) && (curEdge != oldLastEdge->next))
    {
      parentNode = curEdge->parent;
      childNode = curEdge->child;
      if ((parentNode == nil) || (childNode == nil))
	difference = 1;
      else
	difference = childNode->GetLevel() - parentNode->GetLevel();
      if (difference > 1)
	{
	  parentNode->RemoveSuccessor(childNode);
	  childNode->RemovePredecessor(parentNode);
	  deadEdge = curEdge;
	  for (curEdge = firstNode; curEdge->next != deadEdge; curEdge = curEdge->next);
	  curEdge->next = deadEdge->next;
	  curEdge = deadEdge->next;
	  delete deadEdge;
	  lastNode = parentNode;
	  for (increment = parentNode->GetLevel() + 1; increment <= childNode->GetLevel() - 1; increment++)
	    {	
	      if (horizontal)
		fakeDisplayer = new FakeDisplayer(increment, nil, horizontal, FAKE_HEIGHT);
	      else
		fakeDisplayer = new FakeDisplayer(increment, nil, horizontal, nodeWidth);
	      AddToLevel(&this->rows[increment], fakeDisplayer);
	      lastNode->AddSuccessor(fakeDisplayer);
	      fakeDisplayer->AddPredecessor(lastNode);
	      newEdge = new EdgeNode;
	      newEdge->parent = lastNode;
	      newEdge->child = fakeDisplayer;
	      newEdge->next = lastEdge->next;
	      lastEdge->next = newEdge;
	      lastEdge = newEdge;
	      lastNode = fakeDisplayer;
	    }
	  fakeDisplayer->AddSuccessor(childNode);
	  lastNode->AddPredecessor(fakeDisplayer);
	  newEdge = new EdgeNode;
	  newEdge->parent = lastNode;
	  newEdge->child = childNode;
	  newEdge->next = lastEdge->next;
	  lastEdge->next = newEdge;
	  lastEdge = newEdge;
	}
      else
	curEdge = curEdge->next;
    }
}

/*  CleanDisplay:							*
 *    This is the graph layout algorithm itself. This moves the type	*
 *  displayer objects back and forth to try and minimize edge		*
 *  crossings and generally make things look nice.			*/

void HierGraph::CleanDisplay()

{
  HierNode** curNodes;
  int curLevel = 0;
  int curNode;
  int pass;
  int inc;
  boolean goingDown = false;
  int levelCount = 0;

  while ((this->rows[curLevel].levelSize > 0) && (curLevel < MAXLEVELS))
    curLevel++;
  numLevels = curLevel;
  this->rows[numLevels++] = singletons;
  maxRightPos = 0;
  for (curLevel = 0; curLevel < numLevels; curLevel++)
    {
      curNodes = this->rows[curLevel].typeDisps;	// Place types in straight this->rows initially
      for (curNode = 0; curNode < this->rows[curLevel].levelSize; curNode++)
	curNodes[curNode]->SetPos(curNode);
      if (curNode > maxRightPos)
	maxRightPos = curNode;
    }
  maxRightPos += 5;
  margins = round(maxRightPos / 4) + 1;
  maxRightPos += (margins * 2);
  for (pass = 0; pass < PASSES; pass++)			// Start Barycentre passes
    {
      if (goingDown)
	{
	  curLevel = numLevels - 1;
	  goingDown = false;
	}
      else
	{
	  curLevel = 0;
	  goingDown = true;
	}
      while ((curLevel >= 0) && (curLevel < numLevels))
	{
	  curNodes = this->rows[curLevel].typeDisps;
	  if (goingDown)
	    {
	      UpBarycentre(curNodes, this->rows[curLevel].levelSize);
	      UpSort(curNodes, this->rows[curLevel].levelSize);
	      for (curNode = 0; curNode < this->rows[curLevel].levelSize; 
		   curNode++)
		curNodes[curNode]->SetPos(curNode);
	      
	    }
	  else	
	    {
	      DownBarycentre(curNodes, this->rows[curLevel].levelSize);
	      DownSort(curNodes, this->rows[curLevel].levelSize);
	      for (curNode = 0; curNode < this->rows[curLevel].levelSize; 
		   curNode++)
		curNodes[curNode]->SetPos(curNode);
	    }
	  if (goingDown)
	    curLevel++;
	  else
	    curLevel--;
	}
    }
  int prevPriority;
  int curPriority;
  int maxPriority;
  int difference;
  goingDown = false;
  for (curLevel = 0; curLevel < numLevels; curLevel++)
    {
      for (inc = 0; inc < this->rows[curLevel].levelSize; inc++)
	this->rows[curLevel].typeDisps[inc]->SetPos(this->rows[curLevel].typeDisps[inc]->GetPos() + margins);
    }
  for (pass = 0; pass < PASSES; pass++)			// Start Barycentre passes
    {
      if (goingDown)
	{
	  curLevel = numLevels - 1;
	  goingDown = false;
	}
      else
	{
	  curLevel = 0;
	  goingDown = true;
	}
      while ((curLevel >= 0) && (curLevel < numLevels))
	{
	  curNodes = this->rows[curLevel].typeDisps;
	  if (goingDown)
	    UpBarycentre(curNodes, this->rows[curLevel].levelSize);
	  else
	    DownBarycentre(curNodes, this->rows[curLevel].levelSize);
	  prevPriority = FAKE_PRIORITY * 2;
	  inc = 0;
	  while (inc < this->rows[curLevel].levelSize)
	    {
	      maxPriority = 0;
	      for (curNode = 0; curNode < this->rows[curLevel].levelSize; curNode++)
		{
		  curPriority = curNodes[curNode]->GetConnectivity();
		  if ((curPriority > maxPriority) && (curPriority < prevPriority))
		    maxPriority = curNodes[curNode]->GetConnectivity();
		}
	      for (curNode = 0; curNode < this->rows[curLevel].levelSize; curNode++)
		{
		  if (curNodes[curNode]->GetConnectivity() == maxPriority)
		    {
		      if (curNodes[curNode]->GetPredecessors()->Size() == 0)
			difference = round(curNodes[curNode]->GetDownBary());
		      else if (curNodes[curNode]->GetSuccessors()->Size() == 0)
			difference = round(curNodes[curNode]->GetUpBary());
		      else
			difference = round((curNodes[curNode]->GetUpBary() + 
					  curNodes[curNode]->GetDownBary()) / 2);
		      difference -= curNodes[curNode]->GetPos();
		      if (difference != 0)
			MoveNode(curNodes, curNode, difference);
		      inc++;	      
		    }
		}
	      prevPriority = maxPriority;
	    }
	  if (goingDown)
	    curLevel++;
	  else
	    curLevel--;
	}
    }
}

void HierGraph::DownBarycentre(HierNode** level,	// Level to adjust
			       int numNodes)

{
  for (int inc = 0; inc < numNodes; inc++)
    level[inc]->ComputeDownBary();
}

void HierGraph::UpBarycentre(HierNode** level,
			     int numNodes)

{
  for (int inc = 0; inc < numNodes; inc++)
    level[inc]->ComputeUpBary();
}

int HierGraph::MoveNode(HierNode** level, int node, int delta)

{
  int possibleDelta;
  int difference;
  int myPos = level[node]->GetPos();
  if (((node == 0) && (delta < 0)) || 
      ((level[node + 1] == nil) && (delta > 0)))
    possibleDelta = delta;  /* At the beginning or the end */
  else
    {
      if (delta < 0)
	{
	  difference = level[node - 1]->GetPos() - myPos;
	  if (difference > (delta - 1))
	    {
	      possibleDelta = difference + 1;
	      if (level[node]->GetConnectivity() <= level[node - 1]->GetConnectivity())
		possibleDelta += MoveNode(level, node - 1, delta - difference - 1);
	    }
	  else
	    possibleDelta = delta;
	}
      else
	{
	  difference = level[node + 1]->GetPos() - myPos;
	  if (difference < (delta + 1))
	    {
	      possibleDelta = difference - 1;
	      if (level[node]->GetConnectivity() <= level[node + 1]->GetConnectivity())
		possibleDelta += MoveNode(level, node + 1, delta - difference + 1);
	    }
	  else
	    possibleDelta = delta;
	}
    }
  if ((myPos + possibleDelta) < 0)	// We can't go over the edge of the graph
    possibleDelta = 0 - myPos;
  else if ((myPos + possibleDelta) > maxRightPos)
    possibleDelta = maxRightPos - myPos;
  level[node]->SetPos(myPos + possibleDelta);
  return possibleDelta;
}	
	    

/*  DrawDisplay:						*
 *    Once the nodes have been laid out, place their actual	*
 *  canvasses in the window.					*/

void HierGraph::DrawDisplay()

{
  int curLevel;
  int curNode;
  HierNode** curNodes;
  Coord curX;
  Coord curY;
  Coord maxX;
  Coord topY;
  
  const Coord fudge = (nodeWidth / 2) + (2 * FUDGE);
  maxX = 0;
  if (horizontal)
    topY = numLevels * ROW_HEIGHT;
  else
    topY = maxRightPos * (FAKE_HEIGHT + V_PAD) + FUDGE;
  for (curLevel = 0; curLevel < numLevels; curLevel++)
    {
      curNodes = this->rows[curLevel].typeDisps;
      if (horizontal)
	{
	  curY = topY - (curLevel * ROW_HEIGHT) - 15;
	  for (curNode = 0; curNode < this->rows[curLevel].levelSize; curNode++)
	    {
	      curX = (curNodes[curNode]->GetPos() * nodeWidth) + fudge;
	      Insert(curNodes[curNode], curX, curY, TopCenter);
	      if ((curX + nodeWidth) > maxX)
		maxX = curX + nodeWidth;
	    }
	}
      else
	{
	  curX = curLevel * (nodeWidth + H_PAD) + H_PAD;
	  for (curNode = 0; curNode < this->rows[curLevel].levelSize; curNode++)
	    {
	      curY = topY - (curNodes[curNode]->GetPos() * 
			     (FAKE_HEIGHT + V_PAD)) - FAKE_HEIGHT;
	      Insert(curNodes[curNode], curX, curY, CenterLeft);
	      if ((curX + nodeWidth) > maxX)
		maxX = curX + nodeWidth;
	    }
	}
    }
}

HierNode* HierGraph::FindDisp(int what)

{
  IndexNode* cur = displayerIndex->next;
  while ((cur != nil) &&
	 (cur->typeDisp->GetType() != what))
    cur = cur->next;
  if (cur != nil)
    return cur->typeDisp;
  else
    return nil;
}

void HierGraph::Reconfig()

{
  Scene::Reconfig();
  output->SetOverwrite(true);
  Listen(updownEvents);
  FindShape();
  if (handCursor == nil)
    {
      Bitmap* theHand = new Bitmap((void*)hand_bits, hand_width, hand_height, hand_x_hot, hand_y_hot);
      Bitmap* theMask = new Bitmap((void*)hand_mask_bits, hand_mask_height, hand_mask_width);
      handCursor = new Cursor(theHand, theMask, output->GetFgColor(), output->GetBgColor());
    }
}

void HierGraph::FindShape()

{
  Coord maxX, maxY;

  if (horizontal)
    {
      maxX = maxRightPos * (nodeWidth + H_PAD);
      maxY = ROW_HEIGHT * numLevels;
    }
  else
    {
      maxX = numLevels * (nodeWidth + H_PAD);
      maxY = maxRightPos * (FAKE_HEIGHT + V_PAD);
    }
  shape->Rect(maxX + (3 * FUDGE), maxY + (3 * FUDGE));
  shape->Rigid();
  if (perspective != nil)
    {
      lastWidth = perspective->curwidth;
      lastHeight = perspective->curheight;
    }
}

void HierGraph::Redraw(Coord, Coord, Coord, Coord)

{
  Coord parentX, parentY, childX, childY;
  EdgeNode* curEdge = edgeList->next;
  HierNode* prevParent = nil;
  while (curEdge != nil)
    {
      if ((curEdge->parent != nil) && (curEdge->child != nil))
	{
	  prevParent = curEdge->parent;
	  if (horizontal)
	    {
	      parentX = curEdge->parent->GetXCoord();
	      parentY = curEdge->parent->GetBotYCoord();
	      childX = curEdge->child->GetXCoord();
	      childY = curEdge->child->GetTopYCoord();
	    }	
	  else
	    {
	      parentX = curEdge->parent->GetRightXCoord();
	      parentY = curEdge->parent->GetYCoord();
	      childX = curEdge->child->GetLeftXCoord();
	      childY = curEdge->child->GetYCoord();
	    }
	  if ((parentX != -1) && (parentY != -1) && (childX != -1) && (childY != -1))
	    output->Line(canvas, parentX, parentY, childX, childY);
	}
      curEdge = curEdge->next;
    }
}

void HierGraph::Draw()

{
  int inc;
  HierNode** curNodes;
  output->Scale(scaleX, scaleY);
  for (int curLevel = 0; curLevel < numLevels; curLevel++)
    {
      curNodes = this->rows[curLevel].typeDisps;
      for (inc = 0; inc < this->rows[curLevel].levelSize; inc++)
	curNodes[inc]->Draw();
    }
  Redraw(0, 0, xmax, ymax);
  output->Scale(1.0, 1.0);
}

void HierGraph::Resize()

{	
  Scene::Resize();
  DrawDisplay();
  FindShape();
  Draw();
}

void HierGraph::Update()

{
  Scene::Update();
  if (Parent()->GetPerspective() != nil)
    {
      boolean changeIt = true;
      if (lastWidth == 0)
	changeIt = false;
      if ((Parent()->GetPerspective()->curwidth != lastWidth) || 
	  (Parent()->GetPerspective()->curheight != lastHeight))
	{
	  if (changeIt)
	    {
	      scaleX *= Parent()->GetPerspective()->curwidth / lastWidth;
	      scaleY *= Parent()->GetPerspective()->curheight / lastHeight;
	    }
	  else
	    {
	      scaleX = 1.0;
	      scaleY = 1.0;
	    }
	  lastWidth = Parent()->GetPerspective()->curwidth;
	  lastHeight = Parent()->GetPerspective()->curheight;
	}
      Draw();
    }
}

void HierGraph::Handle(Event& anEvent)

{
  if (anEvent.target != this)
    anEvent.target->Handle(anEvent);
  else if ((anEvent.eventType == DownEvent) && (anEvent.button == MIDDLEMOUSE))
    {
      Viewport* thePort = (Viewport*)Parent();
      Cursor* oldCursor = GetCursor();
      SetCursor(handCursor);
      float hierWidth = ((float)shape->width);
      float hierHeight = ((float)shape->height);
      Coord oldX = anEvent.x;
      Coord oldY = anEvent.y;
      float Xpos = thePort->XPos();
      float Ypos = thePort->YPos();
      do 
	{
	  Xpos += (oldX - anEvent.x) / hierWidth;
	  Ypos += (oldY - anEvent.y) / hierHeight;
	  thePort->ScrollTo(Xpos, Ypos);
	  oldX = anEvent.x;
	  oldY = anEvent.y;
	  Poll(anEvent);
	}
      while (anEvent.middlemouse);
      SetCursor(oldCursor);
    }
}

void HierGraph::DoInsert(Interactor* i, boolean, Coord& leftX, Coord& botY)

{
  Shape* itsShape = i->GetShape();
  Coord x1, x2, y1, y2;
  x1 = leftX;
  x2 = x1 + itsShape->width;
  y1 = botY;
  y2 = botY + itsShape->height;
  Place(i, x1, y1, x2, y2);
}

/*  NewRoot:						*
 *    Make a particular type the root of the hierarchy.	*/

void HierGraph::NewRoot(HierNode* t)

{
  CleanUp();
  SetUp();
  MakeLists(0, t, true);
  MakeFakeNodes();
  CleanDisplay();
  DrawDisplay();
  FindShape();
  output->ClearRect(canvas, 0, 0, xmax, ymax);
  Parent()->Change(this);
  Draw();
}

/*  FullView:						*
 *    Lay out the entire hierarhcy.			*/

void HierGraph::FullView()

{
  CleanUp();
  SetUp();
  for (int inc = 0; inc < numRoots; inc++)
        if (rootNodes[inc]->GetSuccessors()->Size() == 0)
      AddToLevel(&singletons, rootNodes[inc]);
    else
      MakeLists(0, rootNodes[inc], true);
  MakeFakeNodes();
  CleanDisplay();
  DrawDisplay();
  FindShape();
  output->ClearRect(canvas, 0, 0, xmax, ymax);
  Parent()->Change(this);
  Draw();
}

void HierGraph::SetAutoUpdate(boolean val)

{
  if (val == true)
    {
      autoUpdating = true;
      if (needsUpdating)
	{
	  FullView();
	  needsUpdating = false;
	}
    }
  else
    autoUpdating = false;
}

void HierGraph::SetShowInteresting(boolean val)

{
  if ((val == true) && (showInteresting == false))
    {
      showInteresting = true;
      FullView();
    }
  else if ((val == false) && (showInteresting == true))
    {
      showInteresting = false;
      FullView();
    }
}

void HierGraph::NewType(HierNode* newNode)

{
  allNodes[numNodes] = newNode;
  numNodes++;
  if (autoUpdating)
    {
      FullView();		// This should be smarter
      needsUpdating = false;
    }
  else
    needsUpdating = true;
}

void HierGraph::GetPostScript(ostream& to,
			      float scaleVal,
			      boolean EPSF,
			      boolean landscape)

{
  float Left = .25;
  float Bot = .25;
  float Width;
  float Height;
  if (landscape)
    {
      Width = 10.5;
      Height = 8.0;
    }
  else
    {
      Width = 8.0;
      Height = 10.5;
    }
  int numRows;
  int numCols;
  if (EPSF)
    numRows = numCols = 1;
  else
    {
      if (landscape)
	{
	  numRows = round((xmax * scaleVal) / inch / Width) + 1;
	  numCols = round((ymax * scaleVal) / inch / Height) + 1;
	}
      else
	{
	  numRows = round((ymax * scaleVal) / inch / Height) + 1;
	  numCols = round((xmax * scaleVal) / inch / Width) + 1;
	}
    }
  int numPages = numRows * numCols;
  to << "%!PS-Adobe-2.0 EPSF-2.0\n";
  to << "%%Title: ENCORE type hierarchy\n";
  to << "%%Creator: ENCORE Graphical User Interface\n";
  to << "%%DocumentFonts: Helvetica-Bold\n";
  to << "%%ProofMode: Substitute\n";
  to << "%%Pages: " << numPages << "\n";
  if (EPSF)
    {
      if (landscape)
	to << "%%BoundingBox: 0 0 " << ymax * scaleVal << " " << xmax * scaleVal << "\n";
      else
	to << "%%BoundingBox: 0 0 " << xmax * scaleVal << " " << ymax * scaleVal << "\n";
    }
  to << "%%EndComments\n\n";
  to << "/inch {72 mul} def\n\n";
  to << "/Rect {\n";
  to << "/t exch def\n";
  to << "/r exch def\n";
  to << "/b exch def\n";
  to << "/l exch def\n";
  to << "newpath\n";
  to << "l b moveto\n";
  to << "l t lineto\n";
  to << "r t lineto\n";
  to << "r b lineto\n";
  to << "closepath\n";
  to << "stroke\n";
  to << "} def\n\n";
  to << "/Line {\n";
  to << "/y2 exch def\n";
  to << "/x2 exch def\n";
  to << "/y1 exch def\n";
  to << "/x1 exch def\n";
  to << "newpath\n";
  to << "x1 y1 moveto\n";
  to << "x2 y2 lineto\n";
  to << "closepath\n";
  to << "stroke\n";
  to << "} def\n\n";
  to << "%%EndProlog\n";
  int curPage = 1;
  int curLevel;
  Coord x1, y1, x2, y2;
  EdgeNode* curEdge;
  HierNode** curNodes;
  int curNode;
  for (int row = 0; row < numRows; row++)
    {
      for (int col = 0; col < numCols; col++)
	{
	  to << "\n%%Page: " << curPage << " " << curPage << "\n\n";
	  to << "/Helvetica-Bold findfont 10 scalefont setfont\n";
	  to << Left << " inch " << Bot << " inch translate\n";
	  to << "gsave\n";
	  if (scaleVal != 1.0)
	    to << scaleVal << " " << scaleVal << " scale\n";
	  if (landscape)
	    {
	      to << "90 rotate\n";
	      to << (Width * (row - 1.0)) / scaleVal << " inch neg\n";
	      to << (Height * col) / scaleVal << " inch neg\n";
	    }
	  else
	    {
	      to << (Width * (col - 1.0)) / scaleVal << " inch neg\n";
	      to << (Height * (row - 1.0)) / scaleVal << " inch neg\n";
	    }
	  to << "translate\n";
	  for (curLevel = 0; curLevel < numLevels; curLevel++)
	    {
	      curNodes = this->rows[curLevel].typeDisps;
	      for (curNode = 0; curNode < this->rows[curLevel].levelSize; curNode++)
		curNodes[curNode]->GetPostScript(to);
	    }
	  curEdge = edgeList;
	  while (curEdge != nil)
	    {
	      if ((curEdge->parent != nil) && (curEdge->child != nil))
		{
		  x1 = curEdge->parent->GetXCoord();
		  y1 = curEdge->parent->GetBotYCoord();
		  x2 = curEdge->child->GetXCoord();
		  y2 = curEdge->child->GetTopYCoord();
		  if ((x1 != -1) && (y1 != -1) && (x2 != -1) && (y2 != -1))
		    to << x1 << " " << y1 << " " << x2 << " " << y2 << " Line\n";
		}
	      curEdge = curEdge->next;
	    }
	  if (!EPSF)
	    to << "gsave showpage grestore\n";
	  to << "grestore\n";
	  curPage++;
	}
    }
  to << "\n%%Trailer\n";
}

void HierGraph::UpSort(HierNode** theList, int size)

{
  register int i;
  register int j;
  int min;
  HierNode* temp;

  i = 0;
  while (i < size - 1)
    {
       min = i;
       j = i + 1;
       while (theList[j] != nil)
	 {
	   if (theList[j]->GetUpBary() < theList[min]->GetUpBary())
	     min = j;
	   j++;
	 }
       temp = theList[min];
       theList[min] = theList[i];
       theList[i] = temp;
       i++;
     }
}

void HierGraph::DownSort(HierNode** theList, int size)

{
  register int i;
  register int j;
  int min;
  HierNode* temp;

  i = 0;
  while (i < size - 1)
    {
       min = i;
       j = i + 1;
       while (theList[j] != nil)
	 {
	   if (theList[j]->GetDownBary() < theList[min]->GetDownBary())
	     min = j;
	   j++;
	 }
       temp = theList[min];
       theList[min] = theList[i];
       theList[i] = temp;
       i++;
     }
}

HierNode::HierNode(char* name,
		   ButtonState* s,
		   int t,
		   CommandPtr toDo,
		   void* what)
     : SquareButton(name, s, t, toDo, what)


{
  whatLevel = -1;
  whatPos = -1;
  myData = t;
  upBary = downBary = 0.0;
  SetClassName("HierNode");
  predecessors = new AdjList();
  successors = new AdjList();
  myName = name;
}

HierNode::~HierNode()

{
  delete predecessors;
  delete successors;
}

void HierNode::Reconfig()

{
  SquareButton::Reconfig();
  SetClassName("HierNode");
}  

void HierNode::ClearLists()

{
  delete predecessors;
  delete successors;
  predecessors = new AdjList();
  successors = new AdjList();
}

float HierNode::ComputeDownBary()

{
  if (successors->Size() == 0)
    downBary = 0.0;
  else
    {
      HierNode* curNode = successors->First()->GetDisp();
      int totalPos = 0;
      for (int inc = 0; inc < successors->Size(); inc++)
	{
	  totalPos += successors->GetCur()->GetDisp()->GetPos();
	  successors->Next();
	}
      downBary = totalPos / successors->Size();
    }
  return downBary;
}

float HierNode::ComputeUpBary()

{
  if (predecessors->Size() == 0)
    upBary = 0.0;
  else
    {
      predecessors->First();
      int totalPos = 0;
      for (int inc = 0; inc < predecessors->Size(); inc++)
	{
	  totalPos += predecessors->GetCur()->GetDisp()->GetPos();
	  predecessors->Next();
	}
      upBary = totalPos / predecessors->Size();
    }
  return upBary;
}

int HierNode::GetConnectivity()

{
  return (predecessors->Size() + successors->Size());
}

FakeDisplayer::FakeDisplayer(int level, int t, boolean horiz, Coord ext)
     : HierNode("X", new ButtonState(), t, nil, nil)

{
  whatLevel = level;
  SetClassName("FakeDisplayer");
  horizontal = horiz;
  extent = ext;
}

void FakeDisplayer::Reconfig()

{
  if (horizontal)
    {
      shape->width = 1;
      shape->height = extent;
    }
  else
    {
      shape->width = extent;
      shape->height = 1;
    }
  Listen(noEvents);
}

void FakeDisplayer::Redraw(Coord, Coord, Coord, Coord)

{
  if (horizontal)
    output->Line(canvas, 0, 0, 0, ymax);
  else
    output->Line(canvas, 0, 0, xmax, 0);
}

void FakeDisplayer::Handle(Event& anEvent)

{
  if (anEvent.target != this)
    anEvent.target->Handle(anEvent);
}

Coord FakeDisplayer::GetXCoord()

{
  Coord thisX = xmax / 2;
  Coord thisY = 0;
  GetRelative(thisX, thisY, Parent());
  return thisX;
}

Coord FakeDisplayer::GetTopYCoord()

{
  Coord thisX = 0;
  Coord thisY = ymax;
  GetRelative(thisX, thisY, Parent());
  return thisY;
}

Coord FakeDisplayer::GetBotYCoord()

{
  Coord thisX = 0;
  Coord thisY = 0;
  GetRelative(thisX, thisY, Parent());
  return thisY;
}

Coord FakeDisplayer::GetYCoord()

{
  Coord thisX = 0;
  Coord thisY = ymax / 2;
  GetRelative(thisX, thisY, Parent());
  return thisY;
}

Coord FakeDisplayer::GetLeftXCoord()

{
  Coord thisX = 0;
  Coord thisY = 0;
  GetRelative(thisX, thisY, Parent());
  return thisX;
}

Coord FakeDisplayer::GetRightXCoord()

{
  Coord thisX = xmax;
  Coord thisY = 0;
  GetRelative(thisX, thisY, Parent());
  return thisX;
}

void FakeDisplayer::GetPostScript(ostream& to)

{
  Coord x1, y1, x2, y2;
  if (horizontal)
    {
      x1 = y1 = x2 = 0;
      y2 = extent;
    }
  else
    {
      x1 = y1 = y2 = 0;
      x2 = extent;
    }
  GetRelative(x1, y1, Parent());
  GetRelative(x2, y2, Parent());
  to << x1 << " " << y1 << " " << x2 << " " << y2 << " Line\n";
}

AdjNode::AdjNode(HierNode* disp)
     
{
  displayer = disp;
  valuep = (void**)displayer;
}

void AdjList::Append(HierNode* newDisp)

{
  AdjNode* newNode = new AdjNode(newDisp);
  BaseList::Append(newNode);
}

void AdjList::Remove(HierNode* rmThis)

{
  First();
  while ((GetCur()->GetDisp() != rmThis) && (!AtEnd()))
    Next();
  if (GetCur()->GetDisp() == rmThis)
    DeleteCur();
}

