/*  $Id: hiergraph.h,v 1.2 90/06/17 18:45:37 gjb Exp $
 *									*
 *  class HierGraph: A class that displays hierarchical graphs, like	*
 *  the ones used for type hierarchies. It handles multiple inheritance,*
 *  but it's a good tree-drawing algorithm as well.			*
 *  It uses an adaptation of						*
 *  an algorithm by Sugiyama to draw directed hierarchic graphs. It's 	*
 *  a little different because it's object-oriented and because most	*
 *  nodes in our graph have many more successors (children) than 	*
 *  predecessors. This may be the kind of code that needs a skull and	*
 *  crossbones on it. At the very least, be forewarned that small 	*
 *  changes to this code may cause random and unpredictable changes	*
 *  to the graph layout.						

Copyright 1991, Brown University, Providence, RI

Permission to use and modify this software and its documentation for any
purpose other than its incorporation into a commercial product is hereby
granted without fee.  Permission to copy and distribute this software and its
documentation for non-commercial use only is also granted without fee,
provided, however, that the above copyright notice appear in all copies, that
both that copyright notice and this permission notice appear in supporting
documentation, that the name of Brown University not be used in advertising or
publicity pertaining to distribution of the software without specific, written
prior permission, and that Brown University be notified by the distributor when
such distributions are done outside of his or her organization. Brown
University makes no representations about the suitability of this software for
any purpose.  It is provided "as is" without express or implied warranty.
Brown University requests notification of any modifications to this software
or its documentation.

Please send the name, organization, and address (postal and electronic) of
recipients of software outside of the receiving organization to:

	Software Librarian
	Computer Science Department, Box 1910
	Brown University
	Providence, RI 02912
		or
	brusd@cs.brown.edu

All electronic notifications will be acknowledged.

*/ 


#ifndef HIERGRAPH_H
#define HIERGRAPH_H

#include <string.h>
#include <stream.h>
#include <stdlib.h>
#include <InterViews/interactor.h>
#include <InterViews/world.h>
#include <InterViews/button.h>
#include <InterViews/box.h>
#include <InterViews/glue.h>
#include <InterViews/event.h>
#include <InterViews/sensor.h>
#include <InterViews/viewport.h>
#include <InterViews/brush.h>
#include "squarebutton.h"
#include "list.h"

class HierNode;

/*  AdjNode: A node in an adjacency list -- see baselist.h */

class AdjNode : public BaseNode

{
protected:
  HierNode* displayer;			// The actual object we want
public:
  AdjNode(HierNode* disp);
  HierNode* GetDisp()
    {
      return displayer;
    }
};

/*  AdjList: An adjacency list for a type display object */

class AdjList : public BaseList 

{
public:
  AdjNode* First()
    {
      return (AdjNode*) BaseList::First();
    }
  AdjNode* Last()
    {
      return (AdjNode*) BaseList::Last();
    }
  AdjNode* Prev()
    {
      return (AdjNode*) BaseList::Prev();
    }
  AdjNode* Next()
    {
      return (AdjNode*) BaseList::Next();
    }
  AdjNode* GetCur()
    {
      return (AdjNode*) BaseList::GetCur();
    }
  void Append(HierNode* newDisp);
  void Remove(HierNode* rmThis);
};

/* Constants */

static const int MAXLEVELS = 100;	// Maximum number of levels in type hierarchy
static const int FAKE_PRIORITY = 5000;	// Big "connectivity" for fake nodes

class HierGraph : public Scene

{
public:
  HierGraph(ButtonState* s,			// Buttonstate for whole thing
	    int num,				// Number of nodes in graph
	    HierNode** theNodes,		// The nodes themselves
	    int numR,				// Number of root nodes
	    HierNode** roots,			// The root nodes
	    boolean horiz = false);		// Whether to make horizontal
  ~HierGraph();					// Deconstructor
  void ReGraph(int, HierNode**, 
	       int, HierNode**);		// Reconstructor
  virtual void Reconfig();			// Find shape and stuff
  void FindShape();				// Really find the shape
  virtual void Redraw(Coord, Coord, Coord, Coord);   // Draw edges
  virtual void Draw();				// Draw everything
  virtual void Resize();			// Recreate canvas
  virtual void Update();			// Update when deiconified
  virtual void Handle(Event& anEvent);		// Handle an event (do nothing)
  void DoInsert(Interactor*, boolean, Coord&, Coord&);  // Insert a node
  void NewRoot(HierNode* t);			// Make node root of hierarchy
  void FullView();				// Show the whole thing again
  void SetShowInteresting(boolean val);		// Show only user-defined types
  void NewType(HierNode* t);			// A new type was created
  void SetAutoUpdate(boolean val);		// Set auto-update mode
  void GetPostScript(ostream& to, float scaleVal, boolean EPSF,
		     boolean landscape);	// Print the type hierarchy
  HierNode* FindDisp(int what);			// Find a displayer with a type
private:
  typedef struct EdgeNode 		
    /* Node in an edge list. I handle these lists myself rather than 	*
     * using baselist. Don't ask why.					*/
    {
      HierNode* parent;			// Parent node for edge
      HierNode* child;			// Child
      EdgeNode* next;			// Next edge in list
    } EdgeNode;
  
  typedef struct IndexNode		/*  Node in the type->display object index */
    {
      int type;				// Type
      HierNode* typeDisp;		// TypeDisplayer that represents it
      IndexNode* next;			// Next node in list
    } IndexNode;
  
  typedef struct Level			/*  List of nodes for a level */
    {
      int levelSize;			// Number of nodes in level
      int allocSize;			// Number of nodes allocated
      HierNode** typeDisps;		// The nodes
    } Level;

  ButtonState* state;				// Which type is selected (one & only one)
  int numNodes;					// Number of nodes in graph
  HierNode** allNodes;				// All the nodes
  int numRoots;					// Number of root nodes
  HierNode** rootNodes;				// The roots of the hierarchy
  EdgeNode* edgeList;				// List of edges
  EdgeNode* lastEdge;				// Last node in edge list
  Level rows[MAXLEVELS];			// Lists of nodes for each level
  Level singletons;				// Special level of singleton types
  IndexNode* displayerIndex;			// Matches displayers to types
  int maxRightPos;				// Maximum right-hand position
  int numLevels;				// Number of levels in graph
  boolean autoUpdating;				// Whether to update automatically
  boolean needsUpdating;			// Display needs updating
  boolean showInteresting;			// Display only "interesting" types
  Cursor* handCursor;				// For moving with the hand
  int lastWidth;				// Last value of curwidth and curheight
  int lastHeight;				// values in Perspective
  float scaleX;					// X scale value
  float scaleY;					// Y scale value
  boolean horizontal;				// Whether to make horizontal graph
  Coord nodeWidth;				// Max. width of a node
  int margins;					// Size of "margins"

  void SetUp();					// Set up type index, etc.
  void MakeFakeNodes();				// Put fake nodes in long edges
  void DownBarycentre(HierNode** level, 
		      int numNodes); 		// Average of successors
  void UpBarycentre(HierNode** level, 
		    int numNodes); 		// Average position of predecessors	
  void MakeDisplay();				// Create the display
  void MakeLists(int level, HierNode* whatType,
		 boolean goOn);			// Recursive level-assigning routine
  void AddToLevel(Level* whatLevel, 		// Add a node to a level
		  HierNode* whatNode);
  void CleanDisplay();				// Find optimum positions for nodes
  void DrawDisplay();				// Place nodes' canvases in window
  void CleanUp();				// Do the real deconstructing
  int MoveNode(HierNode**, int node, int delta); // Move a node either direction
  void UpSort(HierNode**, int);			// Sort by up barycentre
  void DownSort(HierNode**, int);		// And by down barycentre
};

/* One node (a type) in the display. Users should make a subclass of	*
 * this type that uses the data pointer to implement these operations.	*
 *   This object supports clicking					*
 * and double-clicking, and also keeps track of its own predecessors 	*
 * and successors (supetypes and subtypes). Since it's a subclass of	*
 * the InterViews Interactor class, though, it it its own X window.	*
 * This means that drawing a large graph can take quite a while. 	*
 * Writing more custom code that just puts boxes in the canvas for	*
 * the whole window and checks for mouse clicks by finding which box	*
 * the mouse is in (a la Macintosh) would make things a hell of a lot	*
 * faster. In other words, if drawing the hierarchy takes a long time,	*
 * it's not my fault, dammit! Blame it on X.				*/

class HierNode : public SquareButton
{
public:
  HierNode(char* name, ButtonState* s, int t,
	   CommandPtr toDo, void* what);
  ~HierNode();
  virtual void Reconfig();
  virtual Coord GetXCoord() = 0;			// Return global X coord of center
  virtual Coord GetTopYCoord() = 0;			// Return top and bottom center Y coords
  virtual Coord GetBotYCoord() = 0;			
  virtual Coord GetYCoord() = 0;			// Return same stuff for vertical nodes
  virtual Coord GetLeftXCoord() = 0;
  virtual Coord GetRightXCoord() = 0;
  virtual void GetPostScript(ostream& to) = 0;		// Return PostScript to print with
  void ClearLists();					// Empty predecessor and successor lists
  float ComputeDownBary();				// Re-compute down barycentre
  float ComputeUpBary();				// Do same for up barycentre
  virtual int GetConnectivity();			// Number of sub & supertypes
  virtual boolean IsInteresting()			// Return if type is "interesting" or not
    {
      return true;
    }
  int GetType()
    {
      return myData;
    }
  char* getName()
    {
      return myName;
    }
  void AddPredecessor(HierNode* addThis)	// Add super type to adjacency list
    {
      predecessors->Append(addThis);
    }
  void AddSuccessor(HierNode* addThis)	// Add subtype
    {
      successors->Append(addThis);
    }
  void RemovePredecessor(HierNode* rmThis)
    {
      predecessors->Remove(rmThis);
    }
  void RemoveSuccessor(HierNode* rmThis)
    {
      successors->Remove(rmThis);
    }
  virtual void SetLevel(int level)		// Set this displayer's level
    {
      whatLevel = level;
    }
  int GetLevel()				// Return it
    {
      return whatLevel;
    }
  void SetPos(int newPos)			// Set the X coordinate of the left side
    {
      whatPos = newPos;
    }
  int GetPos()					// Return it
    {
      return whatPos;
    }
  AdjList* GetPredecessors()			// Return supertypes
    {
      return predecessors;
    }
  AdjList* GetSuccessors()			// Return subtypes
    {
      return successors;
    }
  float GetDownBary()
    {
      return downBary;
    }
  float GetUpBary()
    {
      return upBary;
    }
protected:
  int whatLevel;			// What level it's on
  int whatPos;				// What position in that level
  AdjList* predecessors;		// List of predecessors (parents)
  AdjList* successors;			// List of successors (children)
  int myData;				// The object it represents
  char* myName;				// The object's name
  float downBary;			// Average of positions of successors
  float upBary;				// Average of positions of preds.
};

/* A TypeDisplayer that helps make long edges look better but doesn't do anything. *
 * It just draws a vertical line rather than a box, and doesn't support mouse 	   *
 * clicks.									   */

class FakeDisplayer : public HierNode

{
public:
  FakeDisplayer(int level, int myData, boolean horiz, Coord ext);
  virtual void Reconfig();
  virtual void Redraw(Coord, Coord, Coord, Coord);
  virtual void Handle(Event& anEvent);
  virtual void GetPostScript(ostream& to);
  virtual Coord GetXCoord();
  virtual Coord GetTopYCoord();
  virtual Coord GetBotYCoord();
  virtual Coord GetYCoord();
  virtual Coord GetLeftXCoord();
  virtual Coord GetRightXCoord();
  virtual int GetWidth()
    {
      return 1;
    }
  virtual int GetConnectivity()
    {
      return FAKE_PRIORITY;
    }
private:
  boolean horizontal;
  Coord extent;
};
  
#endif








