// This may look like C code, but it is really -*- C++ -*-
/* Copyright (c) 1991 by Sanjay Ghemawat */

/*
 * $Revision: 1.4 $
 * $Date: 91/11/04 19:43:55 $
 * $Author: sanjay $
 * $State: Exp $
 *
 * $Log:	List.hP,v $
 * Revision 1.4  91/11/04  19:43:55  sanjay
 * Prepare for release
 * 
 * Revision 1.3  91/09/18  20:33:01  sanjay
 * Added MoveToEnd, MoveToStart, and traversal macros.
 * 
 * Revision 1.2  91/03/28  11:52:12  sanjay
 * Fixed log
 * 
 * Revision 1.1  91/03/28  11:42:42  sanjay
 * Initial deposit
 *
 */

#ifndef _DateItemLISTH
#define _DateItemLISTH

#include "DateItem.h"

/*
 * A DateItemLink has previous and next pointers to other DateItemLinks,
 * and has a boolean that indicates whether it is read cell
 * in the list, or just a dummy header node.
 *
 * A DateItemList is basically a DateItemLink set up so that the boolean
 * indicates that it is a dummy node.
 *
 * A DateItemCell is also a DateItemLink, but has a DateItem field.
 */

class DateItemLink;
class DateItemCell;
class DateItemList;

class DateItemLink {
  public:
    DateItemCell* Next();		/* Obtain next/prev cells.  If next/prev */
    DateItemCell* Prev();		/* field points to a dummy node, return 0 */
  protected:
    DateItemLink(int);		/* Only DateItemCell and DateItemList should */
    virtual ~DateItemLink();		/* create/destroy DateItemLinks */

    void LinkAfter(DateItemCell*);	/* Append the cell after this */

    DateItemLink* next;		/* Next link */
    DateItemLink* prev;		/* Previous link */
    int	     isCell;		/* Is this link a DateItemCell? */

    friend class DateItemCell;
    friend class DateItemList;
};

/*
 * A DateItemCell is a DateItemLink that actually contains an element.
 */
class DateItemCell : public DateItemLink {
  public:
    DateItem& Element();		/* Get cell element */

    DateItemCell* AddBefore(DateItem&);	/* Add a new cell before this */
    DateItemCell* AddAfter(DateItem&);	/* Add a new cella fter this */
  protected:
    DateItemCell(DateItem&);		/* General users should not create DateItemCell */
    virtual ~DateItemCell();

    DateItem element;		/* Cell contents */

    friend class DateItemList;
};

/*
 * A DateItemList is a DateItemLink that does not contain any elements.
 */
class DateItemList : private DateItemLink {
  public:
    DateItemList();			/* Create empty list */
    virtual ~DateItemList();		/* Destroy list */

    DateItemCell* Prepend(DateItem&);	/* Add new element at start */
    DateItemCell* Append(DateItem&);	/* Add new element at end */

    void Remove(DateItemCell*);	/* Remove a cell */
    void MoveToEnd(DateItemCell*);	/* Move to end of list */
    void MoveToStart(DateItemCell*);	/* Move to start of list */

    DateItemCell* First();		/* Return first element - NULL if none */
    DateItemCell* Last();		/* Return last element - NULL if none */

    int Empty();		/* Is list empty */
};

/*
 * List traversal macros.
 *
 * Usage -
 *	DateItemCell* cell;
 *	DateItemList* list;
 *
 *	ForListElements(cell, list) { ... }
 *	ForReverseListElements(cell, list) { ... }
 */
#ifndef ForListElements
#define ForListElements(c,l)	    for (c = (l)->First(); c; c = c->Next())
#define ForReverseListElements(c,l) for (c = (l)->Last();  c; c = c->Prev())
#endif

inline DateItemLink::DateItemLink(int thisIsACell) {
    isCell = thisIsACell;
}

inline DateItemLink::~DateItemLink() { }

inline DateItemCell* DateItemLink::Next() {
    return (next->isCell ? ((DateItemCell*) next) : 0);
}

inline DateItemCell* DateItemLink::Prev() {
    return (prev->isCell ? ((DateItemCell*) prev) : 0);
}

inline void DateItemLink::LinkAfter(DateItemCell* newCell) {
    newCell->next = next;
    newCell->prev = this;
    next->prev = newCell;
    this->next = newCell;
}

inline DateItemCell::DateItemCell(DateItem& e) : DateItemLink(1) {
    element = e;
}

inline DateItemCell::~DateItemCell() {
    prev->next = next;
    next->prev = prev;
}

inline DateItem& DateItemCell::Element() {
    return element;
}

inline DateItemCell* DateItemCell::AddBefore(DateItem& e) {
    DateItemCell* newCell = new DateItemCell(e);
    prev->LinkAfter(newCell);
    return newCell;
}

inline DateItemCell* DateItemCell::AddAfter(DateItem& e) {
    DateItemCell* newCell = new DateItemCell(e);
    this->LinkAfter(newCell);
    return newCell;
}

inline DateItemList::DateItemList() : DateItemLink(0) {
    prev = this;
    next = this;
}

inline DateItemCell* DateItemList::Prepend(DateItem& e) {
    DateItemCell* newCell = new DateItemCell(e);
    this->LinkAfter(newCell);
    return newCell;
}

inline DateItemCell* DateItemList::Append(DateItem& e) {
    DateItemCell* newCell = new DateItemCell(e);
    prev->LinkAfter(newCell);
    return newCell;
}

inline void DateItemList::Remove(DateItemCell* cell) {
    delete cell;
}

inline void DateItemList::MoveToEnd(DateItemCell* cell) {
    cell->prev->next = cell->next;
    cell->next->prev = cell->prev;
    prev->LinkAfter(cell);
}

inline void DateItemList::MoveToStart(DateItemCell* cell) {
    cell->prev->next = cell->next;
    cell->next->prev = cell->prev;
    this->LinkAfter(cell);
}

inline DateItemCell* DateItemList::First() {
    return Next();
}

inline DateItemCell* DateItemList::Last() {
    return Prev();
}

inline int DateItemList::Empty() {
    return prev == this;
}

#endif /* _DateItemLISTH */
