// This may look like C code, but it is really -*- C++ -*-
/* Copyright (c) 1991 by Sanjay Ghemawat */

/*
 * $Revision: 1.4 $
 * $Date: 91/11/04 19:43:55 $
 * $Author: sanjay $
 * $State: Exp $
 *
 * $Log:	List.hP,v $
 * Revision 1.4  91/11/04  19:43:55  sanjay
 * Prepare for release
 * 
 * Revision 1.3  91/09/18  20:33:01  sanjay
 * Added MoveToEnd, MoveToStart, and traversal macros.
 * 
 * Revision 1.2  91/03/28  11:52:12  sanjay
 * Fixed log
 * 
 * Revision 1.1  91/03/28  11:42:42  sanjay
 * Initial deposit
 *
 */

#ifndef _ItemPLISTH
#define _ItemPLISTH

#include "ItemP.h"

/*
 * A ItemPLink has previous and next pointers to other ItemPLinks,
 * and has a boolean that indicates whether it is read cell
 * in the list, or just a dummy header node.
 *
 * A ItemPList is basically a ItemPLink set up so that the boolean
 * indicates that it is a dummy node.
 *
 * A ItemPCell is also a ItemPLink, but has a ItemP field.
 */

class ItemPLink;
class ItemPCell;
class ItemPList;

class ItemPLink {
  public:
    ItemPCell* Next();		/* Obtain next/prev cells.  If next/prev */
    ItemPCell* Prev();		/* field points to a dummy node, return 0 */
  protected:
    ItemPLink(int);		/* Only ItemPCell and ItemPList should */
    virtual ~ItemPLink();		/* create/destroy ItemPLinks */

    void LinkAfter(ItemPCell*);	/* Append the cell after this */

    ItemPLink* next;		/* Next link */
    ItemPLink* prev;		/* Previous link */
    int	     isCell;		/* Is this link a ItemPCell? */

    friend class ItemPCell;
    friend class ItemPList;
};

/*
 * A ItemPCell is a ItemPLink that actually contains an element.
 */
class ItemPCell : public ItemPLink {
  public:
    ItemP& Element();		/* Get cell element */

    ItemPCell* AddBefore(ItemP&);	/* Add a new cell before this */
    ItemPCell* AddAfter(ItemP&);	/* Add a new cella fter this */
  protected:
    ItemPCell(ItemP&);		/* General users should not create ItemPCell */
    virtual ~ItemPCell();

    ItemP element;		/* Cell contents */

    friend class ItemPList;
};

/*
 * A ItemPList is a ItemPLink that does not contain any elements.
 */
class ItemPList : private ItemPLink {
  public:
    ItemPList();			/* Create empty list */
    virtual ~ItemPList();		/* Destroy list */

    ItemPCell* Prepend(ItemP&);	/* Add new element at start */
    ItemPCell* Append(ItemP&);	/* Add new element at end */

    void Remove(ItemPCell*);	/* Remove a cell */
    void MoveToEnd(ItemPCell*);	/* Move to end of list */
    void MoveToStart(ItemPCell*);	/* Move to start of list */

    ItemPCell* First();		/* Return first element - NULL if none */
    ItemPCell* Last();		/* Return last element - NULL if none */

    int Empty();		/* Is list empty */
};

/*
 * List traversal macros.
 *
 * Usage -
 *	ItemPCell* cell;
 *	ItemPList* list;
 *
 *	ForListElements(cell, list) { ... }
 *	ForReverseListElements(cell, list) { ... }
 */
#ifndef ForListElements
#define ForListElements(c,l)	    for (c = (l)->First(); c; c = c->Next())
#define ForReverseListElements(c,l) for (c = (l)->Last();  c; c = c->Prev())
#endif

inline ItemPLink::ItemPLink(int thisIsACell) {
    isCell = thisIsACell;
}

inline ItemPLink::~ItemPLink() { }

inline ItemPCell* ItemPLink::Next() {
    return (next->isCell ? ((ItemPCell*) next) : 0);
}

inline ItemPCell* ItemPLink::Prev() {
    return (prev->isCell ? ((ItemPCell*) prev) : 0);
}

inline void ItemPLink::LinkAfter(ItemPCell* newCell) {
    newCell->next = next;
    newCell->prev = this;
    next->prev = newCell;
    this->next = newCell;
}

inline ItemPCell::ItemPCell(ItemP& e) : ItemPLink(1) {
    element = e;
}

inline ItemPCell::~ItemPCell() {
    prev->next = next;
    next->prev = prev;
}

inline ItemP& ItemPCell::Element() {
    return element;
}

inline ItemPCell* ItemPCell::AddBefore(ItemP& e) {
    ItemPCell* newCell = new ItemPCell(e);
    prev->LinkAfter(newCell);
    return newCell;
}

inline ItemPCell* ItemPCell::AddAfter(ItemP& e) {
    ItemPCell* newCell = new ItemPCell(e);
    this->LinkAfter(newCell);
    return newCell;
}

inline ItemPList::ItemPList() : ItemPLink(0) {
    prev = this;
    next = this;
}

inline ItemPCell* ItemPList::Prepend(ItemP& e) {
    ItemPCell* newCell = new ItemPCell(e);
    this->LinkAfter(newCell);
    return newCell;
}

inline ItemPCell* ItemPList::Append(ItemP& e) {
    ItemPCell* newCell = new ItemPCell(e);
    prev->LinkAfter(newCell);
    return newCell;
}

inline void ItemPList::Remove(ItemPCell* cell) {
    delete cell;
}

inline void ItemPList::MoveToEnd(ItemPCell* cell) {
    cell->prev->next = cell->next;
    cell->next->prev = cell->prev;
    prev->LinkAfter(cell);
}

inline void ItemPList::MoveToStart(ItemPCell* cell) {
    cell->prev->next = cell->next;
    cell->next->prev = cell->prev;
    this->LinkAfter(cell);
}

inline ItemPCell* ItemPList::First() {
    return Next();
}

inline ItemPCell* ItemPList::Last() {
    return Prev();
}

inline int ItemPList::Empty() {
    return prev == this;
}

#endif /* _ItemPLISTH */
