// This may look like C code, but it is really -*- C++ -*-
/* Copyright (c) 1991 by Sanjay Ghemawat */

/*
 * $Revision: 1.4 $
 * $Date: 91/11/04 19:43:55 $
 * $Author: sanjay $
 * $State: Exp $
 *
 * $Log:	List.hP,v $
 * Revision 1.4  91/11/04  19:43:55  sanjay
 * Prepare for release
 * 
 * Revision 1.3  91/09/18  20:33:01  sanjay
 * Added MoveToEnd, MoveToStart, and traversal macros.
 * 
 * Revision 1.2  91/03/28  11:52:12  sanjay
 * Fixed log
 * 
 * Revision 1.1  91/03/28  11:42:42  sanjay
 * Initial deposit
 *
 */

#ifndef _NoteInfoPLISTH
#define _NoteInfoPLISTH

#include "NoteInfoP.h"

/*
 * A NoteInfoPLink has previous and next pointers to other NoteInfoPLinks,
 * and has a boolean that indicates whether it is read cell
 * in the list, or just a dummy header node.
 *
 * A NoteInfoPList is basically a NoteInfoPLink set up so that the boolean
 * indicates that it is a dummy node.
 *
 * A NoteInfoPCell is also a NoteInfoPLink, but has a NoteInfoP field.
 */

class NoteInfoPLink;
class NoteInfoPCell;
class NoteInfoPList;

class NoteInfoPLink {
  public:
    NoteInfoPCell* Next();		/* Obtain next/prev cells.  If next/prev */
    NoteInfoPCell* Prev();		/* field points to a dummy node, return 0 */
  protected:
    NoteInfoPLink(int);		/* Only NoteInfoPCell and NoteInfoPList should */
    virtual ~NoteInfoPLink();		/* create/destroy NoteInfoPLinks */

    void LinkAfter(NoteInfoPCell*);	/* Append the cell after this */

    NoteInfoPLink* next;		/* Next link */
    NoteInfoPLink* prev;		/* Previous link */
    int	     isCell;		/* Is this link a NoteInfoPCell? */

    friend class NoteInfoPCell;
    friend class NoteInfoPList;
};

/*
 * A NoteInfoPCell is a NoteInfoPLink that actually contains an element.
 */
class NoteInfoPCell : public NoteInfoPLink {
  public:
    NoteInfoP& Element();		/* Get cell element */

    NoteInfoPCell* AddBefore(NoteInfoP&);	/* Add a new cell before this */
    NoteInfoPCell* AddAfter(NoteInfoP&);	/* Add a new cella fter this */
  protected:
    NoteInfoPCell(NoteInfoP&);		/* General users should not create NoteInfoPCell */
    virtual ~NoteInfoPCell();

    NoteInfoP element;		/* Cell contents */

    friend class NoteInfoPList;
};

/*
 * A NoteInfoPList is a NoteInfoPLink that does not contain any elements.
 */
class NoteInfoPList : private NoteInfoPLink {
  public:
    NoteInfoPList();			/* Create empty list */
    virtual ~NoteInfoPList();		/* Destroy list */

    NoteInfoPCell* Prepend(NoteInfoP&);	/* Add new element at start */
    NoteInfoPCell* Append(NoteInfoP&);	/* Add new element at end */

    void Remove(NoteInfoPCell*);	/* Remove a cell */
    void MoveToEnd(NoteInfoPCell*);	/* Move to end of list */
    void MoveToStart(NoteInfoPCell*);	/* Move to start of list */

    NoteInfoPCell* First();		/* Return first element - NULL if none */
    NoteInfoPCell* Last();		/* Return last element - NULL if none */

    int Empty();		/* Is list empty */
};

/*
 * List traversal macros.
 *
 * Usage -
 *	NoteInfoPCell* cell;
 *	NoteInfoPList* list;
 *
 *	ForListElements(cell, list) { ... }
 *	ForReverseListElements(cell, list) { ... }
 */
#ifndef ForListElements
#define ForListElements(c,l)	    for (c = (l)->First(); c; c = c->Next())
#define ForReverseListElements(c,l) for (c = (l)->Last();  c; c = c->Prev())
#endif

inline NoteInfoPLink::NoteInfoPLink(int thisIsACell) {
    isCell = thisIsACell;
}

inline NoteInfoPLink::~NoteInfoPLink() { }

inline NoteInfoPCell* NoteInfoPLink::Next() {
    return (next->isCell ? ((NoteInfoPCell*) next) : 0);
}

inline NoteInfoPCell* NoteInfoPLink::Prev() {
    return (prev->isCell ? ((NoteInfoPCell*) prev) : 0);
}

inline void NoteInfoPLink::LinkAfter(NoteInfoPCell* newCell) {
    newCell->next = next;
    newCell->prev = this;
    next->prev = newCell;
    this->next = newCell;
}

inline NoteInfoPCell::NoteInfoPCell(NoteInfoP& e) : NoteInfoPLink(1) {
    element = e;
}

inline NoteInfoPCell::~NoteInfoPCell() {
    prev->next = next;
    next->prev = prev;
}

inline NoteInfoP& NoteInfoPCell::Element() {
    return element;
}

inline NoteInfoPCell* NoteInfoPCell::AddBefore(NoteInfoP& e) {
    NoteInfoPCell* newCell = new NoteInfoPCell(e);
    prev->LinkAfter(newCell);
    return newCell;
}

inline NoteInfoPCell* NoteInfoPCell::AddAfter(NoteInfoP& e) {
    NoteInfoPCell* newCell = new NoteInfoPCell(e);
    this->LinkAfter(newCell);
    return newCell;
}

inline NoteInfoPList::NoteInfoPList() : NoteInfoPLink(0) {
    prev = this;
    next = this;
}

inline NoteInfoPCell* NoteInfoPList::Prepend(NoteInfoP& e) {
    NoteInfoPCell* newCell = new NoteInfoPCell(e);
    this->LinkAfter(newCell);
    return newCell;
}

inline NoteInfoPCell* NoteInfoPList::Append(NoteInfoP& e) {
    NoteInfoPCell* newCell = new NoteInfoPCell(e);
    prev->LinkAfter(newCell);
    return newCell;
}

inline void NoteInfoPList::Remove(NoteInfoPCell* cell) {
    delete cell;
}

inline void NoteInfoPList::MoveToEnd(NoteInfoPCell* cell) {
    cell->prev->next = cell->next;
    cell->next->prev = cell->prev;
    prev->LinkAfter(cell);
}

inline void NoteInfoPList::MoveToStart(NoteInfoPCell* cell) {
    cell->prev->next = cell->next;
    cell->next->prev = cell->prev;
    this->LinkAfter(cell);
}

inline NoteInfoPCell* NoteInfoPList::First() {
    return Next();
}

inline NoteInfoPCell* NoteInfoPList::Last() {
    return Prev();
}

inline int NoteInfoPList::Empty() {
    return prev == this;
}

#endif /* _NoteInfoPLISTH */
