typedef void* Pointer; /* Automatic prepend ()*^*@# */
// This may look like C code, but it is really -*- C++ -*-
/* Copyright (c) 1991 by Sanjay Ghemawat */

#ifndef _PointerARRAYH
#define _PointerARRAYH

/*
 * A resizable array of Pointers.
 *
 * If Pointer is user-defined, it must have the following operations -
 *
 * class Pointer ... {
 *   public:
 *     Pointer();
 *     ~Pointer();
 *     ...
 *     Pointer& operator = (Pointer&);
 * };
 *
 * If UNSAFE_ARRAY not defined when including PointerArray.h, then array bounds
 * and argument range checking will be done.
 *
 */

/*
 * $Revision: 1.3 $
 * $Date: 91/11/28 16:23:54 $
 * $Author: sanjay $
 * $State: Exp $
 *
 * $Log:	Array.hP,v $
 * Revision 1.3  91/11/28  16:23:54  sanjay
 * CFRONT compatibility
 * 
 * Revision 1.2  91/11/04  19:43:48  sanjay
 * Prepare for release
 * 
 * Revision 1.1  91/06/28  20:19:50  sanjay
 * Initial revision
 * 
 */

class PointerArray {
  public:

    /* Constructors */
    PointerArray();				/* Empty array */
    PointerArray(int predict);		/* Empty array with size prediction */
    PointerArray(const Pointer*, int);		/* Initialized with C array */
    PointerArray(const PointerArray&);		/* Initialized with another Array */
    PointerArray(Pointer, int);			/* Fill with n copies of element */

    /* Destructor */
    ~PointerArray();

    PointerArray& operator=(const PointerArray&);	/* Assignment operator */

    /* Array indexing */
    Pointer& operator[](int index) const;
    Pointer& Slot(int index) const;

    /* Other Array operations */
    int  Size() const;			/* Return size; */
    Pointer& High() const;			/* Return last element */
    Pointer* AsPointer() const;		/* Return as pointer to base */
    void Append(Pointer v);			/* Append an element */
    void Append(Pointer, int n);		/* Append n copies of element */
    void Concat(const Pointer*, int);	/* Concatenate C array */
    void Concat(const PointerArray&);	/* Concatenate another Array */
    void Fill(Pointer,int start,int count);	/* Fill specified region with Pointer */
    Pointer  Remove();			/* Remove and return last element */
    void Remove(int num);		/* Remove last num elements */
    void Clear();			/* Remove all elements */

    /* Storage stuff */
    void Reclaim();			/* Reclaim all unused space */
  protected:
    /* Error handlers */
    void BoundError(int) const;
    void RangeError(int) const;

    void BoundCheck(int) const;
    void RangeCheck(int) const;
    void SizeCheck (int) const;

  private:
    Pointer*	store;			/* Actual storage */
    int		alloc;			/* Size of allocated storage */
    int		size;			/* Size of used storage */

    /* Storage enlargers */
    void EnlargeAllocationTo(int s);	/* Enlarge storage to s */
    void EnlargeTo(int s);		/* Enlarge storage to s if necessary */
};

inline void PointerArray::BoundCheck(int val) const {
#ifndef UNSAFE_ARRAY
    if ((val < 0) || (val >= size)) BoundError(val);
#endif
}

inline void PointerArray::RangeCheck(int val) const {
#ifndef UNSAFE_ARRAY
    if (val < 0) RangeError(val);
#endif
}

inline void PointerArray::SizeCheck(int val) const {
#ifndef UNSAFE_ARRAY
    if ((val < 0) || (val > size)) BoundError(val);
#endif
}

inline PointerArray::PointerArray() {
    alloc = 0;
    size  = 0;
}

inline int PointerArray::Size() const {
    return size;
}

inline Pointer& PointerArray::operator[](int index) const {
    BoundCheck(index);
    return store[index];
}

inline Pointer& PointerArray::Slot(int index) const {
    BoundCheck(index);
    return store[index];
}

inline Pointer& PointerArray::High() const {
    return store[size-1];
}

inline Pointer* PointerArray::AsPointer() const {
    return store;
}

inline void PointerArray::Append(Pointer v) {
    if (size >= alloc)
	EnlargeAllocationTo(size+1);
    store[size++] = v;
}

inline Pointer PointerArray::Remove() {
    if (size > 0) size--;
    return store[size];
}

inline void PointerArray::Remove(int num) {
    SizeCheck(num);
    size -= num;
}

inline void PointerArray::Clear() {
    size = 0;
}

#endif /*_PointerARRAYH */
