// This may look like C code, but it is really -*- C++ -*-
/* Copyright (c) 1991 by Sanjay Ghemawat */

#ifndef TIMEH
#define TIMEH

/*
 * $Revision: 1.2 $
 * $Date: 91/11/04 19:44:01 $
 * $Author: sanjay $
 * $State: Exp $
 *
 * $Log:	Time.h,v $
 * Revision 1.2  91/11/04  19:44:01  sanjay
 * Prepare for release
 * 
 * Revision 1.1  91/06/28  20:07:11  sanjay
 * Initial revision
 * 
 */

class Time;
class Duration;
class WeekDay;
class Month;

/*
 * A class to represent system time.
 */
class Time {
  public:
    /*
     * Constructors and assignments.
     */
    Time();			/* Unspecified time */
    Time(Time const&);		/* Copy other time */

    Time& operator=(Time const&);
    static Time Now();

    /* Addition and subtraction */
    friend Time			operator+ (Time, Duration);
    friend Time			operator- (Time, Duration);
    friend Duration		operator- (Time, Time);
    Time&			operator+=(Duration);
    Time&			operator-=(Duration);

    /* Comparisons */
    friend int operator == (Time, Time);
    friend int operator != (Time, Time);
    friend int operator <  (Time, Time);
    friend int operator <= (Time, Time);
    friend int operator >  (Time, Time);
    friend int operator >= (Time, Time);

    /*
     * Various conversions.
     */

    /* Variables for default references */
    static int junkInt;

    void BreakDownDate(int&	mday,
		       WeekDay&	wday,
		       Month&	month,
		       int&	year);

    void BreakDownClock(int& hour     = junkInt,
			int& minute   = junkInt,
			int& second   = junkInt,
			int& millisec = junkInt);

    void BreakDown(int&	    mday,
		   WeekDay& wday,
		   Month&   month,
		   int&	    year,
		   int&     hour     = junkInt,
		   int&	    minute   = junkInt,
		   int&	    second   = junkInt,
		   int&	    millisec = junkInt);

    /* UN*X specific */
    /* Copy UN*X time structure */
    Time(struct timeval const&);
    void Convert(struct timeval&);
  private:
    /*
     * Rep is the number of seconds since some epoch.
     */
    double rep;

    /* Create time with supplied rep */
    Time(double);

    /* Time initialization stuff. */
    static int	  initialized;
    static double offset;

    static void Initialize();
};

/*
 * A class to represent a time interval.
 */
class Duration {
  public:
    /* Constructors and assignments */
    Duration();
    Duration(Duration const&);
    Duration& operator=(Duration const&);

    /* Special durations */
    static Duration Week();
    static Duration Day();
    static Duration Hour();
    static Duration Minute();
    static Duration Second();
    static Duration MilliSecond();
    static Duration MicroSecond();

    /* Addition and subtraction */
    friend Time			operator+ (Time, Duration);
    friend Time			operator- (Time, Duration);
    friend Duration		operator- (Time, Time);

    friend Duration		operator+ (Duration, Duration);
    friend Duration		operator- (Duration, Duration);

    /* Multiplication and division */
    friend Duration		operator* (Duration, int);
    friend Duration		operator* (Duration, double);
    friend Duration		operator/ (Duration, int);
    friend Duration		operator/ (Duration, double);

    Duration&	operator += (Duration);
    Duration&	operator -= (Duration);
    Duration&	operator *= (int);
    Duration&	operator *= (double);
    Duration&	operator /= (int);
    Duration&	operator /= (double);

    /* Comparisons (ints represent seconds) */
    friend int operator == (Duration, Duration);
    friend int operator != (Duration, Duration);
    friend int operator <  (Duration, Duration);
    friend int operator <= (Duration, Duration);
    friend int operator >  (Duration, Duration);
    friend int operator >= (Duration, Duration);

    /* Conversion */
    double Weeks() const;
    double Days() const;
    double Hours() const;
    double Minutes() const;
    double Seconds() const;
    double MilliSeconds() const;
    double MicroSeconds() const;

    /* UN*X Specific */
    Duration(struct timeval const&);
    void Convert(struct timeval&);
  private:
    /*
     * REP is the number of seconds in the Duration.
     */
    double rep;

    Duration(double);
};

/*
 * Time.
 */
inline Time::Time() { }

inline Time::Time(const Time& src) {
    rep = src.rep;
}

inline Time& Time::operator=(const Time& src) {
    rep = src.rep;
    return *this;
}

inline Time::Time(double r) {
    rep = r;
}

inline int operator == (Time a, Time b) {
    return a.rep == b.rep;
}

inline int operator != (Time a, Time b) {
    return a.rep != b.rep;
}

inline int operator <  (Time a, Time b) {
    return a.rep < b.rep;
}

inline int operator <= (Time a, Time b) {
    return a.rep <= b.rep;
}

inline int operator >  (Time a, Time b) {
    return a.rep > b.rep;
}

inline int operator >= (Time a, Time b) {
    return a.rep >= b.rep;
}

/*
 * Duration.
 */
inline Duration::Duration() { }

inline Duration::Duration(const Duration& src) {
    rep = src.rep;
}

inline Duration& Duration::operator=(const Duration& src) {
    rep = src.rep;
    return *this;
}

inline Duration::Duration(double r) {
    rep = r;
}

inline Duration Duration::Week() {
    return Duration(7.0 * 24.0 * 60.0 * 60.0);
}

inline Duration Duration::Day() {
    return Duration(24.0 * 60.0 * 60.0);
}

inline Duration Duration::Hour() {
    return Duration(60.0 * 60.0);
}

inline Duration Duration::Minute() {
    return Duration(60.0);
}

inline Duration Duration::Second() {
    return Duration(1.0);
}

inline Duration Duration::MilliSecond() {
    return Duration(1e-3);
}

inline Duration Duration::MicroSecond() {
    return Duration(1e-6);
}

inline int operator == (Duration a, Duration b) {
    return a.rep == b.rep;
}

inline int operator != (Duration a, Duration b) {
    return a.rep != b.rep;
}

inline int operator <  (Duration a, Duration b) {
    return a.rep < b.rep;
}

inline int operator <= (Duration a, Duration b) {
    return a.rep <= b.rep;
}

inline int operator >  (Duration a, Duration b) {
    return a.rep > b.rep;
}

inline int operator >= (Duration a, Duration b) {
    return a.rep >= b.rep;
}

inline double Duration::Weeks() const {
    return (rep / (7.0 * 24.0 * 60.0 * 60.0));
}

inline double Duration::Days() const {
    return (rep / (24.0 * 60.0 * 60.0));
}

inline double Duration::Hours() const {
    return (rep / (60.0 * 60.0));
}

inline double Duration::Minutes() const {
    return (rep / (60.0));
}

inline double Duration::Seconds() const {
    return (rep);
}

inline double Duration::MilliSeconds() const {
    return (rep * 1000.0);
}

inline double Duration::MicroSeconds() const {
    return (rep * 1000000.0);
}

/*
 * Arithmetic
 */
inline Time operator + (Time t, Duration d) {
    return Time(t.rep + d.rep);
}

inline Time operator - (Time t, Duration d) {
    return Time(t.rep - d.rep);
}

inline Duration operator - (Time a, Time b) {
    return Duration(a.rep - b.rep);
}

inline Duration operator + (Duration a, Duration b) {
    return Duration(a.rep + b.rep);
}

inline Duration operator - (Duration a, Duration b) {
    return Duration(a.rep - b.rep);
}

inline Time& Time::operator += (Duration d) {
    rep += d.Seconds();
    return *this;
}

inline Time& Time::operator -= (Duration d) {
    rep -= d.Seconds();
    return *this;
}

inline Duration operator * (Duration d, int x) {
    return Duration(d.rep * double(x));
}

inline Duration operator * (Duration d, double x) {
    return Duration(d.rep * x);
}

inline Duration operator / (Duration d, int x) {
    return Duration(d.rep / double(x));
}

inline Duration operator / (Duration d, double x) {
    return Duration(d.rep / x);
}

inline Duration& Duration::operator += (Duration d) {
    rep += d.rep;
    return *this;
}

inline Duration& Duration::operator -= (Duration d) {
    rep -= d.rep;
    return *this;
}

inline Duration& Duration::operator *= (int x) {
    rep *= double(x);
    return *this;
}

inline Duration& Duration::operator *= (double x) {
    rep *= x;
    return *this;
}

inline Duration& Duration::operator /= (int x) {
    rep /= double(x);
    return *this;
}

inline Duration& Duration::operator /= (double x) {
    rep /= x;
    return *this;
}

#endif /*TIMEH */
