// This may look like C code, but it is really -*- C++ -*-
/* Copyright (c) 1991 by Sanjay Ghemawat */

/*
 * $Revision: 1.15 $
 * $Date: 91/11/06 18:01:33 $
 * $Author: sanjay $
 * $State: Exp $
 *
 * $Log:	application.h,v $
 * Revision 1.15  91/11/06  18:01:33  sanjay
 * Do not maintain auto-save files for shared calendars. Auto-save for
 * shared calendars writes directly to the shared files.
 * 
 * Cleaned up other IO handling.
 * 
 * Revision 1.14  91/11/04  19:44:18  sanjay
 * Prepare for release
 * 
 * Revision 1.13  91/09/18  20:33:53  sanjay
 * Use OccurCache for listing items on certain dates.
 * 
 * Revision 1.12  91/09/06  10:52:49  sanjay
 * Made Error and FileError public methods.
 * 
 * Revision 1.11  91/08/07  10:59:12  sanjay
 * Added re-read.
 * 
 * Revision 1.10  91/08/06  09:45:11  sanjay
 * Added editing support for included files.
 * 
 * Revision 1.9  91/04/30  18:20:51  sanjay
 * Add comments
 * 
 * Revision 1.8  91/04/11  10:26:34  sanjay
 * Encapsulated calendar file handling
 * 
 * Revision 1.7  91/04/10  11:33:23  sanjay
 * Added printing
 * 
 * Revision 1.6  91/04/03  21:36:29  sanjay
 * Added range listing
 * 
 * Revision 1.5  91/04/03  20:44:24  sanjay
 * Further reminder listing fixes
 * 
 * Revision 1.4  91/04/02  12:33:01  sanjay
 * Improved auto-save.
 * 
 * Revision 1.3  91/04/01  20:30:32  sanjay
 * Better IO error handling
 * 
 * Revision 1.2  91/04/01  11:56:59  sanjay
 * Added include file capability
 * 
 * Revision 1.1  91/03/28  11:41:08  sanjay
 * Initial deposit
 * 
 */

#ifndef _APPLICATIONH
#define _APPLICATIONH

#include <InterViews/defs.h>
#include "Time.h"

/*
 * Application specific data.
 * There is a single instance of this class.
 */

class Alarmer;
class ApptView;
class CalFile;
class Calendar;
class Closure;
class Date;
class EditRepeat;
class FileMap;
class Interactor;
class Item;
class OccurCache;
class PointerArray;
class RemindMgr;
class Timer;
class TimerAction;
class XWorld;

class Application {
  public:
    /*
     * Constructor/destructor.
     */
    Application(int& argc, char* argv[]);
    ~Application();

    /* Views. */
    void Attach(ApptView*);
    void Detach(ApptView*);
    void SetFocus(ApptView*);
    ApptView* Focus();

    /* Items. */
    void Add(Item*, Calendar*);
    void Delete(Item*);
    void Change(Item*);

    int NumItems();
    Item* GetItem(int);

    /* Generate reminders. */
    void GenerateReminders(Date, Date);
    void PrintMonth(Date);
    void PrintDays(Date, int number, int cols, boolean landscape);

    /* Get/set clipboard item if any. */
    Item* GetClipboard();
    void  SetClipboard(Item*);

    /* World. */
    XWorld* World() { return world; }

    /* Calendar ops. */
    Calendar* GetMainCalendar();
    void AddCalendar(char const*);
    void RemoveCalendar(char const*);
    char const* GetName(Calendar*);
    char const* GetShortName(Calendar*);
    Calendar* GetCalendarByName(char const*);

    /* Save/modify ... */
    boolean IsModified();
    boolean Save();		/* Returns true on success */
    boolean ReRead();		/* Returns true on success */
    void Quit();

    /* Run. */
    void Run();

    /*
     * Cache unused EditRepeat.
     */
    void NewEditRepeat(Item*);
    void DeleteEditRepeat(EditRepeat*);

    /*
     * For debugging purposes only.
     */
    void Flush();
    void Sync();

    /*
     * Number of general purpose views open right now.
     */
    int OpenViews();
    void AddView();
    void RemoveView();

    /*
     * Application timer.
     */
    Timer* GetTimer();

    /*
     * Reminder manager.
     */
    RemindMgr* GetRemindMgr();

    /*
     * Item occurrence cache.
     */
    OccurCache* GetOccurCache();

    /*
     * Print command.
     */
    const char* GetPrintCommand();
    void SetPrintCommand(const char*);

    /*
     * Storage for last selected file name.
     */
    void SetSelectedFile(char const*);
    char const* GetLastSelectedDir();

    /* Popup an error message */
    void Error(const char*);

    /* Popup an error message related to specified file */
    void FileError(char const* file, char const* msg);
  protected:
    XWorld*	world;
    Item*	clipboard;
    ApptView*	focus;

    PointerArray& views;
    PointerArray& editRepeats;

    /* Top-level calendar and file information */
    CalFile* calFile;

    /* Calendar map */
    FileMap* calendarMap;

    /* Make a calendar */
    CalFile* MakeCalendar(char const* name, boolean onTTY);

    /* Add calendar items to all views */
    void AddCalendarItems(Calendar*);

    /* Remove calendar items from all views */
    void RemoveCalendarItems(Calendar*);

    /* Has calendar been modified? */
    void Modified(CalFile*);

    /* Auto-save parameters and operations*/
    int autoSaveLevel;		/* Auto-save after this many changes */
    int autoSaveDelay;		/* Auto-save if idle for this long */	

    void AutoSave();		/* Performs auto-save */
    void AutoSaveTimeout();	/* Called to auto-save on timeout */
    Closure* autoSaver;		/* Closure for timer */
    TimerAction* autoSaveHook;	/* Timer action */
    int modCount;		/* Number of changes since last autosave */

    /* Number of open views. Last view cannot be closed */
    int openViews;

    RemindMgr* remindMgr;	/* Managers displayed reminders */
    Alarmer* alarmer;		/* Generates reminders at appropriate times */
    OccurCache* occurCache;	/* Item occurrence cache */

    /* Print command */
    const char* printCommand;

    /* Last selected directory */
    char const* lastSelectedDir;

    void Usage();		/* Print usage message */
};

/*
 * Simple inline operations.
 */

inline int Application::OpenViews() {
    return openViews;
}

inline void Application::AddView() {
    openViews++;
}

inline void Application::RemoveView() {
    openViews--;
}

inline RemindMgr* Application::GetRemindMgr() {
    return remindMgr;
}

inline OccurCache* Application::GetOccurCache() {
    return occurCache;
}

#endif /* _APPLICATIONH */
