// This may look like C code, but it is really -*- C++ -*-
/* Copyright (c) 1991 by Sanjay Ghemawat */

/*
 * $Revision: 1.2 $
 * $Date: 91/11/04 19:44:26 $
 * $Author: sanjay $
 * $State: Exp $
 *
 * $Log:	arglist.h,v $
 * Revision 1.2  91/11/04  19:44:26  sanjay
 * Prepare for release
 * 
 * Revision 1.1  91/03/28  11:41:11  sanjay
 * Initial deposit
 * 
 */

#ifndef _ARGLISTH
#define _ARGLISTH

/*
 * Class for parsing argument lines.
 *
 * Usage:
 * The client of this client creates an ArgList object passing it the
 * command line arguments.  The client can also modify an associated
 * table that maps from strings to closures.  Then the client can call
 * the Parse method to do the actual parse.
 *
 * Model:
 * The object maintains the argument list, and an index into the
 * list.  For each argument in the list, the parser searches for
 * an associated closure in the table supplied by the client.  If
 * such a closure is found, the parser invokes it and then continues
 * scanning after advancing the index once.  If no such closure is
 * found in the table, the parser invokes a special closure (again
 * supplied by the user).
 *
 * Notes:
 * - Options can be added by calling either AddOption() or
 *   AddPrefixOption().
 *
 * - Options added later take precedence over options added
 *   earlier.
 *
 * - These closures should usually do the appropriate processing and
 *   return.
 *
 * - The argument at a certain distance from the current index can be
 *   invoked by calling Fetch(int distance = 0).  A distance of zero
 *   returns the current argument (the one which was used in the table
 *   lookup).  It is an error to try and fetch from beyond the end of
 *   the argument list.
 *
 * - A closure can invoke Advance(int k = 1) if it wishes to skip over 
 *   the next k arguments.
 *
 * - Multiple Advance() calls accumulate.
 *
 * - Calling Advance() does not change the current index immediately.
 *   The change only happens once the closure returns.
 *
 * - Remaining() returns the number of arguments that remain beyond the
 *   current one.
 *
 * - The closure can call AddOption() and AddPrefixOption() to add
 *   new options to the option table.  These new options take effect
 *   when the closure returns.
 *
 * - A closure should not call Parse().
 */

class Closure;
struct ArgListOptionArray;

class ArgList {
  public:
    ArgList(int argc, char* argv[], Closure* defaulClosure);
    virtual ~ArgList();

    /*
     * Table modifications.
     *
     * AddOption       - Call closure if arg == s.
     * AddPrefixOption - Call closure if s is a prefix of arg.
     */
    void AddOption(const char* s, Closure*);
    void AddPrefixOption(const char* s, Closure*);

    void Parse();

    /*
     * Callback interface.
     */
    char* Fetch(int distance = 0);
    void Advance(int k = 1);
    int  Remaining();
  protected:
    char**   args;			/* Argument list */
    int	     argc;			/* List extends from 0..argc-1 */
    int      index;			/* Current index */
    int      pending_advance;		/* Sum of pending advances */
    Closure* defaultClosure;		/* Closure to call on no-match */

    ArgListOptionArray& options;	/* Options - in order of addition */
};

#endif /* _ARGLISTH */
