/* Copyright (c) 1991 by Sanjay Ghemawat */
/*
 * $Revision: 1.9 $
 * $Date: 91/11/28 16:24:22 $
 * $Author: sanjay $
 * $State: Exp $
 *
 * $Log:	calendar.cc,v $
 * Revision 1.9  91/11/28  16:24:22  sanjay
 * CFRONT compatibility
 * 
 * Revision 1.8  91/11/04  19:44:32  sanjay
 * Prepare for release
 * 
 * Revision 1.7  91/11/04  13:27:13  sanjay
 * Use version.h to get version number
 * 
 * Revision 1.6  91/11/04  11:50:36  sanjay
 * Print appropriate error message when file cannot be read.
 * 
 * Revision 1.5  91/08/06  09:48:21  sanjay
 * Calendar does not read included files. It just stores their names.
 * 
 * Revision 1.4  91/06/28  20:09:34  sanjay
 * Moved Array and Time classes into hierarchy.
 * 
 * Revision 1.3  91/04/01  11:57:01  sanjay
 * Added include file capability
 * 
 * Revision 1.2  91/03/28  11:51:46  sanjay
 * Fixed log
 * 
 * Revision 1.1  91/03/28  11:41:58  sanjay
 * Initial deposit
 *
 */

#include <string.h>
#include <stdio.h>

#include "calendar.h"
#include "item.h"
#include "lexer.h"
#include "version.h"

#include "Pointer.Array.h"

static const char opener = '[';
static const char closer = ']';

Calendar::Calendar()
     : items(*(new PointerArray)),
       includes(*(new PointerArray))
{
    readonly = false;
}

Calendar::~Calendar() {
    int i;

    for (i = 0; i < items.Size(); i++) {
	Item* item = (Item*) items[i];
	delete item;
    }
    delete &items;

    for (i = 0; i < includes.Size(); i++) {
	char* includeName = (char*) includes[i];
	delete includeName;
    }
    delete &includes;
}

void Calendar::Add(Item* item) {
    item->SetCalendar(this);
    items.Append((Pointer*) item);
}

void Calendar::Remove(Item* item) {
    for (int i = 0; i < items.Size(); i++) {
	if (items[i] == (Pointer*) item) {
	    /* Found it */
	    delete item;

	    /* Shift the other items over */
	    for (int j = i + 1; j < items.Size(); j++)
		items[j - 1] = items[j];

	    items.Remove();
	    break;
	}
    }
}

boolean Calendar::Read(Lexer* lex) {
    for (int i = 0; i < items.Size(); i++) {
	Item* item = (Item*) items[i];
	delete item;
    }
    items.Clear();

    if (lex->Status() == Lexer::Error) {
	/* No input file at all */
	return true;
    }

    int major, minor;
    const char* modifier;

    if (! lex->SkipWS() ||
	! lex->Skip("Calendar") ||
	! lex->SkipWS()) {
	lex->SetError("file does not contain calendar");
	return false;
    }

    if (! lex->Skip(opener) ||
        ! lex->Skip('v') ||

	! lex->GetNumber(major) ||
	! (major >= 1) ||

	! lex->Skip('.') ||

        ! lex->GetNumber(minor) ||
	! (minor >= 0) ||

	! lex->GetUntil(closer, modifier) ||
	/* Possibly check modifier here */

	! lex->Skip(closer)) {
	lex->SetError("illegal version");
    }

    while (true) {
	char c;

	lex->SkipWS();
	lex->Peek(c);

	switch (lex->Status()) {
	  case Lexer::Eof:
	    return true;
	  case Lexer::Error:
	    return false;
	  default:
	    break;
	}

	char const* keyword;

	if (! lex->GetId(keyword) ||
	    ! lex->SkipWS() ||
	    ! lex->Skip(opener)) {
	    lex->SetError("error reading item header");
	    return false;
	}

	if (strcmp(keyword, "Appt") == 0) {
	    Item* item = new Appointment(this);
	    if (! item->Read(lex)) {
		delete item;
		return false;
	    }
	    Add(item);
	}
	else if (strcmp(keyword, "Note") == 0) {
	    Item* item = new Notice(this);
	    if (! item->Read(lex)) {
		delete item;
		return false;
	    }
	    Add(item);
	}
	else if (strcmp(keyword, "Include") == 0) {
	    /* Read the name */
	    int len;

	    if (! lex->SkipWS() ||
		! lex->GetNumber(len) ||
		! lex->SkipWS() ||
		! lex->Skip(opener)) {
		lex->SetError("error reading included file name");
		return false;
	    }

	    char* name = new char[len + 1];
	    if (! lex->GetText(name, len) ||
		! lex->Skip(closer)) {
		delete name;
		lex->SetError("error reading included file name");
		return false;
	    }
	    name[len] = '\0';
	    Include(name);
	    delete name;
	}
	else {
	    lex->SetError("unrecognized keyword");
	    return false;
	}

	if (! lex->SkipWS() ||
	    ! lex->Skip(closer)) {
	    lex->SetError("incomplete item");
	    return false;
	}
    }
}

void Calendar::Write(FILE* out) const {
    fprintf(out, "Calendar [v%s]\n", Version);

    for (int i = 0; i < includes.Size(); i++) {
	char const* name = (char const*) includes[i];
	fprintf(out, "Include [%d [%s]]\n", strlen(name), name);
    }

    for (i = 0; i < items.Size(); i++) {
	Item* item = (Item*) items[i];

	if (item->AsNotice() != nil) {
	    fprintf(out, "Note [\n");
	}
	else {
	    fprintf(out, "Appt [\n");
	}
	item->Write(out);
	fprintf(out, "]\n");
    }
}

int Calendar::Size() const {
    return items.Size();
}

void Calendar::Include(char const* name) {
    char* name_copy = new char[strlen(name) + 1];
    strcpy(name_copy, name);

    includes.Append((Pointer) name_copy);
}

void Calendar::Exclude(char const* name) {
    for (int i = 0; i < includes.Size(); i++) {
	if (strcmp(name, ((char const*) includes[i])) == 0) {
	    char* includeName = (char*) includes[i];
	    delete includeName;

	    /* Shift other includes over */
	    for (int j = i + 1; j < includes.Size(); j++) {
		includes[j - 1] = includes[j];
	    }
	    includes.Remove();
	    return;
	}
    }
}

int Calendar::NumIncludes() const {
    return includes.Size();
}

char const* Calendar::GetInclude(int i) const {
    return ((char const*) includes[i]);
}

Item* Calendar::Get(int i) const {
    return (Item*) items[i];
}
