/* Copyright (c) 1991 by Sanjay Ghemawat */
/*
 * $Revision: 1.14 $
 * $Date: 91/11/28 16:24:24 $
 * $Author: sanjay $
 * $State: Exp $
 *
 * $Log:	calfile.cc,v $
 * Revision 1.14  91/11/28  16:24:24  sanjay
 * CFRONT compatibility
 * 
 * Revision 1.13  91/11/06  18:01:35  sanjay
 * Do not maintain auto-save files for shared calendars. Auto-save for
 * shared calendars writes directly to the shared files.
 * 
 * Cleaned up other IO handling.
 * 
 * Revision 1.12  91/11/04  19:44:34  sanjay
 * Prepare for release
 * 
 * Revision 1.11  91/09/11  20:20:43  sanjay
 * Do not make non-existent calendars read-only
 * 
 * Revision 1.10  91/09/06  10:52:28  sanjay
 * Unlink auto-save file before auto-saving.
 * 
 * Revision 1.9  91/09/06  10:07:01  sanjay
 * Added #include <stddef.h>
 * 
 * Revision 1.8  91/08/07  16:32:44  sanjay
 * Fixed auto-file name construction bug.
 * 
 * Revision 1.7  91/08/07  10:59:14  sanjay
 * Added re-read.
 * 
 * Revision 1.6  91/08/06  18:19:10  sanjay
 * Disabled access check until I figure out how to handle
 * newly created calendars.
 * 
 * Revision 1.5  91/08/06  13:55:04  sanjay
 * Error popups are now centered on the mouse.
 * 
 * Revision 1.4  91/08/06  09:49:20  sanjay
 * Added GetName() and SetCalendar().
 * Set read-only flag on calendar based on file permissions.
 * 
 *
 */

#include <sys/file.h>
#include <stdlib.h>

#include <stddef.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <stdio.h>
#include <string.h>
#include <os/fs.h>
#include "calfile.h"
#include "calendar.h"
#include "lexer.h"

CalFile::CalFile(Application* a, const char* name) {
    application = a;

    int len = strlen(name);
    char* tmp;

    tmp = new char[len+1];
    strcpy(tmp, name);
    fileName = tmp;

    tmp = new char[len+6];
    strcpy(tmp, name);
    strcat(tmp, ".save");
    autoSaveName = tmp;

    lastModifyValid = false;

    calendar = new Calendar;

    modified = false;
}

CalFile::~CalFile() {
    delete calendar;
    delete (char*) fileName;
    delete (char*) autoSaveName;
}

void CalFile::SetCalendar(Calendar* cal) {
    delete calendar;
    calendar = cal;
}

void CalFile::Modified() {
    modified = true;
}

boolean CalFile::Write() {
    if (WriteTo(calendar, fileName)) {
        modified = false;
        lastModifyValid = GetModifyTime(fileName, lastModifyTime);
        return true;
    }
    else {
        lastModifyValid = false;
        return false;
    }
}

Calendar* CalFile::GetAutoSaveIfNewer() {
    Time autoTime;

    lastModifyValid = GetModifyTime(fileName, lastModifyTime);

    if (lastModifyValid &&
	GetModifyTime(autoSaveName, autoTime) &&
	(lastModifyTime < autoTime)) {
	return ReadFrom(autoSaveName);
    }
    return nil;
}

boolean CalFile::Read() {
    Time newFileTime;
    boolean gotTime = GetModifyTime(fileName, newFileTime);

    Calendar* cal = ReadFrom(fileName);
    if (cal != nil) {
        delete calendar;
        calendar = cal;
	calendar->SetReadOnly(false);

	PerformAccessCheck();

	modified = false;

	lastModifyValid = gotTime;
	if (gotTime) {
	    lastModifyTime = newFileTime;
	}

	return true;
    }
    return false;
}

boolean CalFile::FilePossiblyChanged() {
    if (! lastModifyValid) return true;

    Time t;
    if (! GetModifyTime(fileName, t)) return true;
    return (t != lastModifyTime);
}

boolean CalFile::FileHasChanged() {
    if (! lastModifyValid) return false;

    Time t;
    if (! GetModifyTime(fileName, t)) return true;
    return (t != lastModifyTime);
}

Calendar* CalFile::ReRead() {
    Time newFileTime;
    boolean gotTime = GetModifyTime(fileName, newFileTime);

    Calendar* cal = ReadFrom(fileName);
    Calendar* old = nil;
    if (cal != nil) {
	old = calendar;
	calendar = cal;
	calendar->SetReadOnly(false);

	PerformAccessCheck();

	modified = false;

	lastModifyValid = gotTime;
	if (gotTime) {
	    lastModifyTime = newFileTime;
	}
    }
    return old;
}
	    

boolean CalFile::AutoSave() {
    unlink(autoSaveName);
    return WriteTo(calendar, autoSaveName);
}

Calendar* CalFile::ReadFrom(const char* name, boolean topLevel) {
    Calendar* cal = new Calendar;
    Lexer input(name);

    if (! cal->Read(&input)) {
	delete cal;
	cal = nil;
    }

    return cal;
}

boolean CalFile::WriteTo(Calendar* cal, const char* name) {
    FILE* output = fopen(name, "w");
    if (! output) {
        return false;
    }

    cal->Write(output);
    if (ferror(output)) {
        fclose(output);
        return false;
    }

    fclose(output);
    return true;
}

boolean CalFile::GetModifyTime(char const* file, Time& t) {
    struct stat buf;

    int ret = stat(file, &buf);
    if (ret < 0) return false;

    struct timeval tv;
    tv.tv_sec  = buf.st_mtime;
    tv.tv_usec = 0;

    t = Time(tv);
    return true;
}

void CalFile::PerformAccessCheck() {
    if (access(fileName, W_OK) < 0) {
	switch (errno) {
	  case ENOENT:
	    /* File does not exist */
	    break;
	  case EACCES:
	  case EROFS:
	  case ETXTBSY:
	    /* Permission denied */
	    calendar->SetReadOnly(true);
	    break;
	  default:
	    /* Should not happen if we were successfuly able to read cal */
	    break;
	}
    }
}
