// This may look like C code, but it is really -*- C++ -*-
/* Copyright (c) 1991 by Sanjay Ghemawat */

#ifndef _CALFILEH
#define _CALFILEH

/*
 * $Revision: 1.6 $
 * $Date: 91/11/06 18:01:36 $
 * $Author: sanjay $
 * $State: Exp $
 *
 * $Log:	calfile.h,v $
 * Revision 1.6  91/11/06  18:01:36  sanjay
 * Do not maintain auto-save files for shared calendars. Auto-save for
 * shared calendars writes directly to the shared files.
 * 
 * Cleaned up other IO handling.
 * 
 * Revision 1.5  91/11/04  19:44:35  sanjay
 * Prepare for release
 * 
 * Revision 1.4  91/09/11  20:20:57  sanjay
 * Do not make non-existent calendars read-only
 * 
 * Revision 1.3  91/08/07  10:59:16  sanjay
 * Added re-read.
 * 
 * Revision 1.2  91/08/06  09:48:54  sanjay
 * Added GetName() and SetCalendar().
 * Set read-only flag on calendar based on file permissions.
 * 
 * Revision 1.1  91/04/11  10:26:36  sanjay
 * Encapsulated calendar file handling
 * 
 *
 */

/*
 * Calendar file encapsulation.
 */

#include <InterViews/defs.h>
#include "Time.h"

class Application;
class Calendar;

class CalFile {
  public:
    CalFile(Application*, const char* name);
    ~CalFile();

    /*
     * Get file name.
     */
    char const* GetName();

    /*
     * Get associated calendar.
     */
    Calendar* GetCalendar();

    /*
     * Set calendar.
     */
    void SetCalendar(Calendar*);

    /*
     * Has calendar been modified.
     */
    boolean IsModified();

    /*
     * Called on each modification.
     */
    void Modified();

    enum CalFile_Result {
	Success,
	Cancel,
	Failure
	};

    /*
     * Write calendar out to file.
     * Return true only on success.
     */
    boolean Write();

    /*
     * Read calendar from file.
     */
    boolean Read();

    /*
     * Return auto-save calendar if it is newer than on-disk calendar.
     * Also records the current file time for the calendar.
     */
    Calendar* GetAutoSaveIfNewer();

    /*
     * Auto-save.
     */
    boolean AutoSave();

    /*
     * Can file contents possibly have changed since last read?
     */
    boolean FilePossiblyChanged();

    /*
     * Have file contents actually changed since last read?
     */
    boolean FileHasChanged();

    /*
     * Re-read file contents from disk.
     * Ignores auto-save file.
     * Calendar is not changed on failure.
     * Returns nil on error, old calendar on success.
     */
    Calendar* ReRead();

    /* Read/Write calendar from/to named file */
    static Calendar* ReadFrom(const char*, boolean topLevel = true);
    static boolean   WriteTo(Calendar*, const char*);
  protected:
    Application* application;

    const char* fileName;	/* Calendar file name */
    const char* autoSaveName;	/* auto-save file name */

    Time    lastModifyTime;	/* Last calendar file modify time */
    boolean lastModifyValid;	/* Is lastModifyTime valid? */

    Calendar* calendar;		/* Volatile calendar state */

    boolean modified;		/* Has cal been modified since last write */

    /* Utility routines */
    static boolean GetModifyTime(char const*, Time&);

    void PerformAccessCheck();
};

inline char const* CalFile::GetName() {
    return fileName;
}

inline Calendar* CalFile::GetCalendar() {
    return calendar;
}

inline boolean CalFile::IsModified() {
    return modified;
}

#endif /* _CALFILEH */
