// This may look like C code, but it is really -*- C++ -*-
/* Copyright (c) 1991 by Sanjay Ghemawat */

#ifndef _charARRAYH
#define _charARRAYH

/*
 * A resizable array of chars.
 *
 * If char is user-defined, it must have the following operations -
 *
 * class char ... {
 *   public:
 *     char();
 *     ~char();
 *     ...
 *     char& operator = (char&);
 * };
 *
 * If UNSAFE_ARRAY not defined when including charArray.h, then array bounds
 * and argument range checking will be done.
 *
 */

/*
 * $Revision: 1.3 $
 * $Date: 91/11/28 16:23:54 $
 * $Author: sanjay $
 * $State: Exp $
 *
 * $Log:	Array.hP,v $
 * Revision 1.3  91/11/28  16:23:54  sanjay
 * CFRONT compatibility
 * 
 * Revision 1.2  91/11/04  19:43:48  sanjay
 * Prepare for release
 * 
 * Revision 1.1  91/06/28  20:19:50  sanjay
 * Initial revision
 * 
 */

class charArray {
  public:

    /* Constructors */
    charArray();				/* Empty array */
    charArray(int predict);		/* Empty array with size prediction */
    charArray(const char*, int);		/* Initialized with C array */
    charArray(const charArray&);		/* Initialized with another Array */
    charArray(char, int);			/* Fill with n copies of element */

    /* Destructor */
    ~charArray();

    charArray& operator=(const charArray&);	/* Assignment operator */

    /* Array indexing */
    char& operator[](int index) const;
    char& Slot(int index) const;

    /* Other Array operations */
    int  Size() const;			/* Return size; */
    char& High() const;			/* Return last element */
    char* AsPointer() const;		/* Return as pointer to base */
    void Append(char v);			/* Append an element */
    void Append(char, int n);		/* Append n copies of element */
    void Concat(const char*, int);	/* Concatenate C array */
    void Concat(const charArray&);	/* Concatenate another Array */
    void Fill(char,int start,int count);	/* Fill specified region with char */
    char  Remove();			/* Remove and return last element */
    void Remove(int num);		/* Remove last num elements */
    void Clear();			/* Remove all elements */

    /* Storage stuff */
    void Reclaim();			/* Reclaim all unused space */
  protected:
    /* Error handlers */
    void BoundError(int) const;
    void RangeError(int) const;

    void BoundCheck(int) const;
    void RangeCheck(int) const;
    void SizeCheck (int) const;

  private:
    char*	store;			/* Actual storage */
    int		alloc;			/* Size of allocated storage */
    int		size;			/* Size of used storage */

    /* Storage enlargers */
    void EnlargeAllocationTo(int s);	/* Enlarge storage to s */
    void EnlargeTo(int s);		/* Enlarge storage to s if necessary */
};

inline void charArray::BoundCheck(int val) const {
#ifndef UNSAFE_ARRAY
    if ((val < 0) || (val >= size)) BoundError(val);
#endif
}

inline void charArray::RangeCheck(int val) const {
#ifndef UNSAFE_ARRAY
    if (val < 0) RangeError(val);
#endif
}

inline void charArray::SizeCheck(int val) const {
#ifndef UNSAFE_ARRAY
    if ((val < 0) || (val > size)) BoundError(val);
#endif
}

inline charArray::charArray() {
    alloc = 0;
    size  = 0;
}

inline int charArray::Size() const {
    return size;
}

inline char& charArray::operator[](int index) const {
    BoundCheck(index);
    return store[index];
}

inline char& charArray::Slot(int index) const {
    BoundCheck(index);
    return store[index];
}

inline char& charArray::High() const {
    return store[size-1];
}

inline char* charArray::AsPointer() const {
    return store;
}

inline void charArray::Append(char v) {
    if (size >= alloc)
	EnlargeAllocationTo(size+1);
    store[size++] = v;
}

inline char charArray::Remove() {
    if (size > 0) size--;
    return store[size];
}

inline void charArray::Remove(int num) {
    SizeCheck(num);
    size -= num;
}

inline void charArray::Clear() {
    size = 0;
}

#endif /*_charARRAYH */
