// This may look like C code, but it is really -*- C++ -*-
/* Copyright (c) 1991 by Sanjay Ghemawat */

/*
 * $Revision: 1.2 $
 * $Date: 91/11/04 19:44:59 $
 * $Author: sanjay $
 * $State: Exp $
 *
 * $Log:	field.h,v $
 * Revision 1.2  91/11/04  19:44:59  sanjay
 * Prepare for release
 * 
 * Revision 1.1  91/03/28  11:41:20  sanjay
 * Initial deposit
 * 
 */

#ifndef _FIELDH
#define _FIELDH

#include <InterViews/subject.h>

/*
 * A FieldState is a base class for displaying string representations of
 * various data types.
 */
class FieldState : public Subject {
  public:
    FieldState() { }

    virtual const char* Unparse() const = 0;
};

/*
 * An IntState maintains stuff for displaying an integer
 */
class IntState : public FieldState {
  public:
    IntState(int init)	     { value = init; }

    void Set(int val) { if (value != val) { value = val; Notify(); }}
    int  Get() const  { return value; }

    virtual const char* Unparse() const;
  protected:
    int   value;
};

/*
 * A StringState maintains stuff for displaying strings
 */
class StringState : public FieldState {
  public:
    StringState(const char* init);
    ~StringState();

    void        Set(const char* val);
    const char* Get() const { return value; }

    virtual const char* Unparse() const { return value; }
  protected:
    char*	value;
    int		len;
};

/*
 * Stuff for keeping enumerated types - for example:
 *     typedef enum { false = 0, true = 1} boolean;
 *     int vals[] = { false, true };
 *     char* names[] = { "false", "true" };
 *     EnumState state(true, 2, vals, names);
 */
class EnumState : public FieldState {
  public:
    EnumState(int init, int length, const int* values, const char** names);

    void Set(int val);
    int  Get() const { return valuelist[index]; }

    virtual const char* Unparse() const { return namelist[index]; }
  protected:
    int		 listlength;	/* Length of namelist and valuelist */
    const char** namelist;	/* List of names for enumerated type */
    const int*	 valuelist;	/* List of vals for enumerated type */
    int		 index;		/* Index of current value */

    int Find(int val) const;	/* Return index of val in valuelist */
};

class RangeState : public IntState {
  public:
    RangeState(int i, int m1, int m2) : (i) {
	minValue = m1;
	maxValue = m2;
	if (maxValue < minValue) maxValue = minValue;
	Set(value);
    }

    virtual void Notify() {
	if (value < minValue) value = minValue;
	if (value > maxValue) value = maxValue;
	IntState::Notify();
    }

    int  GetMin()		{ return minValue; }
    int  GetMax()		{ return maxValue; }
    void SetMin(int);
    void SetMax(int);
  protected:
    int minValue;
    int maxValue;
};

class CycleState : public IntState {
  public:
    CycleState(int i, int m1, int m2) : (i) {
	minValue = m1;
	maxValue = m2;
	if (maxValue < minValue) maxValue = minValue;
	Set(value);
    }

    virtual void Notify() {
	if (value < minValue) value = maxValue;
	if (value > maxValue) value = minValue;
	IntState::Notify();
    }
  protected:
    int minValue;
    int maxValue;
};

class MonthState : public CycleState {
  public:
    MonthState(int i, IntState* y = nil) : (i, 1, 12) { year = y; }

    virtual void Notify();
    virtual const char* Unparse() const;
  protected:
    IntState* year;
};

class AmPmState : public EnumState {
    static int values[2];
    static const char* names[2];
  public:
    enum AmOrPm { Am, Pm };
    AmPmState(AmOrPm state) : ((int) state, 2, values, names) { }
};

class HourState : public CycleState {
  public:
    HourState(int i, AmPmState* ap = nil) : (i, 0, 23) { ampm = ap; Notify(); }

    virtual void Notify() {
	CycleState::Notify();
	if (ampm != nil)
	    ampm->Set((int) ((value < 12) ? AmPmState::Am : AmPmState::Pm));
    }
    virtual const char* Unparse() const;
  protected:
    AmPmState* ampm;
};

#endif /* _FIELDH */
