// This may look like C code, but it is really -*- C++ -*-
/* Copyright (c) 1991 by Sanjay Ghemawat */

#ifndef _intARRAYH
#define _intARRAYH

/*
 * A resizable array of ints.
 *
 * If int is user-defined, it must have the following operations -
 *
 * class int ... {
 *   public:
 *     int();
 *     ~int();
 *     ...
 *     int& operator = (int&);
 * };
 *
 * If UNSAFE_ARRAY not defined when including intArray.h, then array bounds
 * and argument range checking will be done.
 *
 */

/*
 * $Revision: 1.3 $
 * $Date: 91/11/28 16:23:54 $
 * $Author: sanjay $
 * $State: Exp $
 *
 * $Log:	Array.hP,v $
 * Revision 1.3  91/11/28  16:23:54  sanjay
 * CFRONT compatibility
 * 
 * Revision 1.2  91/11/04  19:43:48  sanjay
 * Prepare for release
 * 
 * Revision 1.1  91/06/28  20:19:50  sanjay
 * Initial revision
 * 
 */

class intArray {
  public:

    /* Constructors */
    intArray();				/* Empty array */
    intArray(int predict);		/* Empty array with size prediction */
    intArray(const int*, int);		/* Initialized with C array */
    intArray(const intArray&);		/* Initialized with another Array */
    intArray(int, int);			/* Fill with n copies of element */

    /* Destructor */
    ~intArray();

    intArray& operator=(const intArray&);	/* Assignment operator */

    /* Array indexing */
    int& operator[](int index) const;
    int& Slot(int index) const;

    /* Other Array operations */
    int  Size() const;			/* Return size; */
    int& High() const;			/* Return last element */
    int* AsPointer() const;		/* Return as pointer to base */
    void Append(int v);			/* Append an element */
    void Append(int, int n);		/* Append n copies of element */
    void Concat(const int*, int);	/* Concatenate C array */
    void Concat(const intArray&);	/* Concatenate another Array */
    void Fill(int,int start,int count);	/* Fill specified region with int */
    int  Remove();			/* Remove and return last element */
    void Remove(int num);		/* Remove last num elements */
    void Clear();			/* Remove all elements */

    /* Storage stuff */
    void Reclaim();			/* Reclaim all unused space */
  protected:
    /* Error handlers */
    void BoundError(int) const;
    void RangeError(int) const;

    void BoundCheck(int) const;
    void RangeCheck(int) const;
    void SizeCheck (int) const;

  private:
    int*	store;			/* Actual storage */
    int		alloc;			/* Size of allocated storage */
    int		size;			/* Size of used storage */

    /* Storage enlargers */
    void EnlargeAllocationTo(int s);	/* Enlarge storage to s */
    void EnlargeTo(int s);		/* Enlarge storage to s if necessary */
};

inline void intArray::BoundCheck(int val) const {
#ifndef UNSAFE_ARRAY
    if ((val < 0) || (val >= size)) BoundError(val);
#endif
}

inline void intArray::RangeCheck(int val) const {
#ifndef UNSAFE_ARRAY
    if (val < 0) RangeError(val);
#endif
}

inline void intArray::SizeCheck(int val) const {
#ifndef UNSAFE_ARRAY
    if ((val < 0) || (val > size)) BoundError(val);
#endif
}

inline intArray::intArray() {
    alloc = 0;
    size  = 0;
}

inline int intArray::Size() const {
    return size;
}

inline int& intArray::operator[](int index) const {
    BoundCheck(index);
    return store[index];
}

inline int& intArray::Slot(int index) const {
    BoundCheck(index);
    return store[index];
}

inline int& intArray::High() const {
    return store[size-1];
}

inline int* intArray::AsPointer() const {
    return store;
}

inline void intArray::Append(int v) {
    if (size >= alloc)
	EnlargeAllocationTo(size+1);
    store[size++] = v;
}

inline int intArray::Remove() {
    if (size > 0) size--;
    return store[size];
}

inline void intArray::Remove(int num) {
    SizeCheck(num);
    size -= num;
}

inline void intArray::Clear() {
    size = 0;
}

#endif /*_intARRAYH */
