/* Copyright (c) 1991 by Sanjay Ghemawat */
/*
 * $Revision: 1.12 $
 * $Date: 91/11/04 19:45:14 $
 * $Author: sanjay $
 * $State: Exp $
 *
 * $Log:	item.cc,v $
 * Revision 1.12  91/11/04  19:45:14  sanjay
 * Prepare for release
 * 
 * Revision 1.11  91/11/04  11:50:40  sanjay
 * Print appropriate error message when file cannot be read.
 * 
 * Revision 1.10  91/08/06  14:14:09  sanjay
 * Removed source of warnings.
 * 
 * Revision 1.9  91/08/06  09:55:17  sanjay
 * Inherit read-only attribute from calendar.
 * Store calendar pointer in each item.
 * 
 * Revision 1.8  91/06/28  14:32:46  sanjay
 * Make array deletions portable.
 * 
 * Revision 1.7  91/04/23  15:36:47  sanjay
 * Set everyYear to false when reading year range
 * 
 * Revision 1.6  91/04/22  19:47:43  sanjay
 * Removed Date::Midnight() usage
 * 
 * Revision 1.5  91/04/22  07:41:27  sanjay
 * Use WeekDay::Index
 * 
 * Revision 1.4  91/04/14  19:07:30  sanjay
 * Allow deletion of specific occurrences of repeating items
 * 
 * Revision 1.3  91/04/01  11:57:07  sanjay
 * Added include file capability
 * 
 * Revision 1.2  91/03/28  11:51:54  sanjay
 * Fixed log
 * 
 * Revision 1.1  91/03/28  11:42:12  sanjay
 * Initial deposit
 *
 */

#include <string.h>
#include "Year.h"
#include "Month.h"
#include "WeekDay.h"
#include "Date.h"

#include "config.h"
#include "Date.Array.h"
#include "basic.h"
#include "calendar.h"
#include "item.h"
#include "lexer.h"

const int Item::textSize = 2048;
const int Item::defaultRemindStart = 1;

static const char opener = '[';
static const char closer = ']';

Item::Item(Calendar* cal) {
    text = new char[textSize];
    strcpy(text, "");

    calendar = cal;

    deleted = false;
    remindStart = defaultRemindStart;

    datesDeleted = nil;

    /* Random initialization */
    everyYear = false;
    months.Clear();
    days.Clear();
    isWeekly = false;
}

Item::~Item() {
    if (datesDeleted != nil) delete datesDeleted;
    deleteArray(textSize) text;
}

boolean Item::Read(Lexer* lex) {
    while (lex->Status() == Lexer::Valid) {
	char c;
	char const* keyword;

	if (! lex->SkipWS() ||
	    ! lex->Peek(c)) {
	    lex->SetError("incomplete item");
	    return false;
	}

	if (c == closer) {
	    /* Item is over */
	    return true;
	}

	if (! lex->GetId(keyword) ||
	    ! lex->SkipWS() ||
	    ! lex->Skip(opener)) {
	    lex->SetError("error reading item property name");
	    return false;
	}

	if (Parse(lex, keyword)) {
	    if (! lex->SkipWS() ||
		! lex->Skip(closer)) {
		lex->SetError("error reading item property");
		return false;
	    }
	    continue;
	}

	if (lex->Status() != Lexer::Error) {
	    lex->SetError("unrecognized item property name");
	}
	return false;
    }
    lex->SetError("incomplete item");
    return false;
}

boolean Item::Parse(Lexer* lex, char const* keyword) {
    if (strcmp(keyword, "Text") == 0) {
	/* Read text */
	int len;

	if (! lex->SkipWS() ||
	    ! lex->GetNumber(len) ||
	    ! (len >= 0) ||
	    ! (len < textSize) ||
	    ! lex->SkipWS() ||
	    ! lex->Skip(opener) ||
	    ! lex->GetText(text, len) ||
	    ! lex->Skip(closer)) {
	    lex->SetError("error reading item text");
	    return false;
	}

	text[len] = '\0';
	return true;
    }

    if (strcmp(keyword, "Wdays") == 0) {
	isWeekly = true;
	if (! days.Read(lex)) {
	    lex->SetError("error reading weekdays");
	    return false;
	}
	return true;
    }

    if (strcmp(keyword, "Mdays") == 0) {
	isWeekly = false;
	if (! days.Read(lex)) {
	    lex->SetError("error reading monthdays");
	    return false;
	}
	return true;
    }

    if (strcmp(keyword, "Months") == 0) {
	if (! months.Read(lex)) {
	    lex->SetError("error reading set of months");
	    return false;
	}
	return true;
    }

    if (strcmp(keyword, "EveryYear") == 0) {
	everyYear = true;
	return true;
    }

    if (strcmp(keyword, "Years") == 0) {
	everyYear = false;
	if (! lex->SkipWS() ||
	    ! lex->GetNumber(firstYear) ||
	    ! lex->SkipWS() ||
	    ! lex->GetNumber(lastYear)) {
	    lex->SetError("error reading range of years");
	    return false;
	}
	return true;
    }

    if (strcmp(keyword, "Remind") == 0) {
	if (! lex->SkipWS() ||
	    ! lex->GetNumber(remindStart)) {
	    lex->SetError("error reading remind level");
	    return false;
	}
	return true;
    }

    if (strcmp(keyword, "Deleted") == 0) {
	int day, month, year;

	if (! lex->SkipWS() ||
	    ! lex->GetNumber(day) ||
	    ! lex->SkipWS() ||
	    ! lex->GetNumber(month) ||
	    ! lex->SkipWS() ||
	    ! lex->GetNumber(year)) {
	    lex->SetError("error reading deletion date");
	    return false;
	}

	DeleteOn(Date(day, Month::First() + (month - 1), year));
	return true;
    }

    return false;
}

void Item::Write(FILE* out) const {
    fprintf(out, "Text [%d [%s]]\n", strlen(text), text);

    fprintf(out, "%s [", (isWeekly ? "Wdays" : "Mdays"));
    days.Write(out);
    fprintf(out, "]\n");

    fprintf(out, "Months [");
    months.Write(out);
    fprintf(out, "]\n");

    if (everyYear) {
	fprintf(out, "EveryYear []\n");
    }
    else {
	fprintf(out, "Years [%d %d]\n", firstYear, lastYear);
    }

    fprintf(out, "Remind [%d]\n", remindStart);

    if (datesDeleted != nil) {
	for (int i = 0; i < datesDeleted->Size(); i++) {
	    int day, year;
	    Month month;
	    WeekDay wday;

	    ((*datesDeleted)[i]).BreakDown(day, wday, month, year);
	    fprintf(out, "Deleted [%d %d %d]\n", day, month.Index(), year);
	}
    }
}

Calendar* Item::GetCalendar() const {
    return calendar;
}

void Item::SetCalendar(Calendar* cal) {
    calendar = cal;
}

void Item::CopyTo(Item* item) const {
    item->SetCalendar(GetCalendar());

    item->SetText(text);
    if (isWeekly) {
	item->SetWeekDays(days);
    }
    else {
	item->SetMonthDays(days);
    }
    item->SetMonths(months);
    if (everyYear) {
	item->SetEveryYear();
    }
    else {
	item->SetYears(firstYear, lastYear);
    }
    item->remindStart = remindStart;

    if (datesDeleted != nil) {
	if (item->datesDeleted != nil) {
	    item->datesDeleted->Clear();
	}
	else {
	    item->datesDeleted = new DateArray;
	}
	item->datesDeleted->Concat(*datesDeleted);
    }
    else {
	if (item->datesDeleted != nil) {
	    delete item->datesDeleted;
	    item->datesDeleted = nil;
	}
    }
}

const char* Item::GetText() const {
    return text;
}

void Item::SetText(const char* t) {
    int len = strlen(t);
    if (len >= textSize) len = textSize - 1;
    strncpy(text, t, len);
    text[len] = '\0';
}

boolean Item::IsWeekly() const {
    return isWeekly;
}

boolean Item::IsMonthly() const {
    return !isWeekly;
}

SmallIntSet Item::GetWeekDays() const {
    return days;
}

SmallIntSet Item::GetMonthDays() const {
    return days;
}

void Item::SetWeekDays(SmallIntSet s) {
    isWeekly = true;
    days = s;

    if (datesDeleted != nil) {
	delete datesDeleted;
	datesDeleted = nil;
    }
}

void Item::SetMonthDays(SmallIntSet s) {
    isWeekly = false;
    days = s;

    if (datesDeleted != nil) {
	delete datesDeleted;
	datesDeleted = nil;
    }
}

SmallIntSet Item::GetMonths() const {
    return months;
}

void Item::SetMonths(SmallIntSet s) {
    months = s;

    if (datesDeleted != nil) {
	delete datesDeleted;
	datesDeleted = nil;
    }
}

boolean Item::EveryYear() const {
    return everyYear;
}

void Item::SetEveryYear() {
    everyYear = true;

    if (datesDeleted != nil) {
	delete datesDeleted;
	datesDeleted = nil;
    }
}

int Item::FirstYear() const {
    return firstYear;
}

int Item::LastYear() const {
    return lastYear;
}

void Item::SetYears(int first, int last) {
    firstYear = first;
    lastYear = last;
    everyYear = false;

    if (datesDeleted != nil) {
	delete datesDeleted;
	datesDeleted = nil;
    }
}

int Item::GetRemindStart() const {
    return remindStart;
}

void Item::SetRemindStart(int r) {
    remindStart = r;
}

boolean Item::ReadOnly() const {
    return GetCalendar()->ReadOnly();
}

Notice* Item::AsNotice() {
    return nil;
}

Appointment* Item::AsAppointment() {
    return nil;
}

void Item::SetDate(Date date) {
    int d, y;
    Month m;
    WeekDay wd;
    date.BreakDown(d, wd, m, y);

    isWeekly = false;
    days.Clear();
    days.Insert(d);
    months.Clear();
    months.Insert(m.Index());
    everyYear = false;
    firstYear = y;
    lastYear = y;

    if (datesDeleted != nil) {
	delete datesDeleted;
	datesDeleted = nil;
    }
}

boolean Item::IsOn(Date date) const {
    if (IsDeletedOn(date)) return false;

    int d;
    Month m;
    WeekDay wd;
    int y;
    date.BreakDown(d, wd, m, y);

    return ((isWeekly
	     ? days.Member(wd.Index())
	     : days.Member(d)) &&
	    months.Member(m.Index()) &&
	    (everyYear || ((firstYear <= y) && (lastYear >= y))));
}

boolean Item::Deleted() {
    return deleted;
}

void Item::MarkDeleted() {
    deleted = true;
}

Date Item::FirstDate() const {
    return FirstSince(Date::First());
}

Date Item::FirstSince(Date date) const {
    int d, y;
    Month m;
    WeekDay wd;		/* Only maintained if isWeekly */

    date.BreakDown(d, wd, m, y);

    /* Get first possible year */
    int fY = (everyYear ? Year::First() : firstYear);
    if (y < fY) {
	/* Adjust to new first year */
	y = fY;
	m = Month::First();
	d = 1;
	if (isWeekly) wd = Date(d, m, y).GetWDay();
    }

    /* Get last possible year */
    int lY = (everyYear ? Year::Last() : lastYear);

    while (y <= lY) {
	/* ASSERT if (isWeekly) then wd == Date(d, m, y).WDay() */

	/* Search in month m */
	int mSize = m.Size(y);

	if (months.Member(m.Index())) {
	    /* Search for day match */
	    if (isWeekly) {
		/* Search for at most seven days */
		int lastDay = d + 6;
		if (lastDay > mSize) lastDay = mSize;

		while (d <= lastDay) {
		    if (days.Member(wd.Index())) {
			Date xdate = Date(d, m, y);
			if (! IsDeletedOn(xdate)) return xdate;
		    }
		    d++;
		    wd += 1;
		}
	    }
	    else {
		/* Search through rest of month */
		while (d <= mSize) {
		    if (days.Member(d)) {
			Date xdate = Date(d, m, y);
			if (! IsDeletedOn(xdate)) return xdate;
		    }
		    d++;
		}
	    }
	}
	/* ASSERT if (isWeekly) then wd == Date(d, m, y).WDay() */

	/* Advance to next month */
	if (m == Month::Last()) {
	    m = Month::First();
	    y++;
	}
	else {
	    m += 1;
	}

	/* Adjust <d, wd> for beginning of next month */
	if (isWeekly) {
	    wd += mSize - d + 1;
	}
	d = 1;
    }

    /* No occurrence */
    return Date::Last();
}

int Item::OccursBefore(Item* i1, Item* i2) {
    Date d1 = i1->FirstDate();
    Date d2 = i2->FirstDate();

    if (d1 < d2) return 1;
    if (d1 > d2) return 0;

    /* i1->FirstDate() == i2->FirstDate() */

    Appointment* a1 = i1->AsAppointment();
    Appointment* a2 = i2->AsAppointment();

    /* Notices start before all appointments (at time -1) */
    int s1 = (a1 == nil) ? -1 : a1->GetStart();
    int s2 = (a2 == nil) ? -1 : a2->GetStart();

    return s1 < s2;
}

boolean Item::Repeats() const {
    return (everyYear ||
	    (firstYear < lastYear) ||
	    isWeekly ||
	    (months.Size() > 1) ||
	    (days.Size() > 1));
}

void Item::DeleteOn(Date date) {
    if (datesDeleted == nil) {
	datesDeleted = new DateArray;
    }
    datesDeleted->Append(date);
}

boolean Item::IsDeletedOn(Date date) const {
    if (datesDeleted == nil) {
	return false;
    }

    for (int i = 0; i < datesDeleted->Size(); i++) {
	if ((*datesDeleted)[i] == date) {
	    return true;
	}
    }
    return false;
}

Notice::Notice(Calendar* cal) : Item(cal) {
    length = 30;
}

Notice::~Notice() { }

boolean Notice::Parse(Lexer* lex, char const* keyword) {
    if (strcmp(keyword, "Length") == 0) {
	if (! lex->SkipWS() ||
	    ! lex->GetNumber(length)) {
	    lex->SetError("error reading notice display length");
	    return false;
	}
	return true;
    }
    else {
	return Item::Parse(lex, keyword);
    }
}

void Notice::Write(FILE* out) const {
    fprintf(out, "Length [%d]\n", length);
    Item::Write(out);
}

Item* Notice::Clone() const {
    Notice* copy = new Notice(GetCalendar());

    Item::CopyTo(copy);
    copy->SetLength(length);
    return copy;
}

int Notice::GetLength() const {
    return length;
}

void Notice::SetLength(int l) {
    length = l;
}

Notice* Notice::AsNotice() {
    return this;
}

Appointment::Appointment(Calendar* cal) : Item(cal) {
    start = 30;
    length = 30;
}

Appointment::~Appointment() { }

boolean Appointment::Parse(Lexer* lex, char const* keyword) {
    if (strcmp(keyword, "Start") == 0) {
	if (! lex->SkipWS() ||
	    ! lex->GetNumber(start)) {
	    lex->SetError("error reading appointment start time");
	    return false;
	}
	return true;
    }
    else if (strcmp(keyword, "Length") == 0) {
	if (! lex->SkipWS() ||
	    ! lex->GetNumber(length)) {
	    lex->SetError("error reading appointment length");
	    return false;
	}
	return true;
    }
    else {
	return Item::Parse(lex, keyword);
    }
}

void Appointment::Write(FILE* out) const {
    fprintf(out, "Start [%d]\n", start);
    fprintf(out, "Length [%d]\n", length);
    Item::Write(out);
}

Item* Appointment::Clone() const {
    Appointment* copy = new Appointment(GetCalendar());

    Item::CopyTo(copy);
    copy->SetStart(start);
    copy->SetLength(length);
    return copy;
}

int Appointment::GetStart() const {
    return start;
}

void Appointment::SetStart(int s) {
    start = s;
}

int Appointment::GetLength() const {
    return length;
}

void Appointment::SetLength(int l) {
    length = l;
}

int Appointment::GetFinish() const {
    return start + length;
}

Appointment* Appointment::AsAppointment() {
    return this;
}
