// This may look like C code, but it is really -*- C++ -*-
/* Copyright (c) 1991 by Sanjay Ghemawat */

/*
 * $Revision: 1.5 $
 * $Date: 91/11/04 19:45:16 $
 * $Author: sanjay $
 * $State: Exp $
 *
 * $Log:	item.h,v $
 * Revision 1.5  91/11/04  19:45:16  sanjay
 * Prepare for release
 * 
 * Revision 1.4  91/08/06  09:54:58  sanjay
 * Inherit read-only attribute from calendar.
 * Store calendar pointer in each item.
 * 
 * Revision 1.3  91/04/14  19:07:30  sanjay
 * Allow deletion of specific occurrences of repeating items
 * 
 * Revision 1.2  91/04/01  11:57:08  sanjay
 * Added include file capability
 * 
 * Revision 1.1  91/03/28  11:41:25  sanjay
 * Initial deposit
 * 
 */

#ifndef _ITEMH
#define _ITEMH

#include <InterViews/defs.h>
#include <stdio.h>
#include "smallintset.h"
#include "Date.h"

class Lexer;

class Notice;
class Appointment;
class DateArray;
class Calendar;

/*
 * Item
 *
 * Abstract components.
 *
 * Text: string			associated text
 * IsWeekly: bool		Is repetition based on weeks instead of months
 * IsMonthly: bool		!IsWeekly
 * WeekDays: set[WeekDay]	If IsWeekly, the set of week days
 * MonthDays: set[int]		If IsMonthly, the set of month days
 * Months: set[Month]		Set of months.
 * EveryYear: bool		Does this occur every year.
 * FirstYear: int		If !EveryYear, then first year of range.
 * LastYear: int		If !EveryYear, then last year of range.
 * RemindStart: int		No. of days before item when reminder begins.
 * ReadOnly: bool		Is Item read-only
 * DatesDeleted: Array[Date]	Dates on which item has been deleted.
 *
 * AsNotice: Notice*		Typecast to Notice if legal.  Null otherwise.
 * AsAppointment: Appointment*	Typecast to Appointment if legal ... 
 *
 */

class Item {
  public:
    Item(Calendar*);
    virtual ~Item();

    virtual boolean Read(Lexer*);
    virtual boolean Parse(Lexer*, char const* keyword);
    virtual void Write(FILE*) const;

    /* Get/Set containing calendar */
    Calendar* GetCalendar() const;
    void SetCalendar(Calendar*);

    virtual Item* Clone() const = 0;

    const char* GetText() const;
    void SetText(const char*);

    boolean IsWeekly() const;
    boolean IsMonthly() const;

    SmallIntSet GetWeekDays() const;
    SmallIntSet GetMonthDays() const;

    void SetWeekDays(SmallIntSet);
    void SetMonthDays(SmallIntSet);

    SmallIntSet GetMonths() const;
    void SetMonths(SmallIntSet);

    boolean EveryYear() const;
    void SetEveryYear();

    int FirstYear() const;
    int LastYear() const;
    void SetYears(int, int);

    int GetRemindStart() const;
    void SetRemindStart(int);

    boolean ReadOnly() const;

    DateArray* datesDeleted;

    virtual Notice* AsNotice();
    virtual Appointment* AsAppointment();

    void SetDate(Date);
    boolean IsOn(Date) const;

    /* Does item repeat */
    boolean Repeats() const;

    /* Delete particular occurrence */
    void DeleteOn(Date);
    boolean IsDeletedOn(Date) const;

    /* Mark deleted */
    boolean Deleted();
    void MarkDeleted();

    /* Return first occurrence of item */
    Date FirstDate() const;

    /* Return first occurrence of item on or after specified date */
    Date FirstSince(Date) const;

    /* Compare two items */
    static int OccursBefore(Item*, Item*);

    static const int defaultRemindStart;
  protected:
    char* text;

    boolean		isWeekly;	/* Is this weekly */
    SmallIntSet		days;		/* Weekdays or month days */
    SmallIntSet		months;		/* set of months */
    boolean		everyYear;	/* Occurs every year? */
    int			firstYear;	/* Range beginning */
    int			lastYear;	/* Range end */
    int			remindStart;	/* Start reminding early */

    Calendar*		calendar;	/* Containing calendar */

    void CopyTo(Item*) const;

    static const int textSize;

    boolean deleted;
};

/*
 * Notice
 *
 * An Item with just a length.
 * The length is just an indication of how big a window the notice must
 * be displayed in.  The length field can be interpreted just as the length
 * field for Appointments, and if notice->length == appt->length, then
 * the notice will be displayed the same size as the appt.
 */
class Notice : public Item {
  public:
    Notice(Calendar*);
    ~Notice();

    virtual boolean Parse(Lexer*, char const* keyword);
    virtual void Write(FILE*) const;

    virtual Item* Clone() const;

    int GetLength() const;
    void SetLength(int);

    virtual Notice* AsNotice();
  protected:
    int length;
};

/*
 * Appointment
 *
 * An Item with a start time and a length.
 * Both start and length are in minutes.
 * Start is measured from midnight.
 */
class Appointment : public Item {
  public:
    Appointment(Calendar*);
    ~Appointment();

    virtual boolean Parse(Lexer*, char const* keyword);
    virtual void Write(FILE*) const;

    virtual Item* Clone() const;

    int GetStart() const;
    int GetLength() const;
    int GetFinish() const;
    void SetStart(int);
    void SetLength(int);

    virtual Appointment* AsAppointment();
  protected:
    int start;
    int length;
};

#endif /* _ITEMH */
