/* Copyright (c) 1991 by Sanjay Ghemawat */
/*
 * $Revision: 1.6 $
 * $Date: 91/11/04 19:45:31 $
 * $Author: sanjay $
 * $State: Exp $
 *
 * $Log:	monthdisplay.cc,v $
 * Revision 1.6  91/11/04  19:45:31  sanjay
 * Prepare for release
 * 
 * Revision 1.5  91/09/19  14:36:57  sanjay
 * Added resource lookup for holiday coloring.
 * 
 * Revision 1.4  91/09/18  20:34:40  sanjay
 * Show non-empty days in different colors.
 * 
 * Revision 1.3  91/09/18  14:17:15  sanjay
 * Pass Application to monthdisplay and dateeditor to allow monthdisplay
 * appearence to change with number of items on a certain day of the
 * month. The appearence changing has not been implemented yet.
 * 
 * Revision 1.2  91/03/28  11:52:00  sanjay
 * Fixed log
 * 
 * Revision 1.1  91/03/28  11:42:20  sanjay
 * Initial deposit
 *
 */

#include <InterViews/painter.h>
#include <InterViews/font.h>
#include <InterViews/color.h>
#include <InterViews/sensor.h>
#include <InterViews/event.h>
#include "Month.h"
#include "WeekDay.h"
#include "Date.h"
#include "application.h"
#include "field.h"
#include "monthdisplay.h"
#include "occurcache.h"

static const char* columnHeadings[] = {
    "Sun",
    "Mon",
    "Tue",
    "Wed",
    "Thu",
    "Fri",
    "Sat"
    };

static const char* dayString[] = {
      "",  "1",  "2",  "3",  "4",  "5",  "6",  "7",  "8",  "9",
    "10", "11", "12", "13", "14", "15", "16", "17", "18", "19",
    "20", "21", "22", "23", "24", "25", "26", "27", "28", "29",
    "30", "31"
    };

declare(AVForwarder,MonthDisplay);
implement(AVForwarder,MonthDisplay);

MonthDisplay::MonthDisplay(Application* a,
			   IntState* d,
			   IntState* m,
			   IntState* y)
: (7, 8) {
    SetClassName("MonthDisplay");

    application = a;
    view = new AVForwarder(MonthDisplay)(a, this);

    for (int i = 0; i <= 31; i++) {
	numberOfItems[i] = 0;
    }

    day = d;
    month = m;
    year = y;
    maxWidth = round(0.5 * inch);

    holidayColor = nil;
    interestColor = nil;

    holidayPainter = nil;
    interestingPainter = nil;

    ComputeMonthInfo();

    day->Attach(this);
    month->Attach(this);
    year->Attach(this);

    input = new Sensor;
    input->Catch(DownEvent);
    Listen(input);
}

MonthDisplay::~MonthDisplay() {
    delete view;

    Unref(interestingPainter);
    Unref(holidayPainter);
	
    Unref(holidayColor);
    Unref(interestColor);

    day->Detach(this);
    month->Detach(this);
    year->Detach(this);
    Unref(input);
}

void MonthDisplay::Update() {
    if ((displayedMonth != month->Get()) ||
	(displayedYear  != year->Get())) {
	    /* Month changed */
	    ComputeMonthInfo();
	    if (canvas != nil) {
		output->ClearRect(canvas, 0, 0, xmax, FindTop(2));
		Redraw(0, 0, xmax, FindTop(2));
	    }
	}
    else if (displayedDay != day->Get()) {
	int oldDay = displayedDay;
	int row, col;

	displayedDay = day->Get();

	if (canvas != nil) {
	    /* Paint the old day normally */
	    row = (oldDay + dayAdjust - 1) / 7 + 2;
	    col = (oldDay + dayAdjust - 1) % 7;
	    output->ClearRect(canvas, FindLeft(col), FindBottom(row),
			      FindRight(col), FindTop(row));
	    Paint(FindLeft(col), FindBottom(row), FindRight(col), FindTop(row),
		  col, row);

	    /* Paint the new day reversed */
	    row = (displayedDay + dayAdjust - 1) / 7 + 2;
	    col = (displayedDay + dayAdjust - 1) % 7;
	    output->ClearRect(canvas, FindLeft(col), FindBottom(row),
			      FindRight(col), FindTop(row));
	    Paint(FindLeft(col), FindBottom(row), FindRight(col), FindTop(row),
		  col, row);
	}
    }
}

void MonthDisplay::GetExtent(int, int, Coord& w, Coord& h) {
    w = maxWidth;
    h = output->GetFont()->Height();
}

void MonthDisplay::Paint(Coord l, Coord b, Coord r, Coord t, int col, int row){
    if (canvas == nil) return;
    if (row == 0) {
	/* Column heading. */
	RightJustify(((col == 0) || (col == 6)) ? holidayPainter : output,
		     columnHeadings[col], l, b, r, t);
    }
    else if (row == 1) {
	/* Blank row */
    }
    else {
	/* Day display - convert <row, col> to day number */
	int day = (row - 2) * 7 + col + 1 - dayAdjust;
	if ((day < 0) || (day > daysInMonth))
	    day = 0;

	/* Choose the painter */
	Painter* painter = output;
	if (numberOfItems[day] > 0) {
	    painter = interestingPainter;
	}
	else if ((col == 0) || (col == 6)) {
	    painter = holidayPainter;
	}

	if (day == displayedDay) {
	    painter->SetColors(painter->GetBgColor(), painter->GetFgColor());
	    painter->ClearRect(canvas, FindLeft(col), FindBottom(row),
			       FindRight(col), FindTop(row));
	    RightJustify(painter, dayString[day], l, b, r, t);
	    painter->SetColors(painter->GetBgColor(), painter->GetFgColor());
	}
	else {
	    RightJustify(painter, dayString[day], l, b, r, t);
	}
    }
}

void MonthDisplay::Reconfig() {
    Unref(holidayPainter);
    Unref(interestingPainter);

    Unref(holidayColor);
    Unref(interestColor);

    maxWidth = output->GetFont()->Width("Wed");
    LabelGrid::Reconfig();

    /* Create painter for painting holidays */
    const char* holiday = GetAttribute("holidayForeground");
    if (holiday == nil) {
	holiday = "red";
    }
    holidayColor = new Color(holiday);
    if (! holidayColor->Valid()) {
	delete holidayColor;
	holidayColor = output->GetFgColor();
    }
    holidayColor->Reference();

    holidayPainter = new Painter(output);
    holidayPainter->SetColors(holidayColor, output->GetBgColor());
    holidayPainter->Reference();

    /* Create painter for interesting days */
    const char* interest = GetAttribute("interestForeground");
    if (interest == nil) {
	interest = "blue";
    }

    interestColor = new Color(interest);

    if (! interestColor->Valid()) {
	delete interestColor;
	interestColor = output->GetFgColor();
    }
    interestColor->Reference();

    interestingPainter = new Painter(output);
    interestingPainter->SetColors(interestColor, output->GetBgColor());
    interestingPainter->Reference();
}

void MonthDisplay::Handle(Event& e) {
    int row, col;

    switch (e.eventType) {
      case DownEvent:
	row = FindRow(e.y);
	col = FindColumn(e.x);
	if ((row >= 2) && (row <= 7) && (col >= 0) && (col < 7)) {
	    int d = (row - 2) * 7 + col + 1 - dayAdjust;
	    if ((d > 0) && (d  <= daysInMonth)) {
		day->Set(d);
	    }
	}
	break;
      default:
	break;
    }
}

void MonthDisplay::RightJustify(Painter* painter, const char* str,
				Coord l, Coord b, Coord r, Coord t)
{
    painter->Text(canvas, str, r - painter->GetFont()->Width(str), b);
}
    

void MonthDisplay::ComputeMonthInfo() {
    displayedDay   = day->Get();
    displayedMonth = month->Get();
    displayedYear  = year->Get();

    Month month = Month::First() + (displayedMonth - 1);
    Date first = Date(1, month, displayedYear);

    dayAdjust = first.GetWDay().Index() - 1;
    daysInMonth = month.Size(displayedYear);

    if (displayedDay > daysInMonth) {
	/* Prevent another notification for day change */
	displayedDay = daysInMonth;
	day->Set(daysInMonth);
    }

    /* Count number of items on each day of month */
    OccurCache* cache = application->GetOccurCache();
    cache->Preload(first, daysInMonth);
    for (int i = 1; i <= daysInMonth; i++) {
	numberOfItems[i] = cache->NumberOn(first+i-1);
    }
    numberOfItems[0] = 0;
    for (i = daysInMonth+1; i <= 31; i++) {
	numberOfItems[i] = 0;
    }
}

void MonthDisplay::ItemMsg(ApptViewItemMsg, Item*) {
    /* Recompute item counts */

    /* Save old item counts */
    int oldCounts[32];
    for (int i = 1; i <= daysInMonth; i++) {
	oldCounts[i] = numberOfItems[i];
    }

    /* Get new info */
    ComputeMonthInfo();

    /* Redraw changed regions */
    if (canvas != nil) {
	for (i = 1; i <= daysInMonth; i++) {
	    boolean hadItems  = (oldCounts[i] != 0);
	    boolean haveItems = (numberOfItems[i] != 0);

	    if ((hadItems && !haveItems) || (!hadItems && haveItems)) {
		/* Redraw */
		int row = (i + dayAdjust - 1) / 7 + 2;
		int col = (i + dayAdjust - 1) % 7;

		output->ClearRect(canvas,
				  FindLeft(col), FindBottom(row),
				  FindRight(col), FindTop(row));
		Paint(FindLeft(col), FindBottom(row),
		      FindRight(col), FindTop(row), col, row);
	    }
	}
    }
}

void MonthDisplay::Msg(ApptViewMsg) {
    /* Ignore - we are only interested in item changes */
}
