/* Copyright (c) 1991 by Sanjay Ghemawat */
/*
 * $Revision: 1.5 $
 * $Date: 91/11/28 16:25:39 $
 * $Author: sanjay $
 * $State: Exp $
 *
 * $Log:	simpleeditor.cc,v $
 * Revision 1.5  91/11/28  16:25:39  sanjay
 * CFRONT compatibility
 * 
 * Revision 1.4  91/11/04  19:45:55  sanjay
 * Prepare for release
 * 
 * Revision 1.3  91/08/06  14:14:11  sanjay
 * Removed source of warnings.
 * 
 * Revision 1.2  91/03/28  11:52:04  sanjay
 * Fixed log
 * 
 * Revision 1.1  91/03/28  11:42:29  sanjay
 * Initial deposit
 *
 */

#include <InterViews/textbuffer.h>
#include <InterViews/textdisplay.h>
#include <InterViews/sensor.h>
#include <InterViews/event.h>
#include <stdlib.h>
#include <bstring.h>
#include <string.h>
#include <ctype.h>
#include "simpleeditor.h"

static const int maxTextSpace = 2048;

SimpleEditor::SimpleEditor(int r, int c)
     : (r, c, 8, Reversed)
{
    SetClassName("SimpleEditor");

    input = new Sensor;
    input->Catch(KeyEvent);
    input->Catch(UpEvent);
    input->Catch(DownEvent);
    Listen(input);

    region = new char[maxTextSpace];
    TextEditor::Edit(new TextBuffer(region, 0, maxTextSpace));
    modified = false;
}

SimpleEditor::~SimpleEditor() {
    delete text;
    delete region;
    Unref(input);
};

void SimpleEditor::Edit(const char* str) {
    BeginningOfText();
    DeleteText(text->Length());
    InsertText(str, strlen(str));
    BeginningOfText();
    ScrollToSelection();
}

int SimpleEditor::Length() {
    return text->Length();
}

void SimpleEditor::Copy(char* buffer) {
    const char* contents = text->Text();
    bcopy(contents, buffer, text->Length());
    buffer[text->Length()] = '\0';
}

void SimpleEditor::ModifyFlagChanged() {
}

void SimpleEditor::Handle(Event& e) {
    int m1, m2;

    switch (e.eventType) {
      case DownEvent:
	switch (e.button) {
	  case LEFTMOUSE:
	    Select(Locate(e.x, e.y));
	    input->Catch(MotionEvent);
	    Listen(input);
	    ScrollToSelection();
	    break;
	  case MIDDLEMOUSE:
	    if (e.shift) {
		m1 = min(Dot(), Mark());
		m2 = max(Dot(), Mark());

		if (m2 > m1) {
		    char* buffer = new char[m2 - m1];
		    bcopy(text->Text(m1, m2), buffer, m2 - m1);
		    Select(Locate(e.x, e.y));

		    InsertText(buffer, m2 - m1);
		    delete buffer;

		    SetModified(true);
		    ScrollToSelection();
		}
	    }
	    else {
		GrabScroll(e);
	    }
	    break;
	  case RIGHTMOUSE:
	    if (e.shift) {
		SelectMore(Locate(e.x, e.y));
		ScrollToSelection();
	    }
	    else {
		RateScroll(e);
	    }
	    break;
	  default:
	    break;
	}
	break;
      case UpEvent:
	switch (e.button) {
	  case LEFTMOUSE:
	    input->Ignore(MotionEvent);
	    Listen(input);
	    break;
	  default:
	    break;
	}
	break;
      case MotionEvent:
	SelectMore(Locate(e.x, e.y));
	ScrollToSelection();
	break;
      case KeyEvent:
	if (e.meta)
	    HandleMetaChar(e.keystring[0]);
	else
	    HandleChar(e.keystring[0]);
	break;
      default:
	break;
    }
}

void SimpleEditor::HandleChar(char c) {
    int dot, here, column;
    char buffer[1];

    switch (c) {
      case '\001':
	BeginningOfLine();
	ScrollToSelection();
	break;
      case '\002':
	BackwardCharacter(1);
	ScrollToSelection();
	break;
      case '\004':
	DeleteText(1);
	SetModified(true);
	ScrollToSelection();
	break;
      case '\005':
	EndOfLine();
	ScrollToSelection();
	break;
      case '\006':
	ForwardCharacter(1);
	ScrollToSelection();
	break;
      case '\011':
	buffer[0] = c;
	InsertText(buffer, 1);
	SetModified(true);
	ScrollToSelection();
	break;
      case '\012':
	buffer[0] = c;
	InsertText(buffer, 1);
	SetModified(true);
	ScrollToSelection();
	break;
      case '\013':
	here = Dot();
	EndOfLine();
	dot = Dot();
	Select(here);
	DeleteText((dot == here) ? 1 : (dot - here));
	SetModified(true);
	ScrollToSelection();
	break;
      case '\015':
	buffer[0] = '\012';
	InsertText(buffer, 1);
	SetModified(true);
	ScrollToSelection();
	break;
      case '\016':
	/* Get column number */
	here = Dot();
	BeginningOfLine();
	column = here - Dot();
	Select(here);
	
	ForwardLine(1);
	
	/* Position to column */
	here = Dot();
	EndOfLine();
	if ((Dot() - here) > column) {
	    BeginningOfLine();
	    ForwardCharacter(column);
	}
	ScrollToSelection();
	break;
      case '\017':
	buffer[0] = '\012';
	InsertText(buffer, 1);
	BackwardCharacter(1);
	SetModified(true);
	ScrollToSelection();
	break;
      case '\020':
	column = text->LineOffset(Dot());
	BackwardLine(1);
	here = Dot();
	EndOfLine();
	if ((Dot() - here) > column) {
	    BeginningOfLine();
	    ForwardCharacter(column);
	}
	ScrollToSelection();
	break;
      case '\177':
	DeleteText(-1);
	SetModified(true);
	ScrollToSelection();
	break;
      default:
	if (isprint(c)) {
	    buffer[0] = c;
	    InsertText(buffer, 1);
	    SetModified(true);
	    ScrollToSelection();
	}
	break;
    }
}

void SimpleEditor::HandleMetaChar(char c) {
    int here;
    int dot;

    switch (c) {
      case 'b':
	PrevWord();
	ScrollToSelection();
	break;
      case 'd':
	here = Dot();
	NextWord();
	dot = Dot();
	Select(here);
	if (dot > here) DeleteText(dot - here);
	SetModified(true);
	ScrollToSelection();
	break;
      case 'f':
	NextWord();
	ScrollToSelection();
	break;
      case '\177':
	here = Dot();
	PrevWord();
	dot = Dot();
	Select(here);
	if (dot < here) DeleteText(dot - here);
	SetModified(true);
	ScrollToSelection();
	break;
      default:
	break;
    }
}

static inline int IsWhite(char c) {
    return ((c == ' ') || (c == '\t') || (c == '\n'));
}

void SimpleEditor::NextWord() {
    int dot = Dot();

    /*
     * Skip initial white space.
     */
    while (! text->IsEndOfText(dot) && IsWhite(text->Char(dot)))
	dot++;

    /*
     * Skip until whit space.
     */
    while (! text->IsEndOfText(dot) && ! IsWhite(text->Char(dot)))
	dot++;

    Select(dot);
}

void SimpleEditor::PrevWord() {
    int dot = Dot();

    while (! text->IsBeginningOfText(dot) && IsWhite(text->Char(dot - 1)))
	dot--;

    while (! text->IsBeginningOfText(dot) && ! IsWhite(text->Char(dot - 1)))
	dot--;

    Select(dot);
}
