/*
 * Copyright (c) 1987, 1988, 1989, 1990, 1991 Stanford University
 * Copyright (c) 1991 Silicon Graphics, Inc.
 * Copyright (c) 1991 Yufan Hu (yufan@cs.keele.ac.uk)
 *		Dept. of Computer Science, University of Keele, UK
 * 
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * Stanford and Silicon Graphics may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of Stanford and Silicon Graphics.
 *
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
 *
 * IN NO EVENT SHALL STANFORD OR SILICON GRAPHICS OR KEELE UNIVERSITY OR
 * Yufan Hu BE LIABLE FOR ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL 
 * DAMAGES OF ANY KIND, OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF 
 * USE, DATA OR PROFITS, WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, 
 * AND ON ANY THEORY OF LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE 
 * USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include <InterViews/textdisplay.h>
#include <InterViews/textbuffer.h>
#include <InterViews/painter.h>
#include <InterViews/shape.h>
#include <InterViews/bitmap.h>
#include <InterViews/cursor.h>
#include <InterViews/font.h>
#include <InterViews/sensor.h>
#include <InterViews/event.h>
#include <InterViews/world.h>
#include <InterViews/canvas.h>
#include <InterViews/Bitmaps/hand.bm>
#include <InterViews/Bitmaps/handMask.bm>
#include <InterViews/Bitmaps/lfast.bm>
#include <InterViews/Bitmaps/lfastMask.bm>
#include <InterViews/Bitmaps/rfast.bm>
#include <InterViews/Bitmaps/rfastMask.bm>
#include <string.h>
#include <stdio.h>
#include <unistd.h>
#include <ctype.h>

#include "Iediteditor.h"
#include "xselection.h"

static void beep(){
	World::current()->RingBell(1);
}

static Cursor* handCursor;
static Cursor* leftCursor;
static Cursor* rightCursor;

void SelStrEditor :: Reconfig(){
	if (handCursor == nil) {
		Bitmap hand(
			hand_bits, hand_width, hand_height, hand_x_hot, hand_y_hot
		);
		Bitmap handmask(hand_mask_bits, hand_mask_width, hand_mask_height);
		Bitmap left(
			lfast_bits, lfast_width, lfast_height, lfast_x_hot, lfast_y_hot
		);
		Bitmap leftmask(lfast_mask_bits, lfast_mask_width, lfast_mask_height);
		Bitmap right(
			rfast_bits, rfast_width, rfast_height, rfast_x_hot, rfast_y_hot
		);
		Bitmap rightmask(rfast_mask_bits, rfast_mask_width, rfast_mask_height);

		handCursor = new Cursor(
			&hand, &handmask, output->GetFgColor(), output->GetBgColor()
		);
		leftCursor = new Cursor(
			&left, &leftmask, output->GetFgColor(), output->GetBgColor()
		);
		rightCursor = new Cursor(
			&right, &rightmask, output->GetFgColor(), output->GetBgColor()
		);
	}

	const Font* f = output->GetFont();
	shape->Rect(f->Width(sample), f->Height());
	shape->Rigid(hfil, hfil, 0, 0);
	display->LineHeight(f->Height());
	StringEditor::Reconfig();
	display->TabWidth(output->GetFont()->Width("n")*4);
}

SelStrEditor :: SelStrEditor(
	ButtonState* button, const char* sample, const char* done
	) : StringEditor(button, sample, done) {
	_clipboard = new Xselection(this);
	SetClassName("SelStrEditor");
}

SelStrEditor :: SelStrEditor(
	const char* name, ButtonState* button, const char* sample, const char* done
	) : StringEditor(name, button, sample,  done ) { 
	_clipboard = new Xselection(this);
	SetClassName("SelStrEditor");
	SetInstance(name);
}

SelStrEditor::~SelStrEditor(){ delete _clipboard; }

boolean SelStrEditor :: HandleChar(char c ){
	switch(c){
	case '\t':
		InsertText(&c,1); return false;
	default:
		return StringEditor::HandleChar(c);
	}
}

#include "key4code.h"

void SelStrEditor :: Handle(Event& e){
	if(e.target != this) { e.target->Handle(e); return;}
	boolean done = false;
	display->Draw(output, canvas);
	display->CaretStyle(BarCaret);
	do {
		switch (e.eventType) {
		  case KeyEvent:
			if (e.len != 0) {
				done = HandleChar(e.keystring[0]);
			} else {
				switch(e.keycode()){
				case KeyCopy:
					Copy(); break;
				case KeyCut:
					Cut(); break;
				case KeyPaste:
					Paste(); break;
				case KeyLeftArraw:
					if(left > 0) Select(left-1);
					break;
				case KeyRightArraw:
					Select(right+1); break;
				}
			}
			break;
		  case DownEvent:
			if (e.target == this) {
				int origin = display->Left(0, 0);
				int width = display->Width();
				if (e.button == LEFTMOUSE) {
					int start = display->LineIndex(0, e.x);
					do {
						if (e.x < 0) {
							origin = min(0, origin - e.x);
						} else if (e.x > xmax) {
							origin = max(xmax - width, origin - (e.x - xmax));
						}
						display->Scroll(0, origin, ymax);
						DoSelect(start, display->LineIndex(0, e.x));
						Poll(e);
					} while (e.leftmouse);
				} else if (e.button == MIDDLEMOUSE) {
					if(e.control_is_down()){
						Paste(); break;
					}
					Cursor* origCursor = GetCursor();
					SetCursor(handCursor);
					int x = e.x;
					do {
						origin += e.x - x;
						origin = min(0, max(min(0, xmax - width), origin));
						display->Scroll(0, origin, ymax);
						x = e.x;
						Poll(e);
					} while (e.middlemouse);
					SetCursor(origCursor);
				} else if (e.button == RIGHTMOUSE) {
					if(e.control_is_down()){
						Copy(); break;
					}
					Cursor* origCursor = GetCursor();
					int x = e.x;
					do {
						origin += x - e.x;
						origin = min(0, max(min(0, xmax - width), origin));
						display->Scroll(0, origin, ymax);
						if (e.x - x < 0) {
							SetCursor(leftCursor);
						} else {
							SetCursor(rightCursor);
						}
						Poll(e);
					} while (e.rightmouse);
					SetCursor(origCursor);
				}
			} else {
				UnRead(e);
				done = true;
			}
			break;
		case LeaveEvent:
			done = true; break;
		}
		if (!done) {
			Read(e);
		}
	} while (!done);
	display->CaretStyle(NoCaret);
}

void SelStrEditor :: Copy(){
	if(left != right){
		_clipboard->Text(text->Text(left, right));
		Select(left);
	}
}

void SelStrEditor :: Cut() {
	if(left != right) {
		_clipboard->Text(text->Text(left, right));
		HandleChar('\177');
	}
}

void SelStrEditor :: Paste(){
	const char* clip = _clipboard->Text();
	while(*clip && !HandleChar(*clip)) clip++;
}

SelEditor :: SelEditor(
	int rows, int cols, int tabsize, int highlight
	):TextEditor( rows, cols, tabsize, highlight) {
	_clipboard = new Xselection(this);
	dirty = false; readonly = false; wrap = false;
	tabsize_ = tabsize;
	SetClassName("SelEditor");
}

SelEditor :: SelEditor(
	const char* name, int rows, int cols, int tabsize, int highlight
	):TextEditor(name, rows, cols, tabsize, highlight) {
	_clipboard = new Xselection(this);
	dirty = false; readonly = false; wrap = false;
	tabsize_ = tabsize;
	SetClassName("SelEditor");
}

SelEditor::~SelEditor(){ delete _clipboard; }

void SelEditor :: Reconfig(){
	TextEditor::Reconfig();
	display->TabWidth(output->GetFont()->Width("n")*tabsize_);
}

void SelEditor :: InsertText(const char* string, int count){
	if(readonly){ beep(); return; }
	if(count){
		if( wrap && *string != '\n' && count == 1 ) DoWrap(Dot());
		TextEditor :: InsertText(string, count);
		if(indent && *string == '\n' && count == 1 ) DoIndent('\0');
		dirty = true;
	}
}

int SelEditor :: DoIndent(const char nl){
	int line = text->LineNumber(Dot());
	int inserted = 0;
	char buf[256];
	if(nl == '\n') { buf[inserted++] = nl; line++;}
	if(line > 0 && indent){
		line--;
		int dot = Dot();
		for(int prev=text->LineIndex(line);
				inserted < 255 && prev < dot; prev++){
			char c = text->Char(prev);
			if(c == ' ' || c == '\t'){
				buf[inserted++] = c;
			}
			else break;
		}
	}
	if(inserted) 
		TextEditor :: InsertText(buf, inserted);
	return inserted;
}

static boolean LeadingBlanks(TextBuffer* text, int dot){
	while(! text->IsBeginningOfLine(dot)){
		char c = text->Char(--dot);
		if(c != ' ' && c != '\t') return false;
	}
	return true;
}

static boolean WordChar(const char c){
	return isalnum(c) || strchr("-._", c) != nil; /* No-Alpah-Numeric word characters */
}

void SelEditor :: DoWrap(int dot){
	Canvas* canvas = GetCanvas();
	Coord canvasRight = (Coord) canvas->width(); /* Strange Warning here ??? */
	int line = text->LineNumber(dot);
	int index = text->LineOffset(dot);
	Coord textRight = display->Right(line, index);
	if(textRight < canvasRight) return;
	while(WordChar(text->Char(Dot()-1)) && ! text->IsBeginningOfLine(Dot())){
		BeginningOfWord();
		if(!WordChar(text->Char(Dot()-1))) break;
		Select(Dot()-1);
	}
	if(! LeadingBlanks(text, Dot())){
		dot = dot +  DoIndent('\n');
	}
	Select(dot);
}

void SelEditor :: Edit(TextBuffer* buff, int index){
	dirty = false;
	TextEditor::Edit(buff, index);
	if(output){
		display->TabWidth(output->GetFont()->Width("n")*tabsize_);
		Draw();
	}
}

void SelEditor :: DeleteText(int count){
	if(readonly){ beep(); return; }
	if(count){
		if(Dot() != Mark())
			DeleteSelection();
		else 
			TextEditor :: DeleteText( count);
		dirty = true;
	}
}

void SelEditor :: DeleteSelection(){
	if(readonly){ beep(); return; }
	if(Dot() != Mark()) {
		Copy();
		TextEditor :: DeleteSelection();
		dirty = true;
	}
}

void SelEditor :: Copy(){
	int beg = Dot(), end = Mark();
	if(beg>end){
		int tmp = beg; beg = end; end = tmp;
	}
	int sel = end - beg;
	if(sel){
		_clipboard->Text(text->Text(beg, end),sel);
	}
}

void SelEditor :: Cut() {
	if(Dot() != Mark()){
		Copy(); DeleteSelection();
	}
}

const char* SelEditor :: CutBuffer(){
	return _clipboard->Text();
}

void SelEditor :: Paste(const char* buf){
	if(readonly){ beep(); return; }
	TextEditor::DeleteSelection();
	
	const char* s = buf;
	if(s == nil) s = _clipboard->Text();
	InsertText(s, strlen(s));
	ScrollToSelection();
}

void SelEditor::TabSize(int ts) {
	tabsize_ = ts;
	if(output) {
		Reconfig();
		Draw();
	}
}

void SelEditor :: SetFont(const char* font){
	output->SetFont(new Font(font));
	Draw();
}