// -*- c++ -*-
// Interviews tester class
// dave sternlicht mit X consortium

#include "tester.h"

//
// Tester base class
//

Tester::Tester(Session* ses, int repeat) : Glyph() {
  alloc_ = new Allocation();
  ses_ = ses;
  style_ = ses_->style();
  dpy_ = ses_->default_display();
  brush_ = new Brush(BRUSH_SIZE);
  black_ = Color::lookup(ses_->default_display(), "black");
  gray_ = Color::lookup(ses_->default_display(), "gray");
  Resource::ref(black_);
  Resource::ref(gray_);
  syncTime_ = 0.0;
  repeat_ = repeat;
}

void Tester::allocate(Canvas* c, const Allocation& a, Extension& e) {
  Glyph::allocate(c, a, e);
  *alloc_ = a;
  canvas_ = c;
}

Tester::~Tester() {
  Resource::unref(black_); Resource::unref(gray_);   
}

void Tester::init(Canvas* c, int reps) {
  c->fill_rect(alloc_->left(), alloc_->bottom(), 
	       alloc_->right(), alloc_->top(), gray_);  
  dpy_->flush();
}

int Tester::calibrateSyncTest(int seconds, 
			      double* usecperobj, int objects) const {
  const double goal =  2500000.0;   // Try to get up to 2.5 seconds
  const double enough = 2000000.0;  // But settle for 2.0 seconds             
  const double tick =   10000.0;    // Assume clock not faster than .01 seconds
  int reps = 1;
  double usecs = 0.0;
  int exponent;
  // Attempt to get an idea how long each rep lasts by getting enough
  // reps to last more than enough.  Then scale that up to the number of
  // seconds desired.
  for (;;) {
    hardwareSync(canvas_);
    initTimes();
    doHardwareSync(canvas_, reps);
    hardwareSync(canvas_);
    usecs += elapsedTime(syncTime_);

    /* Did we go long enough? */
    if (usecs >= enough) break;
    
    /* Don't let too short a clock make new reps wildly high */
    if (usecs < tick) usecs = tick;
    
    /* Try to get up to goal seconds. */
    reps = (int) (goal * (double)reps / usecs) + 1;
  }
  *usecperobj = usecs / (double) (reps * objects);
  reps = (int) ((double)seconds * 1000000.0 * (double)reps / usecs) + 1;
  
  /* Now round reps up to 1 digit accuracy, so we don't get stupid-looking
    numbers of repetitions. */
  reps--;
  exponent = 1;
  while (reps > 9) {
    reps /= 10;
    exponent *= 10;
  }
  reps = (reps + 1) * exponent;
  printf("Sync time adjustment is %6.4f msecs.\n\n", *usecperobj/1000);
  return reps;
}

int Tester::calibrateTest(int seconds, double* usecperobj,
			       int objects) const {
  const double goal =  2500000.0;   // Try to get up to 2.5 seconds
  const double enough = 2000000.0;  // But settle for 2.0 seconds              
  const double tick =    10000.0;   // Assume clock not faster than .01 seconds
  int reps = 1;
  double usecs = 0.0;
  int exponent;
  printf("calibrating...\n");
  // Attempt to get an idea how long each rep lasts by getting enough
  // reps to last more than enough.  Then scale that up to the number of
  // seconds desired.
  for (;;) {
    // initialization changes an object by definition so cast it non-const.
    ((Tester*)this)->init(canvas_, reps);	
    hardwareSync(canvas_);
    initTimes();
    proc(canvas_, reps);
    hardwareSync(canvas_);
    usecs += elapsedTime(syncTime_);

    /* Did we go long enough? */
    if (usecs >= enough) break;
    
    /* Don't let too short a clock make new reps wildly high */
    if (usecs < tick) usecs = tick;
    
    /* Try to get up to goal seconds. */
    reps = (int) (goal * (double)reps / usecs) + 1;
  }
  *usecperobj = usecs / (double) (reps * objects);
  reps = (int) ((double)seconds * 1000000.0 * (double)reps / usecs) + 1;
  
  /* Now round reps up to 1 digit accuracy, so we don't get stupid-looking
    numbers of repetitions. */
  reps--;
  exponent = 1;
  while (reps > 9) {
    reps /= 10;
    exponent *= 10;
  }
  reps = (reps + 1) * exponent;

  return reps;
}

void Tester::draw(Canvas* c, const Allocation& a) const {
  double usecs = 0.0;
  int reps = 1;
  
  // calibrate sync time
  calibrateSyncTest(1, &((Tester*)this)->syncTime_, 1);

  // do the test
  processTest();

  exit(0);
}

void Tester::processTest() const {
  int j, reps;
  double time, totalTime = 0;
  reps = calibrateTest(5, &time, objects_);
  printf("doing test...\n");
  for (j = 0; j != repeat_; j++) {
    time = doTest(reps);
    totalTime += time;
    reportTimes(time, reps * objects_, False);
  }
  if (repeat_ > 1) 
    reportTimes(totalTime, repeat_ * reps * objects_, True);
}

double Tester::doTest(int reps) const {
  double time;
  ((Tester*)this)->init(canvas_, reps);
  hardwareSync(canvas_);
  initTimes();
  proc(canvas_, reps);
  hardwareSync(canvas_);
  time = elapsedTime(syncTime_);
  cleanup(canvas_);
  return time;
}

double Tester::roundTo3Digits(double d) const
{
    /* It's kind of silly to print out things like ``193658.4/sec'' so just
       junk all but 3 most significant digits. */

    double exponent, sign;

    exponent = 1.0;
    /* the code below won't work if d should happen to be non-positive. */
    if (d < 0.0) {
	d = -d;
	sign = -1.0;
    } else
	sign = 1.0;
    if (d >= 1000.0) {
	do {
	    exponent *= 10.0;
	} while (d/exponent >= 1000.0);
	d = (double)((int) (d/exponent + 0.5));
	d *= exponent;
    } else {
	if (d != 0.0) {
	    while (d*exponent < 100.0) {
	        exponent *= 10.0;
	    }
	}
	d = (double)((int) (d*exponent + 0.5));
	d /= exponent;
    }
    return d * sign;
}

void Tester::reportTimes(double usecs, int n, boolean average) const {
  double msecsperobj, objspersec;
  
  msecsperobj = usecs / (1000.0 * (double)n);
  objspersec = (double) n * 1000000.0 / usecs;

  // Round obj/sec to 3 significant digits.  Leave msec untouched, to
  // allow averaging results from several repetitions.
  objspersec = roundTo3Digits(objspersec);

  if (average) 
    printf("%6d trep @ %7.4f msec (%6.1f/sec): ",
	   n, msecsperobj, objspersec);
  else printf("%6d reps @ %7.4f msec (%6.1f/sec): ",
	      n, msecsperobj, objspersec);
}

void Tester::doHardwareSync(Canvas* c, int reps) const {
  int i;
  for (i = 0; i != reps; i++)
    hardwareSync(c);
}

void Tester::hardwareSync(Canvas* c) const{
  XImage *image;
  image = XGetImage(dpy_->rep()->display_,
		    c->rep()->window_->rep()->xwindow_,
		    WIDTH-1, HEIGHT-1, 1, 1, ~0, ZPixmap);
  XDestroyImage(image);
}

static struct timeval start;

void Tester::initTimes() const
{
    struct timezone foo;
    gettimeofday(&start, &foo);
}

double Tester::elapsedTime(double correction) const
{
    struct timeval stop;
    struct timezone foo;
    
    gettimeofday(&stop, &foo);
    if (stop.tv_usec < start.tv_usec) {
        stop.tv_usec += 1000000;
	stop.tv_sec -= 1;
    }
    return  (double)(stop.tv_usec - start.tv_usec) +
            (1000000.0 * (double)(stop.tv_sec - start.tv_sec)) - correction;
}



