// This file is part of xplot.
// 
// Copyright (C) 1989 Rene' Seindal (seindal@diku.dk)
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.


#include "xplot.hh"
#include "util.hh"
#include "plot.hh"

#include <InterViews/world.h>
#include <InterViews/painter.h>
#include <InterViews/brush.h>
#include <InterViews/font.h>

#include <ctype.h>
#include <math.h>

// enumerator for different lines styles
enum PlotLinemodes {
    PlotLineSolid,
    PlotLineDotted,
    PlotLineLongdashed,
    PlotLineShortdashed,
    PlotLineDotdashed,
    PlotLineMAXMODES,
};

static const char *brush_props[] = {
    "lineSolid",
    "lineDotted",
    "lineLongdashed",
    "lineShortdashed",
    "lineDotdashed",
};

// Brushes shared by all PlotGraphs (will never be deleted)
static Brush *brushes[PlotLineMAXMODES];

// Default area if nothing is specified for a graph
static const int default_minx = 0;
static const int default_miny = 0;
static const int default_maxx = 4096;
static const int default_maxy = 4096;

// Number of command kept in one PlotGraphSegment.
static const int SegmentSize = 1024;


inline int PlotCmd::Code(Coord px, Coord py, Coord x1, Coord y1, Coord x2, Coord y2)
{
    return (((px < x1) ? 0x01 : 0) | ((x2 < px) ? 0x02 : 0) | 
	    ((py < y1) ? 0x04 : 0) | ((y2 < py) ? 0x08 : 0));
}

void PlotCmd::Draw(Canvas *, Painter *,
		   Coord&, Coord&,
		   Coord, Coord, Coord, Coord)
{
}

void PlotMove::Draw(Canvas *, Painter *,
		    Coord& curx, Coord& cury,
		    Coord, Coord, Coord, Coord)
{
    curx = x;
    cury = y;
}

void PlotCont::Draw(Canvas *canvas, Painter *output,
		    Coord& curx, Coord& cury,
		    Coord minx, Coord miny, Coord maxx, Coord maxy)
{
    if ((Code(curx, cury, minx, miny, maxx, maxy) &
	 Code(x, y, minx, miny, maxx, maxy)) == 0)
	output->Line(canvas, curx, cury, x, y);  
    curx = x;
    cury = y;
}

void PlotPoint::Draw(Canvas *canvas, Painter *output,
		     Coord& curx, Coord& cury,
		     Coord minx, Coord miny, Coord maxx, Coord maxy)
{
    if (Code(x, y, minx, miny, maxx, maxy) == 0)
	output->Point(canvas, x, y);
    curx = x;
    cury = y;
}

void PlotLine::Draw(Canvas *canvas, Painter *output,
		    Coord& curx, Coord& cury,
		    Coord minx, Coord miny, Coord maxx, Coord maxy)
{
    if ((Code(x1, y1, minx, miny, maxx, maxy) &
	 Code(x2, y2, minx, miny, maxx, maxy)) == 0)
	output->Line(canvas, x1, y1, x2, y2);
    curx = x2;
    cury = y2;
}

PlotLabel::PlotLabel(const char *t, PlotGraph *g)
{
    text = StringDup(t);
    graph = g;
}

PlotLabel::~PlotLabel()
{
    delete text;
}

void PlotLabel::Draw(Canvas *canvas, Painter *output,
		     Coord& curx, Coord& cury,
		     Coord minx, Coord miny, Coord maxx, Coord maxy)
{
    float textscale = graph->GetTextScale();
    Font *f = output->GetFont();
    int width = round(f->Width(text) * textscale);
    int height = round(f->Height() * textscale);

    if (width <= 5 || height <= 5)	// XXX
	return;

    int codell = Code(curx, cury, minx, miny, maxx, maxy);
    int codelr = Code(curx + width, cury + height, minx, miny, maxx, maxy);
    int codeul = Code(curx + width, cury, minx, miny, maxx, maxy);
    int codeur = Code(curx, cury + height, minx, miny, maxx, maxy);

    if ((codell & codelr) == 0 || (codelr & codeur) == 0
	|| (codeur & codeul) == 0 || (codeul & codell) == 0)
    {
	output->Scale(textscale, textscale);
	textscale = 1.0/textscale;
	output->Text(canvas, text, round(curx*textscale), round(cury*textscale));
	output->Scale(textscale, textscale);
    }
}

void PlotArc::Draw(Canvas *canvas, Painter *output,
		   Coord&, Coord&,
		   Coord minx, Coord miny, Coord maxx, Coord maxy)
{
    float theta0, theta1, delta;
    int i;
    int rx, ry;
    int px, py;				// prev point
    int nx, ny;				// next point
    int curcode, nextcode;
    
    rx = sx - cx;
    ry = sy - cy;
    
    theta0 = atan2((double)ry, (double)rx);
    theta1 = atan2((double)(ey - cx), (double)(ex - cx));
    
    if (theta0 >= theta1)
	theta1 += 2*M_PI;
    
    delta = theta1 - theta0;
    if ((i = round(delta/M_PI*180)) < 4)
	i = 4;
    delta /= i;
    
    px = sx;
    py = sy;
    curcode = Code(sx, sy, minx, miny, maxx, maxy);

    for (theta0 = delta; --i >= 0; theta0 += delta) {
	float c = cos(theta0);
	float s = sin(theta0);
	nx = cx + round(rx*c - ry*s);
	ny = cy + round(rx*s + ry*c);
	nextcode = Code(nx, ny, minx, miny, maxx, maxy);
	if ((curcode & nextcode) == 0)
	    output->Line(canvas, px, py, nx, ny);
	px = nx;
	py = ny;
	curcode = nextcode;
    }
}

void PlotCircle::Draw(Canvas *canvas, Painter *output,
		      Coord& curx, Coord& cury,
		      Coord minx, Coord miny, Coord maxx, Coord maxy)
{
    if (!((cx+r < minx) || (cx-r > maxx) || (cy+r < miny) || (cy-r > maxy)))
	output->Circle(canvas, cx, cy, r);
    curx = cx;
    cury = cy;
}

void PlotLinemod::Draw(Canvas *, Painter *output,
		       Coord&, Coord&,
		       Coord, Coord, Coord, Coord)
{
    output->SetBrush(brush);
}


class PlotGraphSegment {
private:
    PlotGraphSegment *next;		// next segment
public:
    int ncmds;				// number of commands in this segment
    PlotCmd *cmds[SegmentSize];		// actual commands

    PlotGraphSegment(PlotGraphSegment *);
    ~PlotGraphSegment();

    PlotGraphSegment *Next() {
	return next;
    };
};

PlotGraphSegment::PlotGraphSegment(PlotGraphSegment *prev)
{
    ncmds = 0;
    next = nil;
    if (prev != nil)
	prev->next = this;
}

PlotGraphSegment::~PlotGraphSegment()
{
    while (--ncmds >= 0)
	delete cmds[ncmds];
    ncmds = 0;
    if (next) {
	delete next;
	next = nil;
    }
}


PlotGraph::PlotGraph(FILE *fp, PlotGraph *previous)
{
    if (previous == nil) {
	next = prev = nil;
    } else {
	next = previous->next;
	prev = previous;
	if (next != nil)
	    next->prev = this;
	previous->next = this;
    }
    big_endian = IsOn(world->GetAttribute("bigEndian"));
    textscale = 1.0;
    ReadGraph(fp);
}

PlotGraph::~PlotGraph()
{
    delete first;

    if (next != nil)
	next->prev = prev;
    if (prev != nil)
	prev->next = next;
    next = prev = nil;
}

void PlotGraph::ReadGraph(FILE *fp)
{
    PlotCmd *cmd;
    PlotGraphSegment *current;
    int opcode;
    int px, py, px1, py1, px2, py2, r;
    
    current = first = new PlotGraphSegment(nil);
    
    // Copy area size from previous graph or use the default values.
    // May be overridded for this graph by an 's' command later.
    if (prev != nil) {
	minx = prev->minx;
	miny = prev->miny;
	maxx = prev->maxx;
	maxy = prev->maxy;
    } else {
	minx = default_minx;
	miny = default_miny;
	maxx = default_maxx;
	maxy = default_maxy;
    }
    
    while ((opcode = GetCh(fp)) != EOF) {
	
	cmd = nil;
	
	switch (opcode) {
	case 'e':			// end of graph
	    break;
	case 'm':			// move
	    px = GetSi(fp);
	    py = GetSi(fp);
	    cmd = new PlotMove(px, py);
	    break;
	case 'n':			// continued line
	    px = GetSi(fp);
	    py = GetSi(fp);
	    cmd = new PlotCont(px, py);
	    break;
	case 'p':			// point
	    px = GetSi(fp);
	    py = GetSi(fp);
	    cmd = new PlotPoint(px, py);
	    break;
	case 'l':			// line
	    px = GetSi(fp);
	    py = GetSi(fp);
	    
	    px1 = GetSi(fp);
	    py1 = GetSi(fp);
	    cmd = new PlotLine(px, py, px1, py1);
	    break;
	case 't':
	    ;{
		char *s, *l;

		// We ignore labels consisting of whitespace only, since
		// they are often used to break output lines in data
		// generated by graph(1).  A bit kludgy, maybe, but it
		// makes the display look much nicer.
		l = GetStr(fp);
		for (s = l; *s && isspace(*s); s++)
		    ;
		if (*s)
		    cmd = new PlotLabel(l, this);
	    }
	    break;
	case 'a':
	    px = GetSi(fp);
	    py = GetSi(fp);
	    
	    px1 = GetSi(fp);
	    py1 = GetSi(fp);
	    px2 = GetSi(fp);
	    py2 = GetSi(fp);
	    
	    cmd = new PlotArc(px, py, px1, py1, px2, py2);
	    break;
	case 'c':			// circle
	    px = GetSi(fp);
	    py = GetSi(fp);
	    
	    r = GetSi(fp);
	    cmd = new PlotCircle(px, py, r);
	    break;
	case 'f':			// linemode
	    ;{
		char *l = GetStr(fp);
		int linestyle = -1;;

		if (strlen(l) < 4)
		    break;
		switch (l[3]) {		// the fourth character is unique!
		case 'i':
		    if (strcmp(l, "solid") == 0)
			linestyle = PlotLineSolid;
		    break;
		case 't':
		    if (strcmp(l, "dotted") == 0)
			linestyle = PlotLineDotted;
		    break;
		case 'g':
		    if (strcmp(l, "longdashed") == 0)
			linestyle = PlotLineLongdashed;
		    break;
		case 'r':
		    if (strcmp(l, "shortdashed") == 0)
			linestyle = PlotLineShortdashed;
		    break;
		case 'd':
		    if (strcmp(l, "dotdashed") == 0)
			linestyle = PlotLineDotdashed;
		    break;
		}
		if (linestyle >= 0) {
		    if (brushes[linestyle] == nil) {
			int pat;
			char *attr = StringDup(world->GetAttribute(brush_props[linestyle]));

			if (sscanf(attr, "0x%x", &pat) != 1)
			    pat = 0xffff;
			brushes[linestyle] = new Brush(pat, pat != 0xffff);
			delete attr;
		    }
		    cmd = new PlotLinemod(brushes[linestyle]);
		}
	    }
	    break;
	case 's':			// space
	    minx = GetSi(fp);
	    miny = GetSi(fp);
	    maxx = GetSi(fp);
	    maxy = GetSi(fp);
	    break;
	default:
	    fprintf(stderr, "xplot: Undefined plot command 0%o\n", cmd);
	    break;
	}
	if (opcode == 'e')
	    break;
	
	if (cmd != nil) {
	    if (current->ncmds >= SegmentSize)
		current = new PlotGraphSegment(current);
	    current->cmds[current->ncmds++] = cmd;
	}
    }

    // If there is no real commands, trash this ``graph''
    if (first->ncmds == 0) {
	delete first;
	first = nil;
    }

    // This test is kind of crap, but it does catch the common case of
    // (minx, miny) = (0, 0), (maxx, maxy) = (4096, 4096), and the
    // bytesex reversed.
    if ((maxx - minx <= 256 || maxy - miny <= 256)
	&& !IsOn(world->GetAttribute("forceOutput")))
    {
	delete first;
	first = nil;
	fprintf(stderr, "xplot: Tiny graph deleted (ranges from (%d, %d) to (%d, %d)).\n",
		minx, miny, maxx, maxy);
	fprintf(stderr, "xplot: If this is deliberate, use `-force' to override.\n");
    }
}


// GetCh() reads a single byte from stdin, not reading eof more than once,
// and not returning more than 8 bits.

int PlotGraph::GetCh(FILE *fp)
{
    int ch;
    
    if (feof(fp) || (ch = getc(fp)) == EOF)
	return EOF;
    return ch&0377;
}

// GetSi() return a signed short integer, XXX endian and sign extending.
// returns eof on end of file.

int PlotGraph::GetSi(FILE *fp)
{
    int m, n;
    int val;
    
    n = GetCh(fp);
    if ( n == EOF )
	return EOF;
    
    m = GetCh(fp);
    if ( m == EOF )
	return EOF;
    
    val = big_endian ? ((n << 8) | m) : ((m << 8) | n);
    if ( val & 0x8000 )
	val |= ~0x7fff;
    return val;
}

// GetStr() reads a string from stdin.  The string is terminated by EOF,
// newline or '\0'.

char *PlotGraph::GetStr(FILE *fp)
{
    char str[1024];
    char *s = str;
    int ch;
    
    while ( (ch = GetCh(fp)) != EOF && ch != '\n' && ch != '\0' )
	*s++ = ch;
    *s = '\0';
    
    return str;
}


PlotCmd *PlotGraphIterator::Next()
{
    if (curcmd >= cur->ncmds) {
	if (cur->Next() == nil) {
	    cur = nil;
	    return nil;
	} else {
	    cur = cur->Next();
	    curcmd = 0;
	}
    }
    return cur->cmds[curcmd++];
}


PlotGraphSet::PlotGraphSet(FILE *fp)
{
    PlotGraph *next = nil;
    
    nviews = 0;
    nsets++;

    while (!(feof(fp) || ferror(fp))) 
	next = new PlotGraph(fp, next);

    if (next == nil) {
	current = nil;
	return;
    }
    while (next->PrevGraph() != nil) {
	current = next->PrevGraph();
	if (!next->Valid())
	    delete next;
	next = current;
    }
    if (next == nil) {
	current = nil;
	return;
    }
    if (!next->Valid()) {
	current = next->NextGraph();
	delete next;
    } else
	current = next;
}

PlotGraphSet::~PlotGraphSet()
{
    PlotGraph *next;
    
    while (current != nil && current->PrevGraph() != nil)
	current = current->PrevGraph();
    while (current) {
	next = current->NextGraph();
	delete current;
	current = next;
    }
    DeleteViews();
    --nsets;
}

