// This file is part of xplot.
// 
// Copyright (C) 1989 Rene' Seindal (seindal@diku.dk)
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.


#ifndef plot_h
#define plot_h

class PlotCmd;				// Base class for plot commands.
class PlotMove;				// A `move current' command.
class PlotCont;				// A `continued line' command.
class PlotPoint;			// A `set a point' command.
class PlotLine;				// A `line' command.
class PlotLabel;			// A `label' command.
class PlotArc;				// A `circular arc command.
class PlotCircle;			// A `circle' command.
class PlotLinemod;			// A `change linestyle' command.

class PlotGraphSegment;			// A segment of a graph.
class PlotGraph;			// A complete graph.
class PlotGraphIterator;		// An iterator over all command in a graph.
class PlotGraphSet;			// A set of graphs.

#include <InterViews/subject.h>

class Painter;
class Canvas;
class Brush;
class Interactor;

typedef short PlotCoord;


class PlotCmd {
protected:
    int Code(Coord curx, Coord cury,
	     Coord minx, Coord miny, Coord maxx, Coord maxy);
public:
    virtual void Draw(Canvas *, Painter *,
		      Coord& curx, Coord& cury,
		      Coord minx, Coord miny, Coord maxx, Coord maxy);
};

class PlotMove : public PlotCmd {
private:
    PlotCoord x;
    PlotCoord y;
public:
    PlotMove(int xx, int yy) {
	x = xx; y = yy;
    };
    void Draw(Canvas *, Painter *, Coord&, Coord&,
	      Coord minx, Coord miny, Coord maxx, Coord maxy);
};


class PlotCont : public PlotCmd {
private:
    PlotCoord x;
    PlotCoord y;
public:
    PlotCont(int xx, int yy) {
	x = xx; y = yy;
    };
    void Draw(Canvas *, Painter *, Coord&, Coord&,
	      Coord minx, Coord miny, Coord maxx, Coord maxy);
};


class PlotPoint : public PlotCmd {
private:
    PlotCoord x;
    PlotCoord y;
public:
    PlotPoint(int xx, int yy) {
	x = xx; y = yy;
    };
    void Draw(Canvas *, Painter *, Coord&, Coord&,
	      Coord minx, Coord miny, Coord maxx, Coord maxy);
};


class PlotLine : public PlotCmd {
private:
    PlotCoord x1;
    PlotCoord y1;
    PlotCoord x2;
    PlotCoord y2;
public:
    PlotLine(int xx1, int yy1, int xx2, int yy2) {
	x1 = xx1; y1 = yy1;
	x2 = xx2; y2 = yy2;
    };
    void Draw(Canvas *, Painter *, Coord&, Coord&,
	      Coord minx, Coord miny, Coord maxx, Coord maxy);
};

class PlotLabel : public PlotCmd {
private:
    char *text;
    PlotGraph *graph;
public:
    PlotLabel(const char *, PlotGraph *);
    ~PlotLabel();
    void Draw(Canvas *, Painter *, Coord&, Coord&,
	      Coord minx, Coord miny, Coord maxx, Coord maxy);
};


class PlotArc : public PlotCmd {
private:
    PlotCoord cx;
    PlotCoord cy;
    PlotCoord sx;
    PlotCoord sy;
    PlotCoord ex;
    PlotCoord ey;
public:
    PlotArc(int cxx, int cyy, int sxx, int syy, int exx, int eyy) {
	cx = cxx; cy = cyy;
	sx = sxx; sy = syy;
	ex = exx; ey = eyy;
    };
    void Draw(Canvas *, Painter *, Coord&, Coord&,
	      Coord minx, Coord miny, Coord maxx, Coord maxy);
};

class PlotCircle : public PlotCmd {
private:
    PlotCoord cx;
    PlotCoord cy;
    PlotCoord r;
public:
    PlotCircle(int cxx, int cyy, int rr) {
	cx = cxx; cy = cyy;
	r = rr;
    };
    void Draw(Canvas *, Painter *, Coord&, Coord&,
	      Coord minx, Coord miny, Coord maxx, Coord maxy);
};

class PlotLinemod : public PlotCmd {
private:
    Brush *brush;
public:
    PlotLinemod(Brush *b) {
	brush = b;
    };
    void Draw(Canvas *, Painter *, Coord&, Coord&,
	      Coord minx, Coord miny, Coord maxx, Coord maxy);
};


class PlotGraph {
private:
    // Graph list
    PlotGraph *prev;			// previous in file
    PlotGraph *next;			// next in file

    int GetCh(FILE *);			// read a character
    boolean big_endian;			// true iff data is big endian
    int GetSi(FILE *);			// read a short integer
    char *GetStr(FILE *);		// read a string
    void ReadGraph(FILE *);		// read a complete graph
    
    // Plot commands
    PlotGraphSegment *first;		// first segment of plot commands

    friend class PlotGraphIterator;

    // Plot area size
    PlotCoord minx;
    PlotCoord miny;
    PlotCoord maxx;
    PlotCoord maxy;

    // Factor to scale fonts with for this graph.  It is set by
    // Canvas::Redraw for each redraw.
    float textscale;
public:
    PlotGraph(FILE *, PlotGraph *);	// read graph data from file, and insert into list
    ~PlotGraph();			// detach from list

    PlotGraph *NextGraph() {
	return next;
    };
    PlotGraph *PrevGraph() {
	return prev;
    };

    void GetArea(int& minxx, int& minyy, int& maxxx, int& maxyy) {
	minxx = minx; minyy = miny;
	maxxx = maxx; maxyy = maxy;
    };
    void SetTextScale(float x) {
	textscale = x;
    };
    float GetTextScale() {
	return textscale;
    };

    boolean Valid() {
	return first != 0;
    };
};


class PlotGraphIterator {
private:
    PlotGraphSegment *cur;
    int curcmd;
public:
    PlotGraphIterator(PlotGraph *p) {
	cur = p->first;
	curcmd = 0;
    };

    PlotCmd *Next();
};


class PlotGraphSet : public Subject {
private:
    PlotGraph *current;			// current graph
    int nviews;				// number of views on this graphset
    static int nsets;			// total number of graphsets.
public:
    PlotGraphSet(FILE *);
    ~PlotGraphSet();
    
    void Attach(Interactor *i) {
	Subject::Attach(i);
	nviews++;
    };
    void Detach(Interactor *i) {
	Subject::Detach(i);
	--nviews;
    };
    int NViews() {
	return nviews;
    };
    int NSets() {
	return nsets;
    };
    
    PlotGraph *Current() {
	return current;
    };
    PlotGraph *Previous() {
	return current->PrevGraph();
    };
    PlotGraph *Next() {
	return current->NextGraph();
    };
    void Forward() {
	if (Next()) {
	    current = Next();
	    Notify();
	}
    };
    void Backward() {
	if (Previous()) {
	    current = Previous();
	    Notify();
	}
    };
};

#endif /* plot_h */
