#include "peer.h"
#include "reader.h"
#include "writer.h"
#include <Dispatch/rpcstream.h>
#include <osfcn.h>		/* for close */

// If the peer is running, open a connection to it.  If not, create a
// socket so that the peer can open a connection to us.

Peer::Peer(const char* path) : RpcPeer(path),
    writer_(nil),
    reader_(nil),
    local_n_(-1),
    remote_n_(-1)
{
    init(path);
}

// Delete reader_ first to work around a bug.  RpcReader doesn't
// remember which file descriptor it's attached to, so it looks up the
// rpcstream's file descriptor when unlinking itself.  RpcWriter
// closes the rpcstream, so the stored file descriptor will be gone if
// writer_'s deleted first.

Peer::~Peer() {
    delete reader_;
    delete writer_;
}

// If the peer is running, send it our integer.  Then wait for the
// peer to send us its own integer.

int Peer::exchange(int n) {
    local_n_ = n;
    if (writer_) {
	writer_->send(local_n_);
    }
    run();
    return remote_n_;
}

// Store the peer's integer which was just received by Reader and
// relayed to us.  Our job is accomplished, so we can stop running.

void Peer::remote_n(int n) {
    remote_n_ = n;
    quitRunning();
}

// Open a connection to the peer's socket.  We'll send it our integer
// when main() calls exchange() later on.

boolean Peer::createReaderAndWriter(const char* rHost, int rPort) {
    writer_ = new Writer(rHost, rPort);
    if (writer_->server()) {
	reader_ = new Reader(&writer_->server(), this);
	return true;
    } else {
	cerr << "Peer: error opening connection to peer at ";
	cerr << rHost << "." << rPort << "\n";
	delete writer_;
	writer_ = nil;
	return false;
    }
}

// The peer just opened a connection to our socket.  Send it our
// integer since we've been waiting for the peer to appear.

void Peer::createReaderAndWriter(int fd) {
    if (writer_) {
	cerr << "Peer: can't talk to more than one peer\n";
	close(fd);
	return;
    }
    writer_ = new Writer(fd);
    if (writer_->server()) {
	reader_ = new Reader(&writer_->server(), this);
	writer_->send(local_n_);
    } else {
	cerr << "Peer: error accepting connection from peer on ";
	cerr << fd << "\n";
	delete writer_;
	writer_ = nil;
    }
}
