/*
 * Copyright (c) 1991 Redwood Design Automation
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the name of
 * Redwood Design Automation may not be used in any advertising or publicity
 * relating to the software without the specific, prior written permission of
 * Redwood Design Automation.
 *
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
 *
 * IN NO EVENT SHALL REDWOOD DESIGN AUTOMATION BE LIABLE FOR ANY SPECIAL,
 * INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND, OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER OR NOT
 * ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF LIABILITY,
 * ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
 * SOFTWARE.
 */

#include <InterViews/align.h>
#include <InterViews/superpose.h>
#include <InterViews/geometry.h>
#include <InterViews/canvas.h>
#include <InterViews/color.h>
#include <InterViews/brush.h>
#include <InterViews/hit.h>

#include <r.h>
#include <rutArray.h>
#include <ruiBox.h>
#include <ruiTile.h>

//////////////////////////////////////////////////

static float min (float a, float b) {
  if (a < b) return a;
  return b;
}

static Coord max (Coord a, Coord b) {
  if (a > b) return a;
  return b;
}

static DimensionName otherDim (DimensionName dim) {
  if (dim == Dimension_X) return Dimension_Y;
  return Dimension_X;
}

static ruiTile* getBoxTile (DimensionName dim) {
  if (dim == Dimension_X) return new ruiTile(Dimension_X);
  return new ruiTileReversed(Dimension_Y);
}

//////////////////////////////////////////////////

ruiBox::ruiBox(
    DimensionName dim,
    GlyphIndex size
) : Box(
    new Superpose(tile_ = getBoxTile(dim), new Align(otherDim(dim))),
    size
) { 
  dim_ = dim;
  knobsOn_ = true;
  ridgeOn_ = false;
  ridge_ = 0;
}

ruiBox::ruiBox(
    DimensionName dim,
    Glyph* g1, Glyph* g2, Glyph* g3, Glyph* g4, Glyph* g5,
    Glyph* g6, Glyph* g7, Glyph* g8, Glyph* g9, Glyph* g10
) : Box(
    new Superpose(tile_ = getBoxTile(dim), new Align(otherDim(dim))),
    g1, g2, g3, g4, g5, g6, g7, g8, g9, g10
) { 
  dim_ = dim;
  knobsOn_ = true;
  ridgeOn_ = false;
  ridge_ = 0;
  for (int i = 0; i < count(); i++) {
    tile_->insertLast();
  }
}

ruiBox::~ruiBox() { }

//////////////////////////////////////////////////

ruiHorzBox::ruiHorzBox(
    GlyphIndex size
) : ruiBox(
    Dimension_X,
    size
) { }

ruiHorzBox::ruiHorzBox(
    Glyph* g1, Glyph* g2, Glyph* g3, Glyph* g4, Glyph* g5,
    Glyph* g6, Glyph* g7, Glyph* g8, Glyph* g9, Glyph* g10
) : ruiBox(
    Dimension_X,
    g1, g2, g3, g4, g5, g6, g7, g8, g9, g10
) { }

ruiHorzBox::~ruiHorzBox() { }

ruiVertBox::ruiVertBox(
    GlyphIndex size
) : ruiBox(
    Dimension_Y,
    size
) { }

ruiVertBox::ruiVertBox(
    Glyph* g1, Glyph* g2, Glyph* g3, Glyph* g4, Glyph* g5,
    Glyph* g6, Glyph* g7, Glyph* g8, Glyph* g9, Glyph* g10
) : ruiBox(
    Dimension_Y,
    g1, g2, g3, g4, g5, g6, g7, g8, g9, g10
) { }

ruiVertBox::~ruiVertBox() { }

//////////////////////////////////////////////////

ruiKnob* ruiBox::newKnob(Coord xspan, Coord yspan) {
  ruiKnob* knob;
  knob = new ruiKnob(this, xspan, yspan);
  Resource::ref(knob);
  return knob;
}

ruiKnob* ruiBox::newRidge() {
  ruiKnob* ridge;
  if (dim_ == Dimension_X) {
    ridge = newKnob(4, allocation_.top() - allocation_.bottom());
  } else {
    ridge = newKnob(allocation_.right() - allocation_.left(), 4);
  }
  return ridge;
}

void ruiBox::allocateKnobs() { 
  // make enough knobs for current number of components
  int n = (int)count();
  int m = knobs_.size();
  int i;
  for (i = m; i < n; i++) {
    if (dim_ == Dimension_X) {
      knobs_.insertLast(newKnob(4,8));
    } else {
      knobs_.insertLast(newKnob(8,4));
    }
    knobAllocs_.insertLast(new Allocation);
  }
  //
  if (n < 2) return;
  //
  Requisition r;
  knobs_[1]->request(r);
  Coord xnat = r.requirement(Dimension_X).natural();
  Coord ynat = r.requirement(Dimension_Y).natural();
  //
  Coord cx;
  Coord cy;
  if (dim_ == Dimension_X) {
    cy = allocation_.top() - ynat; 
  } else {
    cx = allocation_.right() - xnat;
  }
  for (i = 1; i < count(); i++) {
    if (tile_->flexibleBorder(i)) {
      Coord d;
      knobs_[i]->knobOn();
      d = tile_->getBorderOffset(i);
      if (dim_ == Dimension_X) {
        cx = allocation_.left() + d;
      } else {
        cy = allocation_.top() - d; 
      }
      knobAllocs_[i]->allot_x(Allotment(cx, xnat, .5));
      knobAllocs_[i]->allot_y(Allotment(cy, ynat, .5));
      Extension e;
      knobs_[i]->allocate(canvas_, *knobAllocs_[i], e);
      extension_.extend(e);
    } else {
      knobs_[i]->knobOff();
    }
  }
}

void ruiBox::allocateRidge() {
  ridge_ = newRidge();
  //
  Requisition r;
  ridge_->request(r);
  Coord xnat = r.requirement(Dimension_X).natural();
  Coord ynat = r.requirement(Dimension_Y).natural();
  //
  if (dim_ == Dimension_X) {
    ridgeAlloc_.allot_x(Allotment(xnat, xnat, .5));
    ridgeAlloc_.allot_y(Allotment(allocation_.bottom(), 
      allocation_.top() - allocation_.bottom(), 0)); 
  } else {
    ridgeAlloc_.allot_y(Allotment(ynat, ynat, .5));
    ridgeAlloc_.allot_x(Allotment(allocation_.left(), 
      allocation_.right() - allocation_.left(), 0)); 
  }
  Extension e;
  ridge_->allocate(canvas_,ridgeAlloc_,e);
  extension_.extend(e);
}

void ruiBox::allocate(Canvas* canvas, const Allocation& allocation, 
    Extension& extension) {
  canvas_ = canvas;
  allocation_ = allocation;
  extension_ = extension;
  Box::allocate(canvas,allocation,extension); 
  allocateKnobs();
  allocateRidge();
}

void ruiBox::draw(Canvas* canvas, const Allocation& allocation) const {
  Box::draw(canvas, allocation);
  drawKnobs(canvas, allocation);
  drawRidge();
}

void ruiBox::pick(Canvas* canvas, const Allocation& allocation, 
    int depth, Hit& hit) {
  if (!pickKnobs(canvas, allocation, depth, hit)) {
    Box::pick(canvas, allocation, depth, hit);
  }
}
 
void ruiBox::knobsOn() {
  knobsOn_ = true;
}

void ruiBox::knobsOff() {
  knobsOn_ = false;
}

boolean ruiBox::isKnobsOn() {
  return knobsOn_;
}

int ruiBox::indexOfKnob (ruiKnob* knob) {
  for (int i = 0; i < knobs_.size(); i++) {
    if (knob == knobs_[i]) return i;
  }
  return 0;
}

void ruiBox::pressKnob(ruiKnob* knob, Event&) {
  ridgeOn_ = true;
  knobIndex_ = indexOfKnob(knob);
  int i = knobIndex_;
  knobComponentSize_ = tile_->getComponentSize(knobIndex_);
  Coord gb, ga, sb, sa;
  tile_->getSlacks(i,gb,ga,sb,sa);
  if (dim_ == Dimension_X) {
    knobPressOriginX_ = knobAllocs_[i]->x_allotment().origin();
    //knobPressX_ = event.pointer_x();
    knobPressX_ = knobPressOriginX_; 
    ridgeX_ = knobPressOriginX_;
    Coord kx = (knobAllocs_[i]->left() + knobAllocs_[i]->right())/2;
    knobXmin_ = max(kx-sb, kx-ga);
    knobXmax_ = min(kx+gb, kx+sa);
  } else {
    rAssert(dim_ = Dimension_Y);
    knobPressOriginY_ = knobAllocs_[i]->y_allotment().origin();
    //knobPressY_ = event.pointer_y();
    knobPressY_ = knobPressOriginY_; 
    ridgeY_ = knobPressOriginY_;
    Coord ky = (knobAllocs_[i]->bottom() + knobAllocs_[i]->top())/2;
    knobYmin_ = max(ky-gb, ky-sa);
    knobYmax_ = min(ky+sb, ky+ga);
  }
}

void ruiBox::dragKnob(ruiKnob* knob, Event& event) {
  dragKnobXY(knob, event.pointer_x(), event.pointer_y());
}

void ruiBox::changeKnobAlloc(int i, Coord d) {
  if (dim_ == Dimension_X) {
    knobAllocs_[i]->x_allotment().origin(knobPressOriginX_ + d);
  } else {
    knobAllocs_[i]->y_allotment().origin(knobPressOriginY_ + d);
  }
}

void ruiBox::changeRidgeAllocation() {
  if (dim_ == Dimension_X) {
    ridgeAlloc_.x_allotment().origin(ridgeX_);
  } else {
    ridgeAlloc_.y_allotment().origin(ridgeY_);
  }
}

void ruiBox::damageRidge() {
  canvas_->damage(ridgeAlloc_.left(), ridgeAlloc_.bottom(), 
    ridgeAlloc_.right(), ridgeAlloc_.top());
}

void ruiBox::damageKnob(int i) {
  canvas_->damage(knobAllocs_[i]->left(), knobAllocs_[i]->bottom(),
    knobAllocs_[i]->right(), knobAllocs_[i]->top());
}

void ruiBox::dragKnobXY(ruiKnob*, Coord x, Coord y) {
  int i = knobIndex_; 
  rAssert(i);
  rAssert(knobs_[i]->isOn());
  Coord dsize;
  Coord d;

  damageRidge();
  damageKnob(i);
 
  if (dim_ == Dimension_X) {
    Coord nx = min(knobXmax_, max(x, knobXmin_));
    d = nx - knobPressX_;
    ridgeX_ = nx;
    dsize = -d;  
  } else {
    Coord ny = min(knobYmax_, max(y, knobYmin_));
    d = ny - knobPressY_;
    ridgeY_ = ny;
    dsize = d;
  } 
  tile_->setComponentSize(i, knobComponentSize_ + dsize);
  changeKnobAlloc(i, d);
  changeRidgeAllocation();
 
  damageRidge();
  damageKnob(i);

}
 
void ruiBox::commitKnob(ruiKnob*, Event&) {
  change(0);
  ridgeOn_ = false;
  allocate(canvas_, allocation_, extension_);
  canvas_->damage(allocation_.left(), allocation_.bottom(), 
    allocation_.right(), allocation_.top());
}

void ruiBox::abortKnob(ruiKnob* knob, Event& event) {
  dragKnobXY(knob, knobPressX_, knobPressY_);
  commitKnob(knob, event);
}

boolean ruiBox::pickKnobs(
    Canvas* c, const Allocation&,
    int depth, Hit& h) {
  Coord x = h.left();
  Coord y = h.bottom();
  int count = knobs_.size(); 
  for (int i= 0; i< count; ++i) {
    if (!knobs_[i]->isOn()) continue;
    Allocation& a = *knobAllocs_[i];
      if (
        x >= a.left() && x < a.right()
        && y >= a.bottom() && y < a.top()
      ) { 
        h.begin(depth, this, i);
        knobs_[i]->pick(c, a, depth + 1, h);
        h.end();
        return true;
      }   
    }
  return false;
}

void ruiBox::drawKnobs(
  Canvas* canvas, const Allocation&) const {
  if (!knobsOn_) return;
  for (int i = 1; i < count(); i++) {
    if (knobs_.get(i)->isOn()) {
      knobs_.get(i)->draw(canvas, *knobAllocs_.get(i));
    }
  } 
}

void ruiBox::drawRidge() const {
  if (!ridgeOn_) return;
  ridge_->draw(canvas_,ridgeAlloc_);
}

////////////////////////////////////////////////////

void ruiBox::append(Glyph* g) {
  Box::append(g);
  tile_->insertLast();
}

void ruiBox::prepend(Glyph* g) {
  Box::prepend(g);
  tile_->insert(0);
}

void ruiBox::insert(GlyphIndex i, Glyph* g) {
  Box::insert(i,g);
  tile_->insert((int)i);
}

void ruiBox::remove(GlyphIndex i) {
  Box::remove(i);
  tile_->remove((int)i);
}

void ruiBox::replace(GlyphIndex i, Glyph* g) {
  Box::replace(i,g);
  tile_->change((int)i);
}

void ruiBox::change(GlyphIndex i) {
  Box::change(i);
}
