/*
 * Copyright (c) 1991 Redwood Design Automation
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the name of
 * Redwood Design Automation may not be used in any advertising or publicity
 * relating to the software without the specific, prior written permission of
 * Redwood Design Automation.
 *
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
 *
 * IN NO EVENT SHALL REDWOOD DESIGN AUTOMATION BE LIABLE FOR ANY SPECIAL,
 * INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND, OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER OR NOT
 * ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF LIABILITY,
 * ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
 * SOFTWARE.
 */

#include <string.h>

#include <strstream.h>

#include <InterViews/event.h>
#include <InterViews/glyph.h>
#include <InterViews/handler.h>
#include <InterViews/listener.h>
#include <InterViews/session.h>
#include <InterViews/style.h>
#include <InterViews/window.h>

#include <IV-look/kit.h>

#include <IV-X11/xevent.h>
#include <IV-X11/xwindow.h>

#include <X11/Xatom.h>

#include <ruiDrag.h>

////////////////////
// Drag Utilities //
////////////////////

static void queryPointer(Window* window, int& x, int& y) {
  WindowRep* rep = window->rep();

  XWindow root, child;
  int wx, wy;
  unsigned int keys;
  XQueryPointer(rep->dpy(), rep->xwindow_, &root, &child, &x, &y, &wx, &wy,
    &keys);
}

static XWindow translateCoordinates(XDisplay* display, int& x, int& y) {
  XWindow root = DefaultRootWindow(display);
  XWindow dest = None;
  XWindow prev = root;
  XWindow child;
  int nx, ny;
  XTranslateCoordinates(display, root, root, x, y, &nx, &ny, &child);
  while (child != None) {
    dest = child;
    XTranslateCoordinates(display, prev, dest, x, y, &nx, &ny, &child);
    x = nx;
    y = ny;
    prev = dest;
  }
  return dest;
}

//////////
// Drag //
//////////

unsigned ruiDrag::uid_ = 0;
const char* ruiDrag::dragName_ = "XV_DO_DRAG_LOAD"; // works with OpenWindows
ManagedWindow* ruiDrag::window_ = 0;

void ruiDrag::sendClientMessage(Window* window, char* value, int x, int y) {
  WindowRep* rep = window->rep();

  XWindow destination = translateCoordinates(rep->dpy(), x, y);
  if (! destination) {
    return;
  }

  char* propertyName;
  ostrstream name;
  name << "ruiDrag" << ruiDrag::uid_++;
  propertyName = name.str();

  Atom property = XInternAtom(rep->dpy(), propertyName, False);
  XChangeProperty(rep->dpy(), rep->xwindow_, property,
    XA_STRING, 8, PropModePrepend, (unsigned char *) value, strlen(value));

  XEvent xevent;
  xevent.xclient.type = ClientMessage;
  xevent.xclient.window = destination;
  xevent.xclient.message_type = XInternAtom(rep->dpy(), ruiDrag::dragName_,
    False);
  xevent.xclient.format = 32;
  xevent.xclient.data.l[1] = x;
  xevent.xclient.data.l[2] = y;
  xevent.xclient.data.l[3] = rep->xwindow_;
  xevent.xclient.data.l[4] = property;
  XSendEvent(rep->dpy(), destination, False, NoEventMask, &xevent);
}

void ruiDrag::dragAndDrop(Window* window, Glyph* glyph, char* value) {
  Resource::ref(glyph);

  PopupWindow dragWindow(glyph);
  dragWindow.map();
  WindowRep* rep = dragWindow.rep();

  int x, y;
  queryPointer(window, x, y);
  XMoveWindow(rep->dpy(), rep->xwindow_, x - dragWindow.pwidth() / 2,
    y - dragWindow.pheight() / 2);

  Event event;
  event.display(dragWindow.display());
  XGrabPointer(rep->dpy(), rep->xwindow_, True, (unsigned int) ButtonMotionMask,
    GrabModeAsync, GrabModeAsync, None, None, CurrentTime);

  do {
    event.read();
    if (event.type() == Event::motion) {
      x = event.rep()->xevent_.xmotion.x_root - dragWindow.pwidth() / 2;
      y = event.rep()->xevent_.xmotion.y_root - dragWindow.pheight() / 2;
      XMoveWindow(rep->dpy(), rep->xwindow_, x, y);
    }
  } while (event.type() != Event::up);

  XUngrabPointer(rep->dpy(), CurrentTime);
  dragWindow.unmap();

  sendClientMessage(window, value, x, y);

  Resource::unref(glyph);
}

void ruiDrag::dragAndDrop(Glyph* glyph, char* value) {
  if (! window_) {
    Kit* kit = Kit::instance();
    Style* style = Session::instance()->style();
    window_ = new ApplicationWindow(kit->inset_frame(
      kit->label("drag and drop properties", style), style));
    // !!! need a way to bind without mapping the window.
    window_->iconic(true);
    window_->map();
  }
  dragAndDrop(window_, glyph, value);
}

boolean ruiDrag::value(const Event& event, char*& value) {
  XEvent& xevent = event.rep()->xevent_;

  XDisplay* display = xevent.xclient.display;
  XWindow xwindow = xevent.xclient.data.l[3];
  Atom property = xevent.xclient.data.l[4];

  Atom actual_type;
  int actual_format;
  unsigned long nitems;
  unsigned long bytes_after;
  char* buffer;
  if (XGetWindowProperty(display, xwindow, property, 0L, 256, True,
    AnyPropertyType, &actual_type, &actual_format, &nitems, &bytes_after,
    (unsigned char**) &buffer) != Success) {
    return 0;
  }

  value = new char[strlen(buffer) + 1];
  strcpy(value, buffer);
  XFree(buffer);
  return 1;
}

boolean ruiDrag::isDrag(const Event& event) {
  return isDrag(event.rep()->xevent_);
}

boolean ruiDrag::isDrag(const XEvent& xevent) {
  Atom drag = XInternAtom(xevent.xclient.display, dragName_, True);
  return drag && (xevent.xclient.message_type == drag);
}

void ruiDrag::locate(const XEvent& xevent, int& x, int& y) {
  x = (int) xevent.xclient.data.l[1];
  y = (int) xevent.xclient.data.l[2];
}
