/*
 * Copyright (c) 1991 Redwood Design Automation
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the name of
 * Redwood Design Automation may not be used in any advertising or publicity
 * relating to the software without the specific, prior written permission of
 * Redwood Design Automation.
 *
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
 *
 * IN NO EVENT SHALL REDWOOD DESIGN AUTOMATION BE LIABLE FOR ANY SPECIAL,
 * INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND, OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER OR NOT
 * ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF LIABILITY,
 * ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
 * SOFTWARE.
 */

#include <string.h>

#include <InterViews/action.h>
#include <InterViews/center.h>
#include <InterViews/fixedspan.h>
#include <InterViews/handler.h>
#include <InterViews/listener.h>
#include <InterViews/monoglyph.h>
#include <OS/string.h> // because style.h needs it but doesn't include it
#include <InterViews/style.h>

#include <OS/directory.h>

#include <IV-look/kit.h>
#include <IV-look/menu.h>

#include <ruiBox.h>
#include <ruiDrag.h>
#include <ruiFileBrowse.h>
#include <ruiPanner.h>
#include <ruiScene.h>

///////////////////////
// File Browser Item //
///////////////////////

class ruiFileBrowseItem : public Listener, public Handler {
public:
  ruiFileBrowseItem(ruiFileBrowse* FileBrowse, const char* name, Kit* kit,
    Style* style);
  virtual ~ruiFileBrowseItem();

  void event(Event& event);

protected:
  Glyph* glyph();

  ruiFileBrowse* FileBrowse_;
  char* name_;
  Kit* kit_;
  Style* style_;
};

ruiFileBrowseItem::ruiFileBrowseItem(ruiFileBrowse* FileBrowse,
  const char* name, Kit* kit, Style* style) {
  target(this);
  button(true);

  FileBrowse_ = FileBrowse;
  name_ = new char[strlen(name) + 1];
  strcpy(name_, name);
  kit_ = kit;
  style_ = style;

  body(new Center(glyph(), Dimension_Y, 0));
}

ruiFileBrowseItem::~ruiFileBrowseItem() {
  delete name_;
}

Glyph* ruiFileBrowseItem::glyph() {
  return kit_->outset_frame(kit_->label(name_, style_), style_);
}

void ruiFileBrowseItem::event(Event& event) {
  if (event.type() != Event::down) {
    return;
  }

  if (event.pointer_button() == Event::left) {
    FileBrowse_->open(name_);
  } else {
    rutString path = FileBrowse_->name() & "/" & name_;
    char* pathName = path.newCString();
    ruiDrag::dragAndDrop(glyph(), pathName);
    delete pathName;
  }
}

////////////////////////
// File Browser Scene //
////////////////////////

class ruiFileBrowseScene : public ruiScene {
public:
  ruiFileBrowseScene();
  virtual ~ruiFileBrowseScene();

  void allocate(Canvas* canvas, const Allocation& allocation,
    Extension& extension);

protected:
  Coord xSpacing_;
  Coord ySpacing_;
};

ruiFileBrowseScene::ruiFileBrowseScene() {
  xSpacing_ = 2;
  ySpacing_ = 2;
}

ruiFileBrowseScene::~ruiFileBrowseScene() {
}

void ruiFileBrowseScene::allocate(Canvas* canvas, const Allocation& allocation,
  Extension& extension) {
  Coord x = 0;
  Coord y = 0;
  Coord xMax = 0;
  for (GlyphIndex index = 0; index < count(); ++index) {
    Glyph* glyph = component(index);
    Requisition requisition;
    glyph->request(requisition);
    Coord bottom = y - requisition.requirement(Dimension_Y).natural();
    if ((bottom < -allocation.allotment(Dimension_Y).span()) && (y != 0)) {
      x += xMax + xSpacing_;
      y = 0;
      xMax = 0;
    }
    y -= requisition.requirement(Dimension_Y).natural();
    move(index, x, y);
    xMax = Math::max(xMax, requisition.requirement(Dimension_X).natural());
    y -= ySpacing_;
  }
  ruiScene::allocate(canvas, allocation, extension);
}

//////////////////
// File Browser //
//////////////////

implement(ActionCallback, ruiFileBrowse);

ruiFileBrowse::ruiFileBrowse(Kit* kit, Style* style) {
  kit_ = kit;
  style_ = style;
  scene_ = new ruiFileBrowseScene();
  name_ = ".";
  allocated_ = false;

  reset();

  Menu* menuBar = kit->menubar(style);
  Menu* pulldown;

  pulldown = kit->pulldown(style);
  pulldown->add_item(kit->simple_menu_item("New", style),
    new ActionCallback(ruiFileBrowse)(this, &ruiFileBrowse::reset));
  menuBar->add_item(kit->simple_menubar_item("File", style), pulldown);

  Glyph* scene = new ruiHorzBox(kit->inset_frame(scene_, style),
    kit->inset_frame(new FixedSpan(new ruiPanner(scene_, scene_, style),
    Dimension_X, 75), style));

  body(new ruiVertBox(menuBar, scene));
}

ruiFileBrowse::~ruiFileBrowse() {
}

void ruiFileBrowse::allocate(Canvas* canvas, const Allocation& allocation,
  Extension& extension) {
  MonoGlyph::allocate(canvas, allocation, extension);
  if (! allocated_) {
    scene_->scroll_to(Dimension_X, 0);
    scene_->scroll_to(Dimension_Y, scene_->bottom() - scene_->top());
    allocated_ = true;
  }
}

void ruiFileBrowse::reset() {
  while (scene_->count() > 0) {
    scene_->remove(0);
  }

  char* name = name_.newCString();
  DirectoryInfo* info = Directory::open(name);
  delete name;
  if (info) {
    Directory directory(info);
    for (int i = 0; i < directory.count(); ++i) {
      Glyph* glyph = new ruiFileBrowseItem(this, directory.name(i), kit_,
        style_);
      GlyphIndex index = scene_->count();
      scene_->append(glyph);
      scene_->show(index, true);
    }
  }

  scene_->reallocate();
  scene_->scroll_to(Dimension_X, 0);
  scene_->scroll_to(Dimension_Y, scene_->bottom() - scene_->top());
}

int ruiFileBrowse::open(const char* name) {
  rutString path = name_ & "/" & name;
  char* pathName = path.newCString();
  DirectoryInfo* info = Directory::open(pathName);
  delete pathName;
  if (info) {
    Directory directory(info); // need to delete this to delete info
    name_ = pathName;
    reset();
    return 1;
  }
  return 0;
}

rutString ruiFileBrowse::name() const {
  return name_;
}
