/*
 * Copyright (c) 1991 Redwood Design Automation
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the name of
 * Redwood Design Automation may not be used in any advertising or publicity
 * relating to the software without the specific, prior written permission of
 * Redwood Design Automation.
 *
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
 *
 * IN NO EVENT SHALL REDWOOD DESIGN AUTOMATION BE LIABLE FOR ANY SPECIAL,
 * INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND, OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER OR NOT
 * ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF LIABILITY,
 * ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
 * SOFTWARE.
 */

#include <string.h>

#include <InterViews/action.h>
#include <InterViews/event.h>
#include <InterViews/hit.h>

#include <rutArray.h>

#include <ruiKeyMap.h>

//////////////////
// Key Map Info //
//////////////////

class ruiKeyMapInfo {
public:
  ruiKeyMapInfo(Action* action, char* string);
  ruiKeyMapInfo(Action* action, char keycode, ruiKeyMapModifier modifiers);
  virtual ~ruiKeyMapInfo();

  virtual char* string();
  virtual char keycode();
  virtual ruiKeyMapModifier modifiers();
  virtual Action* action();

protected:
  char* string_;
  char keycode_;
  ruiKeyMapModifier modifiers_;
  Action* action_;
};

rutArrayDefine(ruiKeyMapInfoArray,ruiKeyMapInfo)

ruiKeyMapInfo::ruiKeyMapInfo(Action* action, char* string) {
  string_ = new char[strlen(string) + 1];
  strcpy(string_, string);
  keycode_ = 0;
  action_ = action;
}

ruiKeyMapInfo::ruiKeyMapInfo(Action* action, char keycode, ruiKeyMapModifier
  modifiers) {
  string_ = nil;
  keycode_ = keycode;
  modifiers_ = modifiers;
  action_ = action;
}

ruiKeyMapInfo::~ruiKeyMapInfo() {
  delete string_;
}

char* ruiKeyMapInfo::string() {
  return string_;
}

char ruiKeyMapInfo::keycode() {
  return keycode_;
}

ruiKeyMapModifier ruiKeyMapInfo::modifiers() {
  return modifiers_;
}

Action* ruiKeyMapInfo::action() {
  return action_;
}

/////////////
// Key Map //
/////////////

ruiKeyMap::ruiKeyMap(Glyph* glyph) : MonoGlyph(glyph) {
  info_ = new ruiKeyMapInfoArray;
}

ruiKeyMap::~ruiKeyMap() {
  for (int i = 0; i < info_->size(); ++i) {
    ruiKeyMapInfo* info = info_->get(i);
    delete info;
  }
  delete info_;
}

void ruiKeyMap::pick(Canvas* canvas, const Allocation& allocation, int depth,
  Hit& hit) {
  const Event* event = hit.event();
  if (event && caught(*event)) {
    hit.target(depth, this, 0, this);
  } else {
    MonoGlyph::pick(canvas, allocation, depth, hit);
  }
}

boolean ruiKeyMap::caught(const Event& event) const {
  return lookup(event) != 0;
}

void ruiKeyMap::event(Event& event) {
  ruiKeyMapInfo* info = lookup(event);
  if (info) {
    info->action()->execute();
  }
}

void ruiKeyMap::add(Action* action, char* string) {
  info_->insertLast(new ruiKeyMapInfo(action, string));
}

void ruiKeyMap::add(Action* action, char keycode, ruiKeyMapModifier modifiers) {
  info_->insertLast(new ruiKeyMapInfo(action, keycode, modifiers));
}

ruiKeyMapModifier ruiKeyMap::modifiers(const Event& event) const {
  ruiKeyMapModifier modifiers = 0;
  if (event.control_is_down()) {
    modifiers |= control;
  }
  if (event.meta_is_down()) {
    modifiers |= meta;
  }
  if (event.shift_is_down()) {
    modifiers |= shift;
  }
  if (event.capslock_is_down()) {
    modifiers |= capslock;
  }
  return modifiers;
}

ruiKeyMapInfo* ruiKeyMap::lookup(const Event& event) const {
  const unsigned int length = 8;
  char buffer[length];
  unsigned int count = event.mapkey(buffer, length);
  for (int i = 0; i < info_->size(); ++i) {
    ruiKeyMapInfo* info = info_->get(i);
    if (info->keycode()) {
      if ((info->keycode() == event.keycode()) && (info->modifiers() ==
        modifiers(event))) {
        return info;
      }
    } else {
      if (count && (strncmp(buffer, info->string(), count) == 0)) {
        return info;
      }
    }
  }
  return nil;
}
