/*
 * Copyright (c) 1991 Redwood Design Automation
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the name of
 * Redwood Design Automation may not be used in any advertising or publicity
 * relating to the software without the specific, prior written permission of
 * Redwood Design Automation.
 *
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
 *
 * IN NO EVENT SHALL REDWOOD DESIGN AUTOMATION BE LIABLE FOR ANY SPECIAL,
 * INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND, OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER OR NOT
 * ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF LIABILITY,
 * ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
 * SOFTWARE.
 */

#include <InterViews/box.h>
#include <InterViews/canvas.h>
#include <InterViews/character.h>
#include <InterViews/center.h>
#include <InterViews/color.h>
#include <InterViews/deck.h>
#include <InterViews/fixedspan.h>
#include <InterViews/font.h>
#include <InterViews/glue.h>
#include <InterViews/patch.h>
#include <InterViews/session.h>
#include <InterViews/shapeof.h>
#include <InterViews/style.h>

#include <IV-look/kit.h>

#include <ruiLegend.h>
#include <ruiText.h>

//////////////////
// Legend Check //
//////////////////

class ruiLegendCheck : public Glyph {
public:
  ruiLegendCheck();

  void request(Requisition& requisition) const;
  void draw(Canvas* canvas, const Allocation& allocation) const;

protected:
  Coord span_;
  Color* color_;
};

ruiLegendCheck::ruiLegendCheck() {
  Style* style = Session::instance()->style();
  const Font* font = style->font();
  span_ = font->ascent() + font->descent();
  if (style->value_is_on("monochrome")) {
    color_ = new Color(*style->foreground());
  } else {
    color_ = new Color(.8, 1, 0, 1);
  }
  Resource::ref(color_);
}

void ruiLegendCheck::request(Requisition& requisition) const {
  requisition.requirement(Dimension_X).natural(span_);
}

void ruiLegendCheck::draw(Canvas* canvas, const Allocation&
  allocation) const {
  Coord l = allocation.left();
  Coord b = allocation.bottom();
  Coord r = allocation.right();
  Coord t = allocation.top();
  Coord my = (b + t) / 2;
  Coord mx = (l + r) / 2;
  Coord lx = (l + mx) / 2;
  Coord hy = (my + t) / 2;
  Coord ly = (b + my) / 2;
  canvas->new_path();
  canvas->move_to(l, my);
  canvas->line_to(lx, ly);
  canvas->line_to(r, t);
  canvas->line_to(r, hy);
  canvas->line_to(lx, b);
  canvas->close_path();
  canvas->fill(color_);
}

////////////
// Legend //
////////////

ruiLegend::ruiLegend(Kit* kit, Style* style) {
  Glyph* check = new ruiLegendCheck;
  modified_ = new Deck;
  modified_->append(new ShapeOf(check));
  modified_->append(check);
  modified_->flip_to(0);
  patch_ = new Patch(modified_);
  name_ = new ruiTextLine(16);
  name_->readOnly(true);
  adjuster_ = new ruiTextLineAdjuster(name_);
  Requisition requisition;
  name_->request(requisition);
  Coord span = requisition.requirement(Dimension_X).natural();
  body(new LRBox(
    kit->outset_frame(new Center(kit->inset_frame(new FixedSpan(adjuster_,
    Dimension_X, span), style), Dimension_Y, 0), style),
    kit->outset_frame(new Center(kit->inset_frame(patch_, style),
    Dimension_Y, 0), style)));
}

void ruiLegend::name(const rutString& name) {
  name_->reset();
  name_->write(name);
  name_->scroll_to(Dimension_X, name_->upper(Dimension_X) -
    name_->cur_length(Dimension_X));
}

void ruiLegend::modified(boolean modified) {
  modified_->flip_to(modified ? 1 : 0);
  patch_->redraw();
  // !!! redraw doesn't always work! but damage_all does! why?
  if (patch_->canvas()) {
    patch_->canvas()->damage_all();
  }
}
