#include <InterViews/background.h>
#include <InterViews/canvas.h>
#include <InterViews/color.h>
#include <InterViews/handler.h>
#include <InterViews/style.h>

#include <IV-look/adjustable.h>
#include <IV-look/bevel.h>

#include <ruiPanner.h>

ruiPanner::ruiPanner(
    Adjustable* x, Adjustable* y, Style* s, Direction d
) : Adjuster(x) {
    target(this);
    y_adjust_ = y;
    if (s->value_is_on("monochrome")) {
      thumb_ = new Background(nil, s->foreground());
    } else {
      thumb_ = new Bevel(
	nil, &Bevel::rect, s->light(), s->flat(), s->dark(), 2.0
      );
    }
    thumb_->ref();
    canvas_ = nil;
    xscale_ = 1.0;
    yscale_ = 1.0;
    direction_ = d;
    button(true);
}

ruiPanner::~ruiPanner() {
    Resource::unref(thumb_);
}

void ruiPanner::request(Requisition& req) const {
    Requirement r(72.0, fil, 72.0, 0.0);
    req.require(Dimension_X, r);
    req.require(Dimension_Y, r);
}

void ruiPanner::allocate(Canvas* c, const Allocation& a, Extension& ext) {
    canvas_ = c;
    allocation_ = a;
    ext.xy_extents(a);
    update_all();
}

/*
 * We assume that we can draw the thumb without necessarily
 * having allocated it.  This works because thumb is a bevel
 * with nothing inside it.  If it is changed to something that
 * needs to be allocated, then it should be allocated in Scroller::update_all.
 */

void ruiPanner::draw(Canvas* c, const Allocation&) const {
    c->push_clipping();
    c->clip_rect(allocation_.left(), allocation_.bottom(), allocation_.right(),
      allocation_.top());
    thumb_->draw(c, thumb_allocation_);
    c->pop_clipping();
}

void ruiPanner::update(DimensionName) {
    update_all();
}

void ruiPanner::damage_thumb() {
    if (canvas_ != nil) {
        Coord l = Math::max(allocation_.left(), thumb_allocation_.left());
        Coord b = Math::max(allocation_.bottom(), thumb_allocation_.bottom());
        Coord r = Math::min(allocation_.right(), thumb_allocation_.right());
        Coord t = Math::min(allocation_.top(), thumb_allocation_.top());
	canvas_->damage(l, b, r, t);
    }
}

void ruiPanner::update_all() {
    if (canvas_ != nil) {
	/* first damage the old thumb area */
        damage_thumb();

	/* now compute the new thumb location */
	Adjustable* x = subject();
	Adjustable* y = y_adjust_;
        boolean xzero;
        boolean yzero;
	const DimensionName dx = Dimension_X, dy = Dimension_Y;
	Allocation& a = allocation_;
	Coord a_xlen = x->length(dx);
	if (Math::equal(a_xlen, float(0.0), float(1e-6))) {
	    xscale_ = 1.0;
            xzero = true;
	} else {
	    xscale_ = a.allotment(dx).span() / x->length(dx);
            xzero = false;
	}
	Coord a_ylen = y->length(dy);
	if (Math::equal(a_ylen, float(0.0), float(1e-6))) {
	    yscale_ = 1.0;
            yzero = true;
	} else {
	    yscale_ = a.allotment(dy).span() / y->length(dy);
            yzero = false;
	}
        Coord xoff;
        Coord xlen;
        if (xzero) {
          xoff = 0;
          xlen = a.allotment(dx).span();
        } else {
	  xoff = xscale_ * (x->cur_lower(dx) - x->lower(dx));
	  xlen = xscale_ * x->cur_length(dx);
        }
        Coord yoff;
        Coord ylen;
        if (yzero) {
          yoff = 0;
          ylen = a.allotment(dy).span();
        } else {
	  yoff = yscale_ * (y->cur_lower(dy) - y->lower(dy));
	  ylen = yscale_ * y->cur_length(dy);
        } 
	Allotment ax((direction_ & leftToRight ? a.left() + xoff :
            a.right() - xoff - xlen), xlen, 0.0);
	Allotment ay((direction_ & bottomToTop ? a.bottom() + yoff :
            a.top() - yoff - ylen), ylen, 0.0);
	thumb_allocation_.allot(Dimension_X, ax);
	thumb_allocation_.allot(Dimension_Y, ay);

	/* now damage the new thumb area */
        damage_thumb();
    }
}

void ruiPanner::event(Event& e) {
    if (e.type() != Event::down) {
	return;
    }
    e.grab(this);
    motion(true);
    boolean done = false;
    Adjustable* x = subject();
    Adjustable* y = y_adjust_;
    Coord xbase = x->lower(Dimension_X);
    Coord ybase = y->lower(Dimension_Y);
    Coord px = e.pointer_x();
    Coord py = e.pointer_y();
    Allocation& t = thumb_allocation_;
    Coord dx;
    Coord dy;
    if ((t.left() <= px) && (px <= t.right()) && (t.bottom() <= py) &&
        (py <= t.top())) {
        dx = (direction_ & leftToRight ? px - t.left() : t.right() - px);
        dy = (direction_ & bottomToTop ? py - t.bottom() : t.top() - py);
    } else {
        dx = (t.right() - t.left()) / 2;
        dy = (t.top() - t.bottom()) / 2;
    }
    Allocation& a = allocation_;
    for (;;) {
	switch (e.type()) {
	case Event::up:
	    done = true;
	    /* fall through */
	case Event::motion:
	case Event::down:
	    x->scroll_to(
                Dimension_X, xbase + ((direction_ & leftToRight ? e.pointer_x() -
                a.left() : a.right() - e.pointer_x()) - dx) / xscale_
            );
	    y->scroll_to(
                Dimension_Y, ybase + ((direction_ & bottomToTop ? e.pointer_y() -
                a.bottom() : a.top() - e.pointer_y()) - dy) / yscale_
            );
	    break;
	default:
	    break;
	}
	if (done) {
	    break;
	}
	e.read();
    }
    motion(false);
    e.ungrab(this);
}
