#include <InterViews/canvas.h>
#include <InterViews/hit.h>
#include <InterViews/printer.h>

#include <OS/list.h>
#include <OS/math.h>

#include <ruiScene.h>

//////////////////
// Scene Canvas //
//////////////////

class ruiSceneCanvas : public Canvas {
public:
    ruiSceneCanvas(ruiScene* scene);
    virtual ~ruiSceneCanvas();

    virtual Window* window() const;
 
    virtual void size(Coord width, Coord height);
    virtual void psize(unsigned int width, unsigned int height);

    virtual Coord width() const;
    virtual Coord height() const;
    unsigned int pwidth() const;
    unsigned int pheight() const;
 
    virtual void new_path();
    virtual void move_to(Coord x, Coord y);
    virtual void line_to(Coord x, Coord y);
    virtual void curve_to(
        Coord x, Coord y, Coord x1, Coord y1, Coord x2, Coord y2
    );
    virtual void close_path();
    virtual void stroke(const Color*, const Brush*);
    virtual void line(
        Coord x1, Coord y1, Coord x2, Coord y2, const Color*, const Brush*
    );
    virtual void rect(
        Coord l, Coord b, Coord r, Coord t, const Color*, const Brush*
    );
    virtual void fill(const Color*);
    virtual void fill_rect(Coord l, Coord b, Coord r, Coord t, const Color*);
 
    virtual void character(
        const Font*, long ch, Coord width, const Color*, Coord x, Coord y
    );
    virtual void stencil(const Bitmap*, const Color*, Coord x, Coord y);
    virtual void image(const Raster*, Coord x, Coord y);
 
    virtual void push_transform();
    virtual void transform(const Transformer&);
    virtual void pop_transform();
 
    virtual void push_clipping();
    virtual void clip();
    virtual void clip_rect(Coord l, Coord b, Coord r, Coord t);
    virtual void pop_clipping();
 
    virtual void damage(Coord left, Coord bottom, Coord right, Coord top);
    virtual boolean damaged(
        Coord left, Coord bottom, Coord right, Coord top
    ) const;
    virtual void damage_all();
    virtual boolean any_damage() const;
    virtual void redraw(Coord left, Coord bottom, Coord right, Coord top);
    virtual void repair();
 
    class CanvasRep* rep() const;

    /* anachronisms */
    virtual CanvasLocation status() const;
    unsigned int Width() const;
    unsigned int Height() const;
    virtual void SetBackground(const Color*);

protected:
    friend ruiScene;

    void bind(Canvas* canvas);

    ruiScene* scene_;
    Canvas* canvas_;
};

ruiSceneCanvas::ruiSceneCanvas(ruiScene* scene) {
  scene_ = scene;
  canvas_ = nil;
}

ruiSceneCanvas::~ruiSceneCanvas() {
}
 
void ruiSceneCanvas::bind(Canvas* canvas) {
  canvas_ = canvas;
}

Window* ruiSceneCanvas::window() const {
  return canvas_->window();
}
 
void ruiSceneCanvas::size(Coord width, Coord height) {
  canvas_->size(width, height);
}

void ruiSceneCanvas::psize(unsigned int width, unsigned int height) {
  canvas_->psize(width, height);
}
 
Coord ruiSceneCanvas::width() const {
  return canvas_->width();
}

Coord ruiSceneCanvas::height() const {
  return canvas_->height();
}

unsigned int ruiSceneCanvas::pwidth() const {
  return canvas_->pwidth();
}

unsigned int ruiSceneCanvas::pheight() const {
  return canvas_->pheight();
}
 
void ruiSceneCanvas::new_path() {
  canvas_->new_path();
}

void ruiSceneCanvas::move_to(Coord x, Coord y) {
  float tx, ty;
  scene_->matrix_.transform(x, y, tx, ty);
  canvas_->move_to(tx, ty);
}

void ruiSceneCanvas::line_to(Coord x, Coord y) {
  float tx, ty;
  scene_->matrix_.transform(x, y, tx, ty);
  canvas_->line_to(tx, ty);
}

void ruiSceneCanvas::curve_to(
    Coord x, Coord y, Coord x1, Coord y1, Coord x2, Coord y2
) {
  float tx, ty, tx1, ty1, tx2, ty2;
  scene_->matrix_.transform(x, y, tx, ty);
  scene_->matrix_.transform(x1, y1, tx1, ty1);
  scene_->matrix_.transform(x2, y2, tx2, ty2);
  canvas_->curve_to(tx, ty, tx1, ty1, tx2, ty2);
}

void ruiSceneCanvas::close_path() {
  canvas_->close_path();
}

void ruiSceneCanvas::stroke(const Color* color, const Brush* brush) {
  canvas_->stroke(color, brush);
}

void ruiSceneCanvas::line(Coord x1, Coord y1, Coord x2, Coord y2,
  const Color* color, const Brush* brush) {
  float tx1, ty1, tx2, ty2;
  scene_->matrix_.transform(x1, y1, tx1, ty1);
  scene_->matrix_.transform(x2, y2, tx2, ty2);
  canvas_->line(tx1, ty1, tx2, ty2, color, brush);
}

void ruiSceneCanvas::rect(
    Coord left, Coord bottom, Coord right, Coord top, const Color* color,
    const Brush* brush
) {
  float l, b, r, t;
  scene_->matrix_.transform(left, bottom, l, b);
  scene_->matrix_.transform(right, top, r, t);
  canvas_->rect(l, b, r, t, color, brush);
}

void ruiSceneCanvas::fill(const Color* color) {
  canvas_->fill(color);
}

void ruiSceneCanvas::fill_rect(Coord left, Coord bottom, Coord right, Coord top,
  const Color* color) {
  float l, b, r, t;
  scene_->matrix_.transform(left, bottom, l, b);
  scene_->matrix_.transform(right, top, r, t);
  canvas_->fill_rect(l, b, r, t, color);
}

void ruiSceneCanvas::character(
    const Font* font, long ch, Coord width, const Color* color, Coord x, Coord y
) {
  float tx, ty;
  scene_->matrix_.transform(x, y, tx, ty);
  canvas_->character(font, ch, width, color, tx, ty);
}

void ruiSceneCanvas::stencil(const Bitmap* bitmap, const Color* color,
  Coord x, Coord y) {
  float tx, ty;
  scene_->matrix_.transform(x, y, tx, ty);
  canvas_->stencil(bitmap, color, tx, ty);
}

void ruiSceneCanvas::image(const Raster* raster, Coord x, Coord y) {
  float tx, ty;
  scene_->matrix_.transform(x, y, tx, ty);
  canvas_->image(raster, tx, ty);
}

void ruiSceneCanvas::push_transform() {
  canvas_->push_transform();
}

void ruiSceneCanvas::transform(const Transformer& transformer) {
  canvas_->transform(transformer);
}

void ruiSceneCanvas::pop_transform() {
  canvas_->pop_transform();
}

void ruiSceneCanvas::push_clipping() {
  canvas_->push_transform();
}

void ruiSceneCanvas::clip() {
  canvas_->clip();
}

void ruiSceneCanvas::clip_rect(Coord left, Coord bottom, Coord right,
  Coord top) {
  float l, b, r, t;
  scene_->matrix_.transform(left, bottom, l, b);
  scene_->matrix_.transform(right, top, r, t);
  canvas_->clip_rect(l, b, r, t);
}

void ruiSceneCanvas::pop_clipping() {
  canvas_->pop_clipping();
}

void ruiSceneCanvas::damage(Coord left, Coord bottom, Coord right, Coord top) {
  float l, b, r, t;
  scene_->matrix_.transform(left, bottom, l, b);
  scene_->matrix_.transform(right, top, r, t);
  canvas_->damage(l, b, r, t);
}

boolean ruiSceneCanvas::damaged(
    Coord left, Coord bottom, Coord right, Coord top
) const {
  float l, b, r, t;
  scene_->matrix_.transform(left, bottom, l, b);
  scene_->matrix_.transform(right, top, r, t);
  return canvas_->damaged(l, b, r, t);
}

void ruiSceneCanvas::damage_all() {
  canvas_->damage_all();
}

boolean ruiSceneCanvas::any_damage() const {
  return canvas_->any_damage();
}

void ruiSceneCanvas::redraw(Coord left, Coord bottom, Coord right, Coord top) {
  float l, b, r, t;
  scene_->matrix_.transform(left, bottom, l, b);
  scene_->matrix_.transform(right, top, r, t);
  canvas_->redraw(l, b, r, t);
}

void ruiSceneCanvas::repair() {
  canvas_->repair();
}

class CanvasRep* ruiSceneCanvas::rep() const {
  return canvas_->rep();
}
 
/* anachronisms */
CanvasLocation ruiSceneCanvas::status() const {
  return canvas_->status();
}

unsigned int ruiSceneCanvas::Width() const {
  return canvas_->Width();
}

unsigned int ruiSceneCanvas::Height() const {
  return canvas_->Height();
}

void ruiSceneCanvas::SetBackground(const Color* color) {
  canvas_->SetBackground(color);
}

///////////
// Scene //
///////////

static const int ruiSceneInfoAllocated = 0x01;
static const int ruiSceneInfoExtended = 0x02;
static const int ruiSceneInfoPositioned = 0x04;
static const int ruiSceneInfoHidden = 0x08;

class ruiSceneInfo {
public:
    ruiSceneInfo();

    Glyph* glyph_;
    Allocation newAllocation_;
    Allocation allocation_;
    Extension extension_;
    Coord x_;
    Coord y_;
    int status_;
};

ruiSceneInfo::ruiSceneInfo() {
    glyph_ = nil;
    x_ = 0;
    y_ = 0;
    status_ = 0;
}

declareList(ruiSceneInfo_List,ruiSceneInfo);
implementList(ruiSceneInfo_List,ruiSceneInfo);

static const float epsilon = 0.1;

ruiScene::ruiScene() : Listener(nil, nil) {
    canvas_ = nil;
    sceneCanvas_ = new ruiSceneCanvas(this);
    info_ = new ruiSceneInfo_List();
    defined_ = false;
    curLowerX_ = 0;
    curLowerY_ = 0;
}

ruiScene::~ruiScene() {
    GlyphIndex count = info_->count();
    for (GlyphIndex i = 0; i < count; ++i) {
        ruiSceneInfo& info = info_->item(i);
        Resource::unref(info.glyph_);
    }
    delete info_;
    info_ = nil;
    canvas_ = nil;
    delete sceneCanvas_;
    sceneCanvas_ = nil;
}

void ruiScene::damage(Extension& extension) {
    sceneCanvas_->damage(extension.left(), extension.bottom(),
        extension.right(), extension.top());
}

void ruiScene::reallocate() {
  Extension extension;
  allocate(canvas_, allocation_, extension);
}

Coord ruiScene::eventX(Event& event) {
  Coord x;
  Coord y;
  matrix_.inverse_transform(event.pointer_x(), event.pointer_y(), x, y);
  return x;
}

Coord ruiScene::eventY(Event& event) {
  Coord x;
  Coord y;
  matrix_.inverse_transform(event.pointer_x(), event.pointer_y(), x, y);
  return y;
}

Coord ruiScene::left() {
  return allocation_.left();
}

Coord ruiScene::bottom() {
  return allocation_.bottom();
}

Coord ruiScene::right() {
  return allocation_.right();
}

Coord ruiScene::top() {
  return allocation_.top();
}

void ruiScene::show(GlyphIndex index, boolean showing) {
    ruiSceneInfo& info = info_->item(index);
    if (((info.status_ & ruiSceneInfoHidden) == 0) != showing) {
        if (canvas_ != nil) {
            Extension& b = info.extension_;
            damage(b);
        }
        if (showing) {
            info.status_ &= ~ruiSceneInfoHidden;
        } else {
            info.status_ |= ruiSceneInfoHidden;
        }
    }
}

boolean ruiScene::showing(GlyphIndex index) const {
    return (info_->item(index).status_ & ruiSceneInfoHidden) == 0;
}

void ruiScene::move(GlyphIndex index, Coord x, Coord y) {
    ruiSceneInfo& info = info_->item(index);
    if (
        (info.status_ & ruiSceneInfoAllocated)
        && (info.x_ != x || info.y_ != y)
    ) {
        Extension& b = info.extension_;
        Allocation& na = info.newAllocation_;
        Allotment& ax = na.allotment(Dimension_X);
        Allotment& ay = na.allotment(Dimension_Y);
        Allotment n_ax(x, ax.span(), ax.alignment());
        Allotment n_ay(y, ay.span(), ay.alignment());
        na.allot(Dimension_X, n_ax);
        na.allot(Dimension_Y, n_ay);
        if (canvas_ != nil) {
            damage(b);
        }
        b.xy_extents(fil, -fil, fil, -fil);
        info.glyph_->allocate(sceneCanvas_, na, b);
        info.allocation_ = na;
        if (canvas_ != nil) {
            damage(b);
        }
        boundingBox();
        notify();
    }
    info.x_ = x;
    info.y_ = y;
}

void ruiScene::location(GlyphIndex index, Coord& x, Coord& y) {
    ruiSceneInfo& info = info_->item(index);
    x = info.x_;
    y = info.y_;
}

GlyphIndex ruiScene::count() const {
    return info_->count();
}

Glyph* ruiScene::component(GlyphIndex index) const {
    return info_->item(index).glyph_;
}

void ruiScene::allotment(GlyphIndex index, DimensionName res, Allotment& a) const {
    a = info_->item(index).allocation_.allotment(res);
}

void ruiScene::change(GlyphIndex index) {
    ruiSceneInfo& info = info_->item(index);
    info.status_ &= ~(ruiSceneInfoAllocated);
}

void ruiScene::append(Glyph* glyph) {
    ruiSceneInfo info;
    info.glyph_ = glyph;
    info.status_ = 0;
    info.x_ = 0;
    info.y_ = 0;
    info_->append(info);
    Resource::ref(glyph);
}

void ruiScene::prepend(Glyph* glyph) {
    ruiSceneInfo info;
    info.glyph_ = glyph;
    info.status_ = 0;
    info.x_ = 0;
    info.y_ = 0;
    info_->prepend(info);
    Resource::ref(glyph);
}

void ruiScene::insert(GlyphIndex index, Glyph* glyph) {
    ruiSceneInfo info;
    info.glyph_ = glyph;
    info.status_ = 0;
    info.x_ = 0;
    info.y_ = 0;
    info_->insert(index, info);
    Resource::ref(glyph);
}

void ruiScene::remove(GlyphIndex index) {
    ruiSceneInfo& info = info_->item(index);
    if (canvas_ != nil && (info.status_ & ruiSceneInfoAllocated)) {
        Extension b = info.extension_;
        damage(b);
    }
    Resource::unref(info.glyph_);
    info_->remove(index);
}

void ruiScene::replace(GlyphIndex index, Glyph* glyph) {
    ruiSceneInfo& info = info_->item(index);
    if (canvas_ != nil && (info.status_ & ruiSceneInfoAllocated)) {
        Extension& b = info.extension_;
        damage(b);
    }
    Resource::ref(glyph);
    Resource::unref(info.glyph_);
    info.glyph_ = glyph;
    info.status_ &= ~(ruiSceneInfoAllocated);
}

void ruiScene::boundingBox() {
    defined_ = false;
    left_   = 0;
    bottom_ = 0;
    right_  = 0;
    top_    = 0;
    GlyphIndex count = info_->count();
    for (GlyphIndex index = 0; index < count; ++index) {
        ruiSceneInfo& info = info_->item(index);
        if (info.glyph_ != nil) {
            Allocation& a = info.allocation_;
            Allotment ax = a.allotment(Dimension_X);
            Allotment ay = a.allotment(Dimension_Y);
            if (! defined_) {
                left_   = ax.begin();
                bottom_ = ay.begin();
                right_  = ax.end();
                top_    = ay.end();
                defined_ = true;
            } else {
                left_   = Math::min(left_, ax.begin());
                bottom_ = Math::min(bottom_, ay.begin());
                right_  = Math::max(right_, ax.end());
                top_    = Math::max(top_, ay.end());
            }
        }
    }
}

void ruiScene::place() {
    GlyphIndex count = info_->count();
    for (GlyphIndex index = 0; index < count; ++index) {
        ruiSceneInfo& info = info_->item(index);
        if (info.glyph_ != nil) {
            Allocation& a = info.newAllocation_;
            Requisition s;
            info.glyph_->request(s);
            Allotment ax = Allotment(
                info.x_,
                s.requirement(Dimension_X).natural(),
                s.requirement(Dimension_X).alignment()
            );
            Allotment ay = Allotment(
                info.y_,
                s.requirement(Dimension_Y).natural(),
                s.requirement(Dimension_Y).alignment()
            );
            a.allot(Dimension_X, ax);
            a.allot(Dimension_Y, ay);
        }
    }
}

void ruiScene::request(Requisition& requisition) const {
    Requirement rx(100, fil, 100, 0);
    Requirement ry(100, fil, 100, 0);
    requisition.require(Dimension_X, rx);
    requisition.require(Dimension_Y, ry);
}

void ruiScene::allocate(Canvas* c, const Allocation& allocation, Extension& ext) {
    canvas_ = c;
    allocation_ = allocation;
    sceneCanvas_->bind(c);
    matrix_ = Transformer();
    matrix_.translate(allocation_.left() - curLowerX_, allocation_.bottom() -
      curLowerY_);
    place();
    GlyphIndex count = info_->count();
    for (GlyphIndex index = 0; index < count; ++index) {
        ruiSceneInfo& info = info_->item(index);
        if (info.glyph_ != nil) {
            Allocation& na = info.newAllocation_;
            Allocation& a = info.allocation_;
            Allotment ax = a.allotment(Dimension_X);
            Allotment ay = a.allotment(Dimension_Y);
            Extension& b = info.extension_;
            if (
                !(info.status_ & ruiSceneInfoAllocated)
                || !ax.equals(na.allotment(Dimension_X), epsilon)
                || !ay.equals(na.allotment(Dimension_Y), epsilon)
            ) {
                if (info.status_ & ruiSceneInfoExtended) {
                    damage(b);
                }
                b.xy_extents(fil, -fil, fil, -fil);
                info.glyph_->allocate(sceneCanvas_, na, b);
                a = na;
                damage(b);
            }
            info.status_ |= ruiSceneInfoAllocated|ruiSceneInfoExtended;
            ext.extend(b);
        }
    }
    boundingBox();
    notify();
}

void ruiScene::draw(Canvas*, const Allocation&) const {
    canvas_->push_clipping();
    canvas_->clip_rect(allocation_.left(), allocation_.bottom(),
        allocation_.right(), allocation_.top());
    GlyphIndex count = info_->count();
    for (GlyphIndex index = 0; index < count; ++index) {
        ruiSceneInfo& info = info_->item(index);
        if (info.glyph_ != nil && !(info.status_ & ruiSceneInfoHidden)) {
            Allocation& a = info.allocation_;
            Extension& b = info.extension_;
            if (
                sceneCanvas_->damaged(b.left(), b.bottom(), b.right(), b.top())
            ) {
                info.glyph_->draw(sceneCanvas_, a);
            }
        }
    }
    canvas_->pop_clipping();
}

void ruiScene::print(Printer* printer, const Allocation& allocation) const {
    printer->push_clipping();
    printer->clip_rect(allocation.left(), allocation.bottom(),
        allocation.right(), allocation.top());
    printer->push_transform();
    printer->transform(matrix_);
    GlyphIndex count = info_->count();
    for (GlyphIndex index = 0; index < count; ++index) {
        ruiSceneInfo& info = info_->item(index);
        if (info.glyph_ != nil && !(info.status_ & ruiSceneInfoHidden)) {
            Allocation& a = info.allocation_;
            Extension& b = info.extension_;
            if (printer->damaged(b.left(), b.bottom(), b.right(), b.top())) {
                info.glyph_->print(printer, a);
            }
        }
    }
    printer->pop_transform();
    printer->pop_clipping();
}

void ruiScene::pick(Canvas* c, const Allocation&, int depth, Hit& h) {
    h.push_transform();
    h.transform(matrix_);
    Coord x = h.left();
    Coord y = h.bottom();
    GlyphIndex count = info_->count();
    for (GlyphIndex index = 0; index < count; ++index) {
        ruiSceneInfo& info = info_->item(index);
        if (info.glyph_ != nil && !(info.status_ & ruiSceneInfoHidden)) {
            Extension& b = info.extension_;
            Allocation& a = info.allocation_;
            if (
                x >= b.left() && x < b.right()
                && y >= b.bottom() && y < b.top()
            ) {
                h.begin(depth, this, index);
                info.glyph_->pick(c, a, depth + 1, h);
                h.end();
            }
        }
    }
    h.pop_transform();
}

Coord ruiScene::lower(DimensionName dimension) const {
  return dimension == Dimension_X ? left_ : bottom_;
}

Coord ruiScene::upper(DimensionName dimension) const {
  return dimension == Dimension_X ? right_ : top_;
}

Coord ruiScene::length(DimensionName dimension) const {
  return upper(dimension) - lower(dimension);
}

Coord ruiScene::cur_lower(DimensionName dimension) const {
  return dimension == Dimension_X ? curLowerX_ : curLowerY_;
}

Coord ruiScene::cur_upper(DimensionName dimension) const {
  return cur_lower(dimension) + cur_length(dimension);
}

Coord ruiScene::cur_length(DimensionName dimension) const {
  return allocation_.allotment(dimension).span();
}

void ruiScene::scroll_to(DimensionName dimension, Coord position) {
  *(dimension == Dimension_X ? &curLowerX_ : & curLowerY_) = position;
  matrix_ = Transformer();
  matrix_.translate(allocation_.left() - curLowerX_, allocation_.bottom() -
    curLowerY_);
  if (canvas_) {
    canvas_->damage(allocation_.left(), allocation_.bottom(),
      allocation_.right(), allocation_.top());
  }
  notify();
}
