/*
 * Copyright (c) 1991 Redwood Design Automation
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the name of
 * Redwood Design Automation may not be used in any advertising or publicity
 * relating to the software without the specific, prior written permission of
 * Redwood Design Automation.
 *
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
 *
 * IN NO EVENT SHALL REDWOOD DESIGN AUTOMATION BE LIABLE FOR ANY SPECIAL,
 * INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND, OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER OR NOT
 * ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF LIABILITY,
 * ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
 * SOFTWARE.
 */

#include <InterViews/action.h>
#include <InterViews/background.h>
#include <InterViews/box.h>
#include <InterViews/center.h>
#include <InterViews/color.h>
#include <InterViews/cursor.h>
#include <InterViews/event.h>
#include <InterViews/font.h>
#include <InterViews/glue.h>
#include <InterViews/handler.h>
#include <InterViews/hit.h>
#include <InterViews/label.h>
#include <InterViews/monoglyph.h>
#include <InterViews/session.h>
#include <InterViews/style.h>
#include <InterViews/window.h>

#include <IV-look/button.h>
#include <IV-look/kit.h>

#include <IV-X11/xevent.h>
#include <IV-X11/xwindow.h>

#include <rutArray.h>
#include <rutString.h>

#include <ruiSheet.h>
#include <ruiText.h>

////////////////
// Sheet Item //
////////////////

class ruiSheetItem : public MonoGlyph {
public:
  ruiSheetItem(rutString label, rutString value);
  ~ruiSheetItem();

  rutString value();

protected:
  Glyph* label_;
  ruiTextLine* value_;
};

ruiSheetItem::ruiSheetItem(rutString label, rutString value) {
  Kit* kit = Kit::instance();
  Style* style = Session::instance()->style();

  char* s = label.newCString();
  label_ = new Center(new Label(s, new Font("fixed"), new Color(0, 0, 0, 1)),
    Dimension_Y, 0);
  delete s;
  value_ = new ruiTextLine(value, label.length() + 10);
  value_->location(0, value.length());

  body(new LRBox(label_, kit->inset_frame(new ruiTextLineAdjuster(value_),
    style)));
}

ruiSheetItem::~ruiSheetItem() {
}

rutString ruiSheetItem::value() {
  return value_->value();
}

////////////////
// Sheet Base //
////////////////

ruiSheetBase::ruiSheetBase() {
}

ruiSheetBase::~ruiSheetBase() {
}

/////////////////
// Sheet Glyph //
/////////////////

implement(ActionCallback, ruiSheetGlyph)

ruiSheetGlyph::ruiSheetGlyph() {
  Kit* kit = Kit::instance();
  Style* style = Session::instance()->style();

  box_ = new TBBox;

  Button* ok = kit->simple_push_button("ok", style,
    new ActionCallback(ruiSheetGlyph)(this, &ruiSheetGlyph::okAction));
  Button* cancel = kit->simple_push_button("cancel", style,
    new ActionCallback(ruiSheetGlyph)(this, &ruiSheetGlyph::cancelAction));

  body(new TBBox(box_, new LRBox(new HGlue(), ok, new HGlue(), cancel,
    new HGlue)));
}

ruiSheetGlyph::~ruiSheetGlyph() {
}

int ruiSheetGlyph::addItem(rutString label, rutString value) {
  ruiSheetItem* item = new ruiSheetItem(label, value);
  item_.insertLast(item);
  box_->append(item);
  return item_.size() - 1;
}

void ruiSheetGlyph::getValue(unsigned index, rutString& value) {
  value = item_[index]->value();
}

int ruiSheetGlyph::addItem(rutString label, float value) {
  ruiSheetItem* item = new ruiSheetItem(label, value);
  item_.insertLast(item);
  box_->append(item);
  return item_.size() - 1;
}

void ruiSheetGlyph::getValue(unsigned index, float& value) {
  value = 0;
  item_[index]->value().convert(value);
}

int ruiSheetGlyph::addItem(rutString label, int value) {
  ruiSheetItem* item = new ruiSheetItem(label, value);
  item_.insertLast(item);
  box_->append(item);
  return item_.size() - 1;
}

void ruiSheetGlyph::getValue(unsigned index, int& value) {
  value = 0;
  item_[index]->value().convert(value);
}

boolean ruiSheetGlyph::ok() {
  return ok_;
}

void ruiSheetGlyph::okAction() {
  ok_ = true;
}

void ruiSheetGlyph::cancelAction() {
  ok_ = false;
}

//////////////////
// Sheet Dialog //
//////////////////

ruiSheetDialog::ruiSheetDialog() {
}

ruiSheetDialog::~ruiSheetDialog() {
}

void ruiSheetDialog::dialog(Window* window) {
  if (! window->display()) {
    window->display(Session::instance()->default_display());
  }
  XDisplay* dpy = window->display()->rep()->display_;
  XWindow root;
  XWindow child;
  int x, y;
  int cx, cy;
  unsigned int buttons;
  XQueryPointer(dpy, DefaultRootWindow(dpy), &root, &child, &x, &y, &cx, &cy,
    &buttons);
  y = window->display()->rep()->pheight_ - y;
  window->align(0.5, 0.5);
  window->pplace(x, y);
  window->map();
  window->grab_pointer(ltextCursor);

  Event event;
  event.display(window->display());
  done_ = 0;
  while (! done_) {
    event.read();
    Handler* h = nil;
    if (event.rep()->has_pointer_location()) {
      Window* w = event.rep()->window_;
      if (w == window) {
        Hit hit(&event);
        WindowRep* rep = w->rep();
        rep->glyph_->pick(rep->canvas_, rep->allocation_, 0, hit);
        h = hit.handler();
      }
    }
    if (h != nil) {
      boolean b = Resource::defer(true);
      h->ref();
      h->event(event);
      h->unref();
      Resource::flush();
      Resource::defer(b);
    }
  }

  window->ungrab_pointer();
  window->unmap();
}

void ruiSheetDialog::done(boolean done) {
  done_ = done;
}

///////////
// Sheet //
///////////

class ruiSheetBody : public ruiSheetGlyph {
public:
  ruiSheetBody(ruiSheetDialog* dialog) : ruiSheetGlyph(), dialog_(dialog) {}

protected:
  void okAction() {dialog_->done(true); ruiSheetGlyph::okAction();}
  void cancelAction() {dialog_->done(true); ruiSheetGlyph::cancelAction();}

  ruiSheetDialog* dialog_;
};

ruiSheet::ruiSheet() {
  glyph_ = new ruiSheetBody(this);
  Resource::ref(glyph_);
}

ruiSheet::~ruiSheet() {
  Resource::unref(glyph_);
}

int ruiSheet::addItem(rutString label, rutString value) {
  return glyph_->addItem(label, value);
}

void ruiSheet::getValue(unsigned index, rutString& value) {
  glyph_->getValue(index, value);
}

int ruiSheet::addItem(rutString label, float value) {
  return glyph_->addItem(label, value);
}

void ruiSheet::getValue(unsigned index, float& value) {
  glyph_->getValue(index, value);
}

int ruiSheet::addItem(rutString label, int value) {
  return glyph_->addItem(label, value);
}

void ruiSheet::getValue(unsigned index, int& value) {
  glyph_->getValue(index, value);
}

boolean ruiSheet::popup() {
  Style* style = Session::instance()->style();
  Color* color;
  if (style->value_is_on("monochrome")) {
    color = new Color(*style->background());
  } else {
    color = new Color(.8, .8, .8, 1);
  }
  Glyph* top = new Background(glyph_, color);
  TransientWindow* window = new TransientWindow(top);
  dialog(window);
  delete window;
  return glyph_->ok();
}
