/*
 * Copyright (c) 1991 Redwood Design Automation
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the name of
 * Redwood Design Automation may not be used in any advertising or publicity
 * relating to the software without the specific, prior written permission of
 * Redwood Design Automation.
 *
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
 *
 * IN NO EVENT SHALL REDWOOD DESIGN AUTOMATION BE LIABLE FOR ANY SPECIAL,
 * INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND, OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER OR NOT
 * ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF LIABILITY,
 * ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
 * SOFTWARE.
 */

#include <string.h>

#include <iostream.h>
#include <fstream.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include <InterViews/action.h>
#include <InterViews/box.h>
#include <InterViews/canvas.h>
#include <InterViews/color.h>
#include <InterViews/event.h>
#include <InterViews/fixedspan.h>
#include <InterViews/font.h>
#include <InterViews/handler.h>
#include <InterViews/geometry.h>
#include <InterViews/session.h>
#include <InterViews/style.h>
#include <InterViews/window.h>

#include <IV-look/button.h>
#include <IV-look/kit.h>

#include <OS/math.h>
#include <OS/string.h>

#include <rutString.h>
#include <rutText.h>

#include <ruiDrag.h>
#include <ruiText.h>

implement(ActionCallback, ruiText);

ruiText::ruiText(unsigned initialLines, unsigned initialColumns) {
  target(this);

  Style* style = Session::instance()->style();

  text_ = new rutText;

  selection_.line1_ = 0;
  selection_.line2_ = 0;
  selection_.column1_ = 0;
  selection_.column2_ = 0;
  if (style->value_is_on("monochrome")) {
    selection_.color_ = new Color(*style->foreground(), .1);
  } else {
    selection_.color_ = new Color(.7, .8, 1, 1);
  }
  Resource::ref(selection_.color_);

  insertion_.line_ = 0;
  insertion_.column_ = 0;
  if (style->value_is_on("monochrome")) {
    insertion_.color_ = new Color(*style->foreground());
  } else {
    insertion_.color_ = new Color(1, .5, .5, 1);
  }
  Resource::ref(insertion_.color_);
  insertion_.width_ = 2;

  initialLines_ = Math::max(initialLines, 1u);
  initialColumns_ = Math::max(initialColumns, 1u);

  font_ = new Font(*style->font());
  Resource::ref(font_);
  textColor_ = new Color(*style->foreground());
  Resource::ref(textColor_);

  button(true, Event::any);
  motion(true);
  readOnly_ = false;
  key(! readOnly_);

  canvas_ = 0;
  allocation_ = 0;

  curLowerX_ = 0;
  curUpperX_ = 0;
  curLowerY_ = 0;
  curUpperY_ = 0;

  textBuffer_ = 0;

  needWidth_ = false;
  width_ = 0;
}

ruiText::~ruiText() {
  Resource::unref(selection_.color_);
  Resource::unref(insertion_.color_);
  Resource::unref(font_);
  Resource::unref(textColor_);
  delete text_;
}

inline Coord ruiText::width(char ch) const {
  return ch == '\t' ? font_->width(' ') * 8 : font_->width(ch);
}

Coord ruiText::width(const rutString& line) const {
  Coord lineTotal = 0;
  for (unsigned j = 0; j < line.length(); ++j) {
    lineTotal += width(line[j]);
  }
  return lineTotal;
}

Coord ruiText::width() const {
  if (needWidth_) {
    Coord total = 0;
    for (unsigned i = 0; i < text_->lines(); ++i) {
      const rutString& line = text_->getNth(i);
      total = Math::max(total, width(line));
    }
    ((ruiText*) this)->width_ = total;
    ((ruiText*) this)->needWidth_ = false;
  }
  return width_;
}

void ruiText::readOnly(boolean readOnly) {
  readOnly_ = readOnly;
  key(! readOnly_);
  damage(insertion_);
}

boolean ruiText::readOnly() const {
  return readOnly_;
}

void ruiText::font(const rutString& font) {
  fontName_ = font;
  char* name = font.newCString();
  Resource::unref(font_);
  font_ = new Font(name);
  Resource::ref(font_);
  delete name;
  needWidth_ = true;
  notify();
  damage();
}

rutString ruiText::font() const {
  return fontName_;
}

void ruiText::request(Requisition& requisition) const {
  Requirement rx(width(' ') * initialColumns_, fil,
    width(' ') * (initialColumns_ - 1), 0);
  Requirement ry((font_->ascent() + font_->descent()) * initialLines_, fil,
    (font_->ascent() + font_->descent()) * (initialLines_ - 1), 0);
  requisition.require(Dimension_X, rx);
  requisition.require(Dimension_Y, ry);
}

void ruiText::allocate(Canvas* canvas, const Allocation& allocation,
  Extension& extension) {
  canvas_ = canvas;
  if (! allocation_) {
    allocation_ = new Allocation(allocation);
    curLowerX_ = 0;
    curLowerY_ = 0;
  } else {
    *allocation_ = allocation;
  }
  curUpperX_ = curLowerX_ + allocation_->allotment(Dimension_X).span();
  curUpperY_ = curLowerY_ + allocation_->allotment(Dimension_Y).span();

  extension.extend(allocation);
}

Coord ruiText::columnCoord(const rutString& line, unsigned column) const {
  Coord x;

  x = allocation_->left() - curLowerX_;
  if (line) {
    unsigned to = Math::min(column, unsigned(line.length()));
    rutString part = line.substring(0, to);
    for (unsigned i = 0; i < to; ++i) {
      x += width(line[i]);
    }
    if (column > line.length()) {
      x += font_->width(' ') * (column - line.length());
    }
  } else {
    x += font_->width(' ') * column;
  }

  return x;
}

void ruiText::drawRegion(const ruiTextRegion& region, unsigned i,
  Coord x, Coord y, const rutString& line) const {
  unsigned line1 = region.line1_;
  unsigned line2 = region.line2_;
  unsigned column1 = region.column1_;
  unsigned column2 = region.column2_;

  if ((line1 == i) && (line1 == line2) && (column1 < column2)) {
    canvas_->fill_rect(columnCoord(line, column1), y - font_->descent(),
      columnCoord(line, column2), y + font_->ascent(), region.color_);
  }
  if ((line1 == i) && (line1 < line2)) {
    canvas_->fill_rect(columnCoord(line, column1), y - font_->descent(),
      allocation_->right(), y + font_->ascent(), region.color_);
  }
  if ((line1 < i) && (i < line2)) {
    canvas_->fill_rect(x, y - font_->descent(), allocation_->right(), y +
      font_->ascent(), region.color_);
  }
  if ((line2 == i) && (line1 < line2)) {
    canvas_->fill_rect(0, y - font_->descent(), columnCoord(line, column2),
      y + font_->ascent(), region.color_);
  }
}

void ruiText::drawLocation(const ruiTextLocation& location, unsigned i,
  Coord /* x */, Coord y, const rutString& line) const {
  if (location.line_ == i) {
    Coord lb = columnCoord(line, location.column_);
    canvas_->fill_rect(lb, y - font_->descent(), lb + location.width_,
      y + font_->ascent(), location.color_);
  }
}

void ruiText::drawLine(unsigned /* i */, Coord x, Coord y,
  const rutString& line) const {
  for (unsigned j = 0; j < line.length(); ++j) {
    char ch = line[j];
    Coord w = width(ch);
    if (ch != '\t') {
      canvas_->character(font_, ch, w, textColor_, x, y);
    }
    x += w;
  }
}

boolean ruiText::damaged(unsigned line) const {
  Coord base = allocation_->top() + curLowerY_;
  Coord lineHeight = font_->ascent() + font_->descent();
  Coord top = base - lineHeight * line;
  Coord bottom = base - lineHeight * (line + 1);
  return canvas_->damaged(allocation_->left(), Math::max(bottom,
    allocation_->bottom()), allocation_->right(), Math::min(top,
    allocation_->top()));
}

void ruiText::draw(Canvas*, const Allocation&) const {
  canvas_->push_clipping();
  canvas_->clip_rect(allocation_->left(), allocation_->bottom(),
    allocation_->right(), allocation_->top());

  Coord r = curLowerY_ / (font_->ascent() + font_->descent());
  unsigned i = unsigned(r);
  Coord y = allocation_->top() + (r - i) * (font_->ascent() + font_->descent());
  unsigned max = Math::max(Math::max(selection_.line2_, insertion_.line_),
    text_->lines() ? text_->lines() - 1 : 0);
  for (; i <= max; ++i) {
    y -= font_->ascent();

    if (damaged(i)) {
      Coord x = allocation_->left() - curLowerX_;
      if (i < text_->lines()) {
        const rutString& line = text_->getNth(i);
        drawRegion(selection_, i, x, y, line);
        if (! readOnly_) {
          drawLocation(insertion_, i, x, y, line);
        }
        for (unsigned j = 0; j < annotation_.size(); ++j) {
          drawRegion(*annotation_.get(j), i, x, y, line);
        }
        drawLine(i, x, y, line);
      } else {
        rutString line;
        drawRegion(selection_, i, x, y, line);
        if (! readOnly_) {
          drawLocation(insertion_, i, x, y, line);
        }
        for (unsigned j = 0; j < annotation_.size(); ++j) {
          drawRegion(*annotation_.get(j), i, x, y, line);
        }
        drawLine(i, x, y, line);
      }
    }
  
    y -= font_->descent();
  
    if (y < (allocation_->bottom() - font_->ascent())) {
      break;
    }
  }

  canvas_->pop_clipping();
}
 
boolean ruiText::snap(Event& event, unsigned& line, unsigned& column) const {
  unsigned originalLine = line;
  unsigned originalColumn = column;

  Coord x = event.pointer_x() - allocation_->left() + curLowerX_;
  Coord y = allocation_->top() + curLowerY_ - event.pointer_y();

  line = Math::max(int(y / (font_->ascent() + font_->descent())), 0);
  if (line < text_->lines()) {
    const rutString& string = text_->getNth(line);
    for (unsigned i = 0; i < string.length(); ++i) {
      x -= width(string[i]) / 2.0;
      if (x < 0) {
        break;
      }
      x -= width(string[i]) / 2.0;
    }
    column = i;
    if (i >= string.length()) {
      column += (unsigned) ((x + width(' ') / 2) / width(' '));
    }
  } else {
    column = (unsigned) ((x + width(' ') / 2) / width(' '));
  }
  column = Math::max(column, unsigned(0));

  return (line != originalLine) || (column != originalColumn);
}

void ruiText::dragLoad(Event& event) {
  char* fileName;
  if (ruiDrag::value(event, fileName)) {
    load(fileName);
    delete fileName;
  }
}

void ruiText::eraseLine() {
  if (insertion_.line_ < text_->lines()) {
    const rutString& string = text_->getNth(insertion_.line_);
    rutTextRegion region(insertion_.line_, 0, insertion_.line_,
      string.length());
    rutText* text = text_->cut(region);
    delete text;
    if (width(string) >= width_) {
      needWidth_ = true;
      notify();
    }
  }
  insertion_.column_ = 0;
  repair();
  damage(insertion_);
  repair();
}

void ruiText::backspace() {
  if ((insertion_.column_ <= 0) && (insertion_.line_ > 0)) {
    --insertion_.line_;
    if (insertion_.line_ >= text_->lines()) {
      insertion_.column_ = 0;
    } else {
      const rutString& string = text_->getNth(insertion_.line_);
      insertion_.column_ = string.length();
      rutTextRegion region(insertion_.line_, insertion_.column_,
        insertion_.line_ + 1, 0);
      rutText* text = text_->cut(region);
      delete text;
    }
    needWidth_ = true;
    notify();
    damage(); // !!! could only damage from insertion to end of window
  } else if (insertion_.column_ > 0) {
    boolean shrunk = (width(text_->getNth(insertion_.line_)) >= width_);
    rutTextRegion region(insertion_.line_, insertion_.column_ - 1,
      insertion_.line_, insertion_.column_);
    rutText* text = text_->cut(region);
    delete text;
    --insertion_.column_;
    if (shrunk) {
      needWidth_ = true;
      notify();
    }
    repair();
    damage(insertion_);
    repair();
  }
}

void ruiText::insertChars(char* buffer, unsigned count) {
  // some string of characters
  rutText text;
  text.write(buffer, count);
  rutTextLocation location(insertion_.line_, insertion_.column_);
  text_->paste(location, &text);
  if (text.lines() > 1) {
    ruiTextRegion area;
    area.line1_ = insertion_.line_;
    area.column1_ = 0;
    area.line2_ = insertion_.line_ + text.lines() - 1;
    area.column2_ = 0;
    insertion_.line_ += text.lines() - 1;
    insertion_.column_ = 0;
    damage(); // !!! could only damage from insertion to end of window
    for (unsigned i = 0; i < text.lines(); ++i) {
      width_ = Math::max(width_, width(text.getNth(i)));
    }
    notify();
  } else {
    ruiTextLocation old = insertion_;
    insertion_.column_ += count;
    Coord newWidth = width(text_->getNth(insertion_.line_));
    if (newWidth >= width_) {
      width_ = Math::max(width_, newWidth);
      notify();
    }
    repair();
    damage(old);
    repair();
  }
}

void ruiText::keyPress(Event& event) {
  char buffer[8]; // needs to be dynamically adjusted
  int count = event.mapkey(buffer, 8);
  if (count <= 0) {
    return;
  }

  // return causes a newline
  if (buffer[0] == '\r') {
    buffer[0] = '\n';
  }

  if (buffer[0] == 21) {
    // control u
    eraseLine();
  } else if ((buffer[0] == '\b') || (buffer[0] == 127)) {
    // backspace and delete
    backspace();
  } else {
    insertChars(buffer, count);
  }

  dirty(true);
}

boolean ruiText::caught(const Event& event) const {
  return ruiDrag::isDrag(event) || Listener::caught(event);
}

void ruiText::event(Event& event) {
  if (! readOnly_) {
    if (ruiDrag::isDrag(event)) {
      dragLoad(event);
      return;
    }

    if (event.type() == Event::key) {
      keyPress(event);
      return;
    }
  }

  PointerHandler::event(event);
}

void ruiText::repair() {
  if (canvas_ && canvas_->window()) {
    canvas_->window()->repair();
  }
}

void ruiText::damage(const ruiTextLocation& location) {
  if (! canvas_) {
    return;
  }

  Coord base = allocation_->top() + curLowerY_;
  Coord lineHeight = font_->ascent() + font_->descent();
  Coord top = Math::min(base - lineHeight * location.line_, allocation_->top());
  Coord bottom = Math::max(base - lineHeight * (location.line_ + 1),
    allocation_->bottom());
  if ((allocation_->left() <= allocation_->right()) && (bottom <= top)) {
    canvas_->damage(allocation_->left(), bottom, allocation_->right(), top);
  }
}

void ruiText::damage(const ruiTextRegion& region) {
  if (! canvas_) {
    return;
  }

  Coord base = allocation_->top() + curLowerY_;
  Coord lineHeight = font_->ascent() + font_->descent();
  Coord top = Math::min(base - lineHeight * region.line1_, allocation_->top());
  Coord bottom = Math::max(base - lineHeight * (region.line2_ + 1),
    allocation_->bottom());
  if ((allocation_->left() <= allocation_->right()) && (bottom <= top)) {
    canvas_->damage(allocation_->left(), bottom, allocation_->right(), top);
  }
}

void ruiText::damage() {
  if (! canvas_) {
    return;
  }

  canvas_->damage(allocation_->left(), allocation_->bottom(),
    allocation_->right(), allocation_->top());
}

void ruiText::press(Event& event) {
  PointerHandler::press(event);

  if (event.pointer_button() == Event::left) {
    ruiTextRegion old = selection_;
    snap(event, selection_.line1_, selection_.column1_);
    selection_.line2_ = selection_.line1_;
    selection_.column2_ = selection_.column1_;
    repair();
    damage(old);
    repair();
  } else if (event.pointer_button() == Event::right) {
    ruiTextLocation old = insertion_;
    snap(event, insertion_.line_, insertion_.column_);
    repair();
    damage(old);
    repair();
    damage(insertion_);
    repair();
  }
}

void ruiText::drag(Event& event) {
  PointerHandler::drag(event);

  if (event.left_is_down()) {
    unsigned line = selection_.line2_;
    unsigned column = selection_.column2_;
    if (snap(event, line, column) && ((line > selection_.line1_) ||
      ((line == selection_.line1_) && (column >= selection_.column1_)))) {
      ruiTextRegion area;
      if (selection_.line2_ < line) {
        area.line1_ = selection_.line2_;
        area.column1_ = selection_.column2_;
        area.line2_ = line;
        area.column2_ = column;
      } else {
        area.line1_ = line;
        area.column1_ = column;
        area.line2_ = selection_.line2_;
        area.column2_ = selection_.column2_;
      }
      selection_.line2_ = line;
      selection_.column2_ = column;
      repair();
      damage(area);
      repair();
    }
  } else if (event.right_is_down()) {
    ruiTextLocation old = insertion_;
    if (snap(event, insertion_.line_, insertion_.column_)) {
      repair();
      damage(old);
      repair();
      damage(insertion_);
      repair();
    }
  }
}

Coord ruiText::height() const {
  return text_->lines() * (font_->ascent() + font_->descent());
}

Coord ruiText::lower(DimensionName) const {
  return 0;
}

Coord ruiText::upper(DimensionName dimension) const {
  return (dimension == Dimension_X ? width() : height());
}

Coord ruiText::length(DimensionName dimension) const {
  return upper(dimension) - lower(dimension);
}

Coord ruiText::cur_lower(DimensionName dimension) const {
  return (dimension == Dimension_X ? curLowerX_ : curLowerY_);
}

Coord ruiText::cur_upper(DimensionName dimension) const {
  return (dimension == Dimension_X ? curUpperX_ : curUpperY_);
}

Coord ruiText::cur_length(DimensionName dimension) const {
  return cur_upper(dimension) - cur_lower(dimension);
}

void ruiText::scroll_forward(DimensionName dimension) {
  if (dimension == Dimension_X) {
    scroll_to(dimension, cur_lower(Dimension_X) + width(' '));
  } else {
    scroll_to(dimension, cur_lower(Dimension_Y) +
      (font_->ascent() + font_->descent()));
  }
}

void ruiText::scroll_backward(DimensionName dimension) {
  if (dimension == Dimension_X) {
    scroll_to(dimension, cur_lower(Dimension_X) - width(' '));
  } else {
    scroll_to(dimension, cur_lower(Dimension_Y) -
      (font_->ascent() + font_->descent()));
  }
}

void ruiText::page_forward(DimensionName dimension) {
  scroll_to(dimension, cur_upper(dimension));
}

void ruiText::page_backward(DimensionName dimension) {
  scroll_to(dimension, cur_lower(dimension) - cur_length(dimension));
}

void ruiText::cur_lower(DimensionName dimension, Coord position) {
  if (dimension == Dimension_X) {
    curLowerX_ = position;
    curUpperX_ = position + allocation_->allotment(Dimension_X).span();
  } else {
    curLowerY_ = position;
    curUpperY_ = position + allocation_->allotment(Dimension_Y).span();
  }
}

void ruiText::scroll_to(DimensionName dimension, Coord position) {
// let the user scroll to the right (or down) past the current size.
/*
  if (position > (upper(dimension) - cur_length(dimension))) {
    position = upper(dimension) - cur_length(dimension);
  }
*/
  if (position < lower(dimension)) {
    position = lower(dimension);
  }

  if (position == cur_lower(dimension)) {
    return;
  }

  cur_lower(dimension, position);
  notify();
  damage();
}

void ruiText::reset() {
  text_->clear();

  insertion_.line_ = 0;
  insertion_.column_ = 0;

  selection_.line1_ = 0;
  selection_.column1_ = 0;
  selection_.line2_ = 0;
  selection_.column2_ = 0;

  if (textBuffer_) {
    delete textBuffer_;
  }
  textBuffer_ = 0;

  deannotate();

  needWidth_ = false;
  width_ = 0;

  fileName("");
  dirty(false);
  notify();
}

int ruiText::load(rutString name) {
  char* fn = name.newCString();
  ifstream source(fn);
  struct stat stats;
  if ((! source) || (stat(fn, &stats) == -1)) {
    delete fn;
    return 0;
  }
  delete fn;

  char* buffer = new char[(int) stats.st_size];
  source.read(buffer, (int) stats.st_size);
  if (source.bad()) {
    delete buffer;
    return 0;
  }

  text_->clear();

  insertion_.line_ = 0;
  insertion_.column_ = 0;

  selection_.line1_ = 0;
  selection_.column1_ = 0;
  selection_.line2_ = 0;
  selection_.column2_ = 0;

  if (textBuffer_) {
    delete textBuffer_;
  }
  textBuffer_ = 0;

  deannotate();

  fileName(name);

  text_->write(buffer, (int) stats.st_size);

  delete buffer;

  needWidth_ = true;

  dirty(false);
  notify();
  damage();

  return 1;
}

int ruiText::save() {
  char* fn = fileName_.newCString();
  ofstream dest(fn);
  delete fn;
  if (! dest) {
    return 0;
  }

  for (unsigned i = 0; i < text_->lines(); ++i) {
    const rutString& string = text_->getNth(i);
    char* buffer = string.newCString();
    dest.write(buffer, string.length());
    delete buffer;
    if (i < (text_->lines() - 1)) {
      dest.write("\n", strlen("\n"));
    }
  }

  dirty(false);

  return 1;
}

void ruiText::modified() {
}

boolean ruiText::dirty() {
  return dirty_;
}

void ruiText::dirty(boolean dirty) {
  if (dirty_ != dirty) {
    dirty_ = dirty;
    modified();
  }
}

rutString ruiText::fileName() {
  return fileName_;
}

void ruiText::fileName(rutString fileName) {
  if (fileName_ != fileName) {
    fileName_ = fileName;
    dirty_ = true;
    modified();
  }
}

void ruiText::write(const rutString& string) {
  text_->write(string);
  needWidth_ = true;
  dirty(true);
  notify();
  damage();
}

void ruiText::write(char* buffer, unsigned count) {
  text_->write(buffer, count);
  needWidth_ = true;
  dirty(true);
  notify();
  damage();
}

void ruiText::copy() {
  unsigned line1 = selection_.line1_;
  unsigned line2 = selection_.line2_;
  unsigned column1 = selection_.column1_;
  unsigned column2 = selection_.column2_;

  if ((line1 > line2) || ((line1 == line2) && (column1 >= column2))) {
    return;
  }
  rutTextRegion region(line1, column1, line2, column2);
  if (textBuffer_) {
    delete textBuffer_;
  }
  textBuffer_ = text_->copy(region);
}

void ruiText::cut() {
  unsigned line1 = selection_.line1_;
  unsigned line2 = selection_.line2_;
  unsigned column1 = selection_.column1_;
  unsigned column2 = selection_.column2_;

  if ((line1 > line2) || ((line1 == line2) && (column1 >= column2))) {
    return;
  }
  rutTextRegion region(line1, column1, line2, column2);
  if (textBuffer_) {
    delete textBuffer_;
  }
  textBuffer_ = text_->cut(region);
  insertion_.line_ = selection_.line1_;
  insertion_.column_ = selection_.column1_;
  selection_.line2_ = selection_.line1_;
  selection_.column2_ = selection_.column1_;
  needWidth_ = true;
  dirty(true);
  notify();
  damage();
}

void ruiText::paste() {
  if (! textBuffer_) {
    return;
  }
  rutTextLocation location(insertion_.line_, insertion_.column_);
  text_->paste(location, textBuffer_);
  needWidth_ = true;
  dirty(true);
  notify();
  damage();
}

void ruiText::replaceSelection() {
  unsigned line1 = selection_.line1_;
  unsigned line2 = selection_.line2_;
  unsigned column1 = selection_.column1_;
  unsigned column2 = selection_.column2_;

  rutText* text = 0;
  if ((line1 <= line2) && ((line1 != line2) || (column1 < column2))) {
    rutTextRegion region(line1, column1, line2, column2);
    text = text_->cut(region);
  }
  rutTextLocation location(selection_.line1_, selection_.column1_);
  text_->paste(location, textBuffer_);
  if (textBuffer_) {
    delete textBuffer_;
  }
  textBuffer_ = text;

  selection_.line2_ = selection_.line1_;
  selection_.column2_ = selection_.column1_;
  needWidth_ = true;
  dirty(true);
  notify();
  damage();
}

void ruiText::replaceWith(const rutString& string) {
  unsigned line1 = selection_.line1_;
  unsigned line2 = selection_.line2_;
  unsigned column1 = selection_.column1_;
  unsigned column2 = selection_.column2_;

  rutText* text = 0;
  if ((line1 <= line2) && ((line1 != line2) || (column1 < column2))) {
    rutTextRegion region(line1, column1, line2, column2);
    text = text_->cut(region);
  }
  rutTextLocation location(selection_.line1_, selection_.column1_);
  rutText* buffer = new rutText;
  buffer->write(string);
  text_->paste(location, buffer);
  delete buffer;
  if (textBuffer_) {
    delete textBuffer_;
  }
  textBuffer_ = text;

  selection_.line2_ = selection_.line1_;
  selection_.column2_ = selection_.column1_;
  needWidth_ = true;
  dirty(true);
  notify();
  damage();
}

rutTextRegion ruiText::search(const rutTextLocation& location,
  const rutLexRegExp& regExp) const {
  return text_->search(location, regExp);
}

void ruiText::expose(unsigned line1, unsigned column1, unsigned /*line2*/,
  unsigned /*column2*/) {
  expose(line1, column1);
}

void ruiText::region(unsigned line1, unsigned column1, unsigned line2,
  unsigned column2) {
  ruiTextRegion old = selection_;
  selection_.line1_ = line1;
  selection_.column1_ = column1;
  selection_.line2_ = line2;
  selection_.column2_ = column2;
  damage(old);
  damage(selection_);
  expose(line1, column1, line2, column2);
}

void ruiText::getRegion(unsigned& line1, unsigned& column1, unsigned& line2,
  unsigned& column2) {
  line1 = selection_.line1_;
  column1 = selection_.column1_;
  line2 = selection_.line2_;
  column2 = selection_.column2_;
}

void ruiText::expose(unsigned line, unsigned column) {
  if (! canvas_) {
    return;
  }
  rutString string;
  if (line < text_->lines()) {
    string = text_->getNth(line);
  }
  Coord x = columnCoord(string, column) - allocation_->left() + curLowerX_;
  boolean invisibleX = (x < curLowerX_) || ((x + width(' ')) > curUpperX_);
  Coord y = line * (font_->ascent() + font_->descent());
  boolean invisibleY = (y < curLowerY_) || ((y + font_->ascent() +
    font_->descent()) > curUpperY_);
  if (invisibleX) {
    scroll_to(Dimension_X, x - (curUpperX_ - curLowerX_) / 2);
  }
  if (invisibleY) {
    scroll_to(Dimension_Y, y - (curUpperY_ - curLowerY_) / 2);
  }
}

void ruiText::location(unsigned line, unsigned column) {
  ruiTextLocation old = insertion_;
  insertion_.line_ = line;
  insertion_.column_ = column;
  damage(old);
  damage(insertion_);
  expose(line, column);
}

void ruiText::getLocation(unsigned& line, unsigned& column) {
  line = insertion_.line_;
  column = insertion_.column_;
}

void ruiText::annotate(ruiTextRegion* region) {
  annotation_.insertLast(region);
  damage(*region);
}

void ruiText::deannotate(ruiTextRegion* region) {
  annotation_.removeEntry(region);
  damage(*region);
}

void ruiText::deannotate() {
  while (annotation_.size()) {
    annotation_.removeLast();
  }
  damage();
}

rutTextIterator* ruiText::iterator() {
  rutTextLocation location;
  rutTextIterator* iterator = new rutTextIterator;
  *iterator = text_->iterator(location);
  return iterator;
}

///////////////
// Text Line //
///////////////

ruiTextLine::ruiTextLine(unsigned initialColumns) :
  ruiText(1u, initialColumns) {
}

ruiTextLine::ruiTextLine(rutString line, unsigned initialColumns) :
  ruiText(1u, initialColumns) {
  text_->write(line);
}

ruiTextLine::~ruiTextLine() {
}

void ruiTextLine::request(Requisition& requisition) const {
  ruiText::request(requisition);
  Requirement ry(font_->ascent() + font_->descent(), 0, 0, 0);
  requisition.require(Dimension_Y, ry);
}

void ruiTextLine::keyPress(Event& event) {
  // ignore carrage returns
  char buffer[2];
  int count = event.mapkey(buffer, 2);
  if (buffer[0] == '\r') {
    return;
  }

  ruiText::keyPress(event);
}

rutString ruiTextLine::value() {
  if (! text_->lines()) {
    return "";
  }
  return text_->getNth(0);
}

////////////////////////
// Text Line Adjuster //
////////////////////////

ruiTextLineAdjuster::ruiTextLineAdjuster(ruiTextLine* textLine) :
  Patch(0), Adjuster(textLine) {
  have_ = 0;

  box_ = new LRBox(textLine);
  body(box_);
}

ruiTextLineAdjuster::~ruiTextLineAdjuster() {
}

void ruiTextLineAdjuster::update(DimensionName) {
  update_all();
}

void ruiTextLineAdjuster::update_all() {
  int need = needButtons();
  if (need != have_) {
    if (need) {
      addButtons();
    } else {
      removeButtons();
    }
    reallocate();
    redraw();
    have_ = ! have_;
  }
}

int ruiTextLineAdjuster::needButtons() {
  return allocation().allotment(Dimension_X).span() <
    subject()->length(Dimension_X);
}

void ruiTextLineAdjuster::addButtons() {
  Adjustable* adjustable = subject();
  Kit* kit = Kit::instance();
  Style* style = new Style(Session::instance()->style());

  Button* left = kit->left_mover(adjustable, style);
  Button* right = kit->right_mover(adjustable, style);

  style->prefix("Mover");
  Coord mover_size = 15.0;
  style->find_attribute("mover_size", mover_size);

  box_->prepend(new FixedSpan(left, Dimension_X, mover_size));
  box_->append(new FixedSpan(right, Dimension_X, mover_size));
}

void ruiTextLineAdjuster::removeButtons() {
  box_->remove(2);
  box_->remove(0);
}
