/*
 * Copyright (c) 1991 Redwood Design Automation
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the name of
 * Redwood Design Automation may not be used in any advertising or publicity
 * relating to the software without the specific, prior written permission of
 * Redwood Design Automation.
 *
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
 *
 * IN NO EVENT SHALL REDWOOD DESIGN AUTOMATION BE LIABLE FOR ANY SPECIAL,
 * INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND, OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER OR NOT
 * ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF LIABILITY,
 * ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
 * SOFTWARE.
 */

#ifndef ruiText_h
#define ruiText_h

#include <InterViews/action.h>
#include <InterViews/listener.h>
#include <InterViews/patch.h>
#include <InterViews/ptrhandler.h>

#include <IV-look/adjuster.h>
#include <IV-look/adjustable.h>

#include <rutArray.h>
#include <rutText.h>
#include <rutString.h>

class Requisition;
class Canvas;
class Allocation;
class Extension;
class Font;
class Color;
class Hit;
class rutString;

class ruiTextRegion {
public:
  unsigned line1_;
  unsigned column1_;
  unsigned line2_;
  unsigned column2_;
  Color* color_;
};

rutArrayDefine(ruiTextRegionArray,ruiTextRegion)

class ruiTextLocation {
public:
  unsigned line_;
  unsigned column_;
  Color* color_;
  unsigned width_;
};

class ruiText : public Adjustable, public Listener, public PointerHandler {
public:
  ruiText(unsigned initialLines = 24, unsigned initialColumns = 80);
  ~ruiText();
 
  // Glyph
  void request(Requisition& requisition) const;
  void allocate(Canvas* canvas, const Allocation& allocation, Extension&);
  void draw(Canvas* canvas, const Allocation& allocation) const;

  // PointerHandler
  boolean caught(const Event& event) const;
  void event(Event& event);
  void press(Event& event);
  void drag(Event& event);
  // other handler stuff...
  void dragLoad(Event& event);
  virtual void keyPress(Event& event);

  // Adjustable
  Coord lower(DimensionName dimension) const;
  Coord upper(DimensionName dimension) const;
  Coord length(DimensionName dimension) const;
  Coord cur_lower(DimensionName dimension) const;
  Coord cur_upper(DimensionName dimension) const;
  Coord cur_length(DimensionName dimension) const;
  void scroll_forward(DimensionName dimension);
  void scroll_backward(DimensionName dimension);
  void page_forward(DimensionName dimension);
  void page_backward(DimensionName dimension);
  void scroll_to(DimensionName dimension, Coord position);

  void reset();
  int load(rutString fileName);
  int save();
  rutString fileName();
  void fileName(rutString fileName);

  void write(const rutString& string);
  void write(char* buffer, unsigned count);

  void copy();
  void cut();
  void paste();
  void replaceSelection();
  void replaceWith(const rutString& string);
  rutTextRegion search(const rutTextLocation& location, const rutLexRegExp&
    regExp) const;

  void region(unsigned line1, unsigned column1, unsigned line2,
    unsigned column2);
  void getRegion(unsigned& line1, unsigned& column1, unsigned& line2,
    unsigned& column2);

  void location(unsigned line, unsigned column);
  void getLocation(unsigned& line, unsigned& column);

  void annotate(ruiTextRegion* region);
  void deannotate(ruiTextRegion* region);
  void deannotate();

  void readOnly(boolean readOnly);
  boolean readOnly() const;

  void font(const rutString& font);
  rutString font() const;

  boolean dirty();

  virtual void modified();

  rutTextIterator* iterator();

protected:
  void drawRegion(const ruiTextRegion& region, unsigned i, Coord x, Coord y,
    const rutString& line) const;
  void drawLocation(const ruiTextLocation& location, unsigned i, Coord x,
    Coord y, const rutString& line) const;
  void drawLine(unsigned i, Coord x, Coord y, const rutString& line) const;
  Coord columnCoord(const rutString& line, unsigned column) const;
  boolean snap(Event& event, unsigned& line, unsigned& column) const;
  Coord width(char ch) const;
  Coord width(const rutString& line) const;
  Coord width() const;
  Coord height() const;
  void cur_lower(DimensionName dimension, Coord position);
  void cur_upper(DimensionName dimension, Coord position);
  void repair();
  void damage(const ruiTextLocation& location);
  void damage(const ruiTextRegion& region);
  void damage();
  boolean damaged(unsigned line) const;
  void eraseLine();
  void backspace();
  void insertChars(char* buffer, unsigned count);
  void expose(unsigned line, unsigned column);
  void expose(unsigned line1, unsigned column1, unsigned line2,
    unsigned column2);
  void dirty(boolean dirty);

  rutText* text_;
  boolean readOnly_;

  ruiTextRegion selection_;
  ruiTextLocation insertion_;

  ruiTextRegionArray annotation_;

  rutText* textBuffer_;

  unsigned initialLines_;
  unsigned initialColumns_;

  rutString fontName_;
  Font* font_;
  Color* textColor_;

  Canvas* canvas_;
  Allocation* allocation_;

  Coord curLowerX_;
  Coord curUpperX_;
  Coord curLowerY_;
  Coord curUpperY_;

  rutString fileName_;
  boolean dirty_;

  Coord width_;
  boolean needWidth_;
};

declare(ActionCallback, ruiText);

class ruiTextLine : public ruiText {
public:
  ruiTextLine(unsigned initialColumns = 10);
  ruiTextLine(rutString line, unsigned initialColumns = 10);
  ~ruiTextLine();

  // Glyph
  void request(Requisition& requisition) const;

  // other handler stuff...
  void keyPress(Event& event);

  rutString value();
};

class Box;

class ruiTextLineAdjuster : public Patch, public Adjuster {
public:
  ruiTextLineAdjuster(ruiTextLine* textLine);
  ~ruiTextLineAdjuster();

  void update(DimensionName);
  void update_all();

protected:
  int needButtons();
  void addButtons();
  void removeButtons();

  int have_;
  Box* box_;
};

#endif
