/*
 * Copyright (c) 1991 Redwood Design Automation
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the name of
 * Redwood Design Automation may not be used in any advertising or publicity
 * relating to the software without the specific, prior written permission of
 * Redwood Design Automation.
 *
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
 *
 * IN NO EVENT SHALL REDWOOD DESIGN AUTOMATION BE LIABLE FOR ANY SPECIAL,
 * INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND, OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER OR NOT
 * ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF LIABILITY,
 * ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
 * SOFTWARE.
 */

#include <InterViews/action.h>
#include <InterViews/background.h>
#include <InterViews/center.h>
#include <InterViews/color.h>
#include <InterViews/fixedspan.h>
#include <InterViews/monoglyph.h>
#include <OS/string.h> // because style.h needs it but doesn't include it?
#include <InterViews/style.h>

#include <IV-look/kit.h>
#include <IV-look/menu.h>

#include <rutLex.h>

#include <ruiBox.h>
#include <ruiKeyMap.h>
#include <ruiLegend.h>
#include <ruiPanner.h>
#include <ruiSheet.h>
#include <ruiText.h>
#include <ruiTextEdit.h>

//////////////////////
// Text Editor Text //
//////////////////////

class ruiTextEditText : public ruiText {
public:
  ruiTextEditText(ruiLegend* legend);
  ~ruiTextEditText();

  void modified();

protected:
  ruiLegend* legend_;
};

ruiTextEditText::ruiTextEditText(ruiLegend* legend) {
  legend_ = legend;
}

ruiTextEditText::~ruiTextEditText() {
}

void ruiTextEditText::modified() {
  legend_->modified(dirty_);
  legend_->name(fileName_);
}

//////////////////////
// Text Editor Font //
//////////////////////

class ruiTextEditFont : public Action {
public:
  ruiTextEditFont(ruiTextEdit* text, char* name);

  void execute();

protected:
  ruiTextEdit* text_;
  char* name_;
};

ruiTextEditFont::ruiTextEditFont(ruiTextEdit* text, char* name) {
  text_ = text;
  name_ = name;
}

void ruiTextEditFont::execute() {
  text_->font(name_);
}

/////////////////
// Text Editor //
/////////////////

implement(ActionCallback, ruiTextEdit);

ruiTextEdit::ruiTextEdit(Kit* kit, Style* style) {
  legend_ = new ruiLegend(kit, style);
  text_ = new ruiTextEditText(legend_);

  Menu* menuBar = kit->menubar(style);
  Menu* pulldown;

  pulldown = kit->pulldown(style);
  pulldown->add_item(kit->simple_menu_item("New", style),
    new ActionCallback(ruiText)(text_, &ruiText::reset));
  pulldown->add_item(kit->simple_menu_item("Load", style),
    new ActionCallback(ruiTextEdit)(this, &ruiTextEdit::load));
  pulldown->add_item(kit->simple_menu_item("Save", style),
    new ActionCallback(ruiTextEdit)(this, &ruiTextEdit::save));
  menuBar->add_item(kit->simple_menubar_item("File", style), pulldown);

  pulldown = kit->pulldown(style);
  pulldown->add_item(kit->simple_menu_item("Cut", style),
    new ActionCallback(ruiText)(text_, &ruiText::cut));
  pulldown->add_item(kit->simple_menu_item("Copy", style),
    new ActionCallback(ruiText)(text_, &ruiText::copy));
  pulldown->add_item(kit->simple_menu_item("Paste", style),
    new ActionCallback(ruiText)(text_, &ruiText::paste));
  pulldown->add_item(kit->simple_menu_item("Replace", style),
    new ActionCallback(ruiText)(text_, &ruiText::replaceSelection));
  pulldown->add_item(kit->simple_menu_item("Search", style),
    new ActionCallback(ruiTextEdit)(this, &ruiTextEdit::search));
  pulldown->add_item(kit->simple_menu_item("Search and Replace", style),
    new ActionCallback(ruiTextEdit)(this, &ruiTextEdit::searchAndReplace));
  menuBar->add_item(kit->simple_menubar_item("Edit", style), pulldown);

  pulldown = kit->pulldown(style);
  pulldown->add_item(kit->simple_menu_item("fixed", style),
    new ruiTextEditFont(this, "fixed"));
  pulldown->add_item(kit->simple_menu_item("variable", style),
    new ruiTextEditFont(this, "variable"));
  pulldown->add_item(kit->simple_menu_item("lucidasans", style),
    new ruiTextEditFont(this, "lucidasans"));
  pulldown->add_item(kit->simple_menu_item("12x24", style),
    new ruiTextEditFont(this, "12x24"));
  menuBar->add_item(kit->simple_menubar_item("Font", style), pulldown);

  Color* color;
  if (style->value_is_on("monochrome")) {
    color = new Color(*style->background());
  } else {
    color = new Color(.96, .98, 1, 1);
  }
  Glyph* text = new ruiHorzBox(kit->inset_frame(new Background(text_, color),
    style), kit->inset_frame(new FixedSpan(new ruiPanner(text_, text_, style,
    (ruiPannerDirection) (ruiPanner::leftToRight | ruiPanner::topToBottom)),
    Dimension_X, 75), style));

  ruiKeyMap* keyMap = new ruiKeyMap(new ruiVertBox(new ruiHorzBox(
    new Center(menuBar, Dimension_Y, 0), new Center(legend_, Dimension_Y, 0)),
    text));

  keyMap->add(new ActionCallback(ruiTextEdit)(this, &ruiTextEdit::save),
    char(8));

  keyMap->add(new ActionCallback(ruiText)(text_, &ruiText::cut), char(104));
  keyMap->add(new ActionCallback(ruiText)(text_, &ruiText::copy), char(58));
  keyMap->add(new ActionCallback(ruiText)(text_, &ruiText::paste), char(80));
  keyMap->add(new ActionCallback(ruiText)(text_, &ruiText::replaceSelection),
    char(80), ruiKeyMap::shift);
  keyMap->add(new ActionCallback(ruiTextEdit)(this, &ruiTextEdit::search),
    char(102));
  keyMap->add(new ActionCallback(ruiTextEdit)(this, &ruiTextEdit::again),
    char(10));
  keyMap->add(new ActionCallback(ruiTextEdit)(this, &ruiTextEdit::
    searchAndReplace), char(102), ruiKeyMap::shift);

  keyMap->add(new ActionCallback(ruiTextEdit)(this, &ruiTextEdit::cursorUp),
    char(76));
  keyMap->add(new ActionCallback(ruiTextEdit)(this, &ruiTextEdit::cursorDown),
    char(120));
  keyMap->add(new ActionCallback(ruiTextEdit)(this, &ruiTextEdit::cursorLeft),
    char(98));
  keyMap->add(new ActionCallback(ruiTextEdit)(this, &ruiTextEdit::cursorRight),
    char(100));

  keyMap->add(new ActionCallback(ruiTextEdit)(this, &ruiTextEdit::scrollUp),
    char(76), ruiKeyMap::shift);
  keyMap->add(new ActionCallback(ruiTextEdit)(this, &ruiTextEdit::scrollDown),
    char(120), ruiKeyMap::shift);
  keyMap->add(new ActionCallback(ruiTextEdit)(this, &ruiTextEdit::scrollLeft),
    char(98), ruiKeyMap::shift);
  keyMap->add(new ActionCallback(ruiTextEdit)(this, &ruiTextEdit::scrollRight),
    char(100), ruiKeyMap::shift);

  keyMap->add(new ActionCallback(ruiTextEdit)(this, &ruiTextEdit::pageUp),
    char(77));
  keyMap->add(new ActionCallback(ruiTextEdit)(this, &ruiTextEdit::pageDown),
    char(121));

  keyMap->add(new ActionCallback(ruiTextEdit)(this, &ruiTextEdit::home),
    char(75));
  keyMap->add(new ActionCallback(ruiTextEdit)(this, &ruiTextEdit::end),
    char(119));

  body(keyMap);
}

ruiTextEdit::~ruiTextEdit() {
  delete text_;
}

void ruiTextEdit::load() {
  ruiSheet sheet;
  sheet.addItem("file name", text_->fileName());
  if (sheet.popup()) {
    rutString fileName;
    sheet.getValue(0, fileName);
    text_->load(fileName);
  }
}

void ruiTextEdit::save() {
  ruiSheet sheet;
  sheet.addItem("file name", text_->fileName());
  if (sheet.popup()) {
    rutString fileName;
    sheet.getValue(0, fileName);
    text_->fileName(fileName);
    text_->save();
  }
}

boolean ruiTextEdit::searchPattern(const rutString& pattern) {
  rutLexRegExp regExp(pattern);
  if (! regExp) {
    return false;
  }

  unsigned line;
  unsigned column;
  text_->getLocation(line, column);
  rutTextLocation location(line, column + 1);
  rutTextRegion region = text_->search(location, regExp);
  if (! region) {
    return false;
  }

  text_->location(region.line1(), region.column1());
  text_->region(region.line1(), region.column1(), region.line2(),
    region.column2());

  return true;
}

void ruiTextEdit::search() {
  ruiSheet sheet;
  sheet.addItem("expression", pattern_);
  if (! sheet.popup()) {
    return;
  }

  sheet.getValue(0, pattern_);
  searchPattern(pattern_);
}

void ruiTextEdit::again() {
  searchPattern(pattern_);
}

void ruiTextEdit::searchAndReplace() {
  ruiSheet sheet;
  sheet.addItem("expression", pattern_);
  sheet.addItem("replacement", replacement_);
  if (! sheet.popup()) {
    return;
  }

  sheet.getValue(0, pattern_);
  sheet.getValue(1, replacement_);
  if (searchPattern(pattern_)) {
    text_->replaceWith(replacement_);
  }
}

void ruiTextEdit::cursorUp() {
  unsigned line;
  unsigned column;
  text_->getLocation(line, column);
  if (line > 0) {
    --line;
    text_->location(line, column);
  }
}

void ruiTextEdit::cursorDown() {
  unsigned line;
  unsigned column;
  text_->getLocation(line, column);
  ++line;
  text_->location(line, column);
}

void ruiTextEdit::cursorLeft() {
  unsigned line;
  unsigned column;
  text_->getLocation(line, column);
  if (column > 0) {
    --column;
    text_->location(line, column);
  }
}

void ruiTextEdit::cursorRight() {
  unsigned line;
  unsigned column;
  text_->getLocation(line, column);
  ++column;
  text_->location(line, column);
}

void ruiTextEdit::scrollUp() {
  text_->scroll_backward(Dimension_Y);
}

void ruiTextEdit::scrollDown() {
  text_->scroll_forward(Dimension_Y);
}

void ruiTextEdit::scrollLeft() {
  text_->scroll_backward(Dimension_X);
}

void ruiTextEdit::scrollRight() {
  text_->scroll_forward(Dimension_X);
}

void ruiTextEdit::pageUp() {
  text_->page_backward(Dimension_Y);
}

void ruiTextEdit::pageDown() {
  text_->page_forward(Dimension_Y);
}

void ruiTextEdit::home() {
  text_->scroll_to(Dimension_Y, 0);
}

void ruiTextEdit::end() {
  text_->scroll_to(Dimension_Y, text_->length(Dimension_Y) -
    text_->cur_length(Dimension_Y));
}

void ruiTextEdit::font(char* name) {
  text_->font(name);
}
