/*
 * Copyright (c) 1991 Redwood Design Automation
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the name of
 * Redwood Design Automation may not be used in any advertising or publicity
 * relating to the software without the specific, prior written permission of
 * Redwood Design Automation.
 *
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
 *
 * IN NO EVENT SHALL REDWOOD DESIGN AUTOMATION BE LIABLE FOR ANY SPECIAL,
 * INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND, OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER OR NOT
 * ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF LIABILITY,
 * ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
 * SOFTWARE.
 */

#include <string.h>

#include <iostream.h>
#include <iomanip.h>

#include <OS/math.h>
#include <InterViews/geometry.h>

#include <r.h>
#include <ruiTile.h>

////////////////////////////////////////////

static const Coord minWorthBothering = 1;

static boolean iszero (float x) {
  if (Math::equal(x,(float)0,(float).001)) return true;
  return false;
}

static float min (float a, float b) {
  if (a < b) return a;
  return b;
}

static Coord max (Coord a, Coord b) {
  if (a > b) return a;
  return b;
}

static Coord minmax (Coord xmin, Coord xmax, Coord x) {
  return max(xmin, min(xmax, x));
}

////////////////////////////////////////////

ruiTile::ruiTile(DimensionName dimension) : Layout() {
    dimension_ = dimension;
}

ruiTile::~ruiTile() { }

ruiTileInfo::ruiTileInfo() {
  natural = 0;
  min = 0;
  max = 0;
  current = 0;
  alignment = 0;
  allocated = false;
}

void ruiTileInfo::reset() {
  natural = 0;
  min = 0;
  max = 0;
  current = 0;
  alignment = 0;
  allocated = false;
}

void ruiTile::computeInfo(GlyphIndex count, const Requisition* request) {
  int i;
  rAssert(ainfo_.size() == count);
  for (i = 0; i < count; i++) {
    const Requirement& r = request[i].requirement(dimension_);
    if (r.defined()) {
      ainfo_[i]->natural = r.natural();
      ainfo_[i]->min = max(0, r.natural() - r.shrink());
      ainfo_[i]->max = r.natural() + r.stretch();
      ainfo_[i]->alignment = r.alignment();
    } else {
      ainfo_[i]->natural = 0; 
      ainfo_[i]->min = 0; 
      ainfo_[i]->max = 0; 
      ainfo_[i]->alignment = 0;
    } 
    if (!ainfo_[i]->allocated) {
      ainfo_[i]->current = ainfo_[i]->natural;
      ainfo_[i]->allocated = true;
    }
    if (ainfo_[i]->current < ainfo_[i]->min) {
      ainfo_[i]->current = ainfo_[i]->min;
    }
    if (ainfo_[i]->current > ainfo_[i]->max) {
      ainfo_[i]->current = ainfo_[i]->max;
    }
  }
}

void ruiTile::insert(int i) {
  ainfo_.insert(new ruiTileInfo(), i);
}

void ruiTile::insertLast() {
  ainfo_.insertLast(new ruiTileInfo());
}

void ruiTile::remove(int i) {
  ainfo_.remove(i);
}

void ruiTile::change(int i) {
  if (i >= ainfo_.size()) return;
  ainfo_[i]->reset();
}

void ruiTile::request(
    GlyphIndex count, const Requisition* request, Requisition& result
) {
    Coord natural = 0;
    Coord stretch = 0;
    Coord shrink = 0;
    for (unsigned long index = 0; index < count; ++index) {
        const Requirement& r = request[index].requirement(dimension_);
        if (r.defined()) {
            natural += r.natural();
            stretch += r.stretch();
            shrink += r.shrink();
        }
    }
    Requirement nr(natural, stretch, shrink, 0.0);
    result.require(dimension_, nr);
    requisition_ = result;
}

int ruiTile::computeChangeableCount(boolean growing) {
  int count = ainfo_.size();
  int n = 0;
  for (int i = 0; i < count; i++) {
    if ((growing && !iszero(ainfo_[i]->max - ainfo_[i]->current)) ||
        (!growing && !iszero(ainfo_[i]->min - ainfo_[i]->current))) {
      n++;
    }
  }
  return n;
}

Coord ruiTile::computeChangeableSize (boolean growing) {
  int count = ainfo_.size();
  Coord sum = 0;
  for (int i = 0; i < count; i++) {
    if ((growing && !iszero(ainfo_[i]->max - ainfo_[i]->current)) ||
        (!growing && !iszero(ainfo_[i]->min - ainfo_[i]->current))) {
      sum += ainfo_[i]->current;
    }
  }
  return sum;
}

void ruiTile::proportionalChange (Coord newSize) {
  int count = ainfo_.size();
  Coord oldSize;
  int index;
  while (true) {
    oldSize = computeCurrentSpan();
    if (iszero(oldSize - newSize)) return; //successfully 
    Coord oldChangeableSize = 
      computeChangeableSize(newSize > oldSize);
    if (iszero(oldChangeableSize)) break; //more strategies below
    Coord newChangeableSize = 
      newSize - (oldSize - oldChangeableSize);
    if (newChangeableSize < 0) {
      newChangeableSize = 0;
    }
    double f = newChangeableSize/oldChangeableSize;
    for (index = 0; index < count; index++) {
      ainfo_[index]->current = 
        minmax(ainfo_[index]->min, ainfo_[index]->max, 
               ainfo_[index]->current * f);
    } 
  }
  while (true) {
    int changeableCount = 
      computeChangeableCount(newSize > oldSize);
    if (!changeableCount) return; // give up
    Coord delta = 
      (newSize - oldSize)/changeableCount;
    for (index = 0; index < count; index++) {
      ainfo_[index]->current = 
        minmax(ainfo_[index]->min, ainfo_[index]->max, 
               ainfo_[index]->current + delta);
    } 
    Coord prevOldSize = oldSize;
    oldSize = computeCurrentSpan();
    if (iszero(oldSize - newSize)) return; // successfully
    if (iszero(oldSize - prevOldSize)) return; // give it up
  }
}

void ruiTile::allocate(
  const Allocation& given,
  GlyphIndex count, const Requisition* request, Allocation* result
) {
  int index;

  if (!count) return;
  computeInfo(count, request);
  Coord span = given.allotment(dimension_).span();
  proportionalChange(span);

  Coord p = given.allotment(dimension_).begin();

  for (index = 0; index < count; ++index) {
    Allotment a;
    a.span(ainfo_[index]->current);
    a.alignment(ainfo_[index]->alignment);
    a.origin(p + a.alignment()*a.span());
    p += ainfo_[index]->current;
    result[index].allot(dimension_, a);
  }
}

void ruiTile::moveBorder(int i, Coord delta) {
  ainfo_[i-1]->current += delta;
  ainfo_[i]->current -= delta;
}

boolean ruiTile::canGrow (int i) {
  if (ainfo_[i]->max - ainfo_[i]->current >= minWorthBothering) 
    return true;
  //if (!iszero(ainfo_[i]->max - ainfo_[i]->current)) return true;
  return false;
}

boolean ruiTile::canShrink (int i) {
  if (ainfo_[i]->current - ainfo_[i]->min >= minWorthBothering) 
    return true;
  //if (!iszero(ainfo_[i]->current - ainfo_[i]->min)) return true;
  return false;
}

void ruiTile::getSlacks (int i, 
    Coord& gl, Coord& gr, Coord& sl, Coord& sr) {
  gl = 0; gr = 0; sl = 0; sr = 0;
  int j;
  j = i-1;
  gl += ainfo_[j]->max - ainfo_[j]->current;
  sl += ainfo_[j]->current - ainfo_[j]->min;
  j = i; 
  gr += ainfo_[j]->max - ainfo_[j]->current;
  sr += ainfo_[j]->current - ainfo_[j]->min;
}

// flexible border between i-1 and ith elements
boolean ruiTile::flexibleBorder (int i) {
  if (canGrow(i-1) && canShrink(i)) return true;
  if (canShrink(i-1) && canGrow(i)) return true;
  return false;
}

Coord ruiTile::getBorderOffset (int i) { 
  Coord d = 0;
  for (int j = 0; j < i; j++) {
    d += ainfo_[j]->current;
  }
  return d;
}

Coord ruiTile::getComponentSize (int i) {
  return ainfo_[i]->current;
}

void ruiTile::setComponentSize (int i, Coord newSize) {
  Coord oldSize = ainfo_[i]->current;
  Coord d = newSize - oldSize;
  ainfo_[i]->current += d;
  ainfo_[i-1]->current -= d;
}

ruiTileReversed::ruiTileReversed(DimensionName dimension) :
  ruiTile(dimension) {}

Coord ruiTile::computeCurrentSpan() {
  Coord sum = 0;
  for (int i = 0; i < ainfo_.size(); i++) {
    sum += ainfo_[i]->current;
  }
  return sum;
}

void ruiTileReversed::allocate(
  const Allocation& given,
  GlyphIndex count, const Requisition* request, Allocation* result
) {
  int index;

  if (!count) return;
  computeInfo(count, request);
  Coord span = given.allotment(dimension_).span();
  proportionalChange(span);
    
  Coord p = given.allotment(dimension_).end();
 
  for (index = 0; index < count; ++index) {
    p -= ainfo_[index]->current;
    Allotment a;
    a.span(ainfo_[index]->current);
    a.alignment(ainfo_[index]->alignment);
    a.origin(p + a.alignment()*a.span());
    result[index].allot(dimension_, a);
  }
}
 
