/*
 * Copyright (c) 1991 Redwood Design Automation
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the name of
 * Redwood Design Automation may not be used in any advertising or publicity
 * relating to the software without the specific, prior written permission of
 * Redwood Design Automation.
 *
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
 *
 * IN NO EVENT SHALL REDWOOD DESIGN AUTOMATION BE LIABLE FOR ANY SPECIAL,
 * INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND, OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER OR NOT
 * ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF LIABILITY,
 * ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
 * SOFTWARE.
 */

#include <r.h>
#include <rutArray.h>

rutArrayRef::rutArrayRef(rutArray& array, int index) : array_(array),
  index_(index) {
}

#ifndef cplusplus_2_0
rutArray::
#endif
rutArrayRef& rutArrayRef::operator=(void* entry) {
  array_.put(entry, index_);
  return *this;
}

rutArrayRef::operator void*() const {
  return array_.get(index_);
}

void* rutArrayRef::operator->() const {
  return array_.get(index_);
}

rutArray::rutArray(int limit) {
  rPrecondition(limit >= 0);

  this->size_   = 0;
  this->limit_  = 0;
  this->buffer_ = 0;

  if (limit > 0) {
    this->buffer_ = new void*[limit];
    if (this->buffer_) { 
      this->limit_ = limit;
    }
  }

  invariants();
}

rutArray::rutArray(const rutArray& array) {
  this->size_   = 0;
  this->limit_  = 0;
  this->buffer_ = 0;

  if (array.size_ > 0) {
    this->buffer_ = new void*[array.size_];
    for (int i = 0; i < array.size_; ++i) {
      this->buffer_[i] = array.buffer_[i];
    }
    if (this->buffer_) { 
      this->limit_ = array.size_;
    }
  }

  invariants();
}

rutArray::~rutArray() {
  size_ = 0;
  if (buffer_) {
    delete buffer_;
    buffer_ = 0;
    limit_  = 0;
  }

  invariants();
}

#ifndef cplusplus_2_0
rutArray::
#endif
rutArrayRef rutArray::operator[](int index) {
  return rutArrayRef(*(rutArray*)this, index);
}

int rutArray::size() const {
  return size_;
}

int rutArray::equal(rutArray &array) const {
  int i;

  if (size_ != array.size_) {
    return 0;
  }
  for (i = 0; i < size_; i++) {
    if (buffer_[i] != array.buffer_[i]) {
      return 0;
    }
  }
  return 1;
}

int rutArray::indexOf(void* entry) const {
  int i;

  for (i = 0; i < size_; i++) {
    if (buffer_[i] == entry) {
      return i;
    }
  }

  return -1;
}

void* rutArray::get(int index) const {
  rPrecondition((0 <= index) && (index < size_));

  if ((0 <= index) && (index < size_)) {
    return (void*) buffer_[index];
  }

  invariants();

  return (void*) 0;
}

void* rutArray::getLast() const {
  return get(size_ - 1);
}

void rutArray::put(void* entry, int index) {
  rPrecondition(0 <= index);

  if (index >= 0) {
    newUpperBound(index + 1);
    buffer_[index] = entry;
    if (index >= size_) {
      size_ = index + 1;
    }
  }

  invariants();
}

void rutArray::insert(void* entry, int index) {
  rPrecondition((0 <= index) && (index <= size_));

  newUpperBound(size_ + 1);

  int i;
  for (i = size_ - 1; i >= index; i--) {
    buffer_[i + 1] = buffer_[i];
  }
  buffer_[index] = entry;
  ++size_;

  invariants();
}

void rutArray::insertLast(void* entry) {
  put(entry, size_);
}

void* rutArray::remove(int index) {
  void* entry;

  rPrecondition((0 <= index) && (index < size_));

  entry = 0;
  if ((0 <= index) && (index < size_)) {  
    int i;

    entry = buffer_[index];
    --size_;
    for (i = index; i < size_; i++) {
      buffer_[i] = buffer_[i + 1];
    }
  }

  invariants();

  return entry;
}

void* rutArray::removeLast() {
  return remove(size_ - 1);
}

int rutArray::removeEntry(void* entry) {
  int index;

  index = indexOf(entry);
  if (index >= 0) {
    remove(index);
  }
  return index;
}

void rutArray::reverse() {
  rutArray array(size_);
  int i;

  for (i = size_ - 1; i >= 0; --i) {
    array.insertLast(get(i));
  }
  if (buffer_) {
    delete buffer_;
  }
  buffer_ = array.buffer_;
  limit_ = array.limit_;
  array.buffer_ = 0;
  array.limit_ = 0;
}

void rutArray::reset() {
  size_ = 0;
}

void rutArray::invariants() const {
  rInvariant(size_ >= 0);
  rInvariant(limit_ >= 0);
  rInvariant(size_ <= limit_);
  rInvariant((limit_ == 0) == (buffer_ == 0));
}

void rutArray::newUpperBound(int limit) {
  if (limit > this->limit_) {
    void**oldbuffer_;
    int i;

    if (limit < (this->limit_ * 2)) {
      limit = this->limit_ * 2;
    }
    oldbuffer_ = this->buffer_;
    this->buffer_ = new void*[limit];
    for (i = 0; i < this->size_; i++) {
      this->buffer_[i] = oldbuffer_[i];
    }
    this->limit_ = limit;
    delete oldbuffer_;
  }
}
