/*
 * Copyright (c) 1991 Redwood Design Automation
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the name of
 * Redwood Design Automation may not be used in any advertising or publicity
 * relating to the software without the specific, prior written permission of
 * Redwood Design Automation.
 *
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
 *
 * IN NO EVENT SHALL REDWOOD DESIGN AUTOMATION BE LIABLE FOR ANY SPECIAL,
 * INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND, OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER OR NOT
 * ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF LIABILITY,
 * ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
 * SOFTWARE.
 */

#ifndef rutBase_h
#define rutBase_h

#include <generic.h>

/////////////
// rutBase //
/////////////

// Example of how to use rutBase code generating macros - lets assume
// that you are creating a new class called rutThing that inherits from
// rutOther:
//
// rutThing.h:
//
// class rutThing : public rutOther {
// public:
//   rutBaseCommonDefine(rutThing) // notice that there is NOT a ';'
// };
//
// rutThing.c:
//
// rutBaseCommonBody(rutThing, rutOther) // notice that there is NOT a ';'

class rutBaseDescriptor {
public:
  virtual char* name() = 0;
  virtual unsigned long size() = 0;

  int total();
  int count();

  void increment();
  void decrement();

  static rutBaseDescriptor* head();
  rutBaseDescriptor* next();

protected:
  rutBaseDescriptor();

private:
  virtual ~rutBaseDescriptor();

  friend rutBase;
  long int total_;
  long int count_;

  static rutBaseDescriptor* head_;
  rutBaseDescriptor* next_;
};

class ostream;

class rutBase {
public:
  virtual ~rutBase();

  virtual rutBaseDescriptor *descriptor() const = 0;
  virtual void debugPrint(ostream& out) const;

  // Clients should not access this member function.
  virtual void *downcastVirtual(const rutBaseDescriptor *descriptor) const;

protected:
  rutBase();
};

// Declares all static members/methods.
//
#define rutBaseCommonDefine(className) \
private: \
static rutBaseDescriptor* descriptor_; \
void* downcastVirtual(const rutBaseDescriptor* descriptor) const; \
public: \
static rutBaseDescriptor* descriptorStatic() const; \
virtual rutBaseDescriptor* descriptor() const; \
static className* downcast(const rutBase* base) const; \
static className* isA(const rutBase* base) const;

// Declares descriptor, the static member and defines the body for the
// descriptor static function. Used by the either the rutBaseCommonBody,
// rutBaseCommonBodyMultiple if the class for which the common body is
// being defined is multiply inherited, then rutBaseCommonBodyMultiple
// should be used, currently it handles only two base classes.  If you
// need more create another macro along the same lines.
//
#define rutBaseDescriptorBody(className) \
class name2(rutBaseDescriptor,className) : public rutBaseDescriptor { \
public: \
  name2(rutBaseDescriptor,className)() {} \
\
  char* name() {return "className";} \
  unsigned long size() {return sizeof(className);} \
}; \
\
rutBaseDescriptor* className::descriptor_ = \
  new name2(rutBaseDescriptor,className); \
\
rutBaseDescriptor* className::descriptorStatic() const { \
  return descriptor_; \
} \
\
rutBaseDescriptor* className::descriptor() const { \
  return descriptor_; \
}

// Defines the body for the static downcast function.
//
#define rutBaseDowncastBody(className) \
className* className::downcast(const rutBase* base) const { \
  if (base) { \
    className* object = (className*) base->downcastVirtual(descriptor_); \
    if (! object) { \
      rFail("nonsensical downcast (" << base->descriptor()->name() << \
        " to className)"); \
    } \
    return object; \
  } \
  return 0; \
}

// Defines the body for the static isA function.
//
#define rutBaseIsABody(className) \
className* className::isA(const rutBase* base) const { \
  if (base) { \
    return (className*) base->downcastVirtual(descriptor_); \
  } \
  return 0; \
}

// Defines body for downcastVirtual for single inheritance case.
//
#define rutBaseDowncastVirtualBody(className,baseClass) \
void* className::downcastVirtual(const rutBaseDescriptor* desc) const { \
  if (descriptor_ == desc) { \
     return (void*) this; \
  } \
  return baseClass::downcastVirtual(desc); \
}

// Defines body for downcastVirtual for multiple inheritance case,
// with two base classes.
//
#define rutBaseDowncastVirtual2MultipleBody(className,baseClass1,baseClass2) \
void* className::downcastVirtual(const rutBaseDescriptor* desc) const { \
  void* objToBeReturned =0; \
  if (descriptor_ == desc) { \
     return (void*) this; \
  } \
  if (objToBeReturned = baseClass1::downcastVirtual(desc)) { \
     return objToBeReturned; \
  } \
  return baseClass2::downcastVirtual(desc); \
}

#define rutBaseCommonBody(className,baseClass) \
rutBaseDowncastBody(className) \
rutBaseIsABody(className) \
rutBaseDescriptorBody(className) \
rutBaseDowncastVirtualBody(className,baseClass)

#define rutBaseCommonBodyMultiple(className,baseClass1,baseClass2) \
rutBaseDowncastBody(className) \
rutBaseIsABody(className) \
rutBaseDescriptorBody(className) \
rutBaseDowncastVirtual2MultipleBody(className,baseClass1,baseClass2)

#endif
