/*
 * Copyright (c) 1991 Redwood Design Automation
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the name of
 * Redwood Design Automation may not be used in any advertising or publicity
 * relating to the software without the specific, prior written permission of
 * Redwood Design Automation.
 *
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
 *
 * IN NO EVENT SHALL REDWOOD DESIGN AUTOMATION BE LIABLE FOR ANY SPECIAL,
 * INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND, OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER OR NOT
 * ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF LIABILITY,
 * ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
 * SOFTWARE.
 */

#include <string.h>
#include <ctype.h>

#include <iostream.h>

#include <rutLex.h>
#include <rutLexParse.h>
#include <rutString.h>

//////////////////
// Parse String //
//////////////////

class rutLexParseString : public rutLexParse {
public:
  rutLexParseString();
  virtual ~rutLexParseString();

  rutLexRegExpNode* parse(rutString string);

protected:
  void yyerror(char* error);
  int yylex();

  int index_;
  rutString string_;
};

rutLexParseString::rutLexParseString() {
}

rutLexParseString::~rutLexParseString() {
}

rutLexRegExpNode* rutLexParseString::parse(rutString string) {
  index_ = 0;
  string_ = string;
  if (yyparse()) { // aborted
    return 0;
  }
  return expression_;
}

void rutLexParseString::yyerror(char*) {
}

int rutLexParseString::yylex() {
  if (index_ >= string_.length()) {
    return 0;
  }
  
  char ch = string_[index_++];

  if (isalpha(ch)) {
    yylval.character = ch;
    return alpha;
  }

  if (isdigit(ch)) {
    yylval.character = ch;
    return digit;
  }

  switch (ch) {
    case '<': return openTag;
    case '>': return closeTag;
    case '{': return openClosure;
    case ',': return closureRange;
    case '}': return closeClosure;
    case '*': return zeroOrMore;
    case '+': return oneOrMore;
    case '?': return zeroOrOne;
    case '|': return eUnion;
    case '(': return openGroup;
    case ')': return closeGroup;
    case '[': return openSet;
    case '^': return notSet;
    case '-': return setRange;
    case ']': return closeSet;
    case '.': return any;
  }

  if ((ch == '\\') && (index_ < string_.length())) {
    switch (string_[index_]) {
      case 'b': {
        ch = '\b';
      } break;
      case 'f': {
        ch = '\f';
      } break;
      case 'n': {
        ch = '\n';
      } break;
      case 'r': {
        ch = '\r';
      } break;
      case 't': {
        ch = '\t';
      } break;
      case 'v': {
        ch = '\v';
      } break;
      case '0': {
        ch = '\0';
      } break;
      default: {
        ch = string_[index_++];
      } break;
    }
  }

  yylval.character = ch;
  return other;
}

/////////
// Set //
/////////

rutLexSet::rutLexSet(unsigned size) {
  initialize(size);
}

rutLexSet::rutLexSet(char ch) {
  initialize(256);
  add(ch);
}

rutLexSet::rutLexSet(const rutLexSet& set) {
  initialize(set.size_);
  copy(set);
}

rutLexSet& rutLexSet::operator=(const rutLexSet& set) {
  copy(set);
  return *this;
}

unsigned rutLexSet::cardinality() const {
  return size_;
}

void rutLexSet::add(unsigned ch) {
  bits_[ch / 32] |= mask_[ch % 32];
}

void rutLexSet::add(const rutLexSet* set) {
  unsigned i;
  for (i = 0; i < bound_; ++i) {
    bits_[i] |= set->bits_[i];
  }
}

void rutLexSet::remove(unsigned ch) {
  bits_[ch / 32] &= ~mask_[ch % 32];
}

int rutLexSet::contains(unsigned ch) const {
  return bits_[ch / 32] & mask_[ch % 32];
}

void rutLexSet::invert() {
  unsigned i;
  for (i = 0; i < bound_; ++i) {
    bits_[i] = ~bits_[i];
  }
}

void rutLexSet::add(unsigned lower, unsigned upper) {
  unsigned i;
  for (i = lower; i <= upper; ++i) {
    add(i);
  }
}

int rutLexSet::operator==(const rutLexSet& set) const {
  unsigned i;
  for (i = 0; i < bound_; ++i) {
    if (bits_[i] != set.bits_[i]) {
      return 0;
    }
  }
  return 1;
}

void rutLexSet::initialize(unsigned size) {
  size_ = size;
  bound_ = (size_ + 31) / 32;
  bits_ = new unsigned[bound_];
  unsigned i;
  for (i = 0; i < bound_; ++i) {
    bits_[i] = 0;
  }
}

void rutLexSet::copy(const rutLexSet& set) {
  unsigned i;
  for (i = 0; i < bound_; ++i) {
    bits_[i] = set.bits_[i];
  }
}

ostream& operator<<(ostream& out, const rutLexSet& set) {
  unsigned i;
  for (i = 0; i < set.size_; ++i) {
    if (set.contains(i)) {
      out << " " << i;
    }
  }
  out << endl;
  return out;
}

unsigned rutLexSet::mask_[32] = {
  1u <<  0, 1u <<  1, 1u <<  2, 1u <<  3,
  1u <<  4, 1u <<  5, 1u <<  6, 1u <<  7,
  1u <<  8, 1u <<  9, 1u << 10, 1u << 11,
  1u << 12, 1u << 13, 1u << 14, 1u << 15,
  1u << 16, 1u << 17, 1u << 18, 1u << 19,
  1u << 20, 1u << 21, 1u << 22, 1u << 23,
  1u << 24, 1u << 25, 1u << 26, 1u << 27,
  1u << 28, 1u << 29, 1u << 30, 0x80000000 // why is "1u << 31" signed?
};

///////////////
// Automaton //
///////////////

rutLexAutomaton::rutLexAutomaton() {
  uid_ = 0;

  size_ = 0;
  limit_ = 0;
  states_ = 0;
}

rutLexAutomaton::~rutLexAutomaton() {
  if (states_) {
    delete states_;
  }
}

void rutLexAutomaton::start(rutLexState* state) {
  start_ = state;
}

rutLexState* rutLexAutomaton::start() const {
  return start_;
}

rutLexAutomaton* rutLexAutomaton::DFA() {
  rutLexAutomaton* dfa = new rutLexAutomaton;
  rutLexStateSetSet stateSetSet(dfa);
  rutLexStateSetList unexplored;
  rutLexStateSet* startSet = new rutLexStateSet(size_);
  rutLexState* startState;

  startSet->add(start_);
  startSet->eClosure();
  startState = stateSetSet.add(startSet);
  dfa->start(startState);
  unexplored.add(startSet, startState);
  while (unexplored) {
    rutLexStateSet* stateSet;
    rutLexState* state;
    int ch;

    unexplored.remove(stateSet, state);
    rutLexSet* set = stateSet->transitions();
    for (ch = 0; ch <= 126; ++ch) {
      if (set->contains(ch)) {
	rutLexStateSet* toSet;
	rutLexState* toState;

	toSet = stateSet->move(ch);
	toSet->eClosure();
	toState = stateSetSet.contains(toSet);
	if (! toState) {
	  toState = stateSetSet.add(toSet);
	  unexplored.add(toSet, toState);
	} else {
	  delete toSet;
	}
	state->addTransition(toState, ch);
      }
    }
    delete set;
  }

  return dfa;
}

void rutLexAutomaton::addState(rutLexState* state) {
  state->uid_ = uid_++;

  newUpperBound(size_ + 1);
  states_[size_] = state;
  ++size_;
}

ostream& operator<<(ostream& out, const rutLexAutomaton& automaton) {
  for (int i = 0; i < automaton.size_; ++i) {
    out << *automaton.states_[i];
  }
  return out;
}

void rutLexAutomaton::newUpperBound(unsigned limit) {
  rutLexState** oldStates_;
  int i;
 
  if (limit <= limit_) {
    return;
  }

  if (limit < (limit_ * 2)) {
    limit = limit_ * 2;
  }
  oldStates_ = states_;
  states_ = new rutLexState*[limit];
  for (i = 0; i < size_; i++) {
    states_[i] = oldStates_[i];
  }
  for (i = size_; i < limit; i++) {
    states_[i] = 0;
  }
  limit_ = limit;
  delete oldStates_;
}

////////////////////////
// Automaton Iterator //
////////////////////////

rutLexAutomatonIterator::rutLexAutomatonIterator(rutLexAutomaton* automaton) {
  automaton_ = automaton;
  reset();
}

rutLexAutomatonIterator::~rutLexAutomatonIterator() {};

void rutLexAutomatonIterator::reset() {
  state_ = automaton_->start();
}

int rutLexAutomatonIterator::move(char ch) {
  state_ = state_->move(ch);
  return state_ != 0;
}

int rutLexAutomatonIterator::active() const {
  return state_ != 0;
}

int rutLexAutomatonIterator::accept() const {
  return state_ ? state_->accept() : 0;
}

///////////
// State //
///////////

rutLexState::rutLexState(rutLexAutomaton* automaton) {
  size_ = 0;
  limit_ = 0;
  sets_ = 0;
  states_ = 0;

  accept_ = 0;

  set_ = 0;

  automaton->addState(this);
}

rutLexState::~rutLexState() {
  int i;

  if (! size_) {
    return;
  }

  for (i = 0; i < size_; ++i) {
    if (sets_[i]) {
      delete sets_[i];
    }
  }
  delete sets_;
  delete states_;
}

void rutLexState::accept(int accept) {
  accept_ = accept;
}

int rutLexState::accept() const {
  return accept_;
}

rutLexState* rutLexState::move(char ch) const {
  int i;

  for (i = 0; i < size_; ++i) {
    if (sets_[i]->contains(ch)) {
      return states_[i];
    }
  }
  return 0;
}

void rutLexState::addEdge(rutLexState* to) {
  newUpperBound(size_ + 1);

  sets_[size_] = 0;
  states_[size_] = to;
  ++size_;
}

void rutLexState::addEdge(rutLexState* to, char ch) {
  newUpperBound(size_ + 1);

  sets_[size_] = new rutLexSet(ch);
  states_[size_] = to;
  ++size_;
}

void rutLexState::addEdge(rutLexState* to, rutLexSet set) {
  newUpperBound(size_ + 1);

  sets_[size_] = new rutLexSet(set);
  states_[size_] = to;
  ++size_;
}

void rutLexState::addTransition(rutLexState* to, char ch) {
  int i = findEdge(to);

  if (i < 0) {
    addEdge(to, ch);
  } else {
    sets_[i]->add(ch);
  }
}

rutLexSet& rutLexState::transitions() {
  if (! set_) {
    set_ = new rutLexSet;
    int i;

    for (i = 0; i < size_; ++i) {
      if (sets_[i]) {
        set_->add(sets_[i]);
      }
    }
  }

  return *set_;
}

ostream& operator<<(ostream& out, const rutLexState& state) {
  int i;

  out << "state " << state.uid_ << "\n";
  for (i = 0; i < state.size_; ++i) {
    rutLexState* to = state.states_[i];
    rutLexSet* set = state.sets_[i];
    int ch;

    out << " edge to " << to->uid_;
    if (set) {
      out << " on";
      for (ch = 0; ch <= 126; ++ch) {
        if (set->contains(ch)) {
          out << " " << char(ch);
	}
      }
    }
    out << "\n";
  }
  return out;
}

int rutLexState::findEdge(const rutLexState* to) const {
  int i;

  for (i = 0; i < size_; ++i) {
    if (states_[i] == to) {
      return i;
    }
  }

  return -1;
}

void rutLexState::newUpperBound(unsigned limit) {
  rutLexSet** oldSets_;
  rutLexState** oldStates_;
  int i;
 
  if (limit <= limit_) {
    return;
  }

  if (limit < (limit_ * 2)) {
    limit = limit_ * 2;
  }
  oldSets_ = sets_;
  oldStates_ = states_;
  sets_ = new rutLexSet*[limit];
  states_ = new rutLexState*[limit];
  for (i = 0; i < size_; i++) {
    sets_[i] = oldSets_[i];
    states_[i] = oldStates_[i];
  }
  for (i = size_; i < limit; i++) {
    sets_[i] = 0;
    states_[i] = 0;
  }
  limit_ = limit;
  delete oldSets_;
  delete oldStates_;
}

////////////////
// State List //
////////////////

rutLexStateList::rutLexStateList() {
  size_ = 0;
  limit_ = 0;
  states_ = 0;
}

rutLexStateList::~rutLexStateList() {
  if (states_) {
    delete states_;
  }
}

void rutLexStateList::add(rutLexState* state) {
  newUpperBound(size_ + 1);
  states_[size_++] = state;
}

rutLexState* rutLexStateList::remove() {
  return states_[--size_];
}

rutLexStateList::operator int() const {
  return size_;
}

void rutLexStateList::newUpperBound(unsigned limit) {
  rutLexState** oldStates_;
  int i;
 
  if (limit <= limit_) {
    return;
  }

  if (limit < (limit_ * 2)) {
    limit = limit_ * 2;
  }
  oldStates_ = states_;
  states_ = new rutLexState*[limit];
  for (i = 0; i < size_; i++) {
    states_[i] = oldStates_[i];
  }
  for (i = size_; i < limit; i++) {
    states_[i] = 0;
  }
  limit_ = limit;
  delete oldStates_;
}

///////////////
// State Set //
///////////////

rutLexStateSet::rutLexStateSet(unsigned size) {
  set_ = new rutLexSet(size);
}

rutLexStateSet::~rutLexStateSet() {
  delete set_;
}

int rutLexStateSet::contains(const rutLexState* state) const {
  return set_->contains(state->uid_);
}

void rutLexStateSet::add(rutLexState* state) {
  set_->add(state->uid_);
  rutLexStateList::add(state);
}

int rutLexStateSet::operator==(const rutLexStateSet& stateSet) const {
  return *set_ == *stateSet.set_;
}

rutLexSet* rutLexStateSet::transitions() const {
  rutLexSet* set = new rutLexSet;
  int i;

  for (i = 0; i < size_; ++i) {
    rutLexState* state = states_[i];
    set->add(&state->transitions());
  }

  return set;
}

void rutLexStateSet::eClosure() {
  rutLexStateList unexplored;
  int i;

  for (i = 0; i < size_; ++i) {
    unexplored.add(states_[i]);
  }
  while (unexplored) {
    rutLexState* state = unexplored.remove();

    for (i = 0; i < state->size_; ++i) {
      if (! state->sets_[i]) {
	rutLexState* s = state->states_[i];
        if (! contains(s)) {
          add(s);
	  unexplored.add(s);
	}
      }
    }
  }
}

rutLexStateSet* rutLexStateSet::move(char ch) const {
  rutLexStateSet* stateSet = new rutLexStateSet(set_->cardinality());
  int i;

  for (i = 0; i < size_; ++i) {
    rutLexState* state = states_[i];
    int j;

    for (j = 0; j < state->size_; ++j) {
      if (state->sets_[j] && state->sets_[j]->contains(ch)) {
	rutLexState* to = state->states_[j];
	if (! stateSet->contains(to)) {
	  stateSet->add(to);
	}
      }
    }
  }

  return stateSet;
}

int rutLexStateSet::accept() const {
  int i;

  for (i = 0; i < size_; ++i) {
    rutLexState* state = states_[i];
    if (state->accept()) {
      return state->accept();
    }
  }
  return 0;
}

////////////////////
// State Set List //
////////////////////

rutLexStateSetList::rutLexStateSetList() {
  size_ = 0;
  limit_ = 0;
  stateSets_ = 0;
  states_ = 0;
}

rutLexStateSetList::~rutLexStateSetList() {
  if (stateSets_) {
    delete stateSets_;
  }
  if (states_) {
    delete states_;
  }
}

void rutLexStateSetList::add(rutLexStateSet* stateSet, rutLexState* state) {
  newUpperBound(size_ + 1);
  stateSets_[size_] = stateSet;
  states_[size_] = state;
  ++size_;
}

void rutLexStateSetList::remove(rutLexStateSet*& stateSet,
  rutLexState*& state) {
  --size_;
  stateSet = stateSets_[size_];
  state = states_[size_];
}

rutLexStateSetList::operator int() const {
  return size_;
}

void rutLexStateSetList::newUpperBound(unsigned limit) {
  rutLexStateSet** oldStateSets_;
  rutLexState** oldStates_;
  int i;
 
  if (limit <= limit_) {
    return;
  }

  if (limit < (limit_ * 2)) {
    limit = limit_ * 2;
  }
  oldStateSets_ = stateSets_;
  oldStates_ = states_;
  stateSets_ = new rutLexStateSet*[limit];
  states_ = new rutLexState*[limit];
  for (i = 0; i < size_; i++) {
    stateSets_[i] = oldStateSets_[i];
    states_[i] = oldStates_[i];
  }
  for (i = size_; i < limit; i++) {
    stateSets_[i] = 0;
    states_[i] = 0;
  }
  limit_ = limit;
  delete oldStateSets_;
  delete oldStates_;
}

///////////////////
// State Set Set //
///////////////////

rutLexStateSetSet::rutLexStateSetSet(rutLexAutomaton* automaton) {
  automaton_ = automaton;
}

rutLexStateSetSet::~rutLexStateSetSet() {
  int i;
  for (i = 0; i < size_; ++i) {
    delete stateSets_[i];
  }
}

rutLexState* rutLexStateSetSet::add(rutLexStateSet* stateSet) {
  rutLexState* state = new rutLexState(automaton_);
  state->accept(stateSet->accept());
  rutLexStateSetList::add(stateSet, state);
  return state;
}

rutLexState* rutLexStateSetSet::contains(const rutLexStateSet* stateSet)
  const {
  int i;

  for (i = 0; i < size_; ++i) {
    rutLexStateSet* s = stateSets_[i];
    if (*s == *stateSet) {
      return states_[i];
    }
  }
  return 0;
}

////////////////////////
// Regular Expression //
////////////////////////

rutLexRegExp::rutLexRegExp(const rutString& expression) {
  rutLexRegExpNode* node = parse(expression);
  if (node) {
    dfa_ = DFA(node);
  }
}

rutLexRegExp::rutLexRegExp(const rutLexRegExpNode* node) {
  dfa_ = DFA((rutLexRegExpNode*) node);
}

rutLexRegExp::~rutLexRegExp() {
  delete dfa_;
}

rutLexRegExpNode* rutLexRegExp::parse(const rutString& expression) {
  rutLexParseString parser;
  return parser.parse(expression);
}

rutLexRegExp::operator int() {
  return dfa_ ? 1 : 0;
}

rutLexAutomaton* rutLexRegExp::NFA(rutLexRegExpNode* node) const {
  rutLexAutomaton* nfa = new rutLexAutomaton;
  rutLexState* start = new rutLexState(nfa);
  rutLexState* accept = new rutLexState(nfa);

  nfa->start(start);
  accept->accept(1);
  node->constructNFA(nfa, start, accept);

  return nfa;
}

rutLexAutomaton* rutLexRegExp::DFA(rutLexRegExpNode* node) const {
  rutLexAutomaton* nfa = NFA(node);
  rutLexAutomaton* dfa = nfa->DFA();
  delete nfa;
  return dfa;
}

/////////////////////////////////
// Regular Expression Iterator //
/////////////////////////////////

rutLexRegExpIterator::rutLexRegExpIterator(const rutLexRegExp* regExp) :
  rutLexAutomatonIterator(regExp->dfa_) {
}

rutLexRegExpIterator::~rutLexRegExpIterator() {
}

////////////////////
// Character Node //
////////////////////

rutLexCharNode::rutLexCharNode(char ch) {
  ch_ = ch;
}

void rutLexCharNode::constructNFA(rutLexAutomaton*, rutLexState* start,
  rutLexState* accept) {
  start->addEdge(accept, ch_);
}

/////////////////
// String Node //
/////////////////

rutLexStringNode::rutLexStringNode(char* string) {
  string_ = new char[strlen(string) + 1];
  strcpy(string_, string);
}

void rutLexStringNode::constructNFA(rutLexAutomaton* automaton,
  rutLexState* start, rutLexState* accept) {
  rutLexState* state;
  rutLexState* next;
  int length;
  int i;

  length = strlen(string_);
//  rPrecondition(length > 0);
  for (i = 0, state = start; i < length; ++i, state = next) {
    next = (i < length - 1 ? new rutLexState(automaton) : accept);
    state->addEdge(next, string_[i]);
  }
}

//////////////////////////
// Character Class Node //
//////////////////////////

rutLexCharClassNode::rutLexCharClassNode(rutLexSet set) {
  set_ = set;
}

void rutLexCharClassNode::constructNFA(rutLexAutomaton*, rutLexState* start,
  rutLexState* accept) {
  start->addEdge(accept, set_);
}

////////////////////////
// Concatination Node //
////////////////////////

rutLexConcatNode::rutLexConcatNode(rutLexRegExpNode* expA,
  rutLexRegExpNode* expB) {
  expA_ = expA;
  expB_ = expB;
}

void rutLexConcatNode::constructNFA(rutLexAutomaton* automaton,
  rutLexState* start, rutLexState* accept) {
  rutLexState* state = new rutLexState(automaton);

  expA_->constructNFA(automaton, start, state);
  expB_->constructNFA(automaton, state, accept);
}

////////////////
// Union Node //
////////////////

rutLexUnionNode::rutLexUnionNode(rutLexRegExpNode* expA,
  rutLexRegExpNode* expB) {
  expA_ = expA;
  expB_ = expB;
}

void rutLexUnionNode::constructNFA(rutLexAutomaton* automaton,
  rutLexState* start, rutLexState* accept) {
  rutLexState* startA = new rutLexState(automaton);
  rutLexState* acceptA = new rutLexState(automaton);
  rutLexState* startB = new rutLexState(automaton);
  rutLexState* acceptB = new rutLexState(automaton);

  expA_->constructNFA(automaton, startA, acceptA);
  expB_->constructNFA(automaton, startB, acceptB);

  start->addEdge(startA);
  start->addEdge(startB);
  acceptA->addEdge(accept);
  acceptB->addEdge(accept);
}

////////////////////////
// Exact Closure Node //
////////////////////////

rutLexExactClosureNode::rutLexExactClosureNode(rutLexRegExpNode* exp, int m) {
  exp_ = exp;
  m_ = m;
}

void rutLexExactClosureNode::constructNFA(rutLexAutomaton* automaton,
  rutLexState* start, rutLexState* accept) {
  rutLexState* state;
  rutLexState* next;
  int i;

  for (i = 0, state = start; i < m_; ++i, state = next) {
    next = (i < m_ - 1 ? new rutLexState(automaton) : accept);
    exp_->constructNFA(automaton, state, next);
  }
}

///////////////////////
// Open Closure Node //
///////////////////////

rutLexOpenClosureNode::rutLexOpenClosureNode(rutLexRegExpNode* exp, int m) :
  rutLexExactClosureNode(exp, m) {
}

void rutLexOpenClosureNode::constructNFA(rutLexAutomaton* automaton,
  rutLexState* start, rutLexState* accept) {
  rutLexState* state;
  rutLexState* expStart = new rutLexState(automaton);
  rutLexState* expAccept = new rutLexState(automaton);

  if (m_ > 0) {
    state = new rutLexState(automaton);
    rutLexExactClosureNode::constructNFA(automaton, start, state);
  } else {
    state = start;
  }
  exp_->constructNFA(automaton, expStart, expAccept);
  state->addEdge(expStart);
  state->addEdge(accept);
  expAccept->addEdge(expStart);
  expAccept->addEdge(accept);
}

////////////////////////
// Range Closure Node //
////////////////////////

rutLexRangeClosureNode::rutLexRangeClosureNode(rutLexRegExpNode* exp, int m,
  int n) {
  exp_ = exp;
  m_ = m;
  n_ = n;
}

void rutLexRangeClosureNode::constructNFA(rutLexAutomaton* automaton,
  rutLexState* start, rutLexState* accept) {
}

//////////////
// Tag Node //
//////////////

rutLexTagNode::rutLexTagNode(rutLexRegExpNode* exp) {
  exp_ = exp;
}

void rutLexTagNode::constructNFA(rutLexAutomaton* automaton,
  rutLexState* start, rutLexState* accept) {
  exp_->constructNFA(automaton, start, accept);
}
