/*
 * Copyright (c) 1991 Redwood Design Automation
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the name of
 * Redwood Design Automation may not be used in any advertising or publicity
 * relating to the software without the specific, prior written permission of
 * Redwood Design Automation.
 *
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
 *
 * IN NO EVENT SHALL REDWOOD DESIGN AUTOMATION BE LIABLE FOR ANY SPECIAL,
 * INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND, OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER OR NOT
 * ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF LIABILITY,
 * ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
 * SOFTWARE.
 */

#ifndef rutLex_h
#define rutLex_h

// lexical analysis utilities - Denis

// References:
//
// "Compilers: Principles, Techniques, and Tools"
//   Alfred V. Aho, Ravi Sethi, Jeffery D. Ullman
//   Chapter 3: Lexical Analysis

// Types of regular expressions:
//
// a                   character
// ab                  string
// [a], [^a], [a-b]    character class
// A{m}, A{m,}, A{m,n} closures
// AB                  concatination
// A|B                 union
// (A)                 grouping
// <A>                 tagged grouping
//
// \b -> backspace
// \f -> formfeed
// \n -> newline
// \r -> carriage return
// \t -> horizontal tab
// \v -> vertical tab
// \0 -> null
// \a -> a
//
// .  -> [^\n]
// A* -> A{0,}
// A+ -> A{1,}
// A? -> A{0,1}

// Typical usage:
//
// rutLexRegExp regExp("(a|b)*abb);
// rutLexRegExpIterator ri(&regExp);
// static char string[] = "ababb";
// int i;
// for (i = 0; (i < strlen(string)) && ri.active(); ++i) {
//   ri.move(string[i]);
// }
// cout << (ri.accept() ? "match" : "bogus") << endl;

class ostream;
class rutString;

class rutLexSet {
public:
  rutLexSet(unsigned size = 256);
  rutLexSet(char ch);
  rutLexSet(const rutLexSet& s);

  rutLexSet& operator=(const rutLexSet& s);

  unsigned cardinality() const;
  void add(unsigned ch);
  void add(const rutLexSet* set);
  void remove(unsigned ch);
  int contains(unsigned ch) const;

  void invert();
  void add(unsigned lower, unsigned upper);

  int operator==(const rutLexSet& set) const;

protected:
  void initialize(unsigned size);
  void copy(const rutLexSet& set);

  friend ostream& operator<<(ostream& out, const rutLexSet& set);

  unsigned size_;
  unsigned bound_;
  unsigned* bits_;
  static unsigned mask_[32];
};

class rutLexStateSet;
class rutLexAutomaton;

class rutLexState {
public:
  rutLexState(rutLexAutomaton* automaton);
  ~rutLexState();

  void accept(int accept);
  int accept() const;

  void addEdge(rutLexState* to);
  void addEdge(rutLexState* to, char ch);
  void addEdge(rutLexState* to, rutLexSet set);

  void addTransition(rutLexState* to, char ch);

  rutLexSet& transitions();

  rutLexState* move(char ch) const;

protected:
  friend rutLexAutomaton;
  friend rutLexStateSet;

  int findEdge(const rutLexState* to) const;
  void newUpperBound(unsigned limit);

  friend ostream& operator<<(ostream& out, const rutLexState& set);

  unsigned uid_;

  int accept_;

  unsigned size_;
  unsigned limit_;
  rutLexSet** sets_;
  rutLexState** states_;

  rutLexSet* set_;
};

class rutLexAutomaton {
public:
  rutLexAutomaton();
  ~rutLexAutomaton();

  rutLexAutomaton* DFA();

  void addState(rutLexState* state);

  void start(rutLexState* state);
  rutLexState* start() const;

protected:
  friend ostream& operator<<(ostream& out, const rutLexAutomaton& automaton);

  void newUpperBound(unsigned limit);

  rutLexState* start_;
  rutLexSet* accept_;

  unsigned uid_;

  unsigned size_;
  unsigned limit_;
  rutLexState** states_;
};

class rutLexAutomatonIterator {
public:
  rutLexAutomatonIterator(rutLexAutomaton* automaton);
  virtual ~rutLexAutomatonIterator();

  void reset();
  int move(char ch);
  int active() const;
  int accept() const;

protected:
  rutLexAutomaton* automaton_;
  rutLexState* state_;
};

class rutLexStateList {
public:
  rutLexStateList();
  virtual ~rutLexStateList();

  void add(rutLexState* state);
  rutLexState* remove();
  operator int() const;

protected:
  void newUpperBound(unsigned limit);

  unsigned size_;
  unsigned limit_;
  rutLexState** states_;
};

class rutLexStateSet : public rutLexStateList {
public:
  rutLexStateSet(unsigned size);
  ~rutLexStateSet();

  int contains(const rutLexState* state) const;
  void add(rutLexState* state);

  int operator==(const rutLexStateSet& stateSet) const;

  rutLexSet* transitions() const;
  void eClosure();
  rutLexStateSet* move(char a) const;

  int accept() const;

protected:
  rutLexSet* set_;
};

class rutLexStateSetList {
public:
  rutLexStateSetList();
  virtual ~rutLexStateSetList();

  void add(rutLexStateSet* stateSet, rutLexState* state);
  void remove(rutLexStateSet*& stateSet, rutLexState*& state);

  operator int() const;

protected:
  void newUpperBound(unsigned limit);

  unsigned size_;
  unsigned limit_;
  rutLexStateSet** stateSets_;
  rutLexState** states_;
};

class rutLexStateSetSet : public rutLexStateSetList {
public:
  rutLexStateSetSet(rutLexAutomaton* automaton);
  ~rutLexStateSetSet();

  rutLexState* add(rutLexStateSet* stateSet);
  rutLexState* contains(const rutLexStateSet* stateSet) const;

protected:
  rutLexAutomaton* automaton_;
};

class rutLexRegExpNode;

class rutLexRegExp {
public:
  rutLexRegExp(const rutString& expression);
  rutLexRegExp(const rutLexRegExpNode* node);
  ~rutLexRegExp();

  static rutLexRegExpNode* parse(const rutString& expression);

  operator int();

protected:
  friend class rutLexRegExpIterator;
  rutLexAutomaton* dfa_;

  rutLexAutomaton* DFA(rutLexRegExpNode* node) const;
  rutLexAutomaton* NFA(rutLexRegExpNode* node) const;
};

class rutLexRegExpIterator : public rutLexAutomatonIterator {
public:
  rutLexRegExpIterator(const rutLexRegExp* regExp);
  ~rutLexRegExpIterator();
};

class rutLexRegExpNode {
public:
  virtual void constructNFA(rutLexAutomaton* automaton, rutLexState* start,
    rutLexState* accept) = 0;
};

class rutLexCharNode : public rutLexRegExpNode {
public:
  rutLexCharNode(char ch);

  void constructNFA(rutLexAutomaton* automaton, rutLexState* start,
    rutLexState* accept);

protected:
  char ch_;
};

class rutLexStringNode : public rutLexRegExpNode {
public:
  rutLexStringNode(char* string);

  void constructNFA(rutLexAutomaton* automaton, rutLexState* start,
    rutLexState* accept);

protected:
  char* string_;
};

class rutLexCharClassNode : public rutLexRegExpNode {
public:
  rutLexCharClassNode(rutLexSet set);

  void constructNFA(rutLexAutomaton* automaton, rutLexState* start,
    rutLexState* accept);

protected:
  rutLexSet set_;
};

class rutLexConcatNode : public rutLexRegExpNode {
public:
  rutLexConcatNode(rutLexRegExpNode* expA, rutLexRegExpNode* expB);

  void constructNFA(rutLexAutomaton* automaton, rutLexState* start,
    rutLexState* accept);

protected:
  rutLexRegExpNode* expA_;
  rutLexRegExpNode* expB_;
};

class rutLexUnionNode : public rutLexRegExpNode {
public:
  rutLexUnionNode(rutLexRegExpNode* expA, rutLexRegExpNode* expB);

  void constructNFA(rutLexAutomaton* automaton, rutLexState* start,
    rutLexState* accept);

protected:
  rutLexRegExpNode* expA_;
  rutLexRegExpNode* expB_;
};

class rutLexExactClosureNode : public rutLexRegExpNode {
public:
  rutLexExactClosureNode(rutLexRegExpNode* exp, int m);

  void constructNFA(rutLexAutomaton* automaton, rutLexState* start,
    rutLexState* accept);

protected:
  unsigned m_;
  rutLexRegExpNode* exp_;
};

class rutLexOpenClosureNode : public rutLexExactClosureNode {
public:
  rutLexOpenClosureNode(rutLexRegExpNode* exp, int m);

  void constructNFA(rutLexAutomaton* automaton, rutLexState* start,
    rutLexState* accept);
};

class rutLexRangeClosureNode : public rutLexRegExpNode {
public:
  rutLexRangeClosureNode(rutLexRegExpNode* exp, int m, int n);

  void constructNFA(rutLexAutomaton* automaton, rutLexState* start,
    rutLexState* accept);

protected:
  unsigned m_;
  unsigned n_;
  rutLexRegExpNode* exp_;
};

class rutLexTagNode : public rutLexRegExpNode {
public:
  rutLexTagNode(rutLexRegExpNode* exp);

  void constructNFA(rutLexAutomaton* automaton, rutLexState* start,
    rutLexState* accept);

protected:
  rutLexRegExpNode* exp_;
};

#endif
