/*
 * Copyright (c) 1991 Redwood Design Automation
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the name of
 * Redwood Design Automation may not be used in any advertising or publicity
 * relating to the software without the specific, prior written permission of
 * Redwood Design Automation.
 *
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
 *
 * IN NO EVENT SHALL REDWOOD DESIGN AUTOMATION BE LIABLE FOR ANY SPECIAL,
 * INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND, OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER OR NOT
 * ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF LIABILITY,
 * ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
 * SOFTWARE.
 */

#include <ctype.h>
#include <stdlib.h>
#include <string.h>

#include <iostream.h>
#include <iomanip.h>
#include <strstream.h>

#include <r.h>
#include <rutLex.h>
#include <rutString.h>

/////////////////
// String Link //
/////////////////

rutStringLink::rutStringLink() {
  prev_ = this;
  next_ = this;
}

int rutStringLink::empty() {
  return prev_ == next_;
}

////////////
// String //
////////////

rutBaseCommonBody(rutString, rutBase)

void rutString::initialize() {
  descriptor_->increment();

  area_ = &rutStringAreaDefault;

  length_ = 0;
  buffer_ = 0;

  insert();
}

rutString::rutString() {
  initialize();
}

rutString::rutString(const rutString& string) {
  initialize();

  area_ = string.area_;

  length_ = string.length_;
  buffer_ = string.buffer_;
}

rutString::rutString(const char* value) {
  initialize();

  int l = strlen(value);
  if (l) {
    buffer_ = area_->allocate(l);
    length_ = l;
    strncpy(buffer_, value, length_);
  }
}

rutString::rutString(const char* value, int length) {
  initialize();

  if (length) {
    buffer_ = area_->allocate(length);
    length_ = length;
    strncpy(buffer_, value, length_);
  }
}

rutString::rutString(const char value) {
  initialize();

  buffer_ = area_->allocate(1);
  length_ = 1;
  buffer_[0] = value;
}

rutString::rutString(const int value) {
  initialize();

  ostrstream stream;
  stream << value;
  char* buffer = stream.str();
  int l = stream.pcount();
  buffer_ = area_->allocate(l);
  length_ = l;
  strncpy(buffer_, buffer, length_);
}

rutString::rutString(const float value) {
  initialize();

  ostrstream stream;
  stream << value;
  char* buffer = stream.str();
  int l = stream.pcount();
  buffer_ = area_->allocate(l);
  length_ = l;
  strncpy(buffer_, buffer, length_);
}

rutString::~rutString() {
  descriptor_->decrement();
  remove();
}

void rutString::insert() {
  next_ = area_->head_.next_;
  area_->head_.next_ = this;
  prev_ = next_->prev_;
  next_->prev_ = this;
}

void rutString::remove() {
  prev_->next_ = next_;
  next_->prev_ = prev_;
}

int rutString::length() const {
  return length_;
}

char rutString::getNth(int i) const {
  rPrecondition((0 <= i) && (i < length_));

  return buffer_[i];
}

char rutString::operator[](int i) const {
  return getNth(i);
}

int rutString::operator==(const rutString& string) const {
  return (length_ == string.length_) &&
    (strncmp(buffer_, string.buffer_, length_) == 0);
}

int rutString::operator!=(const rutString& string) const {
  return (length_ != string.length_) ||
    (strncmp(buffer_, string.buffer_, length_) != 0);
}

int rutString::operator>(const rutString& string) const {
  return (length_ > string.length_) || ((length_ == string.length_) &&
    (strncmp(buffer_, string.buffer_, length_) > 0));
}

int rutString::operator<(const rutString& string) const {
  return (length_ < string.length_) || ((length_ == string.length_) &&
    (strncmp(buffer_, string.buffer_, length_) < 0));
}

int rutString::operator>=(const rutString& string) const {
  return (length_ > string.length_) || ((length_ == string.length_) &&
    (strncmp(buffer_, string.buffer_, length_) >= 0));
}

int rutString::operator<=(const rutString& string) const {
  return (length_ < string.length_) || ((length_ == string.length_) &&
    (strncmp(buffer_, string.buffer_, length_) <= 0));
}

static inline int charCompare(char a, char b, rutStringFlag flags) {
  return (flags & rutString::caseSensitive ? a - b :
    (isupper(a) ? tolower(a) : a) - (isupper(b) ? tolower(b) : b));
}

int rutString::match(const rutString& string, rutStringFlag flags) const {
  for (int i = 0; (i < length_) && (i < string.length_) &&
    (charCompare(buffer_[i], string.buffer_[i], flags) == 0); ++i);
  return i;
}

int rutString::match(const rutLexRegExp& regExp) const {
  rutLexRegExpIterator iterator(&regExp);
  for (int i = 0; (i < length_) && iterator.active(); ++i) {
    iterator.move(buffer_[i]);
  }
  return iterator.accept();
}

int rutString::compare(const rutString& string, rutStringFlag flags) const {
  if (length_ == string.length_) {
    for (int i = 0; i < length_; ++i) {
      if (charCompare(buffer_[i], string.buffer_[i], flags) != 0) {
	return charCompare(buffer_[i], string.buffer_[i], flags);
      }
    }
    return 0;
  }
  return length_ - string.length_;
}

rutString rutString::substring(int from, int to) const {
  rPrecondition((0 <= from) && (from <= length_));
  rPrecondition((0 <= to) && (to <= length_));
  rPrecondition(from <= to);

  rutString string;
  if (from < to) {
    string.buffer_ = buffer_ + from;
    string.length_ = to - from;
  }
  return string;
}

rutString rutString::substringExpand(int from, int to) const {
  rPrecondition(from <= to);

  if (to <= length_) {
    return substring(from, to);
  }

  rutString string;
  if (from < to) {
    string = *this;
    while (to > string.length_) {
      string &= ' ';
    }
    string.buffer_ = string.buffer_ + from;
    string.length_ = to - from;
  }
  return string;
}

rutString rutString::substringEnd(int from) const {
  return substring(from, length_);
}

rutString rutString::substringEndExpand(int from) const {
  if (from < length_) {
    return substring(from, length_);
  } else {
    rutString string;
    return string;
  }
}

rutString rutString::operator&(const rutString& string) const {
  rutString result(*this);
  result &= string;
  return result;
}

ostream& operator<<(ostream& out, const rutString& string) {
  for (int i = 0; i < string.length_; ++i) {
    out << string.buffer_[i];
  }
  return out;
}

istream& operator>>(istream& in, rutString& string) {
  int size = 256;
  int index = 0;
  char* buffer = new char[size];
  while (! in.eof()) {
    if (index == size) {
      char* t = new char[size * 2];
      strncpy(t, buffer, size);
      delete buffer;
      buffer = t;
      size *= 2;
    }
    in >> buffer[index];
  }
  string = rutString(buffer, index);
  delete buffer;
  return in;
}

/*
rutPersist& operator<<(rutPersist& persist, const rutString& string) {
  char* buffer = string.newCString();
  persist << buffer;
  delete buffer;
  return persist;
}

rutPersist& operator>>(rutPersist& persist, rutString& string) {
  char* buffer;
  persist >> buffer;
  string = buffer;
  delete buffer;
  return persist;
}
*/

void rutString::operator=(const rutString& string) {
  if (string.area_ != area_) {
    remove();
    area_ = string.area_;
    insert();
  }
  length_ = string.length_;
  buffer_ = string.buffer_;
}

void rutString::operator&=(const rutString& string) {
  if (length_ == 0) {
    buffer_ = string.buffer_;
  } else if ((buffer_ + length_) != string.buffer_) {
    area_->extend(this, string.length_);
    strncpy(&buffer_[length_], string.buffer_, string.length_);
  }
  length_ += string.length_;
}

void rutString::operator&=(const char* string) {
  int delta;

  delta = strlen(string);
  if (delta) {
    area_->extend(this, delta);
    strncpy(&buffer_[length_], string, delta);
    length_ += delta;
  }
}

rutString rutString::removeWord() {
  // remove leading white space
  for (int from = 0; (from < length_) && isspace(buffer_[from]); ++from);
  // find beginning of next white space
  for (int to = from; (to < length_) && (! isspace(buffer_[to])); ++to);

  rutString string = substring(from, to);
  length_ -= to;
  if (length_) {
    buffer_ += to;
  } else {
    buffer_ = 0;
  }
  return string;
}

rutString::operator int() const {
  return length_;
}

int rutString::convert(char& value) const {
  if (length_ != 1) {
    return 0;
  }
  value = buffer_[0];
  return 1;
}

int rutString::convert(int& value) const {
  char* buffer = newCString();
  istrstream text(buffer);
  int v;
  text >> v;
  delete buffer;
  if (! text.fail()) {
    value = v;
  }
  return ! text.fail();
}

int rutString::convert(float& value) const {
  char* buffer = newCString();
  istrstream text(buffer);
  int v;
  text >> v;
  delete buffer;
  if (! text.fail()) {
    value = v;
  }
  return ! text.fail();
}

void rutString::convert(char* buffer) const {
  strncpy(buffer, buffer_, length_);
  buffer[length_] = '\0';
}

void rutString::convert(char* buffer, int length) const {
  strncpy(buffer, buffer_, length);
}

char* rutString::newCString() const {
  char* buffer = new char[length_ + 1];
  strncpy(buffer, buffer_, length_);
  buffer[length_] = '\0';
  return buffer;
}

/*
int rutString::compareStatic(const rutBase* a, const rutBase* b) {
  rutString* sa = rutString::downCast(a);
  rutString* sb = rutString::downCast(b);
  return sa->compare(*(const rutString*) sb);
}

int rutString::compareCaseInsensitive(const rutBase* a, const rutBase* b) {
  rutString* sa = rutString::downCast(a);
  rutString* sb = rutString::downCast(b);
  return sa->compare(*(const rutString*) sb, caseInsensitive);
}

unsigned rutString::hash(const rutBase* object) {
  rutString* string;
  unsigned code = 0;
  int i;

  string = rutString::downCast(object);
  // !!! I'm sure there is a much better hash function -denis !!!
  for (i = 0; i < string->length_; ++i) {
    code = code << 3;
    code += string->getNth(i);
  }

  return code;
}
*/

//////////////////////////
// String Stream Buffer //
//////////////////////////

rutStringStreamBuffer::rutStringStreamBuffer() {
  streambuf::allocate();
}

rutStringStreamBuffer::~rutStringStreamBuffer() {
}

int rutStringStreamBuffer::overflow(int c) {
  // make sure the current position is pointing into the buffer
  if (! pptr()) {
    setp(base(), base());
  }

  // count the number of characters in the buffer
  int count = pptr() - pbase();

  // copy the contents of the buffer to the string
  operator&=(rutString(pbase(), count));

  // reset the buffer
  setp(base(), ebuf());

  // put the character into the buffer
  return (c != EOF ? sputc(c) : zapeof(c));
}

//////////////////
// String Space //
//////////////////

rutStringArea rutStringAreaDefault;

#define bitsizeof(t) (sizeof(t) * 8)

inline int min(int a, int b) {
  return (a) <= (b) ? (a) : (b);
}

rutStringArea::rutStringArea(int size) {
  size_ = size;
  buffer_ = new char[size];
  index_ = 0;
}

rutStringArea::~rutStringArea() {
/*
  // check for dangling strings.
  if (! head_.empty()) {
    printStrings();
  }
  rPrecondition(head_.empty());
*/

  delete buffer_;
}

void rutStringArea::printMarkMap() {
  int i;

  cerr << "mark map:" << endl;
  for (i = 0; i < size_ / bitsizeof(unsigned); ++i) {
    cerr << setbase(16) << setw(8) << markMap_[i];
  }
  cerr << endl;
}

void rutStringArea::printArea() {
  cerr << "string area:" << endl;
  for (int i = 0; i < index_; ++i) {
    cerr << buffer_[i];
  }
  cerr << endl;
}

void rutStringArea::printStrings() {
  cerr << "strings:" << endl;
  for (rutStringLink* link = head_.next_; link != &head_; link = link->next_) {
    rutString* string = (rutString*) link;
    char* buffer = string->newCString();
    cerr << setbase(16) << setw(8) << string->buffer_ << " " <<
      string->length_ << buffer;
    delete buffer;
  }
}

void rutStringArea::statistics(int& size, int& allocated, int& used) {
  size = size_;
  allocated = index_;

  markMap_ = new unsigned[size_ / bitsizeof(unsigned)];
  {int i; for (i = 0; i < size_ / bitsizeof(unsigned); ++i){markMap_[i] = 0;}}

  mark();
  used = 0;
  for (int i = 0; i < index_; ++i) {
    if (markMap_[i / bitsizeof(unsigned)] & (1 << (i % bitsizeof(unsigned)))) {
      ++used;
    }
  }

  delete markMap_;
}

void rutStringArea::check(int length) {
  while (size_ < (index_ + length)) {
    collect();
  }
}

char* rutStringArea::allocate(int length) {
  check(length);
  char* buffer = &buffer_[index_];
  index_ += length;
  return buffer;
}

void rutStringArea::extend(rutString* string, int delta) {
  int next = string->buffer_ - buffer_ + string->length_;
  if ((next == index_) && (next + delta <= size_)) {
    index_ += delta;
  } else {
    char* buffer = allocate(string->length_ + delta);
    strncpy(buffer, string->buffer_, string->length_);
    string->buffer_ = buffer;
  }

  rPostcondition(index_ <= size_);
}

int rutStringArea::occupied(int i) {
  // 0xf should be computed from sizeof(int)
  return markMap_[i / bitsizeof(unsigned)] & (0xf << (i % bitsizeof(unsigned)));
}

int rutStringArea::forwarded(int i) {
  // 0xf should be computed from sizeof(int)
  return ! (markMap_[i / bitsizeof(unsigned)] & (0xf << (i %
    bitsizeof(unsigned))));
}

void rutStringArea::markRange(int index, int length) {
  int fromIndex = index / bitsizeof(unsigned);
  int fromBit   = index % bitsizeof(unsigned);

  int toIndex = (index + length) / bitsizeof(unsigned);
  int toBit   = (index + length) % bitsizeof(unsigned);

  if (fromIndex == toIndex) {
    unsigned mask;

    for (mask = 1 << fromBit; mask != (1 << toBit); mask = mask << 1) {
      markMap_[fromIndex] |= mask;
    }
  } else {
    unsigned mask;

    for (mask = 1 << fromBit; mask != 0; mask = mask << 1) {
      markMap_[fromIndex] |= mask;
    }
    for (int i = fromIndex; i < toIndex; ++i) {
      markMap_[i] = ~0;
    }
    for (mask = 1; mask != (1 << toBit); mask = mask << 1) {
      markMap_[toIndex] |= mask;
    }
  }
}

void rutStringArea::mark() {
  for (rutStringLink* link = head_.next_; link != &head_; link = link->next_) {
    rutString* string = (rutString*) link;
    if (string->length_) {
      markRange(string->buffer_ - buffer_, string->length_);
    }
  }
}

void rutStringArea::forward(int i) {
  // copy to new buffer.
  // use min(i + sizeof(int), index_) so that buffer size will not grow.
  // but we still may not throw away all garbage at the end of the buffer.

  for (int j = i; j < min(i + sizeof(int), (index_ + 4) & ~0x3); ++j) {
    newBuffer_[newIndex_] = buffer_[j];
    ++newIndex_;
  }
  rAssert(! (newIndex_ & 0x3));
  // store forwarding address.
  char* address = &newBuffer_[newIndex_ - sizeof(unsigned)];
  rAssert(! ((unsigned) address & 0x3));
  *((char**) &buffer_[i]) = address;
  markMap_[i / bitsizeof(unsigned)] &= ~(0xf << (i % bitsizeof(unsigned)));
}

char* rutStringArea::forwardAddress(char* buffer) {
  return (char*) (((unsigned) *((char**) ((unsigned) buffer & ~0x3))) |
    ((unsigned) buffer & 0x3));
}

// this only touches live data.
void rutStringArea::copy() {
  for (rutStringLink* link = head_.next_; link != &head_; link = link->next_) {
    rutString* string = (rutString*) link;
    if (! string->length_) {
      continue;
    }

    if (! forwarded(string->buffer_ - buffer_)) {
      // find the start of the range.
      int j = (string->buffer_ - buffer_) & ~0x3;
      while ((j >= sizeof(unsigned)) && occupied(j - sizeof(unsigned))) {
        j -= sizeof(unsigned);
      }
      // forward the range.
      while ((j < index_) && occupied(j)) {
	forward(j);
        j += sizeof(unsigned);
      }
    }
    char* buffer = forwardAddress(string->buffer_);
    buffer = forwardAddress(string->buffer_);
    rAssert(buffer >= newBuffer_);
    string->buffer_ = buffer;
  }
}

void rutStringArea::collect() {
  int size, allocated, used;

  rutStringAreaDefault.statistics(size, allocated, used);
  cerr << "Starting collection (size: " << size << ", allocated " <<
    allocated << ", used: " << used << ")\n";

  markMap_ = new unsigned[size_ / bitsizeof(unsigned)];
  {int i; for (i = 0; i < size_ / bitsizeof(unsigned); ++i){markMap_[i] = 0;}}
  // we need some smart way to grow and shrink the buffer size based on usage
  newBuffer_ = new char[size_ * 2]; // !!! always grows !!!
  newIndex_ = 0;

  mark();
  copy();

  char* tmp = buffer_;
  buffer_ = newBuffer_;
  index_ = newIndex_;
  size_ = size_ * 2;
  delete tmp;
  delete markMap_;

  rutStringAreaDefault.statistics(size, allocated, used);
  cerr << "Finishing collection (size: " << size << ", allocated " <<
    allocated << ", used: " << used << ")\n";

  cerr.flush();
}
