/*
 * Copyright (c) 1991 Redwood Design Automation
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the name of
 * Redwood Design Automation may not be used in any advertising or publicity
 * relating to the software without the specific, prior written permission of
 * Redwood Design Automation.
 *
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
 *
 * IN NO EVENT SHALL REDWOOD DESIGN AUTOMATION BE LIABLE FOR ANY SPECIAL,
 * INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND, OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER OR NOT
 * ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF LIABILITY,
 * ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
 * SOFTWARE.
 */

#ifndef rutString_h
#define rutString_h

#include <iostream.h>

#include <rutArray.h>
#include <rutBase.h>
/*
#include <rutPersist.h>
*/

class rutStringLink {
protected:
  friend class rutString;
  friend class rutStringArea;

  rutStringLink();

  int empty();

  rutStringLink* prev_;
  rutStringLink* next_;
};

class rutStringArea;
class rutLexRegExp;

typedef unsigned rutStringFlag;

class rutString : public rutStringLink, public rutBase {
public:
  rutBaseCommonDefine(rutString)

  enum {caseSensitive = 0x1, caseInsensitive = 0x2};

  rutString();
  rutString(const rutString& string);
  rutString(const char* value);
  rutString(const char* value, int length);
  rutString(const char value);
  rutString(const int value);
  rutString(const float value);
  ~rutString();

  // accessors
  int length() const;
  char getNth(int i) const;
  char operator[](int i) const;

  // standard comparison functions
  int operator==(const rutString& string) const;
  int operator!=(const rutString& string) const;
  int operator>(const rutString& string) const;
  int operator<(const rutString& string) const;
  int operator>=(const rutString& string) const;
  int operator<=(const rutString& string) const;

  // other comparison functions
  int match(const rutString& string, rutStringFlag flags = caseSensitive) const;
  int match(const rutLexRegExp& regExp) const;
  int compare(const rutString& string, rutStringFlag flags = caseSensitive)
    const;

  // functions that allocate new strings
  rutString substring(int from, int to) const;
  rutString substringExpand(int from, int to) const;
  rutString substringEnd(int from) const;
  rutString substringEndExpand(int from) const;
  rutString operator&(const rutString& string) const;

  // mutators
  void operator=(const rutString& string);
  void operator&=(const rutString& string);
  void operator&=(const char* string);

  // mutators that allocate new strings
  rutString removeWord();

  // input/output
  friend ostream& operator<<(ostream& out, const rutString& string);
  friend ostream& operator>>(ostream& out, rutString& string);
/*
  friend rutPersist& operator<<(rutPersist& persist, const rutString& string);
  friend rutPersist& operator>>(rutPersist& persist, rutString& string);
*/

  // conversion functions
  operator int() const; // length, so you can say: if (! string) { ... }
  int convert(char& value) const;
  int convert(int& value) const;
  int convert(float& value) const;
  void convert(char* buffer) const; // null terminated
  void convert(char* buffer, int length) const; // not null terminated

  // conversion functions that allocate memory
  char* newCString() const;

/*
  // collection functions
  static int compareStatic(const rutBase* a, const rutBase* b);
  static int compareCaseInsensitive(const rutBase* a, const rutBase* b);
  static unsigned hash(const rutBase* string);
*/

protected:
  friend class rutStringArea;

  void initialize();
  void insert();
  void remove();

  rutStringArea* area_;

  char* buffer_;
  int length_;
};

rutArrayDefine(rutStringArray,rutString)

class rutStringStreamBuffer : public rutString, public streambuf {
public:
  rutStringStreamBuffer();
  ~rutStringStreamBuffer();

  int overflow(int);
};

class rutStringArea {
public:
  rutStringArea(int size = 8192);
  ~rutStringArea();

  void statistics(int& size, int& allocated, int& used);
  void collect();

protected:
  friend class rutString;

  void check(int length);

  char* allocate(int length);
  void extend(rutString* string, int delta);

  void printMarkMap();
  void printArea();
  void printStrings();

  int occupied(int i);
  int forwarded(int i);
  char* forwardAddress(char* buffer);
  void forward(int i);
  void markRange(int index, int length);
  void mark();
  void copy();

  char* buffer_;
  int size_;
  int index_;

  rutStringLink head_;

  unsigned* markMap_;
  char* newBuffer_;
  int newIndex_;
};

extern rutStringArea rutStringAreaDefault;

#endif
