/*
 * Copyright (c) 1991 Redwood Design Automation
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the name of
 * Redwood Design Automation may not be used in any advertising or publicity
 * relating to the software without the specific, prior written permission of
 * Redwood Design Automation.
 *
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
 *
 * IN NO EVENT SHALL REDWOOD DESIGN AUTOMATION BE LIABLE FOR ANY SPECIAL,
 * INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND, OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER OR NOT
 * ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF LIABILITY,
 * ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
 * SOFTWARE.
 */

#include <iostream.h>

#include <r.h>
#include <rutArray.h>
#include <rutLex.h>
#include <rutString.h>
#include <rutText.h>

static unsigned min(unsigned a, unsigned b) { return a < b ? a : b; }

class rutTextOperator {
};

/////////////////////////
// Text Operator Paste //
/////////////////////////

class rutTextOperatorPaste : public rutTextOperator {
public:
  rutTextOperatorPaste();

  void paste(rutText* text, unsigned line, unsigned column, rutText* textPaste,
    rutTextKind kind = rutText::linear);

protected:
  void pasteRectangular(rutText* text, unsigned line, unsigned column,
    rutText* textPaste);

  void pasteLinear(rutText* text, unsigned line, unsigned column,
    rutText* textPaste);
};

rutTextOperatorPaste::rutTextOperatorPaste() {
}

void rutTextOperatorPaste::paste(rutText* text, unsigned line, unsigned column,
  rutText* textPaste, rutTextKind kind) {
  if ((textPaste->lines() > 1) && (kind == rutText::linear)) {
    pasteLinear(text, line, column, textPaste);
  } else {
    pasteRectangular(text, line, column, textPaste);
  }
}

void rutTextOperatorPaste::pasteRectangular(rutText* text, unsigned line,
  unsigned column, rutText* textPaste) {
  unsigned i;

  for (i = 0; i < textPaste->lines(); ++i) {
    rutString& string = text->getExpand(line);
    rutString prefix = string.substringExpand(0, column);
    rutString suffix = string.substringEndExpand(column);
    string = prefix & textPaste->get(i) & suffix;
    ++line;
  }
}

void rutTextOperatorPaste::pasteLinear(rutText* text, unsigned line,
  unsigned column, rutText* textPaste) {
  rutString& string = text->getExpand(line);
  rutString prefix = string.substringExpand(0, column);
  rutString suffix = string.substringEndExpand(column);
  string = prefix & textPaste->get(0);
  ++line;

  for (unsigned i = 1; (i + 1) < textPaste->lines(); ++i) {
    text->addNth(textPaste->get(i), line);
    ++line;
  }

  text->addNth(textPaste->get(i) & suffix, line);
}

///////////////////////
// Text Operator Cut //
///////////////////////

class rutTextOperatorCut : public rutTextOperator {
public:
  rutTextOperatorCut();

  void cut(rutText* text, unsigned line1, unsigned column1, unsigned line2,
    unsigned column2, rutText* textCut, rutTextKind kind = rutText::linear);

protected:
  void cutRectangular(rutText* text, unsigned line1, unsigned column1,
    unsigned line2, unsigned column2, rutText* textCut);

  void cutLinear(rutText* text, unsigned line1, unsigned column1,
    unsigned line2, unsigned column2, rutText* textCut);
};

rutTextOperatorCut::rutTextOperatorCut() {
}

void rutTextOperatorCut::cut(rutText* text, unsigned line1, unsigned column1,
  unsigned line2, unsigned column2, rutText* textCut, rutTextKind kind) {
  if ((line1 < line2) && (kind == rutText::linear)) {
    cutLinear(text, line1, column1, line2, column2, textCut);
  } else {
    cutRectangular(text, line1, column1, line2, column2, textCut);
  }
}

void rutTextOperatorCut::cutRectangular(rutText* text, unsigned line1,
  unsigned column1, unsigned line2, unsigned column2, rutText* textCut) {
  unsigned line = line1;
  for (unsigned i = line1; i <= line2; ++i) {
    if (line >= text->lines()) {
      rutString string;
      rutString substring = string.substringExpand(column1, column2);
      textCut->add(substring);
    } else {
      rutString& string = text->get(line);
      textCut->add(string.substringExpand(column1, column2));
      rutString prefix = string.substring(0, min(column1, string.length()));
      rutString suffix = string.substringEndExpand(column2);
      string = prefix & suffix;
    }
    ++line;
  }
}

void rutTextOperatorCut::cutLinear(rutText* text, unsigned line1,
  unsigned column1, unsigned line2, unsigned column2, rutText* textCut) {
  rutString* joinString = 0;

  unsigned line = line1;
  if (line < text->lines()) {
    rutString& string = text->get(line);
    textCut->add(string.substringEndExpand(column1));
    string = string.substringExpand(0, column1);
    joinString = &string;
  } else {
    rutString string;
    textCut->add(string.substringEndExpand(column1));
  }
  ++line;

  for (unsigned i = line1 + 1; i < line2; ++i) {
    if (line < text->lines()) {
      textCut->add(text->get(line));
      text->removeNth(line);
    } else {
      rutString string;
      textCut->add(string);
    }
  }

  if (line < text->lines()) {
    rutString& string = text->get(line);
    textCut->add(string.substringExpand(0, column2));
    *joinString &= string.substringEndExpand(column2);
    text->removeNth(line);
  } else {
    rutString string;
    textCut->add(string.substringExpand(0, column2));
  }
}

/////////////////////////
// Text Operator Apply //
/////////////////////////

class rutTextOperatorApply : public rutTextOperator {
public:
  void apply(rutText* text, unsigned line1, unsigned column1, unsigned line2,
    unsigned column2, rutTextKind kind = rutText::linear);

protected:
  void applyRectangular(rutText* text, unsigned line1, unsigned column1,
    unsigned line2, unsigned column2);

  void applyLinear(rutText* text, unsigned line1, unsigned column1,
    unsigned line2, unsigned column2);

  virtual void function(rutString& substring) = 0;
};

void rutTextOperatorApply::apply(rutText* text, unsigned line1,
  unsigned column1, unsigned line2, unsigned column2, rutTextKind kind) {
  if ((line1 < line2) && (kind == rutText::linear)) {
    applyLinear(text, line1, column1, line2, column2);
  } else {
    applyRectangular(text, line1, column1, line2, column2);
  }
}

void rutTextOperatorApply::applyRectangular(rutText* text, unsigned line1,
  unsigned column1, unsigned line2, unsigned column2) {
  unsigned line = line1;
  for (unsigned i = line1; i <= line2; ++i) {
    function(text->getSubstring(line, column1, column2));
    ++line;
  }
}

void rutTextOperatorApply::applyLinear(rutText* text, unsigned line1,
  unsigned column1, unsigned line2, unsigned column2) {
  unsigned line = line1;
  function(text->getSubstringEnd(line, column1));
  ++line;

  for (unsigned i = line1 + 1; i < line2; ++i) {
    if (line < text->lines()) {
      function(text->get(line));
    } else {
      rutString string;
      function(string);
    }
    ++line;
  }

  function(text->getSubstring(line, 0, column2));
}

////////////////////////
// Text Operator Copy //
////////////////////////

class rutTextOperatorCopy : public rutTextOperatorApply {
public:
  rutTextOperatorCopy();
  rutText* textCopy;

  virtual void function(rutString& substring);
};

rutTextOperatorCopy::rutTextOperatorCopy() {
  textCopy = new rutText;
}

void rutTextOperatorCopy::function(rutString& substring) {
  textCopy->add(substring);
}

///////////////////
// Text Iterator //
///////////////////

rutTextIterator::rutTextIterator() {
  text_ = 0;
  line_ = 0;
  column_ = 0;
}

rutTextIterator::rutTextIterator(rutText* text, unsigned line,
  unsigned column) {
  text_ = text;
  line_ = line;
  column_ = column;
}

rutTextIterator::~rutTextIterator() {
}

int rutTextIterator::next(char& ch) {
  if (line_ >= text_->lines()) {
    ch = 0;
    return 0;
  }
  rutString string = text_->get(line_);
  if (column_ >= string.length()) {
    ++line_;
    column_ = 0;
    ch = '\n';
    return 1;
  }
  ch = string[column_];
  ++column_;
  return 1;
}

int rutTextIterator::prev() {
  if (column_ > 0) {
    --column_;
    return 1;
  }
  if (line_ <= 0) {
    return 0;
  }
  --line_;
  rutString string = text_->get(line_);
  column_ = string.length();
  return 1;
}

unsigned rutTextIterator::line() const {
  return line_;
}

unsigned rutTextIterator::column() const {
  return column_;
}

//////////
// Text //
//////////

rutText::rutText() {
  lines_ = new rutStringArray;
}

rutText::~rutText() {
  clear();
  delete lines_;
}

unsigned rutText::lines() const {
  return lines_->size();
}

const rutString& rutText::getNth(unsigned line) const {
  return *lines_->get(line);
}

rutTextIterator rutText::iterator(const rutTextLocation& location) const {
  return rutTextIterator((rutText*) this, location.line_, location.column_);
}

void rutText::clear() {
  while (lines()) {
    rutString* string = lines_->removeLast();
    delete string;
  }
}

void rutText::write(const rutString& string) {
  char* buffer = string.newCString();
  write(buffer, string.length());
}

void rutText::write(char* buffer, unsigned count) {
  if (! lines()) {
    lines_->insertLast(new rutString);
  }
  rutString* string = &get(lines() - 1);

  for (unsigned i = 0; i < count; ++i) {
    if (buffer[i] == '\n') {
      string = new rutString;
      lines_->insertLast(string);
    } else {
      if (! lines()) {
        string = new rutString;
        lines_->insertLast(string);
      }
      *string &= buffer[i];
    }
  }
}

rutText* rutText::cut(const rutTextRegion& region) {
  rutText* textCut = new rutText;
  rutTextOperatorCut operatorCut;
  operatorCut.cut(this, region.line1_, region.column1_, region.line2_,
    region.column2_, textCut, region.kind_);
  return textCut;
}

rutText* rutText::copy(const rutTextRegion& region) {
  rutTextOperatorCopy operatorCopy;
  operatorCopy.apply(this, region.line1_, region.column1_, region.line2_,
    region.column2_, region.kind_);
  return operatorCopy.textCopy;
}

void rutText::paste(const rutTextLocation& location, rutText* textPaste) {
  rutTextOperatorPaste operatorPaste;
  operatorPaste.paste(this, location.line_, location.column_, textPaste,
    location.kind_);
}

rutTextRegion rutText::search(const rutTextLocation& location,
  const rutLexRegExp& regExp) const {
  int found = 0;
  unsigned line;
  unsigned column;
  char ch;
  rutTextIterator textIterator = iterator(location);
  do {
    rutTextIterator searchIterator(textIterator);
    rutLexRegExpIterator regExpIterator(&regExp);
    while (regExpIterator.active() && searchIterator.next(ch)) {
      if (regExpIterator.move(ch) && regExpIterator.accept()) {
        line = searchIterator.line();
        column = searchIterator.column();
        found = 1;
      }
    }
  } while ((! found) && textIterator.next(ch));

  rutTextRegion region;
  if (found) {
    region.line1(textIterator.line());
    region.column1(textIterator.column());
    region.line2(line);
    region.column2(column);
  }
  return region;
}

ostream& operator<<(ostream& out, rutText& text) {
  unsigned i;

  for (i = 0; i < text.lines_->size(); ++i) {
    out << text.get(i) << "\n";
  }
  return out;
}

rutString& rutText::get(unsigned line) const {
  return *lines_->get(line);
}

rutString& rutText::getExpand(unsigned line) {
  if (line >= lines()) {
    while (line >= lines()) {
      lines_->insertLast(new rutString);
    }
  }
  return get(line);
}

rutString rutText::getSubstring(unsigned line, unsigned column1,
  unsigned column2) {
  if (line < lines()) {
    return get(line).substringExpand(column1, column2);
  } else {
    rutString string;
    return string.substringExpand(column1, column2);
  }
}

rutString rutText::getSubstringEnd(unsigned line, unsigned column) {
  if (line < lines()) {
    return get(line).substringEndExpand(column);
  } else {
    rutString string;
    return string.substringEndExpand(column);
  }
}

void rutText::add(rutString string) {
  lines_->insertLast(new rutString(string));
}

void rutText::addNth(rutString string, unsigned line) {
  lines_->insert(new rutString(string), line);
}

rutString rutText::remove() {
  rPrecondition(lines_->size() > 0);
  return removeNth(lines_->size() - 1);
}

rutString rutText::removeNth(unsigned line) {
  rPrecondition((0 <= line) && (line < lines_->size()));

  rutString* string = lines_->remove(line);
  rutString result = *string;
  delete string;
  return result;
}

///////////////////
// Text Location //
///////////////////

rutTextLocation::rutTextLocation(unsigned line, unsigned column,
  rutTextKind kind) {
  line_ = line;
  column_ = column;
  kind_ = kind;
}

rutTextLocation::~rutTextLocation() {
}

void rutTextLocation::kind(rutTextKind kind) {
  kind_ = kind;
}

rutTextKind rutTextLocation::kind() const {
  return kind_;
}

void rutTextLocation::line(unsigned line) {
  line_ = line;
}

unsigned rutTextLocation::line() const {
  return line_;
}

void rutTextLocation::column(unsigned column) {
  column_ = column;
}

unsigned rutTextLocation::column() const {
  return column_;
}

/////////////////
// Text Region //
/////////////////

rutTextRegion::rutTextRegion(unsigned line1, unsigned column1, unsigned line2,
  unsigned column2, rutTextKind kind) {
  line1_ = line1;
  column1_ = column1;
  line2_ = line2;
  column2_ = column2;
  kind_ = kind;
}

rutTextRegion::~rutTextRegion() {
}

void rutTextRegion::kind(rutTextKind kind) {
  kind_ = kind;
}

rutTextKind rutTextRegion::kind() const {
  return kind_;
}

void rutTextRegion::line1(unsigned line) {
  line1_ = line;
}

unsigned rutTextRegion::line1() const {
  return line1_;
}

void rutTextRegion::column1(unsigned column) {
  column1_ = column;
}

unsigned rutTextRegion::column1() const {
  return column1_;
}

void rutTextRegion::line2(unsigned line) {
  line2_ = line;
}

unsigned rutTextRegion::line2() const {
  return line2_;
}

void rutTextRegion::column2(unsigned column) {
  column2_ = column;
}

unsigned rutTextRegion::column2() const {
  return column2_;
}

rutTextRegion::operator int() {
  return (line2_ > line1_) || ((line1_ == line2_) && (column2_ > column1_));
}

////////////////////////
// Text Stream Buffer //
////////////////////////

rutTextStreamBuffer::rutTextStreamBuffer() {
  streambuf::allocate();
}

rutTextStreamBuffer::~rutTextStreamBuffer() {
}

int rutTextStreamBuffer::overflow(int c) {
  int count;

  // make sure the current position is pointing into the buffer
  if (! pptr()) {
    setp(base(), base());
  }

  // count the number of characters in the buffer
  count = pptr() - pbase();

  // copy the contents of the stream buffer to the text buffer
  write(pbase(), count);

  // reset the buffer
  setp(base(), ebuf());

  // put the character into the buffer
  return (c != EOF ? sputc(c) : zapeof(c));
}
