/*
 * Copyright (c) 1990 Stanford University
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided
 * that the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of Stanford not be used in advertising or
 * publicity pertaining to distribution of the software without specific,
 * written prior permission.  Stanford makes no representations about
 * the suitability of this software for any purpose.  It is provided "as is"
 * without express or implied warranty.
 *
 * STANFORD DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS.
 * IN NO EVENT SHALL STANFORD BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 * DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
 * OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION
 * WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

/*
 * Implementation of schematic capture system-specific commands.
 */

#include "scmed.h"
#include "scmclasses.h"
#include "scmcmds.h"
#include "scmcomps.h"
#include "scmdialogs.h"
#include "scmnet.h"
#include "scmpal.h"
#include "scmtools.h"

#include <Unidraw/catalog.h>
#include <Unidraw/clipboard.h>
#include <Unidraw/ctrlinfo.h>
#include <Unidraw/editorinfo.h>
#include <Unidraw/globals.h>
#include <Unidraw/iterator.h>
#include <Unidraw/selection.h>
#include <Unidraw/statevars.h>
#include <Unidraw/unidraw.h>
#include <Unidraw/viewer.h>
#include <Unidraw/Commands/transforms.h>
#include <Unidraw/Components/grcomp.h>
#include <Unidraw/Graphic/graphic.h>
#include <Unidraw/Graphic/util.h>

#include <InterViews/filechooser.h>
#include <InterViews/window.h>

#include <stdio.h>
#include <stream.h>
#include <string.h>

/*****************************************************************************/

static void Warning (Editor* ed, const char* warning) {
    AcknowledgeDialog dialog(warning);
    ed->InsertDialog(&dialog);
    dialog.Acknowledge();
    ed->RemoveDialog(&dialog);
}

/*****************************************************************************/

ClassId AliasCmd::GetClassId () { return ALIAS_CMD; }
boolean AliasCmd::IsA (ClassId id) {return ALIAS_CMD==id || Command::IsA(id);}

AliasCmd::AliasCmd (ControlInfo* c, GraphicComp* comp) : Command(c) {
    _comp = comp;
}

AliasCmd::AliasCmd (Editor* ed, GraphicComp* comp) : Command(ed) {
    _comp = comp;
}

Command* AliasCmd::Copy () {
    Command* newCmd;

    newCmd = new AliasCmd(CopyControlInfo(), GetGraphicComp());
    InitCopy(newCmd);
    return newCmd;
}

void AliasCmd::Execute () {
    SchemEditor* ed = (SchemEditor*) GetEditor();
    GraphicComp* comp = (_comp == nil) ? Command::GetGraphicComp() : _comp;

    if (!comp->IsA(ELEMENT_COMP)) {
        Warning(ed, "Element is a primitive and cannot be aliased.");

    } else {
        ElementComp* defn = (ElementComp*) comp;
        ToolPalette* tp = ed->GetToolPalette();
        DrawingPalette* dtp = ed->GetDrawingPalette();
        SchemEditor* newEd = new SchemEditor(tp, dtp, new AliasComp(defn));
        unidraw->Open(newEd);
    }
}

boolean AliasCmd::Reversible () { return false; }

void AliasCmd::Read (istream& in) {
    Command::Read(in);
    _comp = (GraphicComp*) unidraw->GetCatalog()->ReadComponent(in);
}

void AliasCmd::Write (ostream& out) {
    Command::Write(out); 
    unidraw->GetCatalog()->WriteComponent(GetGraphicComp(), out);
}

/*****************************************************************************/

ClassId NewToolCmd::GetClassId () { return NEWTOOL_CMD; }

boolean NewToolCmd::IsA (ClassId id) {
    return NEWTOOL_CMD == id || Command::IsA(id);
}

NewToolCmd::NewToolCmd (ControlInfo* c) : Command(c) { _dialog = nil; }
NewToolCmd::NewToolCmd (Editor* ed) : Command(ed) { _dialog = nil; }

Command* NewToolCmd::Copy () {
    Command* newCmd;

    newCmd = new NewToolCmd(CopyControlInfo());
    InitCopy(newCmd);
    return newCmd;
}

void NewToolCmd::Execute () {
    Editor* ed = GetEditor();

    if (_dialog == nil) {
        _dialog = new NewToolDialog(".");
    }
    ed->InsertDialog(_dialog);

    if (_dialog->Accept()) {
        ed->RemoveDialog(_dialog);
        const char* name = _dialog->Choice();
        InitNewTool(name, ((NewToolDialog*) _dialog)->Instantiating());

    } else {
        ed->RemoveDialog(_dialog);
    }        
}

static ControlInfo* CreateCtrlInfo (GraphicComp* origComp) {
    GraphicComp* comp = (GraphicComp*) origComp->Copy();
    GraphicView* view = (GraphicView*) comp->Create(COMPONENT_VIEW);
    comp->Attach(view);
    view->Update();
    Graphic* g = view->GetGraphic()->Copy();
    delete view;

    float l, b, r, t, scale;
    const float size = 1*cm;

    g->GetBounds(l, b, r, t);
    scale = min(size/(r - l), size/(t - b));
    ScaleCmd scaleCmd((Editor*) nil, scale, scale);
    comp->Interpret(&scaleCmd);

    return new ControlInfo(comp, "?");
}

void NewToolCmd::InitNewTool (const char* protoName, boolean instantiate) {
    ElementComp* protoComp;
    Catalog* catalog = unidraw->GetCatalog();

    if (catalog->Retrieve(protoName, (Component*&) protoComp)) {
        SchemEditor* scmEd = (SchemEditor*) GetEditor();
        protoComp = (ElementComp*) protoComp->Copy();

        ToolPalette* tools = scmEd->GetToolPalette();
        ControlInfo* ctrlInfo = CreateCtrlInfo(protoComp);
        Tool* tool = instantiate ?
            new InstanceTool(ctrlInfo, protoComp) :
            new GraphicCompTool(ctrlInfo, protoComp);

        char toolName[CHARBUFSIZE];
        strcpy(toolName, protoName);
        strcat(toolName, ".tool");

        tools->InstallOrRemove(toolName, tool);
    }
}

boolean NewToolCmd::Reversible () { return false; }

/*****************************************************************************/

ClassId ToolsCmd::GetClassId () { return TOOLS_CMD; }

boolean ToolsCmd::IsA (ClassId id) {
    return TOOLS_CMD == id || Command::IsA(id);
}

ToolsCmd::ToolsCmd (ControlInfo* c) : Command(c) { _dialog = nil; }
ToolsCmd::ToolsCmd (Editor* ed) : Command(ed) { _dialog = nil; }

Command* ToolsCmd::Copy () {
    Command* newCmd;

    newCmd = new ToolsCmd(CopyControlInfo());
    InitCopy(newCmd);
    return newCmd;
}

void ToolsCmd::Execute () {
    SchemEditor* scmEd = (SchemEditor*) GetEditor();
    ToolPalette* tools = scmEd->GetToolPalette();
    EditorInfo* edInfo = tools->GetEditorInfo();

    if (_dialog == nil) {
        _dialog = new UChooser(
            "", "Schematic tools in catalog:", " Install/Remove ", ""
        );
    }
    InitDialog();
    scmEd->InsertDialog(_dialog);

    if (_dialog->Accept()) {
        scmEd->RemoveDialog(_dialog);
        const char* toolName = _dialog->Choice();
        tools->InstallOrRemove(toolName);

    } else {
        scmEd->RemoveDialog(_dialog);
    }        
}

void ToolsCmd::InitDialog () {
    SchemEditor* scmEd = (SchemEditor*) GetEditor();
    ToolPalette* tools = scmEd->GetToolPalette();
    EditorInfo* edInfo = tools->GetEditorInfo();
    int count = (edInfo == nil) ? 0 : edInfo->Count();

    _dialog->Clear();

    for (int i = 0; i < count; ++i) {
        _dialog->Include(edInfo->GetName(i));
    }
}

boolean ToolsCmd::Reversible () { return false; }

/*****************************************************************************/

ClassId NewViewCmd::GetClassId () { return NEWVIEW_CMD; }

boolean NewViewCmd::IsA (ClassId id) {
    return NEWVIEW_CMD == id || Command::IsA(id);
}

NewViewCmd::NewViewCmd (ControlInfo* c, GraphicComp* comp) : Command(c) {
    _comp = comp;
}

NewViewCmd::NewViewCmd (Editor* ed, GraphicComp* comp) : Command(ed) {
    _comp = comp;
}

Command* NewViewCmd::Copy () {
    Command* newCmd;

    newCmd = new NewViewCmd(CopyControlInfo(), GetGraphicComp());
    InitCopy(newCmd);
    return newCmd;
}

void NewViewCmd::Execute () {
    SchemEditor* ed = (SchemEditor*) GetEditor();
    GraphicComp* comp = (_comp == nil) ? Command::GetGraphicComp() : _comp;

    if (!comp->IsA(GRAPHIC_COMPS)) {
        Warning(ed, "Element is a primitive and cannot be edited.");

    } else {
        ToolPalette* tp = ed->GetToolPalette();
        DrawingPalette* dtp = ed->GetDrawingPalette();
        SchemEditor* newEd = new SchemEditor(tp, dtp, comp);

        *newEd->GetState("ModifStatusVar") = *ed->GetState("ModifStatusVar");

        unidraw->Open(newEd);
    }
}

boolean NewViewCmd::Reversible () { return false; }

void NewViewCmd::Read (istream& in) {
    Command::Read(in);
    _comp = (GraphicComp*) unidraw->GetCatalog()->ReadComponent(in);
}

void NewViewCmd::Write (ostream& out) {
    Command::Write(out); 
    unidraw->GetCatalog()->WriteComponent(GetGraphicComp(), out);
}

/*****************************************************************************/

ClassId DefinitionCmd::GetClassId () { return DEFINITION_CMD; }

boolean DefinitionCmd::IsA (ClassId id) {
    return DEFINITION_CMD == id || Command::IsA(id);
}

DefinitionCmd::DefinitionCmd (ControlInfo* c, GraphicComp* comp) : Command(c) {
    _comp = comp;
}

DefinitionCmd::DefinitionCmd (Editor* ed, GraphicComp* comp) : Command(ed) {
    _comp = comp;
}

Command* DefinitionCmd::Copy () {
    Command* newCmd;

    newCmd = new DefinitionCmd(CopyControlInfo(), GetGraphicComp());
    InitCopy(newCmd);
    return newCmd;
}

void DefinitionCmd::Execute () {
    SchemEditor* ed = (SchemEditor*) GetEditor();
    GraphicComp* comp = (_comp == nil) ? Command::GetGraphicComp() : _comp;

    if (!comp->IsA(ALIAS_COMP)) {
        Warning(ed, "Element is a primitive and has no definition.");

    } else {
        AliasComp* alias = (AliasComp*) comp;
        GraphicComp* defn = alias->GetDefinition();
        SchemEditor* newEd = (SchemEditor*) unidraw->Find(defn);
        
        if (newEd == nil) {
            ToolPalette* tp = ed->GetToolPalette();
            DrawingPalette* dtp = ed->GetDrawingPalette();
            newEd = new SchemEditor(tp, dtp, defn);
            const char* compName = unidraw->GetCatalog()->GetName(defn);

            if (compName != nil) {
                CompNameVar* cv = (CompNameVar*)newEd->GetState("CompNameVar");
                cv->SetName(compName);
            }
            unidraw->Open(newEd);

        } else {
            ed->GetWindow()->raise();
        }
    }
}

boolean DefinitionCmd::Reversible () { return false; }

void DefinitionCmd::Read (istream& in) {
    Command::Read(in);
    _comp = (GraphicComp*) unidraw->GetCatalog()->ReadComponent(in);
}

void DefinitionCmd::Write (ostream& out) {
    Command::Write(out); 
    unidraw->GetCatalog()->WriteComponent(GetGraphicComp(), out);
}

/*****************************************************************************/

ClassId ToggleDwgPaletteCmd::GetClassId () { return TOGGLEDWGPALETTE_CMD; }

boolean ToggleDwgPaletteCmd::IsA (ClassId id) {
    return TOGGLEDWGPALETTE_CMD == id || Command::IsA(id);
}

ToggleDwgPaletteCmd::ToggleDwgPaletteCmd (
    ControlInfo* c, boolean visible
) : Command(c) { 
    _visible = visible;
}

ToggleDwgPaletteCmd::ToggleDwgPaletteCmd (
    Editor* ed, boolean visible
) : Command(ed) {
    _visible = visible;
}

Command* ToggleDwgPaletteCmd::Copy () {
    Command* newCmd;

    newCmd = new ToggleDwgPaletteCmd(CopyControlInfo());
    InitCopy(newCmd);
    return newCmd;
}

void ToggleDwgPaletteCmd::Execute () {
    SchemEditor* scmed = (SchemEditor*) GetEditor();
    DrawingPalette* dwgPal = scmed->GetDrawingPalette();

    if (unidraw->Opened(dwgPal)) {
        dwgPal->Iconify(_visible);
        _visible = !_visible;

    } else {
        unidraw->Open(dwgPal);
        _visible = true;
    }
}

boolean ToggleDwgPaletteCmd::Reversible () { return false; }

/*****************************************************************************/

ClassId NetlistCmd::GetClassId () { return NETLIST_CMD; }

boolean NetlistCmd::IsA (ClassId id) {
    return NETLIST_CMD == id || Command::IsA(id);
}

NetlistCmd::NetlistCmd (ControlInfo* c) : Command(c) { _dialog = nil; }
NetlistCmd::NetlistCmd (Editor* ed) : Command(ed) { _dialog = nil; }
NetlistCmd::~NetlistCmd () { if (_dialog != nil) delete _dialog; }

Command* NetlistCmd::Copy () {
    Command* newCmd = new NetlistCmd(CopyControlInfo());
    InitCopy(newCmd);
    return newCmd;
}

void NetlistCmd::Execute () {
    GraphicComp* comps = GetGraphicComp();
    filebuf fbuf;
    boolean ok;

    if (_dialog == nil) {
        _dialog = new NetlistDialog;
    }

    _editor->InsertDialog(_dialog);

    do {
        if (!_dialog->Accept()) {
            break;
        }
        ok = fbuf.open((char*) _dialog->Name(), output) != 0;

        if (ok) {
            ostream out(&fbuf);
            NetlistView* ev = (NetlistView*) comps->Create(NETLIST_VIEW);
            comps->Attach(ev);
            ev->Flatten(_dialog->Flattening());
            ev->Update();
            ok = ev->Emit(out);
            delete ev;

        } else {
            _dialog->SetTitle("Couldn't output!");
        }
    } while (!ok);

    _editor->RemoveDialog(_dialog);
    _dialog->SetTitle("");
}

boolean NetlistCmd::Reversible () { return false; }

/*****************************************************************************/

ClassId InfoCmd::GetClassId () { return INFO_CMD; }

boolean InfoCmd::IsA (ClassId id) {
    return INFO_CMD == id || Command::IsA(id);
}

InfoCmd::InfoCmd (ControlInfo* c, GraphicComp* comp) : Command(c) {
    _comp = comp;
}

InfoCmd::InfoCmd (Editor* ed, GraphicComp* comp) : Command(ed) { _comp = comp;}

Command* InfoCmd::Copy () {
    Command* newCmd = new InfoCmd(CopyControlInfo(), GetGraphicComp());
    InitCopy(newCmd);
    return newCmd;
}

void InfoCmd::Execute () {
    Editor* ed = GetEditor();
    GraphicComp* comp = (_comp == nil) ? Command::GetGraphicComp() : _comp;

    if (comp->IsA(ELEMENT_COMP) || comp->IsA(NODE_COMP)) {
        Modify(comp);

    } else {
        Warning(ed, "Element has no schematic information.");
    }
}

void InfoCmd::Modify (GraphicComp* comp) {
    Editor* ed = GetEditor();

    InfoDialog info(comp);
    ed->InsertDialog(&info);

    if (info.Accept()) {
        ed->RemoveDialog(&info);
        info.ApplyChanges();
        comp->Notify();
        unidraw->Update();

    } else {
        ed->RemoveDialog(&info);
    }
}

boolean InfoCmd::Reversible () { return false; }

void InfoCmd::Read (istream& in) {
    Command::Read(in);
    _comp = (GraphicComp*) unidraw->GetCatalog()->ReadComponent(in);
}

void InfoCmd::Write (ostream& out) {
    Command::Write(out); 
    unidraw->GetCatalog()->WriteComponent(GetGraphicComp(), out);
}

/*****************************************************************************/

ClassId TruthTableCmd::GetClassId () { return TRUTHTABLE_CMD; }

boolean TruthTableCmd::IsA (ClassId id) {
    return TRUTHTABLE_CMD == id || Command::IsA(id);
}

TruthTableCmd::TruthTableCmd (ControlInfo* c, GraphicComp* comp) : Command(c) {
    _comp = comp;
}

TruthTableCmd::TruthTableCmd (Editor* ed, GraphicComp* comp) : Command(ed) {
    _comp = comp;
}

Command* TruthTableCmd::Copy () {
    Command* newCmd = new TruthTableCmd(CopyControlInfo(), GetGraphicComp());
    InitCopy(newCmd);
    return newCmd;
}

void TruthTableCmd::Execute () {
    Editor* ed = GetEditor();
    GraphicComp* comp = (_comp == nil) ? Command::GetGraphicComp() : _comp;

    if (comp->IsA(ELEMENT_COMP)) {
        comp->Interpret(this);

    } else {
        Warning(ed, "Element cannot define a truth table.");
    }
}

boolean TruthTableCmd::Reversible () { return false; }

void TruthTableCmd::Read (istream& in) {
    Command::Read(in);
    _comp = (GraphicComp*) unidraw->GetCatalog()->ReadComponent(in);
}

void TruthTableCmd::Write (ostream& out) {
    Command::Write(out); 
    unidraw->GetCatalog()->WriteComponent(GetGraphicComp(), out);
}
