/*
 * Copyright (c) 1990 Stanford University
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided
 * that the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of Stanford not be used in advertising or
 * publicity pertaining to distribution of the software without specific,
 * written prior permission.  Stanford makes no representations about
 * the suitability of this software for any purpose.  It is provided "as is"
 * without express or implied warranty.
 *
 * STANFORD DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS.
 * IN NO EVENT SHALL STANFORD BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 * DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
 * OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION
 * WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

/*
 * Schematic capture system-specific component definitions.
 */

#include "scmclasses.h"
#include "scmcomps.h"
#include "scmdialogs.h"
#include "scmmanips.h"
#include "scmtransfns.h"
#include "scmtruth.h"
#include "scmvars.h"

#include <Unidraw/catalog.h>
#include <Unidraw/clipboard.h>
#include <Unidraw/editor.h>
#include <Unidraw/iterator.h>
#include <Unidraw/manips.h>
#include <Unidraw/selection.h>
#include <Unidraw/transfns.h>
#include <Unidraw/unidraw.h>
#include <Unidraw/viewer.h>
#include <Unidraw/Commands/edit.h>
#include <Unidraw/Commands/macro.h>
#include <Unidraw/Commands/transforms.h>
#include <Unidraw/Graphic/lines.h>
#include <Unidraw/Graphic/picture.h>
#include <Unidraw/Graphic/ustencil.h>
#include <Unidraw/Graphic/util.h>
#include <Unidraw/Tools/tool.h>

#include <InterViews/bitmap.h>
#include <InterViews/event.h>
#include <InterViews/rubcurve.h>
#include <InterViews/rubgroup.h>
#include <InterViews/transformer.h>

#include <math.h>
#include <stdlib.h>
#include <stream.h>
#include <string.h>

#include "bulboff.bm"
#include "bulboffMask.bm"
#include "bulbon.bm"
#include "bulbonMask.bm"
#include "swchoff.bm"
#include "swchoffMask.bm"
#include "swchon.bm"
#include "swchonMask.bm"

/*****************************************************************************/

ElementComp::ElementComp () {
    _elemVar = new ElementVar("elem");
    _transfn = nil;
}

ElementComp::~ElementComp () {
    delete _elemVar;
    if (_transfn != nil) delete _transfn;
}

void ElementComp::Interpret (Command* cmd) {
    if (cmd->IsA(TRUTHTABLE_CMD)) {
        Editor* ed = cmd->GetEditor();
        TruthTable* tt = CreateTruthTable();

        if (tt == nil) {
            AcknowledgeDialog dialog(
                "Schematic has no inputs and/or outputs."
            );
            ed->InsertDialog(&dialog);
            dialog.Acknowledge();
            ed->RemoveDialog(&dialog);
            
        } else {
            TruthTableDialog dialog(tt);
            ed->InsertDialog(&dialog);

            if (dialog.Accept()) {
                ed->RemoveDialog(&dialog);
                InterpretTruthTable(tt);

            } else {
                ed->RemoveDialog(&dialog);
            }
        }

    } else {
        GraphicComps::Interpret(cmd);
    }
}

ElementVar* ElementComp::GetElemVar () { return _elemVar; }
TransferFunct* ElementComp::GetTransferFunct () { return _transfn; }

NodeVar* ElementComp::GetNodeVar (int n) {
    Iterator i;
    int j = 0;

    for (First(i); !Done(i); Next(i)) {
        if (GetComp(i)->IsA(NODE_COMP)) {
            NodeComp* nodeComp = (NodeComp*) GetComp(i);

            if (j++ == n) {
                return nodeComp->GetNodeVar();
            }
        }
    }
    return nil;
}

StateVar* ElementComp::GetState (const char* name) {
    StateVar* stateVar = nil;

    if (strcmp(name, "ElementVar") == 0) {
        stateVar = _elemVar;

    } else if (strlen(name) >= strlen("NodeVar")) {
        char buf[CHARBUFSIZE];
        strcpy(buf, name);
        int suffix = strlen("NodeVar");
        buf[suffix] = '\0';

        if (strcmp(buf, "NodeVar") == 0) {
            stateVar = GetNodeVar(atoi(&name[suffix]));
        }
    }

    return stateVar;
}

int ElementComp::NodeCount () {
    Iterator i;
    int j = 0;

    for (First(i); !Done(i); Next(i)) {
        if (GetComp(i)->IsA(NODE_COMP)) {
            ++j;
        }
    }
    return j;
}

void ElementComp::Read (istream& in) {
    GraphicComps::Read(in);
    Catalog* catalog = unidraw->GetCatalog();

    delete _elemVar;
    _elemVar = (ElementVar*) catalog->ReadStateVar(in);
    _transfn = (TF_TruthTable*) catalog->ReadTransferFunct(in);
}

void ElementComp::Write (ostream& out) {
    GraphicComps::Write(out);
    Catalog* catalog = unidraw->GetCatalog();

    catalog->WriteStateVar(_elemVar, out);
    catalog->WriteTransferFunct(_transfn, out);
}

ClassId ElementComp::GetClassId () { return ELEMENT_COMP; }

boolean ElementComp::IsA (ClassId id) {
    return ELEMENT_COMP == id || GraphicComps::IsA(id);
}

static void IOCount (ElementComp* elem, int& inputs, int& outputs) {
    Iterator i;
    inputs = outputs = 0;

    for (elem->First(i); !elem->Done(i); elem->Next(i)) {
        if (elem->GetComp(i)->IsA(NODE_COMP)) {
            NodeComp* nodeComp = (NodeComp*) elem->GetComp(i);
            NodeVar* nodeVar = nodeComp->GetNodeVar();
            TransMethod transMethod = nodeComp->GetTransMethod();

            if (transMethod == In) {
                ++inputs;
            } else if (transMethod == Out) {
                ++outputs;
            }
        }
    }
}

TruthTable* ElementComp::CreateTruthTable () {
    TruthTable* tt = nil;

    if (_transfn == nil) {
        Iterator i;
        int inputs = 0, outputs = 0;
        IOCount(this, inputs, outputs);

        if (inputs > 0 && outputs > 0) {
            int input = 0, output = 0; 
            tt = new TruthTable(inputs, outputs);

            for (First(i); !Done(i); Next(i)) {
                if (GetComp(i)->IsA(NODE_COMP)) {
                    NodeComp* nodeComp = (NodeComp*) GetComp(i);
                    NodeVar* nodeVar = nodeComp->GetNodeVar();
                    TransMethod transMethod = nodeComp->GetTransMethod();

                    if (transMethod == In) {
                        tt->SetInputVar(nodeVar, input++);

                    } else if (transMethod == Out) {
                        tt->SetOutputVar(nodeVar, output++);
                    }
                }
            }
        }
    } else {
        tt = _transfn->GetTruthTable()->Copy();
    }
    return tt;
}

void ElementComp::InterpretTruthTable (TruthTable* tt) {
    if (_transfn != nil) {
        delete _transfn;
    }
    _transfn = new TF_TruthTable(tt->Copy());
    Notify();
    unidraw->Update();
}

/*****************************************************************************/

ElementView::ElementView (ElementComp* subj) : GraphicViews(subj) { }
ElementComp* ElementView::GetElemComp () { return (ElementComp*) GetSubject();}
ClassId ElementView::GetClassId () { return ELEMENT_VIEW; }

boolean ElementView::IsA (ClassId id) {
    return ELEMENT_VIEW == id || GraphicViews::IsA(id);
}

/*****************************************************************************/

AliasComp::AliasComp (ElementComp* defn) {
    _defn = defn;

    if (_defn != nil) {
        Iterator i;
        
        for (defn->First(i); !defn->Done(i); defn->Next(i)) {
            GraphicComp* comp = _defn->GetComp(i);

            if (comp->IsA(NODE_COMP)) {
                Append((GraphicComp*) comp->Copy());
            }
        }
    }        
}

void AliasComp::Read (istream& in) {
    ElementComp::Read(in);
    _defn = (ElementComp*) unidraw->GetCatalog()->ReadComponent(in);
}

void AliasComp::Write (ostream& out) {
    ElementComp::Write(out);
    unidraw->GetCatalog()->WriteComponent(_defn, out);
}

ClassId AliasComp::GetClassId () { return ALIAS_COMP; }

boolean AliasComp::IsA (ClassId id) {
    return ALIAS_COMP == id || ElementComp::IsA(id);
}

/*****************************************************************************/

AliasView::AliasView (AliasComp* subj) : ElementView(subj) { }
AliasComp* AliasView::GetAliasComp () { return (AliasComp*) GetSubject(); }
ClassId AliasView::GetClassId () { return ALIAS_VIEW; }

boolean AliasView::IsA (ClassId id) {
    return ALIAS_VIEW == id || ElementView::IsA(id);
}

/*****************************************************************************/

NodeComp::NodeComp (PinGraphic* pingr) : PinComp(pingr) {
    NodeComp::SetBinding(new NodeVar("node")); 
}

Component* NodeComp::Copy () { 
    NodeComp* copy = new NodeComp((PinGraphic*) GetGraphic()->Copy());
    copy->_mobility = _mobility;
    copy->SetTransMethod(GetTransMethod());
    *copy->GetNodeVar() = *GetNodeVar();
    return copy;
}

ClassId NodeComp::GetClassId () { return NODE_COMP; }

boolean NodeComp::IsA (ClassId id) {
    return NODE_COMP == id || PinComp::IsA(id);
}

NodeVar* NodeComp::GetNodeVar () { return (NodeVar*) GetBinding(); }

/*****************************************************************************/

NodeView::NodeView (NodeComp* subj) : PinView(subj) { }

ClassId NodeView::GetClassId () { return NODE_VIEW; }

boolean NodeView::IsA (ClassId id) {
    return NODE_VIEW == id || PinView::IsA(id);
}

PinComp* NodeView::NewSubject (PinGraphic* pingr) {return new NodeComp(pingr);}

/*****************************************************************************/

static MultiLine* CreateMultiLine (Coord* x, Coord* y, int n) {
    int last = n-1;
    Transformer* t = new Transformer;

    if (x[0] == x[last]) {
        ++x[last];
        t->Translate(-x[0], 0);
        t->Scale(0, 1);
        t->Translate(x[0], 0);
    }
    if (y[0] == y[last]) {
        ++y[last];
        t->Translate(0, -y[0]);
        t->Scale(1, 0);
        t->Translate(0, y[0]);
    }
    MultiLine* ml = new S_MultiLine(x, y, n, stdgraphic);
    ml->SetTransformer(t);
    Unref(t);
    return ml;
}

WireComp::WireComp () { }

WireComp::WireComp (Coord* x, Coord* y, int n, Graphic* gr) {
    PinGraphic* pin1 = new PinGraphic(x[0], y[0]);
    PinGraphic* pin2 = new PinGraphic(x[n-1], y[n-1]);

    pin1->SetBrush(psnonebr);
    pin2->SetBrush(psnonebr);

    _node1 = new NodeComp(pin1);
    _node2 = new NodeComp(pin2);
    
    Graphic* pict = new Picture(gr);
    MultiLine* ml = CreateMultiLine(x, y, n);
    pict->Append(ml, pin1, pin2);
    SetGraphic(pict);

    _transfn = new TF_Direct(_node1->GetNodeVar(), _node2->GetNodeVar());
}

WireComp::~WireComp () {
    delete _node1;
    delete _node2;    
    delete _transfn;
}

void WireComp::Interpret (Command* cmd) {
    if (cmd->IsA(DELETE_CMD) || cmd->IsA(CUT_CMD)) {
        _node1->Interpret(cmd);
        _node2->Interpret(cmd);

    } else {
        GraphicComp::Interpret(cmd);
    }
}

void WireComp::Uninterpret (Command* cmd) {
    if (cmd->IsA(DELETE_CMD) || cmd->IsA(CUT_CMD)) {
        _node1->Uninterpret(cmd);
        _node2->Uninterpret(cmd);

    } else {
        GraphicComp::Uninterpret(cmd);
    }
}

void WireComp::Update () {
    MultiLine* ml = GetMultiLine();
    const Coord* x, *y;
    int n = ml->GetOriginal(x, y);
    int last = n-1;

    Transformer context, total(ml->GetTransformer());
    ml->Parent()->TotalTransformation(context);
    total.Postmultiply(&context);

    float cx1, cx2, mlx1, mlx2, cy1, cy2, mly1, mly2;
    _node1->GetCenter(cx1, cy1);
    _node2->GetCenter(cx2, cy2);
    total.Transform(float(x[0]), float(y[0]), mlx1, mly1);
    total.Transform(float(x[last]), float(y[last]), mlx2, mly2);

    if (
        round(cx1) != round(mlx1) || round(cx2) != round(mlx2) ||
        round(cy1) != round(mly1) || round(cy2) != round(mly2)
    ) {
        context.Invert();
        Transformer* t = new Transformer;
        ml->SetTransformer(t);
        Unref(t);

        float sx = (cx2 - cx1) / float(x[last] - x[0]);
        float sy = (cy2 - cy1) / float(y[last] - y[0]);

        t->Translate(-x[0], -y[0]);
        t->Scale(sx, sy);
        t->Translate(cx1, cy1);
        t->Postmultiply(&context);

        Notify();
    }
}

void WireComp::First (Iterator& i) { i.SetValue(_node1); }
void WireComp::Last (Iterator& i) { i.SetValue(_node2); }
boolean WireComp::Done (Iterator i) { return i.GetValue() == nil; }

void WireComp::Next (Iterator& i) { 
    void* v = i.GetValue();

    if (v == nil) {
        i.SetValue(_node1);
    } else if (v == _node1) {
        i.SetValue(_node2);
    } else {
        i.SetValue(nil);
    }
}

void WireComp::Prev (Iterator& i) { 
    void* v = i.GetValue();

    if (v == nil) {
        i.SetValue(_node2);
    } else if (v == _node1) {
        i.SetValue(nil);
    } else {
        i.SetValue(_node1);
    }
}

TransferFunct* WireComp::GetTransferFunct () { return _transfn; }
GraphicComp* WireComp::GetComp (Iterator i){return (GraphicComp*)i.GetValue();}
void WireComp::SetComp (GraphicComp* gc, Iterator& i) { i.SetValue(gc); }

void WireComp::SetMobility (Mobility m) {
    _node1->SetMobility(m);
    _node2->SetMobility(m);
}

void WireComp::GetNodes (NodeComp*& node1, NodeComp*& node2) {
    node1 = _node1;
    node2 = _node2;
}

MultiLine* WireComp::GetMultiLine () {
    Iterator i;
    Graphic* gr = GetGraphic();
    gr->First(i);
    return (MultiLine*) gr->GetGraphic(i);
}

Component* WireComp::Copy () { return unidraw->GetCatalog()->Copy(this); }

void WireComp::Read (istream& in) {
    GraphicComp::Read(in);
    Catalog* catalog = unidraw->GetCatalog();
    Coord* x, *y;
    int count;

    ReadVertices(in, x, y, count);
    S_MultiLine* ml = new S_MultiLine(x, y, count);
    delete x;
    delete y;
    PSColor* fg = ReadColor(in);
    PSColor* bg = ReadColor(in);
    ml->SetColors(fg, bg);
    ml->SetBrush(ReadBrush(in));
    ml->SetTransformer(ReadTransformer(in));

    _node1 = (NodeComp*) catalog->ReadComponent(in);
    _node2 = (NodeComp*) catalog->ReadComponent(in);
    _transfn = (TF_Direct*) catalog->ReadTransferFunct(in);

    Graphic* pict = new Picture;
    pict->FillBg(ReadBgFilled(in));
    fg = ReadColor(in);
    bg = ReadColor(in);
    pict->SetColors(fg, bg);
    pict->SetBrush(ReadBrush(in));
    pict->SetTransformer(ReadTransformer(in));

    pict->Append(ml, _node1->GetGraphic(), _node2->GetGraphic());
    SetGraphic(pict);
}

void WireComp::Write (ostream& out) {
    GraphicComp::Write(out);
    Catalog* catalog = unidraw->GetCatalog();
    MultiLine* ml = GetMultiLine();
    const Coord* x, *y;
    int count = ml->GetOriginal(x, y);

    WriteVertices(x, y, count, out);

    WriteColor(ml->GetFgColor(), out);
    WriteColor(ml->GetBgColor(), out);
    WriteBrush(ml->GetBrush(), out);
    WriteTransformer(ml->GetTransformer(), out);

    catalog->WriteComponent(_node1, out);
    catalog->WriteComponent(_node2, out);
    catalog->WriteTransferFunct(_transfn, out);

    Graphic* pict = GetGraphic();
    WriteBgFilled(pict->BgFilled(), out);
    WriteColor(pict->GetFgColor(), out);
    WriteColor(pict->GetBgColor(), out);
    WriteBrush(pict->GetBrush(), out);
    WriteTransformer(pict->GetTransformer(), out);
}

ClassId WireComp::GetClassId () { return WIRE_COMP; }

boolean WireComp::IsA (ClassId id) {
    return WIRE_COMP == id || GraphicComp::IsA(id);
}

/*****************************************************************************/

WireView::WireView (WireComp* subj) : GraphicView(subj) {
    _nodeView1 = _nodeView2 = nil;
}

WireView::~WireView () {
    if (_nodeView1 != nil) {
        Graphic* parent = GetGraphic();
        parent->Remove(_nodeView1->GetGraphic());
        parent->Remove(_nodeView2->GetGraphic());

        delete _nodeView1;
        delete _nodeView2;
    }
}

void WireView::Update () {
    WireComp* wireComp = GetWireComp();
    Graphic* wire = GetGraphic();

    IncurDamage(wire);
    *wire = *wireComp->GetGraphic();
    UpdateMultiLine();
    IncurDamage(wire);
    EraseHandles();
}

Manipulator* WireView::CreateManipulator (
    Viewer* v, Event& e, Transformer* rel, Tool* tool
) {
    Manipulator* m = nil;

    if (tool->IsA(GRAPHIC_COMP_TOOL)) {
        m = CreateWireCompManip(v, e, rel, tool);

    } else if (tool->IsA(MOVE_TOOL)) {
        Coord* x, *y;
        int n;
        GetVertices(x, y, n);
	Rubberband* rub = new SlidingLineList(nil, nil, x, y, n, e.x, e.y);
        delete x;
        delete y;
        m = new DragManip(v, rub, rel, tool);

    } else if (tool->IsA(RESHAPE_TOOL)) {
        // unimplemented

    }
    return m;
}

Command* WireView::InterpretManipulator (Manipulator* m) {
    DragManip* dm = (DragManip*) m;
    Editor* ed = dm->GetViewer()->GetEditor();
    Tool* tool = dm->GetTool();
    Transformer* rel = dm->GetTransformer();
    Command* cmd = nil;

    if (tool->IsA(GRAPHIC_COMP_TOOL)) {
        cmd = InterpWireCompManip(dm);

    } else if (tool->IsA(MOVE_TOOL)) {
        int dummy;
        Coord* xorig, *yorig, *xcur, *ycur;
        float fx0, fy0, fx1, fy1;

        SlidingLineList* sll = (SlidingLineList*) dm->GetRubberband();
        sll->GetOriginal(xorig, yorig, dummy);
        sll->GetCurrent(xcur, ycur, dummy);

        if (rel != nil) {
            rel->InvTransform(float(xorig[0]), float(yorig[0]), fx0, fy0);
            rel->InvTransform(float(xcur[0]), float(ycur[0]), fx1, fy1);
        }
        delete xorig; delete yorig; delete xcur; delete ycur;

        cmd = new MoveCmd(ed, fx1-fx0, fy1-fy0);

    } else if (tool->IsA(RESHAPE_TOOL)) {
        // unimplemented
    }
    return cmd;
}

void WireView::First (Iterator& i) { i.SetValue(_nodeView1); }
void WireView::Last (Iterator& i) { i.SetValue(_nodeView2); }
boolean WireView::Done (Iterator i) { return i.GetValue() == nil; }

void WireView::Next (Iterator& i) { 
    void* v = i.GetValue();

    if (v == nil) {
        i.SetValue(_nodeView1);
    } else if (v == _nodeView1) {
        i.SetValue(_nodeView2);
    } else {
        i.SetValue(nil);
    }
}

void WireView::Prev (Iterator& i) { 
    void* v = i.GetValue();

    if (v == nil) {
        i.SetValue(_nodeView2);
    } else if (v == _nodeView1) {
        i.SetValue(nil);
    } else {
        i.SetValue(_nodeView1);
    }
}

GraphicView* WireView::GetView (Iterator i){return (GraphicView*)i.GetValue();}
void WireView::SetView (GraphicView* gv, Iterator& i) { i.SetValue(gv); }
WireComp* WireView::GetWireComp () { return (WireComp*) GetSubject(); }

Graphic* WireView::GetGraphic () {
    Graphic* gr = GraphicView::GetGraphic();

    if (gr == nil) {
        WireComp* wireComp = GetWireComp();
        gr = new Picture(wireComp->GetGraphic());
        SetGraphic(gr);

        NodeComp* n1, *n2;
        wireComp->GetNodes(n1, n2);
        _nodeView1 = (NodeView*) n1->Create(COMPONENT_VIEW);
        _nodeView2 = (NodeView*) n2->Create(COMPONENT_VIEW);

        n1->Attach(_nodeView1);
        n2->Attach(_nodeView2);
        _nodeView1->Update();
        _nodeView2->Update();

        Graphic* ml = wireComp->GetMultiLine()->Copy();
        gr->Append(ml, _nodeView1->GetGraphic(), _nodeView2->GetGraphic());
    }
    return gr;
}

ClassId WireView::GetClassId () { return WIRE_VIEW; }
boolean WireView::IsA (ClassId id){return WIRE_VIEW==id||GraphicView::IsA(id);}

void WireView::CreateHandles () {
    Coord x[2], y[2];
    Viewer* v = GetViewer();
    
    if (v != nil) {
        GetEndpoints(x[0], y[0], x[1], y[1]);
        _handles = new RubberHandles(nil, nil, x, y, 2, 0, HANDLE_SIZE);
        v->InitRubberband(_handles);
    }
}

MultiLine* WireView::GetMultiLine () {
    Iterator i;
    Graphic* gr = GetGraphic();

    gr->First(i);
    return (MultiLine*) gr->GetGraphic(i);
}

void WireView::GetEndpoints (Coord& x0, Coord& y0, Coord& x1, Coord& y1) {
    float x, y;

    _nodeView1->GetGraphic()->GetCenter(x, y);
    x0 = round(x); y0 = round(y);

    _nodeView2->GetGraphic()->GetCenter(x, y);
    x1 = round(x); y1 = round(y);
}

void WireView::GetVertices (Coord*& x, Coord*& y, int& n) {
    const Coord* vx, *vy;
    Transformer total;
    MultiLine* ml = GetMultiLine();

    n = ml->GetOriginal(vx, vy);
    ml->TotalTransformation(total);
    ArrayDup(vx, vy, n, x, y);
    total.TransformList(x, y, n);
}

void WireView::UpdateMultiLine () {
    Graphic* gr = GetGraphic();
    MultiLine* ml = GetMultiLine();

    gr->Remove(ml);
    delete ml;
    gr->Prepend(GetWireComp()->GetMultiLine()->Copy());
}

Manipulator* WireView::CreateWireCompManip (
    Viewer* v, Event& e, Transformer* rel, Tool* tool
) {
    Selection* s = v->GetSelection();
    GraphicView* views = v->GetGraphicView();
    ConnectorView* target = views->ConnectorIntersecting(
        e.x-SLOP, e.y-SLOP, e.x+SLOP, e.y+SLOP
    );
    Coord x[1], y[1];
    x[0] = e.x;
    y[0] = e.y;

    s->Clear();

    if (target != nil) {
        float fx, fy, tx, ty;
        target->GetConnector()->GetCenter(fx, fy);
        rel->Transform(fx, fy, tx, ty);
        x[0] = round(tx);
        y[0] = round(ty);
    }
    RubberWire* rw = new RubberWire(nil, nil, x, y, 1);
    return new WireManip(v, rw, rel, tool);
}

Command* WireView::InterpWireCompManip (Manipulator* m) {
    Viewer* v = m->GetViewer();
    Editor* ed = v->GetEditor();
    GraphicView* view = v->GetGraphicView();
    BrushVar* brVar = (BrushVar*) ed->GetState("BrushVar");
    WireManip* wm = (WireManip*) m;
    Transformer* rel = wm->GetTransformer();
    RubberWire* rw = (RubberWire*) wm->GetRubberband();
    MacroCmd* macroCmd = new MacroCmd(ed);
    ConnectorView* target1, *target2;

    if (rel != nil) {
        rel = new Transformer(rel);
        rel->Invert();
    }

    FullGraphic gs(stdgraphic);
    if (brVar != nil) {
        gs.SetBrush(brVar->GetBrush());
    }
    gs.SetTransformer(rel);

    Coord* x, *y;
    int n;
    rw->GetCurrent(x, y, n);
    WireComp* wireComp = new WireComp(x, y, n, &gs);
    Coord x0 = x[0], y0 = y[0];
    Coord x1 = x[n-1], y1 = y[n-1];
    delete x;
    delete y;
    NodeComp* n1, *n2;
    wireComp->GetNodes(n1, n2);

    macroCmd->Append(new PasteCmd(ed, new Clipboard(wireComp)));

    target1 = view->ConnectorIntersecting(x0-SLOP, y0-SLOP, x0+SLOP, y0+SLOP);
    target2 = view->ConnectorIntersecting(x1-SLOP, y1-SLOP, x1+SLOP, y1+SLOP);

    if (target1 != nil) {
        macroCmd->Append(new ConnectCmd(ed, n1, target1->GetConnector()));
    }
    if (target2 != nil) {
        macroCmd->Append(new ConnectCmd(ed, n2, target2->GetConnector()));
    }
    return macroCmd;
}

/*****************************************************************************/

LogicComp::LogicComp () { }

LogicComp::LogicComp (PinGraphic* pin) {
    _nodeComp = new NodeComp(pin);
    _nodeComp->SetMobility(Fixed);

    Graphic* pict = new Picture;
    pict->Append(pin);
    SetGraphic(pict);
}

LogicComp::~LogicComp () {
    Graphic* parent = GetGraphic();

    if (parent != nil) {
        parent->Remove(_nodeComp->GetGraphic());
    }
    delete _nodeComp;
}

void LogicComp::Update () { Notify(); }
void LogicComp::First (Iterator& i) { i.SetValue(_nodeComp); }
void LogicComp::Last (Iterator& i) { i.SetValue(_nodeComp); }

void LogicComp::Next (Iterator& i) {
    void* v = i.GetValue();

    if (v == nil) {
        i.SetValue(_nodeComp);
    } else {
        i.SetValue(nil);
    }
}

void LogicComp::Prev (Iterator& i) { Next(i); }
boolean LogicComp::Done (Iterator i) { return i.GetValue() == nil; }
GraphicComp* LogicComp::GetComp (Iterator i){return(GraphicComp*)i.GetValue();}
void LogicComp::SetComp (GraphicComp* gc, Iterator& i) { i.SetValue(gc); }

void LogicComp::Read (istream& in) {
    GraphicComp::Read(in);
    Catalog* catalog = unidraw->GetCatalog();
    Graphic* pict = new Picture;

    _nodeComp = (NodeComp*) catalog->ReadComponent(in);
    pict->SetTransformer(catalog->ReadTransformer(in));
    pict->Append(_nodeComp->GetGraphic());
    SetGraphic(pict);
}

void LogicComp::Write (ostream& out) {
    GraphicComp::Write(out);
    Catalog* catalog = unidraw->GetCatalog();

    catalog->WriteComponent(_nodeComp, out);
    catalog->WriteTransformer(GetGraphic()->GetTransformer(), out);
}

ClassId LogicComp::GetClassId () { return LOGIC_COMP; }

boolean LogicComp::IsA (ClassId id) {
    return LOGIC_COMP == id || GraphicComp::IsA(id);
}

/*****************************************************************************/

LogicView::LogicView (LogicComp* subj) : GraphicView(subj) { _nodeView = nil; }

LogicView::~LogicView () {
    Graphic* parent = GraphicView::GetGraphic();

    if (parent != nil) {
        parent->Remove(_nodeView->GetGraphic());
    }
    delete _nodeView;
}

void LogicView::First (Iterator& i) { i.SetValue(_nodeView); }
void LogicView::Last (Iterator& i) { i.SetValue(_nodeView); }

void LogicView::Next (Iterator& i) {
    void* v = i.GetValue();

    if (v == nil) {
        i.SetValue(_nodeView);
    } else {
        i.SetValue(nil);
    }
}

void LogicView::Prev (Iterator& i) { Next(i); }
boolean LogicView::Done (Iterator i) { return i.GetValue() == nil; }
LogicComp* LogicView::GetLogicComp () { return (LogicComp*) GetSubject(); }
void LogicView::SetView (GraphicView* gv, Iterator& i) { i.SetValue(gv); }

GraphicView* LogicView::GetView (Iterator i) {
    return (GraphicView*) i.GetValue(); 
}

UStencil* LogicView::CreateStencil () { return nil; }

void LogicView::AlignStencil () {
    Graphic* nodegr = _nodeView->GetGraphic();
    Graphic* stencil = GetStencil();
    nodegr->Align(BottomCenter, stencil, BottomCenter);
}

UStencil* LogicView::GetStencil () {
    Graphic* gr = GetGraphic();
    Iterator i;
    gr->First(i);
    gr->Next(i);
    return (UStencil*) gr->GetGraphic(i);
}

Graphic* LogicView::GetGraphic () {
    Graphic* gr = GraphicView::GetGraphic();

    if (gr == nil) {
        LogicComp* logicComp = GetLogicComp();

        NodeComp* nodeComp = logicComp->GetNodeComp();
        _nodeView = (NodeView*) nodeComp->Create(COMPONENT_VIEW);
        nodeComp->Attach(_nodeView);
        _nodeView->Update();

        Graphic* stencil = CreateStencil();
        Graphic* nodegr = _nodeView->GetGraphic();
        gr = new Picture;
        gr->Append(nodegr, stencil);
        
        SetGraphic(gr);
    }
    return gr;
}

void LogicView::Update () {
    LogicComp* logicComp = GetLogicComp();
    Graphic* gr = GetGraphic();
    Graphic* oldStencil = GetStencil();
    Graphic* newStencil = CreateStencil();
    Iterator i;

    IncurDamage(gr);
    gr->SetGraphic(oldStencil, i);
    gr->InsertBefore(i, newStencil);
    gr->Remove(i);
    delete oldStencil;

    *gr = *logicComp->GetGraphic();
    _nodeView->Update();
    AlignStencil();
    IncurDamage(gr);
    EraseHandles();
}

/*****************************************************************************/

BulbComp::BulbComp () { }

BulbComp::BulbComp (Coord x, Coord y) : LogicComp(
    new PinGraphic(x, y, stdgraphic)
) {
    NodeComp* node = GetNodeComp();
    node->GetPin()->SetBrush(psnonebr);
    node->SetTransMethod(In);
}

ClassId BulbComp::GetClassId () { return BULB_COMP; }
boolean BulbComp::IsA (ClassId id){return BULB_COMP==id || LogicComp::IsA(id);}
    
Component* BulbComp::Copy () {
    Coord x, y;
    PinGraphic* pin = _nodeComp->GetPin();
    pin->GetOriginal(x, y);

    BulbComp* copy = new BulbComp(x, y);
    Transformer* t = GetGraphic()->GetTransformer();
    copy->GetGraphic()->SetTransformer(new Transformer(t));
    return copy;
}

/*****************************************************************************/

BulbView::BulbView (BulbComp* subj) : LogicView(subj) { }
BulbComp* BulbView::GetBulbComp () { return (BulbComp*) GetSubject(); }

UStencil* BulbView::CreateStencil () {
    NodeComp* node = GetBulbComp()->GetNodeComp();
    NodeVar* nodeVar = node->GetNodeVar();
    Bitmap* image, *mask;

    if (nodeVar->GetLogicLevel() == Low) {
        image = new Bitmap(bulboff_bits, bulboff_width, bulboff_height);
        mask = new Bitmap(
            bulboffMask_bits, bulboffMask_width, bulboffMask_height
        );

    } else {
        image = new Bitmap(bulbon_bits, bulbon_width, bulbon_height);
        mask = new Bitmap(bulbonMask_bits, bulbonMask_width,bulbonMask_height);
    }
    return new UStencil(image, mask, stdgraphic);
}

/*****************************************************************************/

SwitchComp::SwitchComp () { }

SwitchComp::SwitchComp (Coord x, Coord y) : LogicComp(
    new PinGraphic(x, y,stdgraphic)
) {
    NodeComp* node = GetNodeComp();
    node->GetPin()->SetBrush(psnonebr);
    node->SetTransMethod(Out);
}

ClassId SwitchComp::GetClassId () { return SWITCH_COMP; }

boolean SwitchComp::IsA (ClassId id) {
    return SWITCH_COMP == id || LogicComp::IsA(id);
}
    
Component* SwitchComp::Copy () {
    Coord x, y;
    PinGraphic* pin = _nodeComp->GetPin();
    pin->GetOriginal(x, y);

    SwitchComp* copy = new SwitchComp(x, y);
    Transformer* t = GetGraphic()->GetTransformer();
    copy->GetGraphic()->SetTransformer(new Transformer(t));
    return copy;
}

/*****************************************************************************/

SwitchView::SwitchView (SwitchComp* subj) : LogicView(subj) { }
SwitchComp* SwitchView::GetSwitchComp () { return (SwitchComp*) GetSubject(); }

Manipulator* SwitchView::CreateManipulator (
    Viewer* v, Event& e, Transformer* t, Tool* tool
) {
    Manipulator* m = nil;

    if (tool->IsA(MANIPULATE_TOOL)) {
        SwitchComp* subj = GetSwitchComp();
        NodeComp* nodeComp = subj->GetNodeComp();
        NodeVar* nodeVar = nodeComp->GetNodeVar();
        LogicLevel newLevel = (nodeVar->GetLogicLevel() == Low) ? High : Low;
        nodeVar->SetLogicLevel(newLevel);
        nodeComp->Transmit();
        subj->Notify();
        unidraw->Update();

    } else {
        m = GraphicView::CreateManipulator(v, e, t, tool);
    }
    return m;
}

UStencil* SwitchView::CreateStencil () {
    NodeComp* node = GetSwitchComp()->GetNodeComp();
    NodeVar* nodeVar = node->GetNodeVar();
    Bitmap* image, *mask;

    if (nodeVar->GetLogicLevel() == Low) {
        image = new Bitmap(swchoff_bits, swchoff_width, swchoff_height);
        mask=new Bitmap(swchoffMask_bits,swchoffMask_width,swchoffMask_height);

    } else {
        image = new Bitmap(swchon_bits, swchon_width, swchon_height);
        mask = new Bitmap(swchonMask_bits, swchonMask_width,swchonMask_height);
    }
    return new UStencil(image, mask, stdgraphic);
}

/*****************************************************************************/

class InstanceNotifier : public ComponentView {
public:
    InstanceNotifier(ElementComp* original, InstanceComp*);
    virtual void Update();
private:
    InstanceComp* _instance;
};

InstanceNotifier::InstanceNotifier (
    ElementComp* o, InstanceComp* i
) : ComponentView(o) {
    _instance = i;
}

void InstanceNotifier::Update () { _instance->InstanceUpdate(); }

/*****************************************************************************/

InstanceComp::InstanceComp (ElementComp* original) {
    if (original == nil) {
        _notifier = nil;
    } else {
        _notifier = new InstanceNotifier(original, this);
        InstanceUpdate();
    }
}

InstanceComp::~InstanceComp () { delete _notifier; }
ClassId InstanceComp::GetClassId () { return INSTANCE_COMP; }

boolean InstanceComp::IsA (ClassId id) {
    return INSTANCE_COMP == id || ElementComp::IsA(id);
}
    
void InstanceComp::InstanceUpdate () {
    ElementComp* orig = GetOriginal();
    *_elemVar = *orig->GetElemVar();
    Iterator i;

    for (First(i); !Done(i);) {
        GraphicComp* comp = GetComp(i);
        Remove(i);
        delete comp;
    }
    for (orig->First(i); !orig->Done(i); orig->Next(i)) {
        Append((GraphicComp*) orig->GetComp(i)->Copy());
    }

    delete _transfn;
    _transfn = nil;
    TF_TruthTable* origtf = (TF_TruthTable*) orig->GetTransferFunct();

    if (origtf != nil) {
        TruthTable* tt = CreateTruthTable();
        *tt = *origtf->GetTruthTable();
        _transfn = new TF_TruthTable(tt);
    }
    Notify();
}

ElementComp* InstanceComp::GetOriginal () {
    return (ElementComp*) _notifier->GetSubject();
}

void InstanceComp::Read (istream& in) {
    ElementComp::Read(in);
    ElementComp* orig= (ElementComp*) unidraw->GetCatalog()->ReadComponent(in);
    _notifier = new InstanceNotifier(orig, this);
    InstanceUpdate();
}

void InstanceComp::Write (ostream& out) {
    ElementComp::Write(out);
    unidraw->GetCatalog()->WriteComponent(GetOriginal(), out);
}
