/*
 * Copyright (c) 1990 Stanford University
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided
 * that the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of Stanford not be used in advertising or
 * publicity pertaining to distribution of the software without specific,
 * written prior permission.  Stanford makes no representations about
 * the suitability of this software for any purpose.  It is provided "as is"
 * without express or implied warranty.
 *
 * STANFORD DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS.
 * IN NO EVENT SHALL STANFORD BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 * DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
 * OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION
 * WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

/*
 * Schematic capture system-specific component declarations.
 */

#ifndef scmcomps_h
#define scmcomps_h

#include "scmrubs.h"
#include <Unidraw/Components/pin.h>

class ElementVar;
class MultiLine;
class NodeVar;
class TF_Direct;
class TF_TruthTable;
class TruthTable;
class UStencil;

class ElementComp : public GraphicComps {
public:
    ElementComp();
    virtual ~ElementComp();

    virtual void Interpret(Command*);

    virtual ElementVar* GetElemVar();
    virtual NodeVar* GetNodeVar(int index);
    virtual StateVar* GetState(const char*);
    virtual TransferFunct* GetTransferFunct();

    int NodeCount();

    virtual void Read(istream&);
    virtual void Write(ostream&);
    virtual ClassId GetClassId();
    virtual boolean IsA(ClassId);
protected:
    virtual TruthTable* CreateTruthTable();
    virtual void InterpretTruthTable(TruthTable*);
protected:
    ElementVar* _elemVar;
    TF_TruthTable* _transfn;
};

class ElementView : public GraphicViews {
public:
    ElementView(ElementComp* = nil);

    ElementComp* GetElemComp();

    virtual ClassId GetClassId();
    virtual boolean IsA(ClassId);
};

class AliasComp : public ElementComp {
public:
    AliasComp(ElementComp* = nil);

    ElementComp* GetDefinition();

    virtual void Read(istream&);
    virtual void Write(ostream&);
    virtual ClassId GetClassId();
    virtual boolean IsA(ClassId);
protected:
    ElementComp* _defn;
};

inline ElementComp* AliasComp::GetDefinition () { return _defn; }

class AliasView : public ElementView {
public:
    AliasView(AliasComp* = nil);

    AliasComp* GetAliasComp();

    virtual ClassId GetClassId();
    virtual boolean IsA(ClassId);
};

class NodeComp : public PinComp {
public:
    NodeComp(PinGraphic* = nil);

    NodeVar* GetNodeVar();

    virtual Component* Copy();
    virtual ClassId GetClassId();
    virtual boolean IsA(ClassId);
private:
    friend class NLNode;
    UList* GetConnections();
};

inline UList* NodeComp::GetConnections () { return _cnxns; }

class NodeView : public PinView {
public:
    NodeView(NodeComp* = nil);

    virtual ClassId GetClassId();
    virtual boolean IsA(ClassId);
protected:
    virtual PinComp* NewSubject(PinGraphic*);
};

class WireComp : public GraphicComp {
public:
    WireComp();
    WireComp(Coord* x, Coord* y, int count, Graphic*);
    virtual ~WireComp();

    virtual void Interpret(Command*);
    virtual void Uninterpret(Command*);

    virtual void Update();
    virtual void First(Iterator&);
    virtual void Last(Iterator&);
    virtual void Next(Iterator&);
    virtual void Prev(Iterator&);
    virtual boolean Done(Iterator);

    virtual GraphicComp* GetComp(Iterator);
    virtual void SetComp(GraphicComp*, Iterator&);
    virtual void SetMobility(Mobility);
    virtual TransferFunct* GetTransferFunct();

    void GetNodes(NodeComp*&, NodeComp*&);
    MultiLine* GetMultiLine();

    virtual Component* Copy();
    virtual void Read(istream&);
    virtual void Write(ostream&);
    virtual ClassId GetClassId();
    virtual boolean IsA(ClassId);
private:
    NodeComp* _node1, *_node2;
    TF_Direct* _transfn;
};

class WireView : public GraphicView {
public:
    WireView(WireComp* = nil);
    virtual ~WireView();

    virtual void Update();
    virtual Manipulator* CreateManipulator(Viewer*,Event&,Transformer*,Tool*);
    virtual Command* InterpretManipulator(Manipulator*);

    virtual void First(Iterator&);
    virtual void Last(Iterator&);
    virtual void Next(Iterator&);
    virtual void Prev(Iterator&);
    virtual boolean Done(Iterator);

    virtual GraphicView* GetView(Iterator);
    virtual void SetView(GraphicView*, Iterator&);

    WireComp* GetWireComp();
    virtual Graphic* GetGraphic();

    virtual ClassId GetClassId();
    virtual boolean IsA(ClassId);
protected:
    virtual void CreateHandles();
private:
    MultiLine* GetMultiLine();
    void GetEndpoints(Coord&, Coord&, Coord&, Coord&);
    void GetVertices(Coord*&, Coord*&, int&);
    void UpdateMultiLine();

    Manipulator* CreateWireCompManip(Viewer*, Event&, Transformer*, Tool*);
    Command* InterpWireCompManip(Manipulator*);
private:
    NodeView* _nodeView1, *_nodeView2;
};

class LogicComp : public GraphicComp {
public:
    virtual ~LogicComp();
    virtual void Update();

    virtual void First(Iterator&);
    virtual void Last(Iterator&);
    virtual void Next(Iterator&);
    virtual void Prev(Iterator&);
    virtual boolean Done(Iterator);

    virtual GraphicComp* GetComp(Iterator);
    virtual void SetComp(GraphicComp*, Iterator&);
    NodeComp* GetNodeComp();

    virtual void Read(istream&);
    virtual void Write(ostream&);
    virtual ClassId GetClassId();
    virtual boolean IsA(ClassId);
protected:
    LogicComp();
    LogicComp(PinGraphic*);
protected:
    NodeComp* _nodeComp;
};

inline NodeComp* LogicComp::GetNodeComp () { return _nodeComp; }

class LogicView : public GraphicView {
public:
    virtual ~LogicView();

    virtual void Update();
    virtual Graphic* GetGraphic();
    LogicComp* GetLogicComp();

    virtual void First(Iterator&);
    virtual void Last(Iterator&);
    virtual void Next(Iterator&);
    virtual void Prev(Iterator&);
    virtual boolean Done(Iterator);

    virtual GraphicView* GetView(Iterator);
    virtual void SetView(GraphicView*, Iterator&);
protected:
    LogicView(LogicComp* = nil);
    virtual UStencil* CreateStencil();
    virtual void AlignStencil();
    UStencil* GetStencil();
protected:
    NodeView* _nodeView;
};

class BulbComp : public LogicComp {
public:
    BulbComp();
    BulbComp(Coord, Coord);

    virtual Component* Copy();
    virtual ClassId GetClassId();
    virtual boolean IsA(ClassId);
};

class BulbView : public LogicView {
public:
    BulbView(BulbComp* = nil);

    BulbComp* GetBulbComp();
protected:
    virtual UStencil* CreateStencil();
};

class SwitchComp : public LogicComp {
public:
    SwitchComp();
    SwitchComp(Coord, Coord);

    virtual Component* Copy();
    virtual ClassId GetClassId();
    virtual boolean IsA(ClassId);
};

class SwitchView : public LogicView {
public:
    SwitchView(SwitchComp* = nil);

    virtual Manipulator* CreateManipulator(Viewer*,Event&,Transformer*,Tool*);
    SwitchComp* GetSwitchComp();
protected:
    virtual UStencil* CreateStencil();
};

class InstanceComp : public ElementComp {
public:
    InstanceComp(ElementComp* original = nil);
    virtual ~InstanceComp();

    ElementComp* GetOriginal();

    virtual void Read(istream&);
    virtual void Write(ostream&);
    virtual ClassId GetClassId();
    virtual boolean IsA(ClassId);
private:
    friend class InstanceNotifier;
    void InstanceUpdate(); 
private:
    InstanceNotifier* _notifier;
};

#endif
