/*
 * Copyright (c) 1990 Stanford University
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided
 * that the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of Stanford not be used in advertising or
 * publicity pertaining to distribution of the software without specific,
 * written prior permission.  Stanford makes no representations about
 * the suitability of this software for any purpose.  It is provided "as is"
 * without express or implied warranty.
 *
 * STANFORD DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS.
 * IN NO EVENT SHALL STANFORD BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 * DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
 * OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION
 * WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

/*
 * Implementation of schematic capture system-specific dialog boxes.
 */

#include "scmclasses.h"
#include "scmcomps.h"
#include "scmdialogs.h"
#include "scmtruth.h"
#include "scmvarviews.h"

#include <InterViews/adjuster.h>
#include <InterViews/border.h>
#include <InterViews/box.h>
#include <InterViews/button.h>
#include <InterViews/filebrowser.h>
#include <InterViews/font.h>
#include <InterViews/frame.h>
#include <InterViews/glue.h>
#include <InterViews/message.h>
#include <InterViews/painter.h>
#include <InterViews/scroller.h>
#include <InterViews/sensor.h>
#include <InterViews/streditor.h>
#include <InterViews/tray.h>

#include <stdlib.h>
#include <string.h>

/*****************************************************************************/

InfoDialog::InfoDialog (Component* comp) : BasicDialog(new ButtonState) {
    input = new Sensor;
    Ref(input);
    input->Catch(KeyEvent);

    _elemVarView = nil;
    _nodeVarView = nil;
    _nodes = 0;

    Init(comp);
    Insert(Interior());
}

InfoDialog::~InfoDialog () {
    delete _nodeVarView;
}

boolean InfoDialog::Accept () {
    Event e;
    int v = 0;

    state->SetValue(0);

    while (v == 0) {
	Read(e);
	Forward(e);
	state->GetValue(v);
    }
    return v == '\r';
}

void InfoDialog::Handle (Event& e) {
    if (e.eventType == KeyEvent && e.len != 0) {
        char c = e.keystring[0];

        if (c == '\r' || c == '\007') {
            state->SetValue(c);

        } else if (_elemVarView != nil) {
            _elemVarView->Handle(e);

        } else if (_nodes > 0) {
            _nodeVarView[0]->Handle(e);
        }
    }
}

void InfoDialog::ApplyChanges () {
    if (_elemVarView != nil) {
        _elemVarView->ChangeSubject();
    }
    for (int i = 0; i < _nodes; ++i) {
        _nodeVarView[i]->ChangeSubject();
    }
}

void InfoDialog::Init (Component* comp) {
    if (comp->IsA(ELEMENT_COMP)) {
        SetTitle("Element Information");
        ElementComp* elemComp = (ElementComp*) comp;
        _elemVarView = new ElementVarView(elemComp->GetElemVar(), state);
        _nodes = elemComp->NodeCount();
        _nodeVarView = new NodeVarView*[_nodes];

        for (int i = 0; i < _nodes; ++i) {
            _nodeVarView[i] = new NodeVarView(elemComp->GetNodeVar(i), state);
        }

    } else if (comp->IsA(NODE_COMP)) {
        SetTitle("Node Information");
        NodeVar* nodeVar = (NodeVar*) ((NodeComp*) comp)->GetBinding();
        _nodes = 1;
        _nodeVarView = new NodeVarView*[_nodes];
        _nodeVarView[0] = new NodeVarView(nodeVar, state);
    }
}

Interactor* InfoDialog::Interior () {
    const int space = round(.5*cm);

    return new MarginFrame(
        new VBox(
            _title,
            new VGlue(space/2, 0),
            new VBox(
                new HBorder,
                new VGlue(2, 0),
                new HBorder
            ),
            new VGlue(space, 0),
            VarViews(space),
            new VGlue(space, 0),
            new HBox(
                new HGlue,
                new PushButton("Cancel", state, '\007'),
                new HGlue(space, 0),
                new PushButton("  OK  ", state, '\r')
            )
        ), space, space/2, 0
    );
}

Interactor* InfoDialog::VarViews (int space) {
    VBox* varViews = new VBox;

    if (_elemVarView != nil) {
        varViews->Insert(ElemPart(space));
    }
    if (_elemVarView != nil && _nodes > 0) {
        varViews->Insert(new VGlue(space/2, 0));
    }
    if (_nodes > 0) {
        varViews->Insert(NodePart(space));
    }
    return varViews;
}

Interactor* InfoDialog::ElemPart (int space) {
    HBox* elemPart = new HBox(
        new Message("Name:", Left),
        new HGlue(space, 0),
        new Frame(new MarginFrame(_elemVarView, 2))
    );
    elemPart->Align(Center);
    return elemPart;
}

Interactor* InfoDialog::NodePart (int space) {
    Interactor* varViews;

    if (_elemVarView == nil) {
        HBox* hbox = new HBox(
            new Message("Name:", Left),
            new HGlue(space, 0),
            _nodeVarView[0]
        );
        hbox->Align(Center);
        varViews = hbox;

    } else {
        VBox* vbox = new VBox(
            new Message("Nodes:", Left, 0, hfil),
            new VGlue(space/2, 0)
        );
        for (int i = 0; i < _nodes; ++i) {
            vbox->Insert(_nodeVarView[i]);

            if (i != _nodes-1) {
                vbox->Insert(new VGlue(space, 0));
            }
        }
        varViews = vbox;
    }
    return varViews;
}

/*****************************************************************************/

NewToolDialog::NewToolDialog (
    const char* dir, boolean instantiate
) : FileChooser(new ButtonState, dir, 10, 24, Center) {
    _instantiate = new ButtonState(instantiate);
    input = new Sensor;
    Ref(input);
    input->Catch(KeyEvent);

    Init("", "Create schematic tool for:");
    Insert(Interior());
}

boolean NewToolDialog::Instantiating () {
    int v;
    _instantiate->GetValue(v);
    return v;
}

const char* NewToolDialog::Name () { return _sedit->Text(); }

Interactor* NewToolDialog::Interior () {
    const int space = round(.5*cm);
    VBox* titleblock = new VBox(
        new HBox(title, new HGlue),
        new HBox(subtitle, new HGlue)
    );

    return new MarginFrame(
        new VBox(
            titleblock,
            new VBox(
                new VGlue(space, 0),
                new Frame(new MarginFrame(_sedit, 2)),
                new VGlue(space, 0),
                new Frame(AddScroller(browser())),
                new VGlue(space, 0),
                new HBox(
                    new CheckBox("Instantiate", _instantiate, true, false),
                    new HGlue
                ),
                new VGlue(space, 0)
            ),
            new HBox(
                new VGlue(space, 0),
                new HGlue,
                new PushButton("Cancel", state, '\007'),
                new HGlue(space, 0),
                new PushButton("Create", state, '\r')
            )
        ), space, space/2, 0
    );
}

/*****************************************************************************/

NetlistDialog::NetlistDialog (const char* dir, boolean flatten) : FileChooser(
    new ButtonState, dir, 10, 24, Center
) {
    _flatten = new ButtonState(flatten);
    input = new Sensor;
    Ref(input);
    input->Catch(KeyEvent);

    Init("", "Generate netlist and save as:");
    Insert(Interior());
}

boolean NetlistDialog::Flattening () {
    int v;
    _flatten->GetValue(v);
    return v;
}

const char* NetlistDialog::Name () { return _sedit->Text(); }

Interactor* NetlistDialog::Interior () {
    const int space = round(.5*cm);
    VBox* titleblock = new VBox(
        new HBox(title, new HGlue),
        new HBox(subtitle, new HGlue)
    );

    return new MarginFrame(
        new VBox(
            titleblock,
            new VBox(
                new VGlue(space, 0),
                new Frame(new MarginFrame(_sedit, 2)),
                new VGlue(space, 0),
                new Frame(AddScroller(browser())),
                new VGlue(space, 0),
                new HBox(
                    new CheckBox("Flatten", _flatten, true, false),
                    new HGlue
                ),
                new VGlue(space, 0)
            ),
            new HBox(
                new VGlue(space, 0),
                new HGlue,
                new PushButton("Cancel", state, '\007'),
                new HGlue(space, 0),
                new PushButton("Generate & Save", state, '\r')
            )
        ), space, space/2, 0
    );
}

/*****************************************************************************/

class VMessage : public Message {
public:
    VMessage(const char*, Alignment = Center, int = 0, int = 0, int = 0);
protected:
    virtual void Reconfig();
    virtual void Redraw(Coord, Coord, Coord, Coord);
};

VMessage::VMessage (
    const char* msg, Alignment a, int pad, int hstr, int vstr
) : Message(msg, a, pad, hstr, vstr) { }

void VMessage::Reconfig () {
    const char* a = GetAttribute("text");
    if (a != nil) {
        delete text;
	text = new char[strlen(a)+1];
        strcpy(text, a);
    }
    a = GetAttribute("padding");
    if (a != nil) {
        pad = atoi(a);
    }
    const Font* f = output->GetFont();
    shape->width = pad + f->Height() + pad;
    shape->height = pad + f->Width(text) + pad;
    shape->hshrink = pad + pad;
    shape->vshrink = pad + pad;

    Painter* p = new Painter(output);
    Ref(p);
    Unref(output);
    output = p;
    output->SetTransformer(nil);
    output->Rotate(90);
    output->Translate(pad + f->Height(), 0);
}   

void VMessage::Redraw (Coord l, Coord b, Coord r, Coord t) {
    Coord x = 0, y = 0;
    Align(alignment, shape->width, shape->height, x, y);
    output->Clip(canvas, l, b, r, t);
    if (highlighted) {
        output->SetColors(output->GetBgColor(), output->GetFgColor());
    }
    output->ClearRect(canvas, l, b, r, t);
    output->Text(canvas, text, x + pad, y + pad);
    if (highlighted) {
        output->SetColors(output->GetBgColor(), output->GetFgColor());
    }
    output->NoClip();
}

/*****************************************************************************/

TruthTableDialog::TruthTableDialog (TruthTable* tt) : BasicDialog(
    new ButtonState, "Truth table:"
) {
    input = new Sensor;
    Ref(input);
    input->Catch(KeyEvent);

    _truthTable = tt;
    _truthEd = new TruthTableEditor(state, _truthTable);

    Insert(Interior());
}

boolean TruthTableDialog::Accept () {
    Event e;
    int v = 0;

    state->SetValue(0);

    while (v == 0) {
	Read(e);
	Forward(e);
	state->GetValue(v);
    }
    return v == '\r';
}

void TruthTableDialog::Handle (Event& e) {
    if (e.eventType == KeyEvent && e.len != 0) {
        char c = e.keystring[0];

        if (c == '\r' || c == '\007') {
            state->SetValue(c);
        } else {
            _truthEd->Handle(e);
        }
    }
}

Interactor* TruthTableDialog::Interior () {
    const int space = round(.5*cm);

    return new MarginFrame(
        new VBox(
            _title,
            new VGlue(space, 0),
            Editor(),
            new VGlue(space, 0),
            new HBox(
                new HGlue,
                new PushButton("Cancel", state, '\007'),
                new HGlue(space, 0),
                new PushButton("  OK  ", state, '\r')
            )
        ), space, space/2, 0
    );
}

Interactor* TruthTableDialog::IOLabels () {
    HBox* hbox = new HBox;
    int i;

    for (i = 0; i < _truthTable->Inputs(); i++) {
        NodeVar* inputVar = _truthTable->GetInputVar(i);
        hbox->Insert(new VMessage(inputVar->GetName()));
    }
    hbox->Insert(new HGlue);

    for (i = 0; i < _truthTable->Outputs(); i++) {
        NodeVar* outputVar = _truthTable->GetOutputVar(i);
        hbox->Insert(new VMessage(outputVar->GetName()));
    }
    return hbox;
}

Interactor* TruthTableDialog::Editor () {
    Tray* t = new Tray;

    Interactor* iolabels = new VBox(IOLabels(), new VGlue(2, 0, 0));
    Interactor* edFrame = new Frame(new VBox(new HGlue(0, 0), _truthEd));
    Interactor* scrollbar = new HBox(
        new VBox(
            new HBorder,
            new UpMover(_truthEd, 1),
            new HBorder,
            new VScroller(_truthEd),
            new HBorder,
            new DownMover(_truthEd, 1),
            new HBorder
        ),
        new VBorder
    );

    t->Align(Left, edFrame, iolabels);
    t->Align(Right, edFrame, iolabels);
    t->HBox(t, edFrame, scrollbar, t);

    t->Align(Top, edFrame, scrollbar);
    t->Align(Bottom, edFrame, scrollbar);
    t->VBox(t, iolabels, edFrame, t);
            
    return new HBox(
        new HGlue,
        t,
        new HGlue
    );
}
