/*
 * Copyright (c) 1990 Stanford University
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided
 * that the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of Stanford not be used in advertising or
 * publicity pertaining to distribution of the software without specific,
 * written prior permission.  Stanford makes no representations about
 * the suitability of this software for any purpose.  It is provided "as is"
 * without express or implied warranty.
 *
 * STANFORD DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS.
 * IN NO EVENT SHALL STANFORD BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 * DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
 * OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION
 * WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

/*
 * Schematic editor main class implementation.
 */

#include "scmclasses.h"
#include "scmed.h"
#include "scmcmds.h"
#include "scmcomps.h"
#include "scmkybd.h"
#include "scmpal.h"

#include <Unidraw/catalog.h>
#include <Unidraw/ctrlinfo.h>
#include <Unidraw/grid.h>
#include <Unidraw/keymap.h>
#include <Unidraw/kybd.h>
#include <Unidraw/selection.h>
#include <Unidraw/statevars.h>
#include <Unidraw/stateviews.h>
#include <Unidraw/uctrls.h>
#include <Unidraw/unidraw.h>
#include <Unidraw/upage.h>
#include <Unidraw/viewer.h>

#include <Unidraw/Commands/align.h>
#include <Unidraw/Commands/catcmds.h>
#include <Unidraw/Commands/edit.h>
#include <Unidraw/Commands/struct.h>
#include <Unidraw/Commands/transforms.h>
#include <Unidraw/Commands/viewcmds.h>

#include <Unidraw/Components/grcomp.h>
#include <Unidraw/Components/grview.h>

#include <Unidraw/Graphic/polygons.h>

#include <InterViews/border.h>
#include <InterViews/frame.h>
#include <InterViews/glue.h>
#include <InterViews/menu.h>
#include <InterViews/message.h>
#include <InterViews/panner.h>
#include <InterViews/shape.h>
#include <InterViews/tray.h>
#include <InterViews/window.h>
#include <InterViews/world.h>

#include <string.h>

/*****************************************************************************/

static const float PAGE_WIDTH = 17;
static const float PAGE_HEIGHT = 14;
static const float VIEWER_WIDTH = 6;
static const float VIEWER_HEIGHT = 5;
static const float DEFAULT_XINCR = 8;
static const float DEFAULT_YINCR = 8;

/*****************************************************************************/

inline void InsertSeparator (PulldownMenu* pdm) {
    pdm->GetScene()->Insert(
        new VBox(
            new VGlue(2, 0, 0),
            new HBorder,
            new VGlue(2, 0, 0)
        )
    );
}

inline PulldownMenu* MakePulldown (const char* name) {
    return new PulldownMenu(new Message(name, Center, round(.1*cm)));
}

/*****************************************************************************/

SchemEditor::SchemEditor (
    ToolPalette* tp, DrawingPalette* dtp, GraphicComp* comp
) {
    SetClassName("SchemEditor");

    ManagedWindow* window = new ApplicationWindow(this);
    SetWindow(window);

    _comp = (comp == nil) ? (new ElementComp) : comp;
    _toolPal = tp;
    _dwgPal = dtp;

    _keymap = new KeyMap;
    _keymap->Register(tp->GetKeyMap());
    _keymap->Register(dtp->GetKeyMap());
    
    InitStateVars();
    InitViewer();

    Insert(new Frame(Interior()));
    GetKeyMap()->Execute(CODE_SELECT);
}

SchemEditor::~SchemEditor () {
    delete _keymap;
}

void SchemEditor::InitStateVars () {
    _name = new CompNameVar(_comp);
    _modifStatus = new ModifStatusVar(_comp);
    _gravity = new GravityVar;
    _magnif = new MagnifVar;
}

Component* SchemEditor::GetComponent () { return _comp; }
Viewer* SchemEditor::GetViewer (int id) { return (id == 0) ? _viewer : nil; }
KeyMap* SchemEditor::GetKeyMap () { return _keymap; }
Tool* SchemEditor::GetCurTool () { return _toolPal->GetCurTool(); }
Selection* SchemEditor::GetSelection () { return _toolPal->GetSelection(); }

StateVar* SchemEditor::GetState (const char* name) {
    if (strcmp(name, "CompNameVar") == 0) {
        return _name;
    } else if (strcmp(name, "ModifStatusVar") == 0) {
        return _modifStatus;
    } else if (strcmp(name, "MagnifVar") == 0) {
        return _magnif;
    } else if (strcmp(name, "GravityVar") == 0) {
        return _gravity;
    } else {
        StateVar* stateVar = _dwgPal->GetState(name);
        return (stateVar == nil) ? _toolPal->GetState(name) : stateVar;
    }
}

void SchemEditor::SetComponent (Component* comp) {
    if (comp == nil || comp->IsA(GRAPHIC_COMP)) {
        _comp = (GraphicComp*) comp;
    }
}

void SchemEditor::SetViewer (Viewer* v, int i) { 
    if (i == 0) {
        _viewer = v;
    }
}

void SchemEditor::SetKeyMap (KeyMap* k) { _keymap = k; }
void SchemEditor::SetSelection (Selection* s) { _toolPal->SetSelection(s); }
ToolPalette* SchemEditor::GetToolPalette () { return _toolPal; }
DrawingPalette* SchemEditor::GetDrawingPalette () { return _dwgPal; }

Interactor* SchemEditor::Interior () {
    Interactor* commands = Commands();
    HBorder* hborder = new HBorder;
    VBorder* vborder = new VBorder;
    int gap = round(.1*cm);

    _tray = new Tray(
        new VBox(
            new HBox(
                new HGlue(gap, 0, 0),
                new ModifStatusVarView(_modifStatus),
                new CompNameVarView(_name, Left),
                new GravityVarView(_gravity, Right),
                new MagnifVarView(_magnif, Right)
            ),
            new HBorder,
            new HBox(Commands(), new HGlue),
            new HBorder,
            _viewer
        )
    );
    _tray->Align(BottomRight, new Frame(new Panner(_viewer)));
    _tray->Propagate(false);

    return _tray;
}

Interactor* SchemEditor::Commands () {
    MenuBar* commands = new MenuBar;

    commands->Include(FileMenu());
    commands->Include(EditMenu());
    commands->Include(StructureMenu());
    commands->Include(AlignMenu());
    commands->Include(ViewMenu());

    return commands;
}

void SchemEditor::InitViewer () {
    const float sw = PAGE_WIDTH * inches;
    const float sh = PAGE_HEIGHT * inches;
    const int vw = round(VIEWER_WIDTH * inches);
    const int vh = round(VIEWER_HEIGHT * inches);

    GraphicView* view = (GraphicView*) _comp->Create(COMPONENT_VIEW);
    _comp->Attach(view);
    view->Update();

    UPage* page = new UPage(sw, sh);
    Grid* grid = new Grid(sw, sh, DEFAULT_XINCR, DEFAULT_YINCR);
    grid->Visibility(false);

    _viewer = new Viewer(this, view, page, grid, vw, vh);
}

void SchemEditor::Include (Command* cmd, PulldownMenu* pdm) {
    ControlInfo* ctrlInfo = cmd->GetControlInfo();
    UControl* ctrl = new CommandControl(ctrlInfo);
    _keymap->Register(ctrl);
    pdm->Include(ctrl);
    cmd->SetEditor(this);
}

PulldownMenu* SchemEditor::FileMenu () {
    PulldownMenu* pdm = MakePulldown("File");

    Include(
        new NewCompCmd(
            new ControlInfo("New", KLBL_NEWCOMP, CODE_NEWCOMP), new ElementComp
        ),
        pdm
    );
    Include(
        new RevertCmd(new ControlInfo("Revert", KLBL_REVERT, CODE_REVERT)),
        pdm
    );
    InsertSeparator(pdm);
    Include(
        new ViewCompCmd(
            new ControlInfo("Open...", KLBL_VIEWCOMP, CODE_VIEWCOMP)
        ), pdm
    );
    Include(
        new SaveCompCmd(new ControlInfo("Save", KLBL_SAVECOMP, CODE_SAVECOMP)),
        pdm
    );
    Include(
        new SaveCompAsCmd(
            new ControlInfo("Save As...", KLBL_SAVECOMPAS, CODE_SAVECOMPAS)
        ), pdm
    );
    Include(
        new NewToolCmd(
            new ControlInfo("New Tool...", KLBL_NEWTOOL, CODE_NEWTOOL)
        ), pdm
    );
    Include(
        new ToolsCmd(
            new ControlInfo("Tools...", KLBL_TOOLS, CODE_TOOLS)
        ), pdm
    );
    Include(
        new NetlistCmd(
            new ControlInfo("Netlist...", KLBL_NETLIST, CODE_NETLIST)
        ), pdm
    );
    Include(
        new PrintCmd(
            new ControlInfo("Print...", KLBL_PRINT, CODE_PRINT)
        ), pdm
    );
    InsertSeparator(pdm);
    Include(
        new QuitCmd(
            new ControlInfo("Quit", KLBL_QUIT, CODE_QUIT)
        ), pdm
    );

    return pdm;
}

PulldownMenu* SchemEditor::EditMenu () {
    PulldownMenu* pdm = MakePulldown("Edit");

    Include(
        new UndoCmd(
            new ControlInfo("Undo", KLBL_UNDO, CODE_UNDO)
        ), pdm
    );
    Include(
        new RedoCmd(
            new ControlInfo("Redo", KLBL_REDO, CODE_REDO)
        ), pdm
    );
    InsertSeparator(pdm);
    Include(
        new CutCmd(
            new ControlInfo("Cut", KLBL_CUT, CODE_CUT)
        ), pdm
    );
    Include(
        new CopyCmd(
            new ControlInfo("Copy", KLBL_COPY, CODE_COPY)
        ), pdm
    );
    Include(
        new PasteCmd(
            new ControlInfo("Paste", KLBL_PASTE,CODE_PASTE)
        ), pdm
    );
    Include(
        new DupCmd(
            new ControlInfo("Duplicate", KLBL_DUP, CODE_DUP)
        ), pdm
    );
    Include(
        new DeleteCmd(
            new ControlInfo("Delete", KLBL_DEL, CODE_DEL)
        ), pdm
    );
    Include(
        new SlctAllCmd(
            new ControlInfo("Select All", KLBL_SLCTALL, CODE_SLCTALL)
        ), pdm
    );
    InsertSeparator(pdm);
    Include(
        new ScaleCmd(
            new ControlInfo("Flip Horizontal", KLBL_HFLIP, CODE_HFLIP),
            -1.0, 1.0
        ), pdm
    );
    Include(
        new ScaleCmd(
            new ControlInfo("Flip Vertical", KLBL_VFLIP, CODE_VFLIP),
            1.0, -1.0
        ), pdm
    );
    Include(
        new RotateCmd(
            new ControlInfo("90 Clockwise", KLBL_CW90, CODE_CW90), -90.0
        ), pdm
    );
    Include(
        new RotateCmd(
            new ControlInfo("90 CounterCW", KLBL_CCW90, CODE_CCW90), 90.0
        ), pdm
    );
    InsertSeparator(pdm);
    Include(
        new InfoCmd(new ControlInfo("Info...", KLBL_INFO, CODE_INFO)), pdm
    );
    Include(
        new TruthTableCmd(
            new ControlInfo("Truth Table...", KLBL_TRUTHTABLE, CODE_TRUTHTABLE)
        ), pdm
    );
    Include(
        new MobilityCmd(
            new ControlInfo("Fix", KLBL_FIX, CODE_FIX), Fixed
        ), pdm
    );
    Include(
        new MobilityCmd(
            new ControlInfo("Floating", KLBL_FLOAT, CODE_FLOAT), Floating
        ), pdm
    );

    return pdm;
}

PulldownMenu* SchemEditor::StructureMenu () {
    PulldownMenu* pdm = MakePulldown("Structure");

    Include(
        new GroupCmd(
            new ControlInfo("Group", KLBL_GROUP, CODE_GROUP)
        ), pdm
    );
    Include(
        new UngroupCmd(
            new ControlInfo("Ungroup", KLBL_UNGROUP, CODE_UNGROUP)
        ), pdm
    );
    Include(
        new FrontCmd(
            new ControlInfo("Bring to Front", KLBL_FRONT, CODE_FRONT)
        ), pdm
    );
    Include(
        new BackCmd(
            new ControlInfo("Send to Back", KLBL_BACK, CODE_BACK)
        ), pdm
    );

    return pdm;
}

PulldownMenu* SchemEditor::AlignMenu () {
    PulldownMenu* pdm = MakePulldown("Align");

    Include(
        new AlignCmd(
            new ControlInfo("Left Sides", KLBL_ALGNLEFT, CODE_ALGNLEFT), 
            Left, Left
        ), pdm
    );
    Include(
        new AlignCmd(
            new ControlInfo("Right Sides", KLBL_ALGNRIGHT, CODE_ALGNRIGHT),
            Right, Right
        ), pdm
    );
    Include(
        new AlignCmd(
            new ControlInfo("Tops", KLBL_ALGNTOP, CODE_ALGNTOP), Top, Top
        ), pdm
    );
    Include(
        new AlignCmd(
            new ControlInfo("Bottoms", KLBL_ALGNBOT, CODE_ALGNBOT),	
            Bottom, Bottom
        ), pdm
    );
    Include(
        new AlignCmd(
            new ControlInfo("Horiz Centers", KLBL_ALGNHCTR, CODE_ALGNHCTR),
            HorizCenter, HorizCenter
        ), pdm
    );
    Include(
        new AlignCmd(
            new ControlInfo("Vert Centers", KLBL_ALGNVCTR, CODE_ALGNVCTR),
            VertCenter, VertCenter
        ), pdm
    );
    Include(
        new AlignCmd(
            new ControlInfo("Centers", KLBL_ALGNCTR, CODE_ALGNCTR),	
            Center, Center
        ), pdm
    );
    Include(
        new AlignCmd(
            new ControlInfo("Abut Left", KLBL_ABUTLEFT, CODE_ABUTLEFT), 
            Left, Right
        ), pdm
    );
    Include(
        new AlignCmd(
            new ControlInfo("Abut Right", KLBL_ABUTRIGHT, CODE_ABUTRIGHT),
            Right, Left
        ), pdm
    );
    Include(
        new AlignCmd(
            new ControlInfo("Abut Up", KLBL_ABUTUP, CODE_ABUTUP),
            Top, Bottom
        ), pdm
    );
    Include(
        new AlignCmd(
            new ControlInfo("Abut Down", KLBL_ABUTDOWN, CODE_ABUTDOWN),
            Bottom, Top
        ), pdm
    );

    return pdm;
}

PulldownMenu* SchemEditor::ViewMenu () {
    PulldownMenu* pdm = MakePulldown("View");

    Include(
        new NewViewCmd(
            new ControlInfo("New View", KLBL_NEWVIEW, CODE_NEWVIEW)
        ), pdm
    );
    Include(
        new CloseEditorCmd(
            new ControlInfo("Close View", KLBL_CLOSEEDITOR, CODE_CLOSEEDITOR)
        ), pdm
    );
    Include(
        new ToggleDwgPaletteCmd(
            new ControlInfo(
                "Show/Hide Drawing Palette",
                KLBL_TOGGLEDWGPALETTE, CODE_TOGGLEDWGPALETTE
            )
        ), pdm
    );
    InsertSeparator(pdm);
    Include(
        new AliasCmd(
            new ControlInfo("Create Alias", KLBL_ALIAS, CODE_ALIAS)
        ), pdm
    );
    Include(
        new DefinitionCmd(
            new ControlInfo("Definition", KLBL_DEFINITION, CODE_DEFINITION)
        ), pdm
    );
    InsertSeparator(pdm);
    Include(
        new NormSizeCmd(
            new ControlInfo("Normal Size", KLBL_NORMSIZE, CODE_NORMSIZE)
        ), pdm
    );
    Include(
        new RedToFitCmd(
            new ControlInfo("Reduce to Fit", KLBL_REDTOFIT, CODE_REDTOFIT)
        ), pdm
    );
    Include(
        new CenterCmd(
            new ControlInfo("Center Page", KLBL_CENTER, CODE_CENTER)
        ), pdm
    );
    InsertSeparator(pdm);
    Include(
        new GridCmd(
            new ControlInfo("Grid on/off", KLBL_GRID, CODE_GRID)
        ), pdm
    );
    Include(
        new GridSpacingCmd(
            new ControlInfo("Grid Spacing...", KLBL_GRIDSPC, CODE_GRIDSPC)
        ), pdm
    );
    Include(
        new GravityCmd(
            new ControlInfo("Gravity on/off", KLBL_GRAVITY, CODE_GRAVITY)
        ), pdm
    );

    return pdm;
}
