/*
 * Copyright (c) 1990 Stanford University
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided
 * that the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of Stanford not be used in advertising or
 * publicity pertaining to distribution of the software without specific,
 * written prior permission.  Stanford makes no representations about
 * the suitability of this software for any purpose.  It is provided "as is"
 * without express or implied warranty.
 *
 * STANFORD DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS.
 * IN NO EVENT SHALL STANFORD BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 * DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
 * OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION
 * WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

/*
 * Implementation of NetListView and subclasses.
 */

#include "scmclasses.h"
#include "scmcomps.h"
#include "scmnet.h"
#include "scmvars.h"

#include <Unidraw/iterator.h>
#include <Unidraw/path.h>
#include <Unidraw/ulist.h>

#include <stream.h>
#include <string.h>

/*****************************************************************************/

static const int indentation = 4;

/*****************************************************************************/

boolean NetlistView::_flattening;
int NetlistView::_totalIndent;

ClassId NetlistView::GetClassId () { return NETLIST_VIEW; }

boolean NetlistView::IsA (ClassId id) {
    return NETLIST_VIEW == id || PreorderView::IsA(id);
}

NetlistView::NetlistView (GraphicComp* subj) : PreorderView(subj) { }

NetlistView::NetlistView (
    GraphicComp* subj, boolean flattening
) : PreorderView(subj) {
    Flatten(flattening);
}

boolean NetlistView::Emit (ostream& out) {
    _totalIndent = 0;
    return PreorderView::Emit(out);
}

GraphicComp* NetlistView::GetGraphicComp () {
    return (GraphicComp*) GetSubject();
}

void NetlistView::BeginBlock (ostream& out) { 
    out << "{\n";
    _totalIndent += indentation; 
}

void NetlistView::EndBlock (ostream& out) {
    _totalIndent -= indentation;
    Indent(out);
    out << "}\n";
}

void NetlistView::Indent (ostream& out) {
    for (int i = 0; i < _totalIndent; ++i) {
        out << " ";
    }
}

NetlistView* NetlistView::CreateNetlistView (GraphicComp* comp) {
    NetlistView* nlv = (NetlistView*) comp->Create(NETLIST_VIEW);

    if (nlv != nil) {
        comp->Attach(nlv);
        nlv->Update();
    }
    return nlv;
}

/*****************************************************************************/

ClassId NLElement::GetClassId () { return NL_ELEMENT; }

boolean NLElement::IsA (ClassId id) {
    return NL_ELEMENT == id || NetlistView::IsA(id);
}

NLElement::NLElement (ElementComp* subj) : NetlistView(subj) {
    _views = new UList;
}

NLElement::~NLElement () {
    DeleteViews();
    delete _views;
}

void NLElement::Identification (ostream& out) {
    Indent(out);
    out << GetElemComp()->GetElemVar()->Identification() << " ";
}

boolean NLElement::Definition (ostream& out) {
    Identification(out);
    BeginBlock(out);
    boolean status = NetlistView::Definition(out);
    EndBlock(out);

    return status;
}    

void NLElement::Update () {
    DeleteViews();

    GraphicComp* comps = GetGraphicComp();
    Iterator i;

    for (comps->First(i); !comps->Done(i); comps->Next(i)) {
        GraphicComp* comp = comps->GetComp(i);
        NetlistView* nlv = CreateNetlistView(comp);

        if (nlv != nil) {
            _views->Append(new UList(nlv));
        }
    }
}

ElementComp* NLElement::GetElemComp () { return (ElementComp*) GetSubject(); }
NLElement* NLElement::View (UList* r) { return (NLElement*) (*r)(); }
UList* NLElement::Elem (Iterator i) { return (UList*) i.GetValue(); }
void NLElement::First (Iterator& i) { i.SetValue(_views->First()); }
void NLElement::Last (Iterator& i) { i.SetValue(_views->Last()); }
void NLElement::Next (Iterator& i) { i.SetValue(Elem(i)->Next()); }
void NLElement::Prev (Iterator& i) { i.SetValue(Elem(i)->Prev()); }
boolean NLElement::Done (Iterator i) { return Elem(i) == _views->End(); }
ExternView* NLElement::GetView (Iterator i) { return View(Elem(i)); }

void NLElement::SetView (ExternView* ev, Iterator& i) {
    i.SetValue(_views->Find(ev));
}

void NLElement::DeleteView (Iterator& i) {
    UList* doomed = Elem(i);
    ExternView* view = GetView(i);

    Next(i);
    _views->Remove(doomed);
    SetParent(view, nil);
    delete doomed;
    delete view;
}

void NLElement::DeleteViews () {
    Iterator i;

    First(i);
    while (!Done(i)) {
        DeleteView(i);
    }
}    

/*****************************************************************************/

ClassId NLAlias::GetClassId () { return NL_ALIAS; }

boolean NLAlias::IsA (ClassId id) {
    return NL_ALIAS == id || NLElement::IsA(id);
}

NLAlias::NLAlias (AliasComp* subj) : NLElement(subj) { }
AliasComp* NLAlias::GetAliasComp () { return (AliasComp*) GetSubject(); }

boolean NLAlias::Definition (ostream& out) {
    AliasComp* alias = GetAliasComp();
    boolean ok = true;

    if (_flattening) {
        ok = FlatDefinition(alias, out);

    } else {
        Identification(out);
        BeginBlock(out);
        ok = NetlistView::Definition(out);

        ElementComp* defn = alias->GetDefinition(); 
        NetlistView* nlv = CreateNetlistView(defn);
        ok = nlv->Definition(out);
        delete nlv;

        EndBlock(out);
    }
    return ok;
}

boolean NLAlias::FlatDefinition (AliasComp* alias, ostream& out) {
    ElementComp* defn = alias->GetDefinition(); 
    boolean ok = true;
    Iterator i;

    for (defn->First(i); !defn->Done(i); defn->Next(i)) {
        GraphicComp* comp = defn->GetComp(i);
        NetlistView* nlv = CreateNetlistView(comp);

        if (nlv != nil) {
            ok = ok & nlv->Definition(out);
            delete nlv;
        }
    }
    return ok;
}        

/*****************************************************************************/

ClassId NLNode::GetClassId () { return NL_NODE; }

boolean NLNode::IsA (ClassId id) {
    return NL_NODE == id || NetlistView::IsA(id);
}

NLNode::NLNode (NodeComp* subj) : NetlistView(subj) { }
NodeComp* NLNode::GetNodeComp () { return (NodeComp*) GetSubject(); }

boolean NLNode::Definition (ostream& out) {
    Path path;

    Identification(out);
    BeginBlock(out);
    boolean status = Connections(GetNodeComp(), out, &path);
    EndBlock(out);

    return status;
}

boolean NLNode::Identify (NodeComp* node, ostream& out, Path* path) {
    boolean status = true;

    if (!path->Visited(node)) {
        path->Visit(node);
        Component* parent = node->GetParent();

        if (parent->IsA(WIRE_COMP)) {
            status = IdentifyWireNode(node, out, path);

        } else if (parent->IsA(ALIAS_COMP)) {
            status = IdentifyAliasNode(node, out);

        } else if (parent->IsA(ELEMENT_COMP)) {
            status = IdentifyElemNode(node, out);

        } else {
            status = Connections(node, out, path);
        }
    }
    return status;
}

boolean NLNode::IdentifyWireNode (NodeComp* node, ostream& out, Path* path) {
    WireComp* wireComp = (WireComp*) node->GetParent();
    NodeComp* n1, *n2;
    wireComp->GetNodes(n1, n2);
    node = (node == n1) ? n2 : n1;
    return Connections(node, out, path);
}

NodeComp* NLNode::FindCorrespDefNode (NodeComp* node, ElementComp* defn) {
    Iterator i;
    NodeVar* nodeVar = node->GetNodeVar();
    NodeComp* defNode = nil;

    for (defn->First(i); !defn->Done(i); defn->Next(i)) {
        GraphicComp* comp = defn->GetComp(i);

        if (comp->IsA(NODE_COMP)) {
            defNode = (NodeComp*) comp;
            NodeVar* defNodeVar = defNode->GetNodeVar();

            if (strcmp(defNodeVar->GetName(), nodeVar->GetName()) == 0) {
                break;
            } else {
                defNode = nil;
            }
        }
    }
    return defNode;
}

boolean NLNode::IdentifyAliasNode (NodeComp* node, ostream& out) {
    boolean status;

    if (_flattening) {
        AliasComp* alias = (AliasComp*) node->GetParent();
        ElementComp* defn = alias->GetDefinition();
        NodeComp* defNode = FindCorrespDefNode(node, defn);

        if (defNode == nil) {
            status = IdentifyElemNode(node, out);
        } else {
            Path path;
            status = Identify(defNode, out, &path);
        }
    } else {
        status = IdentifyElemNode(node, out);
    }
    return status;
}

boolean NLNode::IdentifyElemNode (NodeComp* node, ostream& out) {
    ElementComp* elem = (ElementComp*) node->GetParent();
    ElementVar* elemVar = elem->GetElemVar();
    NodeVar* nodeVar = node->GetNodeVar();

    Indent(out);
    out << elemVar->Identification() << "." << nodeVar->GetName() << ";\n";
    return out.good();
}

inline NodeComp* Node (UList* r) { return (NodeComp*) (*r)(); }

boolean NLNode::Connections (NodeComp* node, ostream& out, Path* path) {
    boolean status = true;
    UList* cnxns = node->GetConnections();

    for (UList* u = cnxns->First(); u != cnxns->End(); u = u->Next()) {
        status = Identify(Node(u), out, path);
    }        
    return status;
}

void NLNode::Identification (ostream& out) {
    Indent(out);
    out << GetNodeComp()->GetNodeVar()->GetName() << " ";
}

/*****************************************************************************/

ClassId NLGraphicComps::GetClassId () { return NL_GRAPHIC_COMPS; }

boolean NLGraphicComps::IsA (ClassId id) {
    return NL_GRAPHIC_COMPS == id || NetlistView::IsA(id);
}

NLGraphicComps::NLGraphicComps (GraphicComps* subj) : NetlistView(subj) {
    _views = new UList;
}

NLGraphicComps::~NLGraphicComps () {
    DeleteViews();
    delete _views;
}

UList* NLGraphicComps::Elem (Iterator i) { return (UList*) i.GetValue(); }

void NLGraphicComps::DeleteView (Iterator& i) {
    UList* doomed = Elem(i);
    ExternView* view = GetView(i);

    Next(i);
    _views->Remove(doomed);
    SetParent(view, nil);
    delete doomed;
    delete view;
}

void NLGraphicComps::DeleteViews () {
    Iterator i;

    First(i);
    while (!Done(i)) {
        DeleteView(i);
    }
}    

void NLGraphicComps::Update () {
    DeleteViews();

    GraphicComp* comps = GetGraphicComp();
    Iterator i;

    for (comps->First(i); !comps->Done(i); comps->Next(i)) {
        GraphicComp* comp = comps->GetComp(i);
        NetlistView* nlv = CreateNetlistView(comp);

        if (nlv != nil) {
            _views->Append(new UList(nlv));
        }
    }
}
