/*
 * Copyright (c) 1990 Stanford University
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided
 * that the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of Stanford not be used in advertising or
 * publicity pertaining to distribution of the software without specific,
 * written prior permission.  Stanford makes no representations about
 * the suitability of this software for any purpose.  It is provided "as is"
 * without express or implied warranty.
 *
 * STANFORD DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS.
 * IN NO EVENT SHALL STANFORD BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 * DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
 * OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION
 * WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

/*
 * ToolsPalette and DrawingPalette implementation.
 */

#include "scmclasses.h"
#include "scmcomps.h"
#include "scmkybd.h"
#include "scmpal.h"
#include "scmtools.h"

#include <Unidraw/catalog.h>
#include <Unidraw/ctrlinfo.h>
#include <Unidraw/editorinfo.h>
#include <Unidraw/iterator.h>
#include <Unidraw/keymap.h>
#include <Unidraw/kybd.h>
#include <Unidraw/selection.h>
#include <Unidraw/statevars.h>
#include <Unidraw/stateviews.h>
#include <Unidraw/uctrls.h>
#include <Unidraw/unidraw.h>

#include <Unidraw/Commands/brushcmd.h>
#include <Unidraw/Commands/colorcmd.h>
#include <Unidraw/Commands/font.h>
#include <Unidraw/Commands/patcmd.h>

#include <Unidraw/Components/ellipse.h>
#include <Unidraw/Components/line.h>
#include <Unidraw/Components/polygon.h>
#include <Unidraw/Components/rect.h>
#include <Unidraw/Components/spline.h>
#include <Unidraw/Components/text.h>

#include <Unidraw/Graphic/ellipses.h>
#include <Unidraw/Graphic/lines.h>
#include <Unidraw/Graphic/polygons.h>
#include <Unidraw/Graphic/pspaint.h>
#include <Unidraw/Graphic/splines.h>

#include <Unidraw/Tools/connect.h>
#include <Unidraw/Tools/grcomptool.h>
#include <Unidraw/Tools/magnify.h>
#include <Unidraw/Tools/move.h>
#include <Unidraw/Tools/reshape.h>
#include <Unidraw/Tools/rotate.h>
#include <Unidraw/Tools/scale.h>
#include <Unidraw/Tools/select.h>
#include <Unidraw/Tools/stretch.h>

#include <InterViews/border.h>
#include <InterViews/box.h>
#include <InterViews/frame.h>
#include <InterViews/glue.h>
#include <InterViews/menu.h>
#include <InterViews/message.h>
#include <InterViews/shape.h>
#include <InterViews/window.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/*****************************************************************************/

static const char* brAttrib = "brush";
static const char* fontAttrib = "font";
static const char* patAttrib = "pattern";
static const char* fgAttrib = "fgcolor";
static const char* bgAttrib = "bgcolor";

static const char* initBrAttrib = "initialbrush";
static const char* initFontAttrib = "initialfont";
static const char* initPatAttrib = "initialpattern";
static const char* initFgAttrib = "initialfgcolor";
static const char* initBgAttrib = "initialbgcolor";

/*****************************************************************************/

inline void InsertSeparator (PulldownMenu* pdm) {
    pdm->GetScene()->Insert(
        new VBox(
            new VGlue(2, 0, 0),
            new HBorder,
            new VGlue(2, 0, 0)
        )
    );
}

inline PulldownMenu* MakePulldown (const char* name) {
    return new PulldownMenu(new Message(name, Center, round(.1*cm)));
}

/*****************************************************************************/

ToolPalette::ToolPalette () {
    SetClassName("ToolPalette");

    ManagedWindow* window = new ApplicationWindow(this);
    SetWindow(window);

    _curCtrl = new ControlState;
    _selection = new Selection;
    _keymap = new KeyMap;

    InitEditorInfo();
    Insert(new Frame(Interior()));

    if (_initEdInfo) {
        Catalog* catalog = unidraw->GetCatalog();
        const char* attrib = catalog->GetAttribute("setup");
        catalog->Save(_edInfo, attrib);
    }
}

ToolPalette::~ToolPalette () {
    delete _keymap;
    delete _selection;

    if (_edInfo != nil) {
        delete _edInfo;
    }
}

KeyMap* ToolPalette::GetKeyMap () { return _keymap; }
Selection* ToolPalette::GetSelection () { return _selection; }

Tool* ToolPalette::GetCurTool () {
    UControl* c = (UControl*) _curCtrl->Selection();
    return (Tool*) c->GetControlInfo()->GetOwner();
}

void ToolPalette::SetKeyMap (KeyMap* k) { _keymap = k; }
void ToolPalette::SetSelection (Selection* s) { _selection = s; }

void ToolPalette::InstallOrRemove (const char* toolName, Tool* tool) {
    UControl* ctrl = FindControl(toolName);
    Catalog* catalog = unidraw->GetCatalog();
    const char* attrib = catalog->GetAttribute("setup");

    if (ctrl == nil) {
        if (tool == nil) {
            catalog->Retrieve(toolName, tool);
        }
        ctrl = new VPanelControl(tool->GetControlInfo(), _curCtrl);
        _compTools->Insert(ctrl);

        if (_edInfo != nil) {
            _edInfo->Register(toolName);
            catalog->Save(tool, toolName);
            catalog->Save(_edInfo, attrib);
        }
        _compTools->Change(ctrl);

    } else {
        if (_edInfo != nil) {
            _edInfo->UnregisterName(toolName);
            catalog->Save(_edInfo, attrib);
        }
        _compTools->Remove(ctrl);
        _compTools->Change(ctrl);

        tool = (Tool*) ctrl->GetControlInfo()->GetOwner();

        if (tool == GetCurTool()) {
            GetKeyMap()->Execute(CODE_SELECT);
        }
        _keymap->Unregister(ctrl);
        delete ctrl;
    }
}

Interactor* ToolPalette::Interior () { return Tools(); }

void ToolPalette::InitEditorInfo () {
    Catalog* catalog = unidraw->GetCatalog();
    const char* edInfoFile = catalog->GetAttribute("setup");

    if (edInfoFile == nil) {
        _edInfo = nil;
        _initEdInfo = false;

    } else if (catalog->Retrieve(edInfoFile, _edInfo)) {
        _initEdInfo = false;

    } else {
        _edInfo = new EditorInfo;
        _initEdInfo = true;
    }
}

Interactor* ToolPalette::Tools () {
    _compTools = new VBox;
    _tools = new VBox;

    InitTools();

    if (_edInfo == nil || _initEdInfo) {
        InitCompTools();
    } else {
        ReadCompTools();
    }
    HBox* palette = new HBox(
        new VBox(_tools, new VGlue),
        new VBorder,
        new VBox(_compTools, new VGlue)
    );
    palette->Propagate(false);
    return palette;
}

void ToolPalette::InitTools () {
    Include(
        new SelectTool(new ControlInfo("Select", KLBL_SELECT, CODE_SELECT))
    );
    Include(
        new MoveTool(new ControlInfo("Move", KLBL_MOVE, CODE_MOVE))
    );
    Include(
        new ConnectTool(new ControlInfo("Connect", KLBL_CONNECT,CODE_CONNECT))
    );
    Include(
        new ExamineTool(new ControlInfo("Examine", KLBL_EXAMINE, CODE_EXAMINE))
    );
    Include(
        new MagnifyTool(new ControlInfo("Magnify", KLBL_MAGNIFY, CODE_MAGNIFY))
    );
    Include(
        new ManipulateTool(new ControlInfo("Manipulate",KLBL_MANIP,CODE_MANIP))
    );
}

static const unit = 15;

static Coord xClosed[] = { unit/5, unit, unit, unit*3/5, 0 };
static Coord yClosed[] = { 0, unit/5, unit*3/5, unit, unit*2/5 };
static const int nClosed = 5;

static Coord xOpen[] = { 0, unit/2, unit/2, unit };
static Coord yOpen[] = { 0, unit/4, unit*3/4, unit };
static const int nOpen = 4;

static Coord xWire[] = { 0, unit/2, unit/2, unit };
static Coord yWire[] = { 0, 0, unit, unit };
static const int nWire = 4;

void ToolPalette::InitCompTools () {
    PinGraphic* pin = new PinGraphic(0, 0, stdgraphic);
    NodeComp* nodeComp = new NodeComp(pin);
    Include(
        new GraphicCompTool(
            new ControlInfo(nodeComp, KLBL_PIN, CODE_PIN), nodeComp
        ), "Node.tool"
    );

/*
    SlotGraphic* hslot = new SlotGraphic(0, 0, unit, stdgraphic);
    HSlotComp* hslotComp = new HSlotComp(hslot);
    Include(
        new GraphicCompTool(
            new ControlInfo(hslotComp), hslotComp
        ), "HSlot.tool"
    );

    SlotGraphic* vslot = new SlotGraphic(0, 0, unit, stdgraphic);
    VSlotComp* vslotComp = new VSlotComp(vslot);
    Include(
        new GraphicCompTool(
            new ControlInfo(vslotComp), vslotComp
        ), "VSlot.tool"
    );

    PadGraphic* pad = new PadGraphic(0, 0, unit, unit*4/5, stdgraphic);
    PadComp* padComp = new PadComp(pad);
    Include(
        new GraphicCompTool(
            new ControlInfo(padComp), padComp
        ), "Pad.tool"
    );
*/

    WireComp* wireComp = new WireComp(xWire, yWire, nWire, stdgraphic);
    Include(
        new GraphicCompTool(
            new ControlInfo(wireComp, KLBL_LINK, CODE_LINK), wireComp
        ), "Wire.tool"
    );

    BulbComp* bulbComp = new BulbComp(0, 0);
    Include(
        new GraphicCompTool(
            new ControlInfo(bulbComp, KLBL_BULB, CODE_BULB), bulbComp
        ), "Bulb.tool"
    );
    SwitchComp* switchComp = new SwitchComp(0, 0);
    Include(
        new GraphicCompTool(
            new ControlInfo(switchComp, KLBL_SWITCH, CODE_SWITCH), switchComp
        ), "Switch.tool"
    );
}

void ToolPalette::ReadCompTools () {
    for (int i = 0; i < _edInfo->Count(); ++i) {
        const char* toolName = _edInfo->GetName(i);
        const char* info = _edInfo->GetInfo(i);

        Tool* tool;
        unidraw->GetCatalog()->Retrieve(toolName, tool);
        Include(tool, toolName);
    }
}

void ToolPalette::Include (Tool* tool, const char* toolName) {
    ControlInfo* ctrlInfo = tool->GetControlInfo();
    UControl* ctrl = new VPanelControl(ctrlInfo, _curCtrl);
    _keymap->Register(ctrl);
    
    if (tool->IsA(GRAPHIC_COMP_TOOL)) {
        _compTools->Insert(ctrl);

        if (_initEdInfo && *toolName != '\0') {
            _edInfo->Register(toolName);
            unidraw->GetCatalog()->Save(tool, toolName);
        }

    } else {
        _tools->Insert(ctrl);
    }
}

UControl* ToolPalette::FindControl (const char* toolName) {
    const int nOrig = 30;
    Interactor* origCtrls[nOrig];
    int n;
    Interactor** ctrls = origCtrls;
    UControl* ctrl = nil;

    _compTools->GetComponents(origCtrls, nOrig, ctrls, n);
    
    for (int i = 0; i < n; ++i) {
        UControl* testCtrl = (UControl*) ctrls[i];
        Tool* testTool = (Tool*) testCtrl->GetControlInfo()->GetOwner();
        const char* testToolName = unidraw->GetCatalog()->GetName(testTool);

        if (testToolName != nil && strcmp(toolName, testToolName) == 0) {
            ctrl = testCtrl;
            break;
        }
    }
    if (ctrls != origCtrls) {
        delete ctrls;
    }
    return ctrl;
}

/*****************************************************************************/

DrawingPalette::DrawingPalette (ToolPalette* tp) {
    SetClassName("DrawingPalette");

    ManagedWindow* window = new ApplicationWindow(this);
    SetWindow(window);

    _tools = tp;
    _keymap = new KeyMap;

    InitStateVars();
    Insert(new Frame(Interior()));
}

DrawingPalette::~DrawingPalette () {
    delete _keymap;
}

Component* DrawingPalette::GetComponent () {
    Component* comp = nil;

    if (unidraw->Opened(_tools)) {
        Selection* s = GetSelection();

        if (!s->IsEmpty()) {
            Iterator i;
            s->First(i);
            comp = s->GetView(i)->GetGraphicComp();
        }
    }
    return comp;
}

KeyMap* DrawingPalette::GetKeyMap () { return _keymap; }
Selection* DrawingPalette::GetSelection (){return _tools->GetSelection();}
Tool* DrawingPalette::GetCurTool () { return _tools->GetCurTool(); }
boolean DrawingPalette::DependsOn (Component*) { return false; }

void DrawingPalette::SetKeyMap (KeyMap* k) { _keymap = k; }
void DrawingPalette::SetSelection (Selection* s){_tools->SetSelection(s);}

void DrawingPalette::Iconify (boolean iconify) {
    ManagedWindow* w = GetWindow();

    if (iconify) {
        w->iconify();
    } else {
        w->deiconify();
    }
}

StateVar* DrawingPalette::GetState (const char* name) {
    if (strcmp(name, "FontVar") == 0) {
        return _font;
    } else if (strcmp(name, "BrushVar") == 0) {
        return _brush;
    } else if (strcmp(name, "PatternVar") == 0) {
        return _pattern;
    } else if (strcmp(name, "ColorVar") == 0) {
        return _color;
    } else {
        return Editor::GetState(name);
    }
}

Interactor* DrawingPalette::Interior () {
    Interactor* tools = Tools();
    Interactor* commands = Commands();

    return new VBox(
        new HBox(
            new BrushVarView(_brush, _color),
            new VBorder,
            new PatternVarView(_pattern, _color),
            new VBorder,
            new HGlue,
            new FontVarView(_font, Right)
        ),
        new HBorder,
        new HBox(
            commands,
            new HGlue
        ),
        new HBorder,
        tools
    );
}

void DrawingPalette::InitStateVars () {
    _font = new FontVar;
    _brush = new BrushVar;
    _pattern = new PatternVar;
    _color = new ColorVar;

    Catalog* catalog = unidraw->GetCatalog();

    const char* brIndex = catalog->GetAttribute(initBrAttrib);
    const char* fontIndex = catalog->GetAttribute(initFontAttrib);
    const char* patIndex = catalog->GetAttribute(initPatAttrib);
    const char* fgIndex = catalog->GetAttribute(initFgAttrib);
    const char* bgIndex = catalog->GetAttribute(initBgAttrib);

    _brush->SetBrush(catalog->ReadBrush(brAttrib, atoi(brIndex)));
    _font->SetFont(catalog->ReadFont(fontAttrib, atoi(fontIndex)));
    _pattern->SetPattern(catalog->ReadPattern(patAttrib, atoi(patIndex)));
    _color->SetColors(
        catalog->ReadColor(fgAttrib, atoi(fgIndex)),
        catalog->ReadColor(bgAttrib, atoi(bgIndex))
    );
}

Interactor* DrawingPalette::Commands () {
    MenuBar* commands = new MenuBar;

    commands->Include(FontMenu());
    commands->Include(BrushMenu());
    commands->Include(PatternMenu());
    commands->Include(ColorMenu("FgColor", fgAttrib));
    commands->Include(ColorMenu("BgColor", bgAttrib));

    return commands;
}

PulldownMenu* DrawingPalette::FontMenu () {
    Catalog* catalog = unidraw->GetCatalog();
    PulldownMenu* pdm = MakePulldown("Font");

    int i = 1;
    PSFont* font = catalog->ReadFont(fontAttrib, i);

    while (font != nil) {
        TextGraphic* text = new TextGraphic(
            font->GetPrintFontAndSize(), stdgraphic
        );
        text->SetFont(font);

        Include(new FontCmd(new ControlInfo(new TextComp(text)), font), pdm);
        font = catalog->ReadFont(fontAttrib, ++i);
    }
    return pdm;
}

static float MENU_WIDTH = 1;   /* in cm */
static float MENU_HEIGHT = 0.5;

PulldownMenu* DrawingPalette::BrushMenu () {
    Catalog* catalog = unidraw->GetCatalog();
    PulldownMenu* pdm = MakePulldown("Brush");

    int i = 1;
    PSBrush* br = catalog->ReadBrush(brAttrib, i);

    while (br != nil) {
        ControlInfo* ctrlInfo;

        if (br->None()) {
            ctrlInfo = new ControlInfo("None");

        } else {
            Line* line = new Line(0, 0, round(MENU_WIDTH*cm), 0, stdgraphic);
            line->SetBrush(br);
            ctrlInfo = new ControlInfo(new LineComp(line));
        }
        
        Include(new BrushCmd(ctrlInfo, br), pdm);
        br = catalog->ReadBrush(brAttrib, ++i);
    }
    return pdm;
}

PulldownMenu* DrawingPalette::PatternMenu () {
    Catalog* catalog = unidraw->GetCatalog();
    PulldownMenu* pdm = MakePulldown("Pattern");

    int i = 1;
    PSPattern* pat = catalog->ReadPattern(patAttrib, i);
    
    while (pat != nil) {
        ControlInfo* ctrlInfo;

        if (pat->None()) {
            ctrlInfo = new ControlInfo("None");

        } else {
            SF_Rect* sfr = new SF_Rect(
                0, 0, round(MENU_WIDTH*cm), round(MENU_HEIGHT*cm), stdgraphic
            );
            sfr->SetPattern(pat);
            ctrlInfo = new ControlInfo(new RectComp(sfr));
        }
        
        Include(new PatternCmd(ctrlInfo, pat), pdm);
        pat = catalog->ReadPattern(patAttrib, ++i);
    }
    return pdm;
}

PulldownMenu* DrawingPalette::ColorMenu (const char* name, const char* attrib){
    Catalog* catalog = unidraw->GetCatalog();
    PulldownMenu* pdm = MakePulldown(name);

    int i = 1;
    PSColor* color = catalog->ReadColor(attrib, i);
    
    while (color != nil) {
        ControlInfo* ctrlInfo;
        Coord w = round(MENU_WIDTH*cm);
        Coord h = round(MENU_HEIGHT*cm);

        SF_Rect* sfr = new SF_Rect(0, 0, w, h, stdgraphic);
	sfr->SetColors(color, color);
	ctrlInfo = new ControlInfo(new RectComp(sfr), color->GetName());

        if (strcmp(attrib, fgAttrib) == 0) {
            Include(new ColorCmd(ctrlInfo, color, nil), pdm);
        } else {
            Include(new ColorCmd(ctrlInfo, nil, color), pdm);
        }
	color = catalog->ReadColor(attrib, ++i);
    }
    return pdm;
}

Interactor* DrawingPalette::Tools () {
    _dwgCompTools = new HBox;
    _dwgTools = new HBox;

    InitTools();

    _dwgCompTools->Insert(new HGlue);
    _dwgTools->Insert(new HGlue);

    return new VBox(
        _dwgCompTools,
        new HBorder,
        _dwgTools
    );
}

void DrawingPalette::InitTools () {
    Include(
        new ScaleTool(new ControlInfo("Scale", KLBL_SCALE, CODE_SCALE))
    );
    Include(
        new StretchTool(new ControlInfo("Stretch", KLBL_STRETCH,CODE_STRETCH))
    );
    Include(
        new RotateTool(new ControlInfo("Rotate", KLBL_ROTATE, CODE_ROTATE))
    );
    Include(
        new ReshapeTool(new ControlInfo("Reshape", KLBL_RESHAPE,CODE_RESHAPE))
    );

    TextGraphic* text = new TextGraphic("Text", stdgraphic);
    TextComp* textComp = new TextComp(text);
    Include(
        new GraphicCompTool(
            new ControlInfo("Text", KLBL_TEXT, CODE_TEXT), textComp
        )
    );

    Line* line = new Line(0, 0, unit, unit, stdgraphic);
    LineComp* lineComp = new LineComp(line);
    Include(
        new GraphicCompTool(
            new ControlInfo(lineComp, KLBL_LINE, CODE_LINE), lineComp
        )
    );

    SF_MultiLine* ml = new SF_MultiLine(xOpen, yOpen, nOpen, stdgraphic);
    ml->SetPattern(psnonepat);
    MultiLineComp* mlComp = new MultiLineComp(ml);
    Include(
        new GraphicCompTool(
            new ControlInfo(mlComp, KLBL_MULTILINE, CODE_MULTILINE), mlComp
        )
    );

    SFH_OpenBSpline* spline = new SFH_OpenBSpline(
        xOpen, yOpen, nOpen, stdgraphic
    );
    spline->SetPattern(psnonepat);
    SplineComp* splineComp = new SplineComp(spline);
    Include(
        new GraphicCompTool(
            new ControlInfo(splineComp, KLBL_SPLINE, CODE_SPLINE), splineComp
        )
    );

    SF_Rect* rect = new SF_Rect(0, 0, unit, unit*4/5, stdgraphic);
    rect->SetPattern(psnonepat);
    RectComp* rectComp = new RectComp(rect);
    Include(
        new GraphicCompTool(
            new ControlInfo(rectComp, KLBL_RECT, CODE_RECT), rectComp
        )
    );

    SF_Ellipse* ellipse = new SF_Ellipse(0, 0, unit*2/3, unit*2/5, stdgraphic);
    ellipse->SetPattern(psnonepat);
    EllipseComp* ellipseComp = new EllipseComp(ellipse);
    Include(
        new GraphicCompTool(
            new ControlInfo(ellipseComp, KLBL_ELLIPSE, CODE_ELLIPSE),
            ellipseComp
        )
    );

    SF_Polygon* polygon = new SF_Polygon(xClosed, yClosed, nClosed,stdgraphic);
    polygon->SetPattern(psnonepat);
    PolygonComp* polygonComp = new PolygonComp(polygon);
    Include(
        new GraphicCompTool(
            new ControlInfo(polygonComp, KLBL_POLY, CODE_POLY), polygonComp
        )
    );

    SFH_ClosedBSpline* cspline = new SFH_ClosedBSpline(
        xClosed, yClosed, nClosed, stdgraphic
    );
    cspline->SetPattern(psnonepat);
    ClosedSplineComp* csplineComp = new ClosedSplineComp(cspline);
    Include(
        new GraphicCompTool(
            new ControlInfo(csplineComp, KLBL_CSPLINE,CODE_CSPLINE),csplineComp
        )
    );
}

void DrawingPalette::Include (Command* cmd, PulldownMenu* pdm) {
    ControlInfo* ctrlInfo = cmd->GetControlInfo();
    UControl* ctrl = new CommandControl(ctrlInfo);
    _keymap->Register(ctrl);
    pdm->Include(ctrl);
    cmd->SetEditor(this);
}

void DrawingPalette::Include (Tool* tool) {
    ControlInfo* ctrlInfo = tool->GetControlInfo();
    UControl* ctrl = new HPanelControl(ctrlInfo, _tools->GetControlState());
    _keymap->Register(ctrl);
    
    if (tool->IsA(GRAPHIC_COMP_TOOL)) {
        _dwgCompTools->Insert(ctrl);
    } else {
        _dwgTools->Insert(ctrl);
    }
}
