/*
 * Copyright (c) 1990 Stanford University
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided
 * that the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of Stanford not be used in advertising or
 * publicity pertaining to distribution of the software without specific,
 * written prior permission.  Stanford makes no representations about
 * the suitability of this software for any purpose.  It is provided "as is"
 * without express or implied warranty.
 *
 * STANFORD DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS.
 * IN NO EVENT SHALL STANFORD BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 * DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
 * OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION
 * WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

/*
 * Implementation of schematic capture system-specific rubberbands.
 */

#include "scmrubs.h"

#include <InterViews/painter.h>

#include <math.h>

/*****************************************************************************/

RubberWire::RubberWire (
    Painter* p, Canvas* c, Coord px[], Coord py[], int n, boolean manhattan
) : GrowingVertices(p, c, px, py, n) {
    _manhattan = manhattan;
    CheckBufs();
    x[n] = x[n-1];
    y[n] = y[n-1];
}

void RubberWire::AppendVertex (Coord vx, Coord vy) {
    ConstrainVertices(vx, vy);
    GrowingVertices::AppendVertex(x[count], y[count]);
}

void RubberWire::Draw () {
    if (!drawn && count > 0) {
        ConstrainVertices(trackx, tracky);
        output->MultiLine(canvas, x, y, count+1);
        drawn = true;
    }
}

void RubberWire::Orthonormalize (Coord tx, Coord ty) {
    int last = count-1;
    boolean vert = abs(tx - x[last]) < abs(ty - y[last]);
    x[count] = vert ? x[last] : tx;
    y[count] = vert ? ty : y[last];
}

void RubberWire::ConstrainVertices (Coord tx, Coord ty) {
    CheckBufs();

    if (_manhattan) {
        int last = count-1;

        if (count > 1 && x[last] == x[last-1]) {
            x[count] = tx;
            y[count] = ty;
            y[last] = ty;
            
        } else if (count > 1 && y[last] == y[last-1]) {
            x[count] = tx;
            y[count] = ty;
            x[last] = tx;

        } else {
            Orthonormalize(tx, ty);
        }
    } else {
        x[count] = tx;
        y[count] = ty;
    }
}

void RubberWire::Manhattan (boolean m) { 
    if (_manhattan != m) {
        boolean wasDrawn = drawn;
        Erase();
        _manhattan = m;

        if (wasDrawn) {
            Draw();
        }
    }
}

