/*
 * Copyright (c) 1990 Stanford University
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided
 * that the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of Stanford not be used in advertising or
 * publicity pertaining to distribution of the software without specific,
 * written prior permission.  Stanford makes no representations about
 * the suitability of this software for any purpose.  It is provided "as is"
 * without express or implied warranty.
 *
 * STANFORD DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS.
 * IN NO EVENT SHALL STANFORD BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 * DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
 * OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION
 * WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

/*
 * Implementation of schematic capture system-specific tools.
 */

#include "scmclasses.h"
#include "scmcmds.h"
#include "scmcomps.h"
#include "scmmanips.h"
#include "scmtools.h"

#include <Unidraw/clipboard.h>
#include <Unidraw/iterator.h>
#include <Unidraw/manips.h>
#include <Unidraw/selection.h>
#include <Unidraw/viewer.h>
#include <Unidraw/Commands/edit.h>
#include <Unidraw/Components/grview.h>
#include <Unidraw/Graphic/graphic.h>

#include <InterViews/event.h>
#include <InterViews/menu.h>
#include <InterViews/rubrect.h>
#include <InterViews/transformer.h>

#include "math.h"

/*****************************************************************************/

class PopupItem : public MenuItem {
public:
    PopupItem(const char*, Alignment, Command*);

    virtual void Do();
private:
    Command* _cmd;
};

PopupItem::PopupItem (
    const char* str, Alignment al, Command* cmd
) : MenuItem(str, al) {
    _cmd = cmd;
}

void PopupItem::Do () { _cmd->Execute(); }

/*****************************************************************************/

ClassId ExamineTool::GetClassId () { return EXAMINE_TOOL; }
boolean ExamineTool::IsA (ClassId id) {return EXAMINE_TOOL==id||Tool::IsA(id);}

ExamineTool::ExamineTool (ControlInfo* m) : Tool(m) {
    _aliasCmd = new AliasCmd;
    _defnCmd = new DefinitionCmd;
    _infoCmd = new InfoCmd;
    _newViewCmd = new NewViewCmd;
    _truthTableCmd = new TruthTableCmd;
}

ExamineTool::~ExamineTool () {
    delete _aliasCmd;
    delete _defnCmd;
    delete _infoCmd;
    delete _newViewCmd;
    delete _truthTableCmd;
}

Tool* ExamineTool::Copy () { return new ExamineTool(CopyControlInfo()); }

Manipulator* ExamineTool::CreateManipulator (Viewer* v, Event& e,Transformer*){
    GraphicView* views = v->GetGraphicView();
    Selection* s = v->GetSelection(), *newSel;
    Manipulator* m = nil;

    s->Clear();
    newSel = views->ViewIntersecting(e.x-SLOP, e.y-SLOP, e.x+SLOP, e.y+SLOP);

    if (!newSel->IsEmpty()) {
        Iterator i;
        newSel->First(i);
	GraphicView* gv = newSel->GetView(i);

        s->Append(gv);
        s->Update();

        InitCmds(v, gv->GetGraphicComp());
        m = CreatePopupManip(v);
    }
    delete newSel;
    return m;
}

void ExamineTool::InitCmds (Viewer* v, GraphicComp* comp) {
    Editor* ed = v->GetEditor();

    _newViewCmd->SetGraphicComp(comp);
    _infoCmd->SetGraphicComp(comp);
    _aliasCmd->SetGraphicComp(comp);
    _defnCmd->SetGraphicComp(comp);
    _truthTableCmd->SetGraphicComp(comp);

    _newViewCmd->SetEditor(ed);
    _infoCmd->SetEditor(ed);
    _aliasCmd->SetEditor(ed);
    _defnCmd->SetEditor(ed);
    _truthTableCmd->SetEditor(ed);
}    

Manipulator* ExamineTool::CreatePopupManip (Viewer* v) {
    PopupMenu* popup = new PopupMenu;
    
    popup->Include(new PopupItem("Create Alias", Center, _aliasCmd));
    popup->Include(new PopupItem("Definition", Center, _defnCmd));
    popup->Include(new PopupItem("Info...", Center, _infoCmd));
    popup->Include(new PopupItem("Truth Table...", Center, _truthTableCmd));
    popup->Include(new PopupItem("View", Center, _newViewCmd));

    return new PopupManip(v, popup, this);
}

/*****************************************************************************/

ClassId ManipulateTool::GetClassId () { return MANIPULATE_TOOL; }

boolean ManipulateTool::IsA (ClassId id) {
    return MANIPULATE_TOOL == id || Tool::IsA(id);
}

ManipulateTool::ManipulateTool (ControlInfo* m) : Tool(m) { }
Tool* ManipulateTool::Copy () { return new ManipulateTool(CopyControlInfo()); }

Manipulator* ManipulateTool::CreateManipulator (
    Viewer* v, Event& e, Transformer* rel
) {
    GraphicView* views = v->GetGraphicView();
    Selection* s = v->GetSelection(), *newSel;
    GraphicView* gv;
    Manipulator* m = nil;
    Iterator i;

    newSel = views->ViewIntersecting(e.x-SLOP, e.y-SLOP, e.x+SLOP, e.y+SLOP);
    if (!newSel->IsEmpty()) {
        s->Clear();
        newSel->First(i);
	gv = newSel->GetView(i);
        m = gv->CreateManipulator(v, e, rel, this);
    }
    delete newSel;
    return m;
}

Command* ManipulateTool::InterpretManipulator (Manipulator* m) {
    Selection* s;
    Command* cmd = nil;
    Iterator i;
    
    if (m != nil) {
        s = m->GetViewer()->GetSelection();
        s->First(i);
        cmd = s->GetView(i)->InterpretManipulator(m);
    }
    return cmd;
}

/*****************************************************************************/

InstanceTool::InstanceTool () { }

InstanceTool::InstanceTool (
    ControlInfo* m, ElementComp* p
) : GraphicCompTool(m, p) { }

ClassId InstanceTool::GetClassId () { return INSTANCE_TOOL; }

boolean InstanceTool::IsA (ClassId id) {
    return INSTANCE_TOOL == id || GraphicCompTool::IsA(id);
}

Tool* InstanceTool::Copy () {
    return new InstanceTool(CopyControlInfo(), GetOriginal());
}

Manipulator* InstanceTool::CreateManipulator (
    Viewer* v, Event& e, Transformer* rel
) {
    GraphicView* views = v->GetGraphicView();
    Selection* s = v->GetSelection();
    SlidingRect* sr;
    Coord x0, y0, x1, y1, halfw, halfh;

    s->Clear();
    GetOriginal()->GetGraphic()->GetBox(x0, y0, x1, y1);
    if (rel != nil) {
        rel->Transform(x0, y0);
        rel->Transform(x1, y1);
        halfw = abs(x1 - x0) / 2;
        halfh = abs(y1 - y0) / 2;
    }
    v->Constrain(e.x, e.y);
    sr = new SlidingRect(
        nil, nil, e.x - halfw, e.y - halfh, e.x + halfw, e.y + halfh, e.x, e.y
    );
    return new DragManip(v, sr, rel, this, Gravity);
}

Command* InstanceTool::InterpretManipulator (Manipulator* m) {
    DragManip* dm = (DragManip*) m;
    Editor* ed = dm->GetViewer()->GetEditor();
    Transformer* rel = dm->GetTransformer();
    SlidingRect* slidingRect = (SlidingRect*) dm->GetRubberband();
    Coord cx0, cy0, cx1, cy1;
    float cx, cy, px, py;
    GraphicComp* newComp = new InstanceComp(GetOriginal());
    Graphic* grcomp = newComp->GetGraphic();

    slidingRect->GetCurrent(cx0, cy0, cx1, cy1);
    px = float(cx0 + cx1) / 2;
    py = float(cy0 + cy1) / 2;
    rel->InvTransform(px, py, cx, cy);
    grcomp->GetCenter(px, py);
    grcomp->Translate(cx - px, cy - py);
    return new PasteCmd(ed, new Clipboard(newComp));
}
