/*
 * Copyright (c) 1990 Stanford University
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided
 * that the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of Stanford not be used in advertising or
 * publicity pertaining to distribution of the software without specific,
 * written prior permission.  Stanford makes no representations about
 * the suitability of this software for any purpose.  It is provided "as is"
 * without express or implied warranty.
 *
 * STANFORD DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS.
 * IN NO EVENT SHALL STANFORD BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 * DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
 * OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION
 * WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

/*
 * Implementation of schematic capture system-specific transfer functions.
 */

#include "scmclasses.h"
#include "scmcomps.h"
#include "scmtransfns.h"
#include "scmtruth.h"

#include <Unidraw/catalog.h>
#include <Unidraw/statevars.h>
#include <Unidraw/unidraw.h>

#include <stream.h>

/*****************************************************************************/

TF_TruthTable::TF_TruthTable (TruthTable* tt) { _truth = tt; }
TF_TruthTable::~TF_TruthTable () { delete _truth; }

unsigned TF_TruthTable::GetInputs () {
    unsigned inputs = 0;
    unsigned mask = 1;

    for (int i = 0; i < _truth->Inputs(); ++i) {
        if (_truth->GetInputVar(i)->GetLogicLevel() == High) {
            inputs |= mask;
        }
        mask <<= 1;
    }
    return inputs;
}

void TF_TruthTable::Evaluate (Path* path) {
    unsigned inputs = GetInputs();

    for (int i = 0; i < _truth->Outputs(); ++i) {
        NodeVar* outputVar = _truth->GetOutputVar(i);
        LogicLevel output = _truth->GetOutput(inputs, i);

        if (outputVar->GetLogicLevel() != output) {
            outputVar->SetLogicLevel(output);
            outputVar->GetBinding()->Transmit(path);
        }
    }
}

ClassId TF_TruthTable::GetClassId () { return TF_TRUTHTABLE; }

boolean TF_TruthTable::IsA (ClassId id) {
    return TF_TRUTHTABLE == id || TransferFunct::IsA(id); 
}

TransferFunct* TF_TruthTable::Copy () {
    return new TF_TruthTable(_truth->Copy());
}

void TF_TruthTable::Read (istream& in) {
    TransferFunct::Read(in);
    Catalog* catalog = unidraw->GetCatalog();
    int i, inputs, outputs;

    in >> inputs >> outputs;
    _truth = new TruthTable(inputs, outputs);

    for (i = 0; i < inputs; ++i) {
        NodeVar* nodeVar = (NodeVar*) catalog->ReadStateVar(in);
        _truth->SetInputVar(nodeVar, i);
    }
    for (i = 0; i < outputs; ++i) {
        NodeVar* nodeVar = (NodeVar*) catalog->ReadStateVar(in);
        _truth->SetOutputVar(nodeVar, i);
    }

    int inputCombos = _truth->InputCombos();

    for (unsigned inputVals = 0; inputVals < inputCombos; ++inputVals) {
        for (i = 0; i < _truth->Outputs(); ++i) {
            int level;
            in >> level;
            _truth->SetOutput(inputVals, i, LogicLevel(level));
        }
    }
}

void TF_TruthTable::Write (ostream& out) {
    TransferFunct::Write(out);
    Catalog* catalog = unidraw->GetCatalog();
    int i;

    out << _truth->Inputs() << " " << _truth->Outputs() << " ";

    for (i = 0; i < _truth->Inputs(); ++i) {
        catalog->WriteStateVar(_truth->GetInputVar(i), out);
    }
    for (i = 0; i < _truth->Outputs(); ++i) {
        catalog->WriteStateVar(_truth->GetOutputVar(i), out);
    }
    
    int inputCombos = _truth->InputCombos();

    for (unsigned inputVals = 0; inputVals < inputCombos; ++inputVals) {
        for (i = 0; i < _truth->Outputs(); ++i) {
            out << _truth->GetOutput(inputVals, i) << " ";
        }
    }
}
