/*
 * Copyright (c) 1990 Stanford University
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided
 * that the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of Stanford not be used in advertising or
 * publicity pertaining to distribution of the software without specific,
 * written prior permission.  Stanford makes no representations about
 * the suitability of this software for any purpose.  It is provided "as is"
 * without express or implied warranty.
 *
 * STANFORD DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS.
 * IN NO EVENT SHALL STANFORD BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 * DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
 * OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION
 * WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

/*
 * Implementation of TruthTable and TruthTableEditor.
 */

#include "scmtruth.h"

#include <InterViews/button.h>
#include <InterViews/font.h>
#include <InterViews/painter.h>
#include <InterViews/sensor.h>
#include <InterViews/textbuffer.h>

/*****************************************************************************/

static const int gap = 2;

/*****************************************************************************/

static int twoToThe (int n) {
    int result = 1;

    for (int i = 0; i < n; ++i) {
        result *= 2;
    }
    return result;
}

/*****************************************************************************/

TruthTable::TruthTable (int inputs, int outputs) {
    _inputs = inputs;
    _outputs = outputs;
    _inVar = new NodeVar*[Inputs()];
    _outVar = new NodeVar*[Outputs()];

    int inputCombos = InputCombos();
    _output = new LogicLevel*[inputCombos];

    for (int i = 0; i < inputCombos; ++i) {
        _output[i] = new LogicLevel[Outputs()];

        for (int j = 0; j < Outputs(); ++j) {
            _output[i][j] = Low;
        }
    }
}

TruthTable::~TruthTable () {
    delete _inVar;
    delete _outVar;
    
    for (int i = 0; i < Outputs(); ++i) {
        delete _output[i];
    }
    delete _output;
}

void TruthTable::SetInputVar (NodeVar* node, int index) {_inVar[index] = node;}
void TruthTable::SetOutputVar (NodeVar* node, int index) {_outVar[index]=node;}
NodeVar* TruthTable::GetInputVar (int index) { return _inVar[index]; }
NodeVar* TruthTable::GetOutputVar (int index) { return _outVar[index]; }
int TruthTable::InputCombos () { return twoToThe(Inputs()); }

void TruthTable::SetOutput (unsigned inputs, int index, LogicLevel output) {
    _output[inputs][index] = output;
}

LogicLevel TruthTable::GetOutput (unsigned inputs, int index) {
    return _output[inputs][index];
}

TruthTable& TruthTable::operator = (TruthTable& tt) {
    int inputCombos = InputCombos();

    for (unsigned i = 0; i < inputCombos; ++i) {
        for (int j = 0; j < Outputs(); ++j) {
            SetOutput(i, j, tt.GetOutput(i, j));
        }
    }
    return *this;
}

TruthTable* TruthTable::Copy () {
    TruthTable* copy = new TruthTable(_inputs, _outputs);
    int j;

    for (j = 0; j < Inputs(); ++j) {
        copy->SetInputVar(GetInputVar(j), j);
    }
    for (j = 0; j < Outputs(); ++j) {
        copy->SetOutputVar(GetOutputVar(j), j);
    }
    for (unsigned i = 0; i < InputCombos(); ++i) {
        for (j = 0; j < Outputs(); ++j) {
            copy->SetOutput(i, j, GetOutput(i, j));
        }
    }
    return copy;
}

/*****************************************************************************/

TruthTableEditor::TruthTableEditor (
    ButtonState* bs, TruthTable* tt, const char* done
) : TextEditor(8, tt->Inputs() + gap + tt->Outputs(), 0, Reversed) {
    SetClassName("TruthTableEditor");
    _leftMargin = tt->Inputs() + gap;
    _lineWidth = _leftMargin + tt->Outputs();
    _totalLines = tt->InputCombos();

    _state = bs;
    _truth = tt;
    _done = done;
    _bufsize = _totalLines * (_lineWidth+1);    // add one for newlines
    _buf = new char[_bufsize];

    input = new Sensor;
    input->Catch(DownEvent);
    input->Catch(KeyEvent);
    Ref(input);

    UpdateTextBuffer();
    Select(LimitIndex(0));
}

TruthTableEditor::~TruthTableEditor () {
    delete text;
    delete _buf;
    delete _truth;
}

void TruthTableEditor::Handle (Event& e) {
    if (e.eventType == KeyEvent) {
        HandleKeyEvent(e);

    } else {
        boolean done = false;

        do {
            switch (e.eventType) {
            case DownEvent:
                done = HandleDownEvent(e);
                break;

            case KeyEvent:
                done = HandleKeyEvent(e);
                break;
            }
            if (!done) {
                Read(e);
            }
        } while (!done);
    }
}

boolean TruthTableEditor::HandleKeyEvent (Event& e) {
    boolean done = false;

    if (e.len != 0) {
        done = HandleChar(e.keystring[0]);
    }
    return done;
}

boolean TruthTableEditor::HandleDownEvent (Event& e) {
    boolean done = true;

    if (e.target == this) {
        switch (e.button) {
            case LEFTMOUSE:
                Select(LimitIndex(Locate(e.x, e.y)));
                done = false; 
                break;

            case MIDDLEMOUSE:   GrabScroll(e); break;
            case RIGHTMOUSE:    RateScroll(e); break;
        }

    } else {
        UnRead(e);
    }
    return done;
}

boolean TruthTableEditor::HandleChar (char c) {
    boolean done = false;

    switch (c) {
    case '\010':
    case '\177':
        AdvanceChar(-1);
        DeleteText(1);
        InsertText('x');
        AdvanceChar(-1);
        UpdateTruthTable();
        break;

    case '\004':
    case 'x':
    case ' ':    c = 'x'; // fall through
    case '0':
    case '1':
        DeleteText(1);
        InsertText(c);
        AdvanceChar(0);
        UpdateTruthTable();
        break;

    case '\006':  AdvanceChar(1); break;
    case '\002':  AdvanceChar(-1); break;
    case '\016':  AdvanceLine(1); break;
    case '\020':  AdvanceLine(-1); break;

    default:
        for (int i = 0; _done[i] != '\0'; ++i) {
            if (c == _done[i]) {
                _state->SetValue(c);
                done = true;
            }
        }
        break;
    }
    return done;
}

int TruthTableEditor::LimitIndex (int index) {
    int offset = text->LineOffset(index);

    if (offset == 0) {
        if (text->IsEndOfText(index)) {
            index = text->EndOfPreviousWord(index) - 1;
        } else {
            index = text->BeginningOfNextWord(index);
        }

    } else if (offset < _leftMargin) {
        if (text->LineNumber(index) == 0) {
            index = text->BeginningOfNextWord(index);
        } else {
            index = text->EndOfPreviousLine(index) - 1;
        }

    } else if (offset >= _lineWidth) {
        if (text->LineNumber(index) == _totalLines-1) {
            --index;
        } else {
            index = text->BeginningOfNextWord(index+1);
        }
    }
    return index;
}

void TruthTableEditor::InsertText (char c) {
    char string[1];
    *string = c;
    TextEditor::InsertText(string, 1);
}    

void TruthTableEditor::AdvanceChar (int n) { Select(LimitIndex(Dot()+n)); }

void TruthTableEditor::AdvanceLine (int n) { 
    int line = text->LineNumber(Dot());
    int offset = text->LineOffset(Dot());
    Select(text->LineIndex(line+n) + offset);
}

static const char* Binary (unsigned n, int width) {
    char buf[CHARBUFSIZE];
    unsigned mask = 1 << (width-1);

    for (int i = 0; i < width; ++i) {
        buf[i] = (n & (mask >> i)) ? '1' : '0';
    }
    return buf;
}

void TruthTableEditor::UpdateTextBuffer () {
    if (text != nil) {
        delete text;
        text = nil;
    }
    Edit(new TextBuffer(_buf, 0, _bufsize));
    Select(0);

    for (unsigned input = 0; input < _totalLines; ++input) {
        TextEditor::InsertText(
            Binary(input, _truth->Inputs()), _truth->Inputs()
        );
        for (int i = 0; i < gap; ++i) {
            InsertText(' ');
        }

        for (int output = 0; output < _truth->Outputs(); ++output) {
            LogicLevel level = _truth->GetOutput(input, output);
            char c;

            switch (level) {
                case Low:       c = '0'; break;
                case High:      c = '1'; break;
                case DontCare:  c = 'x'; break;
            }
            InsertText(c);
        }
        InsertText('\n');
    }
}

void TruthTableEditor::UpdateTruthTable () {
    for (unsigned input = 0; input < _totalLines; ++input) {
        for (int output = 0; output < _truth->Outputs(); ++output) {
            int index = text->LineIndex(input) + _leftMargin + output;
            char c = text->Char(index);
            LogicLevel level;

            switch (c) {
                case '0':  level = Low; break;
                case '1':  level = High; break;
                default:   level = DontCare; break;
            }
            _truth->SetOutput(input, output, level);
        }
    }
}

