/*
 * Copyright (c) 1990 Stanford University
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided
 * that the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of Stanford not be used in advertising or
 * publicity pertaining to distribution of the software without specific,
 * written prior permission.  Stanford makes no representations about
 * the suitability of this software for any purpose.  It is provided "as is"
 * without express or implied warranty.
 *
 * STANFORD DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS.
 * IN NO EVENT SHALL STANFORD BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 * DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
 * OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION
 * WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

/*
 * Implementation of schematic capture system-specific state variable views.
 */

#include "scmvars.h"
#include "scmvarviews.h"

#include <Unidraw/Components/connector.h>

#include <InterViews/box.h>
#include <InterViews/button.h>
#include <InterViews/frame.h>
#include <InterViews/glue.h>
#include <InterViews/streditor.h>

#include <string.h>

/*****************************************************************************/

ElementVarView::ElementVarView (
    ElementVar* elemVar, ButtonState* bs
) : StateVarView(elemVar) {
    StringEditor* sedit = new StringEditor(bs, "mmmmmmmmmmmmmmmmmmmmm");
    Insert(sedit);
}

void ElementVarView::ChangeSubject () {
    GetSubject()->SetName(GetStringEditor()->Text());
}

void ElementVarView::Init () {
    GetStringEditor()->Message(GetSubject()->GetName());
}

boolean ElementVarView::Stale () {
    const char* subjText = GetSubject()->GetName();
    const char* viewText = GetStringEditor()->Text();

    return strcmp(viewText, subjText) != 0;
}

StringEditor* ElementVarView::GetStringEditor () {
    return (StringEditor*) interior();
}

void ElementVarView::Handle (Event& e) { GetStringEditor()->Handle(e); }

/*****************************************************************************/

NodeVarView::NodeVarView (
    NodeVar* nodeVar, ButtonState* bs
) : StateVarView(nodeVar) {
    Connector* conn = nodeVar->GetBinding();
    TransMethod transMethod = (conn == nil) ? In : conn->GetTransMethod();
    
    const int space = round(.25*cm);
    _sedit = new StringEditor(bs, "mmmmmmmmmmmmmmmmmmmmm");
    _iobs = new ButtonState(transMethod);
    _in = new RadioButton("input", _iobs, In);
    _out = new RadioButton("output", _iobs, Out);
    _inout = new RadioButton("bi-directional", _iobs, InOut);

    HBox* hbox = new HBox(
        new Frame(new MarginFrame(_sedit, 2)),
        new HGlue(space, space, 0),
        _in,
        new HGlue(space, space, 0),
        _out,
        new HGlue(space, space, 0),
        _inout
    );
    hbox->Align(Center);

    Insert(hbox);
}

void NodeVarView::ChangeSubject () {
    NodeVar* nodeVar = GetSubject();
    Connector* conn = nodeVar->GetBinding();

    if (conn != nil) {
        unsigned transMethod;
        _iobs->GetValue((void*&) transMethod);
        conn->SetTransMethod(TransMethod(transMethod));
    }

    nodeVar->SetName(_sedit->Text());
}

void NodeVarView::Init () {
    NodeVar* nodeVar = GetSubject();
    Connector* conn = nodeVar->GetBinding();

    _iobs->SetValue(conn->GetTransMethod());
    _sedit->Message(nodeVar->GetName());
}

boolean NodeVarView::Stale () {
    boolean staleTransMethod = false;
    NodeVar* nodeVar = GetSubject();
    Connector* conn = nodeVar->GetBinding();

    if (conn != nil) {
        unsigned viewTransMethod;
        _iobs->GetValue((void*&) viewTransMethod);

        staleTransMethod = viewTransMethod != conn->GetTransMethod();
    }

    const char* subjText = GetSubject()->GetName();
    const char* viewText = _sedit->Text();

    return staleTransMethod || strcmp(viewText, subjText) != 0;
}

void NodeVarView::Handle (Event& e) { _sedit->Handle(e); }
