/* Copyright (c) 1991 by Sanjay Ghemawat. */

#include "board.h"
#include "generator.h"

Board::Board(ObjectGen* og, int ht, int wd, int mr) {
    generator = og;
    height = ht;
    width = wd;
    maxrest = mr;
    assert(height > 0);
    assert(width > 0);
    assert(maxrest >= 0);
    filled = new boolean[height];
    square = new Square*[height];
    for (int i = 0; i < height; i++) square[i] = new Square[width];
    Clear();
}

Board::~Board() {
    for (int i = 0; i < height; i++)
	delete[width] square[i];
    delete[height] filled;
    delete[height] square;
}

void Board::SetGenerator(ObjectGen* g) {
    generator = g;
}

void Board::Tick() {
    if (blocked) return;
    if (falling) {
	/* An object is falling */
	EraseFalling();
	if (LegalAt(x, y+1, rot)) {
	    y++;
	    atrest = 0;
	    DrawFalling();
	    return;
	}
	/* Could not fall another row */
	DrawFalling();
	atrest++;
	if (atrest > maxrest) {
	    /* Finalize falling object */
	    atrest = 0;
	    falling = false;
	    /* Add one point for each object allowed to fall */
	    score++;
	    /* Compute list of filled arrays */
	    for (int i = 0; i < height; i++) {
		Square* row = square[i];
		filled[i] = true;
		for (int j = 0; j < width; j++) {
		    if (row[j] == Empty) {
			filled[i] = false;
			break;
		    }
		}
	    }
	    return;
	}
	return;
    }
    /* Nothing was falling - start next object */
    falling = true;
    atrest = 0;
    x = width / 2;
    y = 0;
    ftype = (*generator)();
    rot = R0;
    if (LegalAt(x, y, rot)) {
	DrawFalling();
    }
    else {
	/* Could not put object in */
	falling = false;
	blocked = true;
    }
}

void Board::RotateFalling() {
    Rotation newrot = (Rotation) ((((int) rot) + 1) % NumRotations);
    if (falling) {
	EraseFalling();
	if (LegalAt(x, y, newrot)) {
	    rot = newrot;
	    atrest = 0;
	}
	DrawFalling();
    }
}

void Board::LeftShift() {
    if (falling) {
	EraseFalling();
	if (LegalAt(x-1, y, rot)) {
	    x--;
	    atrest = 0;
	}
	DrawFalling();
    }
}

void Board::RightShift() {
    if (falling) {
	EraseFalling();
	if (LegalAt(x+1, y, rot)) {
	    x++;
	    atrest = 0;
	}
	DrawFalling();
    }
}

void Board::FallAllTheWay() {
    if (falling) {
	EraseFalling();
	while (LegalAt(x, y+1, rot)) {
	    y++;
	    /* Add one point for each square that the object is allowed to drop */
	    /* without control. */
	    score++;
	}
	DrawFalling();
	atrest = maxrest;
	Tick();
    }
}

/* Delete all filled rows */
void Board::KillFilledRows() {
    int row = height-1;
    int numkilled = 0;

    while (row >= 0) {
	/* Check if row can be killed */
	boolean kill = true;
	Square* r = square[row];

	for (int i = 0; i < width; i++) {
	    if (r[i] == Empty) {
		kill = false;
		break;
	    }
	}

	if (kill) {
	    /* Kill row */
	    for (int i = row-1; i >= 0; i--)
		square[i+1] = square[i];
	    for (i = 0; i < width; i++)
		r[i] = Empty;
	    square[0] = r;
	    numkilled++;
	}
	else {
	    row--;
	}
    }
    for (int i = 0; i < height; i++)
	filled[i] = false;

    /* Score 10 points for each row filled */
    score += 10 * numkilled;
}

void Board::Clear() {
    for (int i = 0; i < height; i++) {
	filled[i] = false;
	for (int j = 0; j < width; j++)
	    square[i][j] = Empty;
    }
    falling = false;
    blocked = false;
    score = 0;
}

boolean Board::LegalAt(int nx, int ny, Rotation nrot) {
    for (int i = 0; i < NumObjectDivs; i++) {
	ObjectDiv od = objects[(int) ftype][(int) nrot][i];

	/* Disallow if point to left or right of box */
	int dx = nx + od.x;
	if ((dx < 0) || (dx >= width)) return false;

	/* Disallow if point below box */
	int dy = ny + od.y;
	if (dy >= height) return false;

	/* Allow if above box */
	if (dy < 0) continue;

	/* Disallow if point already taken */
	if (square[dy][dx] != Empty) return false;
    }
    return true;
}

boolean Board::LegalPoint(int x, int y) {
    boolean b = ((x >= 0) && (y >= 0) && (x < width) && (y < height));
    return b;
}

void Board::DrawFalling() {
    for (int i = 0; i < NumObjectDivs; i++) {
	ObjectDiv od = objects[(int) ftype][(int) rot][i];

	if (LegalPoint(x+od.x, y+od.y)) square[y+od.y][x+od.x] = Occupied;
    }
}

void Board::EraseFalling() {
    for (int i = 0; i < NumObjectDivs; i++) {
	ObjectDiv od = objects[(int) ftype][(int) rot][i];

	if (LegalPoint(x+od.x, y+od.y)) square[y+od.y][x+od.x] = Empty;
    }
}
