/* Copyright (c) 1991 by Sanjay Ghemawat. */

#include <InterViews/border.h>
#include <InterViews/box.h>
#include <InterViews/painter.h>
#include <InterViews/glue.h>
#include <InterViews/canvas.h>
#include <InterViews/sensor.h>
#include <stdlib.h>
#include "pboard.h"
#include "board.h"
#include "command.h"
#include "label.h"
#include "field.h"
#include "game.h"
#include "generator.h"

/*
 * Various generators.
 */
static RandomGen	randomGen;
static CycleGen		cycleGen;
static RepeaterGen	repeaterGen;
static Alternator	alternatorGen;

extern "C" {
    extern int strcasecmp(const char*, const char*);
    extern void gettimeofday(Time*, void*);
};

/* An ObjectDisplay displays a tetris object */
class ObjectDisplay {
  public:
    PBoard*	printer;			/* Board to display the object */

    ObjectDisplay(int size);
    ~ObjectDisplay();

    void Show(ObjectType);
    void Clear();
  private:
    Square**	source;			/* Array for displaying object */

    /* Object descriptions; */
    static int obj_inited;		/* SHould be init to zero */
    static int obj_width, obj_height;	/* Max dimensions of objects */
    static int obj_x, obj_y;		/* Coordinates of objects */
};

void ObjectDisplay::Show(ObjectType ot) {
    Clear();

    /* Draw object */
    for (int i = 0; i < NumObjectDivs; i++) {
	source[obj_y+objects[ot][R0][i].y][obj_x+objects[ot][R0][i].x] =
	    Occupied;
    }
    printer->Refresh();
}

void ObjectDisplay::Clear() {
    for (int y = 0; y < obj_height; y++)
	for (int x = 0; x < obj_width; x++)
	    source[y][x] = Empty;
}

ObjectDisplay::ObjectDisplay(int size) {
    if (!obj_inited) {
        /* Find size needed for displaying all objects. */
        obj_width = 0;
        obj_height = 0;
        int min_x = 0;
        int min_y = 0;
        int max_x = 0;
        int max_y = 0;
        for (int i = 0; i < NumObjectTypes; i++) {
            /* get object i without rotation */
            for (int j = 0; j < NumObjectDivs; j++) {
                min_x = min(min_x, objects[i][R0][j].x);
                max_x = max(max_x, objects[i][R0][j].x);
                min_y = min(min_y, objects[i][R0][j].y);
                max_y = max(max_y, objects[i][R0][j].y);
            }
        }
        obj_width  = max_x - min_x + 1;
        obj_height = max_y - min_y + 1;
        obj_x = -min_x;
        obj_y = -min_y;
        obj_inited = true;
    }
    source = new Square*[obj_height];
    for (int y = 0; y < obj_height; y++) {
        source[y] = new Square[obj_width];
        for (int x = 0; x < obj_width; x++)
            source[y][x] = Empty;
    }
    printer = new PBoard(source, obj_height, obj_width, size);
}

ObjectDisplay::~ObjectDisplay() {
    delete printer;
    for (int y = 0; y < obj_height; y++)
        delete source[y];
    delete source;
}

/* Normalize a given struct Time. */
static void NormalizeTime(Time& t) {
    int sec = t.usec / 1000000;
    t.sec += sec;
    t.usec -= (sec * 1000000);
}

/* Level: 1 should be in range */
#define MIN_LEVEL	0
#define MAX_LEVEL	10

/* Array of tick times in milliseconds for different levels. */
static int leveltime[] = {
    1000,				/* 0 */
     800,				/* 1 */
     600,				/* 2 */
     400,				/* 3 */
     300,				/* 4 */
     250,				/* 5 */
     200,				/* 6 */
     150,				/* 7 */
     100,				/* 8 */
      75,				/* 9 */
      50,				/* 10 */
};

/* Array of thresholds for switching levels */
/*    threshold[i] indicates number of rows needed to be killed */
/*    in the whole game so far to go from level i to level i+1. */
static int threshold[] = {
    10,					/* 0 */
    20,					/* 1 */
    30,					/* 2 */
    40,					/* 3 */
    50,					/* 4 */
    60,					/* 5 */
    70,					/* 6 */
    80,					/* 7 */
    90,					/* 8 */
    100,				/* 9 */
    100000000				/* 10 */
    };

/* State of game */
static const int Stopped = 0, Running = 1, Frozen = 2;

static int   StateValueList[] = { Stopped,   Running,   Frozen };
static char* StateNameList[]  = { "Stopped", "Running", "Frozen" };
static int   StateListLength  = sizeof(StateValueList) / sizeof(int);

Game::Game() : (nil, 0) {
    input = new Sensor();
    input->Catch(KeyEvent);
    input->Catch(DownEvent);
    input->Catch(EnterEvent);
    input->Catch(LeaveEvent);
    Listen(input);

    /* Create command menu */
    PulldownMenu* commands = new PulldownMenu(" Command ");
    commands->Include(new Command("Run",   "^R", this, &Game::Start));
    commands->Include(new Command("Stop",  "^S", this, &Game::Stop));
    commands->Include(new Command("Freeze","f",  this, &Game::Freeze));
    commands->Include(new Command("Thaw",  "t",  this, &Game::Thaw));
    commands->Include(new Control(new HBorder));
    commands->Include(new Command("Quit",  "^Q", this, &Game::Quit));

    /* Create Generator menu */
    PulldownMenu* generators = new PulldownMenu(" Generator ");
    generators->Include(new Command("Random",	  this, &Game::UseRandom));
    generators->Include(new Command("Repeater",	  this, &Game::UseRepeater));
    generators->Include(new Command("Cycle",	  this, &Game::UseCycle));
    generators->Include(new Command("Alternator", this, &Game::UseAlternator));

    /* Create option menu */
    PulldownMenu* options = new PulldownMenu(" Options ");
    options->Include(new Command("Toggle Next", "n", this, &Game::ToggleNext));

    /* Create the menu bar */
    MenuBar* menubar = new MenuBar();
    menubar->Include(commands);
    menubar->Include(generators);
    menubar->Include(options);
    menubar->Include(new Control(new HGlue));

    finished = false;
    state = new EnumState(Stopped, StateListLength, StateValueList,
			  StateNameList);

    level = new IntState(1);
    score = new IntState(0);
    genstate = new StringState("XXXXXX");
    block_size = 20;
    bwidth = DEF_WIDTH;
    bheight = DEF_HEIGHT;
    showing_next = false;
    sync = true;
    init_level = level->Get();
    increase_levels = true;
    rows_killed = 0;
    killcount = new IntState(rows_killed);
    on_canvas = true;

    const char* prop;
    int value;

    if (prop = GetAttribute("Level")) {
	value = atoi(prop);
	SetLevel(value, true);
    }

    /* Get generator type */
    if (prop = GetAttribute("Generator")) {
	if (strcasecmp(prop, "random") == 0) generator = &randomGen;
	else if (strcasecmp(prop, "cycle") == 0) generator = &cycleGen;
	else if (strcasecmp(prop, "repeat") == 0) generator = &repeaterGen;
	else if (strcasecmp(prop, "alt") == 0) generator = &alternatorGen;
	else generator = &randomGen;
    }
    else generator = &repeaterGen;
    genstate->Set(generator->name);

    /* Should game be allowed to get harder as rows are killed off? */
    if (prop = GetAttribute("GetHarder")) {
	if (strcasecmp(prop, "on" ) == 0) increase_levels = true;
	if (strcasecmp(prop, "off") == 0) increase_levels = false;
    }

    /* Should the next object be displayed */
    if (prop = GetAttribute("Next")) {
	if (strcasecmp(prop, "on" ) == 0) showing_next = true;
	if (strcasecmp(prop, "off") == 0) showing_next = false;
    }

    /* Should the game be synchronized with the display */
    if (prop = GetAttribute("Sync")) {
	if (strcasecmp(prop, "on" ) == 0) sync = true;
	if (strcasecmp(prop, "off") == 0) sync = false;
    }

    /* Block size, Board width and Board height */
    if (prop = GetAttribute("Size")) {
	value = atoi(prop);
	if (value >= 10) block_size = value;
    }
    if (prop = GetAttribute("Width")) {
	value = atoi(prop);
	if (value > 3) bwidth = value;
    }
    if (prop = GetAttribute("Height")) {
	value = atoi(prop);
	if (value > 5) bheight = value;
    }

    Painter* bgPaint = new Painter(stdpaint);
    bgPaint->SetColors(stdpaint->GetBgColor(), stdpaint->GetFgColor());

    /* Create the components of the game. */
    board = new Board(generator, bheight, bwidth);
    pboard = new PBoard(board->Contents(),
			board->Height(),
			board->Width(),
			block_size);
    next = new ObjectDisplay(block_size);

/* Macros to generate 'Label: Value' lines. */
#define ENTRY(l,r)   new HBox(new Label(l), new HGlue(), new Label(r,Right))
#define ENTRYS(l,r)  ENTRY(new StringState(l),r)
#define ENTRYSS(l,r) ENTRY(new StringState(l),new StringState(r))

    HBox* nextbox = new HBox(new Label(new StringState("Next ")),
			     new HGlue(), next->printer);
    nextbox->Align(Center);
    VBox* stats = new VBox(ENTRYS("Level ",	level),
			   ENTRYS("Generator ",	genstate),
			   ENTRYS("Score ",	score),
			   ENTRYS("Status ",	state),
			   ENTRYS("Killed ",	killcount),
			   nextbox);

    VBox* helper = new VBox(ENTRYSS("Left",	"j"),
			    ENTRYSS("Right",	"l"),
			    ENTRYSS("Rotate",	"k"),
			    ENTRYSS("Fall",	"space")
			    );

    VBox* rcol = new VBox(stats, new VGlue(bgPaint), helper);
    rcol->Propagate(false);

    HBox* body = new HBox(new Frame(pboard),
			  new HGlue(bgPaint),
			  rcol,
			  new HGlue(bgPaint));
    body->Propagate(false);

    VBox* interior = new VBox(menubar, new HBorder(2), body);
    Insert(interior);
}

Game::~Game()
{
    delete state;
    delete level;
    delete score;
    delete genstate;
    delete killcount;
}

/* Adjust input sensor so that if a read happens right now, it will */
/* finish before we are supposed to let the falling object drop another */
/* round. */
void Game::AdjustTicker() {
    if ((state->Get() == Running) && on_canvas) {
	Time delta;
	delta.sec = 0;
	delta.usec = leveltime[level->Get()] * 1000;

	if (sync) {
	    /* Do a sync and update delta by the time needed to do the sync */
	    Time now, after_sync;

	    gettimeofday(&now, 0);
	    Sync();
	    gettimeofday(&after_sync, 0);
	    after_sync.sec  -= now.sec;
	    after_sync.usec -= now.usec;
	    if (after_sync.usec < 0) {
		after_sync.sec--;
		after_sync.usec += 1000000;
	    }
	    NormalizeTime(after_sync);

	    /* If delta has already elapsed, then dont wait */
	    if (after_sync.sec > delta.sec)
		delta.sec = delta.usec = 0;
	    else if ((after_sync.sec == delta.sec) &&
		     (after_sync.usec >= delta.usec))
		delta.sec = delta.usec = 0;
	}
	else {
	    Flush();
	}
	NormalizeTime(delta);
	input->CatchTimer(delta.sec, delta.usec);
    }
}

void Game::Flash() {
    Turn(Occupied);
    Turn(Empty);
    Turn(Occupied);
    Turn(Empty);
    Turn(Occupied);
}

void Game::Turn(Square s) {
    for (int i = 0; i < bheight; i++) {
	if (board->Filled(i)) {
	    for (int j = 0; j < bwidth; j++)
		board->Set(i, j, s);
	}
    }
    pboard->Refresh();

    /* Wait for .05 seconds */
    input->CatchTimer(0, 50000);
    Flush();
    Event e;
    while (true) {
	Read(e);
	if (e.target == this) break;
    }
    input->Ignore(TimerEvent);
}

void Game::SetLevel(int l, boolean set_by_user) {
    if ((l >= MIN_LEVEL) && (l <= MAX_LEVEL)) {
	level->Set(l);
	if (set_by_user) init_level = l;
    }
}

void Game::Handle(Event& e) {
    switch (e.eventType) {
      case TimerEvent:
	if ((state->Get() == Running) && on_canvas) board->Tick();
	break;
      case EnterEvent:
	on_canvas = true;
	break;
      case LeaveEvent:
	on_canvas = false;
	break;
      case KeyEvent:
	switch (e.keystring[0]) {
	  case 'j':
	    board->LeftShift();
	    break;
	  case 'l':
	    board->RightShift();
	    break;
	  case 'k':
	    board->RotateFalling();
	    break;
	  case ' ':
	    board->FallAllTheWay();
	    break;
	  case '\022':		/* Control-R */
	    Start();
	    break;
	  case '\023':		/* Control-S */
	    Stop();
	    break;
	  case '\021':		/* Control-Q */
	    Quit();
	    break;
	  case '+':
	    Harder();
	    break;
	  case '-':
	    Easier();
	    break;
	  case 'f':
	    Freeze();
	    break;
	  case 't':
	    Thaw();
	    break;
	  case 'n':
	    ToggleNext();
	    break;
	}
	break;
      case DownEvent:
	switch (e.button) {
 	  case LEFTMOUSE:
 	    board->LeftShift();
 	    break;
 	  case MIDDLEMOUSE:
 	    board->RotateFalling();
 	    break;
 	  case RIGHTMOUSE:
 	    board->RightShift();
 	    break;
 	}
 	break;
      default:
	break;
    }
}

void Game::Run() {
    state->Set(Stopped);
    finished = false;
    while (!finished) {
	Event e;

	pboard->Refresh();
	if (showing_next) next->Show(generator->next);
	score->Set(board->Score());
	if (board->Blocked()) Stop();
	else {
	    for (int i = 0; i < bheight; i++) {
		if (board->Filled(i)) {
		    rows_killed++;
		}
	    }
	    killcount->Set(rows_killed);
	    for (i = 0; i < bheight; i++) {
		if (board->Filled(i)) {
		    Flash();
		    board->KillFilledRows();
		    pboard->Refresh();
		    Sync();
		    break;
		}
	    }
	    if (increase_levels && (rows_killed >= threshold[level->Get()])) {
		/* Increase level because enough rows killed */
		SetLevel(level->Get()+1, false);
	    }
	}
	AdjustTicker();
	Read(e);
	if (e.target == nil) continue;
	e.target->Handle(e);
    }
}

void Game::Start() {
    if (state->Get() == Stopped) {
	rows_killed = 0;
	killcount->Set(rows_killed);
	SetLevel(init_level, false);
	state->Set(Running);
	pboard->Clear();
	board->Clear();
	AdjustTicker();
    }
}

void Game::Stop() {
    if (state->Get() != Stopped) {
	state->Set(Stopped);
	input->Ignore(TimerEvent);
    }
}

void Game::Freeze() {
    if (state->Get() == Running) {
	state->Set(Frozen);
	input->Ignore(TimerEvent);
    }
}

void Game::Thaw() {
    if (state->Get() == Frozen) {
	state->Set(Running);
	AdjustTicker();
    }
}

void Game::Quit() {
    state->Set(Stopped);
    input->Ignore(TimerEvent);
    finished = true;
}

void Game::Harder() {
    SetLevel(level->Get()+1, true);
}

void Game::Easier() {
    SetLevel(level->Get()-1, true);
}

void Game::ToggleNext() {
    showing_next = ! showing_next;
    if (! showing_next) {
	next->Clear();
	next->printer->Refresh();
    }
}

void Game::UseRandom() {
    generator = &randomGen;
    board->SetGenerator(generator);
    genstate->Set(generator->name);
}

void Game::UseCycle() {
    generator = &cycleGen;
    board->SetGenerator(generator);
    genstate->Set(generator->name);
}

void Game::UseRepeater() {
    generator = &repeaterGen;
    board->SetGenerator(generator);
    genstate->Set(generator->name);
}

void Game::UseAlternator() {
    generator = &alternatorGen;
    board->SetGenerator(generator);
    genstate->Set(generator->name);
}
